section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory "cuc_csp_refinement"
imports "stable_failures_traces_conformance_parallel"
  "CSP-Prover-5-1-2016/CSP_F_Main"
begin

subsection {* Preparations *}

paragraph {* CSP - CUC conversion *}

text {* The conversion takes a map to allow for different events in the CSP and CUC world. *}

text {* A question is how to handle Ticks. In general, one idea is to simply append a tick if the
CUC state is normal and points out of the domain. Otherwise, no tick is added. The question is,
how to differentiate between CSP STOP and SKIP.
*}

fun CUCtoCSPtrace :: "'e definitions.trace \<Rightarrow> ('e definitions.event \<Rightarrow> 'ce) \<Rightarrow> 'ce Trace.trace"
("CV\<^sub>t\<^sub>r _ _" [90])
where
"(CV\<^sub>t\<^sub>r [] _) = <>" |
"(CV\<^sub>t\<^sub>r (x # xs) CV) = <Ev (CV x)> ^^^ (CV\<^sub>t\<^sub>r xs CV)"

definition CSPtoCUCtrace :: "'ce Trace.trace \<Rightarrow> ('ce \<Rightarrow> 'e definitions.event) \<Rightarrow> 'e definitions.trace"
where
"CSPtoCUCtrace xs CV = map (\<lambda> x . case x of Ev x' \<Rightarrow> CV x') (Rep_trace xs)"

text {* We also need to convert the refusal set. *}

definition CUCtoCSPrefusal :: "'e definitions.event set \<Rightarrow> ('e definitions.event \<Rightarrow> 'ce) \<Rightarrow> 'ce Trace.event set"
("CV\<^sub>X _ _" [90]) where
"(CV\<^sub>X X' CV) = {x . \<exists> x' . x' \<in> X' \<and> x = Ev (CV x')}"

definition CSPtoCUCrefusal :: "'ce Trace.event set \<Rightarrow> ('ce \<Rightarrow> 'e definitions.event) \<Rightarrow> 'e definitions.event set"
where
"(CSPtoCUCrefusal X' CV) = {x . \<exists> x' . Ev x' \<in> X' \<and> x = (CV x')}"

text {* Transform the connecting property. As we do not have a notion of tick in CUC, we simply forbid it. *}

definition CUCConnToCSPConn :: "(('e definitions.trace \<times> 'e definitions.event set) \<Rightarrow> bool) \<Rightarrow> ('ce \<Rightarrow> 'e) \<Rightarrow> ('ce Set_F.failure \<Rightarrow> bool)" where
"CUCConnToCSPConn Conn CV = (\<lambda> (tr, X) . noTick tr \<and> Tick \<notin> X \<and> Conn (CSPtoCUCtrace tr CV, {x . \<exists> x' . Ev x' \<in> X \<and> x = CV x'}))"

definition CUCtoCSPStableFailure :: "('s, 'e) parallelFailure \<Rightarrow> ('e \<Rightarrow> 'ce) \<Rightarrow> 'ce definitions.event Set_F.failure"
("CV\<^sub>f _ _" [90])
where
"(CV\<^sub>f s CV) = (CV\<^sub>t\<^sub>r (fst s) CV, CV\<^sub>X (snd (snd s)) CV)"

text {* A sanity check: Converting a CUC trace into the CSP world and back should result in the original trace, if
the event projection is bijective. *}

lemma CSPtoCUCtrace_inj:
  assumes inj: "inj CV"
  shows "CSPtoCUCtrace (CV\<^sub>t\<^sub>r tr CV) (inv CV) = tr"
  unfolding CSPtoCUCtrace_def
proof (induct tr, auto)
  show "Rep_trace <> = []"
    by (simp add: Abs_trace_inverse nilt_def)
next
  show "\<And> a tr . map (\<lambda>x. case x of Ev x' \<Rightarrow> (inv CV) x') (Rep_trace CV\<^sub>t\<^sub>r tr CV) = tr \<Longrightarrow>
              map (\<lambda>x. case x of Ev x' \<Rightarrow> (inv CV) x') (Rep_trace (<Ev (CV a)> ^^^ CV\<^sub>t\<^sub>r tr CV)) = a # tr"
    by (simp add: Abs_trace_inverse appt_head inj)
qed

text {* The same property should hold for CSP -> CUC -> CSP. *}

lemma CUCtoCSPtrace_bij:
  assumes bij: "bij CV" and noTick: "noTick tr"
  shows "CV\<^sub>t\<^sub>r (CSPtoCUCtrace tr (inv CV)) CV = tr"
proof -
  have "\<exists> tr' . CV\<^sub>t\<^sub>r tr' CV = tr" using noTick
  proof (induct tr)
    fix y
    show "y \<in> trace \<Longrightarrow> noTick (Abs_trace y) \<Longrightarrow> \<exists>tr'. CV\<^sub>t\<^sub>r tr' CV = Abs_trace y"
    proof (induct y, auto)
      show "\<exists>tr'. CV\<^sub>t\<^sub>r tr' CV = Abs_trace []"
        by (metis CUCtoCSPtrace.simps(1) nilt_def)
    next
      fix a y
      assume
        ba1: "y \<in> trace \<Longrightarrow> noTick (Abs_trace y) \<Longrightarrow> \<exists>tr'. CV\<^sub>t\<^sub>r tr' CV = Abs_trace y" and
        ba2: "a # y \<in> trace" and
        ba3: "noTick (Abs_trace (a # y))"
      then obtain tr' where tr': "CV\<^sub>t\<^sub>r tr' CV = Abs_trace y"
        by (metis Abs_trace_inverse CUCtoCSPtrace.simps(1) decompo_head_in_trace insert_iff list.simps(15) nilt_def noTick_def sett_def)
      obtain e where e: "Ev e = a"
        by (metis ba2 ba3 decompo_head_in_trace event.exhaust not_noTick_Tick)
      have CVeq: "\<And> e . CV (inv CV e) = e"
        by (simp add: bij bij_is_surj surj_f_inv_f)
      have eq: "CV\<^sub>t\<^sub>r [(inv CV) e] CV = <Ev e>"
        unfolding "CUCtoCSPtrace.simps" CVeq
        by simp
      thus "\<exists>tr'. CV\<^sub>t\<^sub>r tr' CV = Abs_trace (a # y)"
        using tr' unfolding e[symmetric]
        by (metis (no_types) Abs_trace_inverse CUCtoCSPtrace.simps(2) eq appt_head ba2 decompo_head_in_trace tr' CVeq)
    qed
  qed
  then obtain tr' where tr': "CV\<^sub>t\<^sub>r tr' CV = tr" by blast
  show "CV\<^sub>t\<^sub>r (CSPtoCUCtrace tr (inv CV)) CV = tr"
    unfolding tr'[symmetric]
    using CSPtoCUCtrace_inj bij bij_is_inj
    by force 
qed

text {* Converting the refusal set CUC -> CSP -> CUC is an identity, too. *}

lemma CUCtoCSPrefusal_inv:
  assumes inj: "inj CV"
  shows "{x. \<exists>x'. Ev x' \<in> CV\<^sub>X X CV \<and> x = inv CV x'} = X"
unfolding CUCtoCSPrefusal_def using inj
by fastforce

text {* There is no tick in a CUC->CSP converted trace, nor in a refusal set. *}

lemma CUCtoCSPtrace_noTick: "noTick CV\<^sub>t\<^sub>r tr CV" proof (induct tr, auto) qed
lemma CUCtoCSPrefusal_noTick: "Tick \<notin> CV\<^sub>X X CV" unfolding CUCtoCSPrefusal_def by simp

text {* We can use the trace restriction on both cuc and csp side. *}

lemma rest_tr_cv_distr:
  "inj CV \<Longrightarrow> CV\<^sub>t\<^sub>r (tr \<restriction> \<alpha>) CV = CV\<^sub>t\<^sub>r tr CV rest-tr (CV ` \<alpha>)"
proof (induct tr, auto)
  fix a tr
  assume
    inj: "inj CV" and
    not: "a \<notin> \<alpha>"
  hence "CV a \<notin> CV ` \<alpha>"
    by (simp add: inj_image_mem_iff)
  thus "CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha> = <Ev (CV a)> rest-tr CV ` \<alpha> ^^^ CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha>"
    by auto
qed

text {* The connecting property will in both worlds show the same behaviour. *}

lemma CUCConnToCSPConn_eq:
  assumes bij: "bij CV"
  shows "Conn (tr, X) = CUCConnToCSPConn Conn (inv CV) (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV)"
proof -
  have inj: "inj CV" using bij by (simp add: bij_def)  
  show ?thesis
    unfolding CUCConnToCSPConn_def
  proof (auto simp add: CSPtoCUCtrace_inj inj CUCtoCSPrefusal_noTick CUCtoCSPrefusal_inv CUCtoCSPtrace_noTick bij) qed
qed

text {* Instead of converting the trace and refusal set from CUC -> CSP, we cann instead convert the
connecting property from CUC -> CSP. *}

lemma CUCConnToCSPConn_imp:
  assumes
    bij: "bij CV" and
    cuc: "\<forall>tr X . Conn (tr, X) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f sndF ([[$spec]]Ff [[PN]]Ffix)"
  shows "\<forall> tr X . CUCConnToCSPConn Conn (inv CV) (tr, X) \<longrightarrow> (tr, X) :f sndF ([[$spec]]Ff [[PN]]Ffix)"
using cuc
proof (auto)
  fix tr X
  assume
    aa3: "\<forall>tr X. Conn (tr, X) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f failures ($spec) [[PN]]Ffix" and
    aa4: "CUCConnToCSPConn Conn (inv CV) (tr, X)"
  from aa4[unfolded CUCConnToCSPConn_def] have
    pre: "noTick tr \<and> Tick \<notin> X \<and> Conn (CSPtoCUCtrace tr (inv CV), {x. \<exists>x'. Ev x' \<in> X \<and> x = inv CV x'})"
    by fast
  from CUCtoCSPtrace_bij[OF bij pre[THEN cj2_1]] have
    tr: "tr = CV\<^sub>t\<^sub>r CSPtoCUCtrace tr (inv CV) CV" ..
  have inj: "inj CV" using bij unfolding bij_def ..
  have X: "X = CV\<^sub>X {x. \<exists>x'. Ev x' \<in> X \<and> x = inv CV x'} CV"
    unfolding CUCtoCSPrefusal_def
  proof (auto)
    show "\<And>x. x \<in> X \<Longrightarrow> \<exists>x'. (\<exists>x'a. Ev x'a \<in> X \<and> x' = inv CV x'a) \<and> x = Ev (CV x')"
      using bij pre[THEN cj3_2]
      by (metis UNIV_I bij_betw_inv_into_right event_Tick_or_Ev)
  next
    show "\<And>x'a. Ev x'a \<in> X \<Longrightarrow> Ev (CV (inv CV x'a)) \<in> X"
      using bij pre[THEN cj3_2]
      by (simp add: bij_is_surj surj_f_inv_f)
  qed
  have "(CV\<^sub>t\<^sub>r CSPtoCUCtrace tr (inv CV) CV, CV\<^sub>X {x. \<exists>x'. Ev x' \<in> X \<and> x = inv CV x'} CV) :f failures ($spec) [[PN]]Ffix"
    using aa3 pre[THEN cj3_3] by blast
  thus "(tr, X) :f failures ($spec) [[PN]]Ffix"
    unfolding tr[symmetric] X[symmetric]
    .
qed

definition memFailure :: "'a Trace.trace \<times> ('a Trace.trace \<times> 'a Trace.event set) \<Rightarrow> 'a domF \<Rightarrow> bool" ("_ \<in>\<^sub>t _" [51,52]) where
  "(memFailure x F) = ((fst x) :t (fstF F) \<and> (snd x) :f (sndF F))" 

text {* Converting a set of typed events in CSP world to a set of events. This ways we
effectively can use the elements directly, but we lose the information, whether a TICK was
in the original set. (As it is CSP internal only, we need no conversion function) *}

definition eventSetToSet :: "'e Trace.event set \<Rightarrow> 'e set"
("CV\<^sub>e\<^sub>s _" [90])
where
"(CV\<^sub>e\<^sub>s S) = { x . Ev x \<in> S }"

lemma eventSetToSetDistributive: "CV\<^sub>e\<^sub>s A \<union> CV\<^sub>e\<^sub>s B = CV\<^sub>e\<^sub>s (A \<union> B)"
  unfolding eventSetToSet_def by blast

text {* Because of the removed tick, we have to add it in some scenarios. *}

definition tickify_failureset :: "'a set \<Rightarrow> 'a Trace.event set" ("'(_')\<^sub>t\<^sub>i\<^sub>c\<^sub>k")
  where
"(A)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = insert Tick (Ev ` A)"

text {* With added tick, the conversion function together with the event creation function
is the identity. *}

lemma evInjOfCVes: "insert Tick x = insert Tick (Ev ` (CV\<^sub>e\<^sub>s x))"
  unfolding eventSetToSet_def image_def
  using event_Tick_or_Ev by fast

paragraph {* Connection to CSP *}

(* just to be safe ... *)
lemma unfoldFtofailues: "sndF ([[P]]Ff [[ProcName]]Ffix) = failures P [[ProcName]]Ffix"
  by auto
  
lemma unfoldTtotraces: "[[P]]Tf [[ProcName]]Tfix = traces P [[ProcName]]Tfix"
  by (simp add: semTf_def)

text {* Make sure, that we use the right definition for alphabetized parallel failures.
This definition is taken from Steve Schneider's book \emph{Concurrent and Real Time Systems: The CSP approach}, page 180.
*}
  
lemma SteveSchneiderAlphaParFailuresEqualToCSPVersion:
  shows (* In rest-tr, tick is preserved per default, so no extending with tick. *)
  "{(tr, X) . \<exists> X\<^sub>1 X\<^sub>2
    . X \<inter> (A \<union> B \<union> {Tick}) = (X\<^sub>1 \<inter> (A \<union> {Tick})) \<union> (X\<^sub>2 \<inter> (B \<union> {Tick}))
    \<and> (tr rest-tr CV\<^sub>e\<^sub>s A, X\<^sub>1) :f (failures P PN)
    \<and> (tr rest-tr CV\<^sub>e\<^sub>s B, X\<^sub>2) :f (failures Q PN)
    \<and> sett tr \<subseteq> (A \<union> B \<union> {Tick})}f =
    failures (P |[CV\<^sub>e\<^sub>s A,CV\<^sub>e\<^sub>s B]| Q) PN
  "                     
unfolding failures_Alpha_parallel 
using eventSetToSetDistributive
unfolding eventSetToSetDistributive evInjOfCVes[symmetric]
by auto

text {* Another version of the previous lemma, where the used refusal sets @{term A} and @{term B}
are simply sets of events (and thus not containing Tick).
*}

lemma SteveSchneiderAlphaParFailuresEqualToCSPVersion2:
  shows (* In rest-tr, tick is preserved per default, so no extending with tick. *)
  "{(tr, X) . \<exists> X\<^sub>1 X\<^sub>2
    . (X \<inter> (A \<union> B)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) = (X\<^sub>1 \<inter> (A)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<union> (X\<^sub>2 \<inter> (B)\<^sub>t\<^sub>i\<^sub>c\<^sub>k)
    \<and> (tr rest-tr A, X\<^sub>1) :f (failures P PN)
    \<and> (tr rest-tr B, X\<^sub>2) :f (failures Q PN)
    \<and> sett tr \<subseteq> (A \<union> B)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}f =
    failures (P |[A,B]| Q) PN
  "
unfolding failures_Alpha_parallel tickify_failureset_def
by auto
  
lemma SteveSchneiderAlphaParTracesEqualToCSPVersion:
  shows (* In rest-tr, tick is preserved per default, so no extending with tick. *)
  "{tr .
      (tr rest-tr CV\<^sub>e\<^sub>s A) :t (traces P PN)
    \<and> (tr rest-tr CV\<^sub>e\<^sub>s B) :t (traces Q PN)
    \<and> sett tr \<subseteq> (A \<union> B \<union> {Tick})}t =
    traces (P |[CV\<^sub>e\<^sub>s A,CV\<^sub>e\<^sub>s B]| Q) PN
  "                     
unfolding traces_Alpha_parallel 
using eventSetToSetDistributive
unfolding eventSetToSetDistributive evInjOfCVes[symmetric]
by auto

lemma SteveSchneiderAlphaParTracesEqualToCSPVersion2:
  shows (* In rest-tr, tick is preserved per default, so no extending with tick. *)
  "{tr .
      (tr rest-tr A) :t (traces P PN)
    \<and> (tr rest-tr B) :t (traces Q PN)
    \<and> sett tr \<subseteq> (A \<union> B)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}t =
    traces (P |[A,B]| Q) PN
  "
unfolding traces_Alpha_parallel tickify_failureset_def
by auto

text {* Helper Lemma to handle the custom defined datatype of failures in the CSP package. *}

lemma memberF: (* also keep in mind: CollectF_open_memF *) 
  assumes "S \<in> setF" and "x \<in> S"
  shows "x :f (Abs_setF S)"
by (simp add: Abs_setF_inverse assms(1) assms(2) memF_def)

subsection {* Connecting Property and Refinement *}

text {* In the following definitions, we always need to carry the process name function as it is used in the
CSP Prover. For its documentation, refer to the CSP Prover package.
In general, we have two basic concepts of the CUC-CSP refinement. One with the user definable connecting property,
the other one without a connecting property. Also, both for SF semantics and T semantics.
A further split is due to the starting property, which in general can be generic. There is also a predefined precondition,
which fixes the starting program counter to be 1, the state to be a normal state, and the trace to be empty. The overall idea is,
that each code probably starts the execution at some program counter, which we simply fixed arbitrarily. Also, the trace should always
be empty in the beginning. Further assumptions concerning the state (i.e. initial values of the registers) can be made
explicit via some initializing instruction(s). Hence, this fixed starting property should be sufficient. Still, we want
to praise genericity and thus allow for arbitrary starting properties.
A third parameter, being required in any case, is the conversion function to relate the CUC event type to the CSP event type.
*}

paragraph {* Refinement with Conn, generic Pre *}

text {* The basic parallel CUC-CSP semantics refinement for SF and T with the connecting property.
Also, we allow to define the condition for the starting states.
*}

definition spec_impl_refinement_sf :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) parallelFailure \<Rightarrow> bool) \<Rightarrow> ('a Set_F.failure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_, _, _')=F _")
where
"spec_impl_refinement_sf Spec ProcNames Pre Conn CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x' . Pre x'}) \<longrightarrow> (Conn (CV\<^sub>f (tr, s, X) CV))) \<and>
  (\<forall> tr X . Conn (tr, X) \<longrightarrow> (tr, X) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) parallelTraceState \<Rightarrow> bool) \<Rightarrow> ('a Trace.trace \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_, _, _')=T _")
where
"spec_impl_refinement_t Spec ProcNames Pre Conn CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>T{x' . Pre x'}) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV))) \<and>
  (\<forall> tr . Conn (tr) \<longrightarrow> tr :t [[Spec]]Tf [[ProcNames]]Tfix))
"

text {* The following two definitions are reformulated special cases of the previous one for the case of sequential code.
They allow for easier reasoning.
*}

definition (in SFsemSigma) spec_impl_refinement_sf_seq :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) stable_failures.failure \<Rightarrow> bool) \<Rightarrow> ('a Set_F.failure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_sf_seq Spec ProcNames Pre Conn CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> (sfsem Impl {x' . Pre x'})) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV))) \<and>
  (\<forall> tr X . Conn (tr, X) \<longrightarrow> (tr, X) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t_seq :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) trace_state \<Rightarrow> bool) \<Rightarrow> ('a Trace.trace \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_t_seq Spec ProcNames Pre Conn CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> (densem Impl {x' . Pre x'})) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV))) \<and>
  (\<forall> tr . Conn (tr) \<longrightarrow> (tr) :t [[Spec]]Tf [[ProcNames]]Tfix))
"

paragraph {* Refinement with Conn, predefined Pre *}

text {* The following definitions are as the previous ones, except for the start condition being
fixed to the minimal required one as described in previously. *}

definition spec_impl_refinement_sf_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('a Set_F.failure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(, _, _')=F _")
where
"spec_impl_refinement_sf_pre_set Spec ProcNames Conn CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x' . PreStd\<^sub>S\<^sub>F x'}) \<longrightarrow> (Conn (CV\<^sub>f (tr, s, X) CV))) \<and>
  (\<forall> tr X . Conn (tr, X) \<longrightarrow> (tr, X) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('a Trace.trace \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(, _, _')=T _")
where
"spec_impl_refinement_t_pre_set Spec ProcNames Conn CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>T{x' . PreStd\<^sub>T x'}) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV))) \<and>
  (\<forall> tr . Conn (tr) \<longrightarrow> tr :t [[Spec]]Tf [[ProcNames]]Tfix))
"

definition (in SFsemSigma) spec_impl_refinement_sf_seq_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('a Set_F.failure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_sf_seq_pre_set Spec ProcNames Conn CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> (sfsem Impl {x' . PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q x'})) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV))) \<and>
  (\<forall> tr X . Conn (tr, X) \<longrightarrow> (tr, X) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t_seq_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('a Trace.trace \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_t_seq_pre_set Spec ProcNames Conn CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> (densem Impl {x' . PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q x'})) \<longrightarrow> (Conn (CV\<^sub>t\<^sub>r tr CV))) \<and>
  (\<forall> tr . Conn (tr) \<longrightarrow> (tr) :t [[Spec]]Tf [[ProcNames]]Tfix))
"

paragraph {* Refinement without Conn, generic Pre *}

text {* The following definitions work without the connecting property. This is the ultimate goal, as it eliminates the intermediate step
and allows for composability, such that only the CUC-CSP refinement has to be shown for the sequential subcomponents. *}

definition spec_impl_refinement_sf_no_conn :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) parallelFailure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_, _')=F _")
where
"spec_impl_refinement_sf_no_conn Spec ProcNames Pre CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x' . Pre x'}) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t_no_conn :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) parallelTraceState \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_, _')=T _")
where
"spec_impl_refinement_t_no_conn Spec ProcNames Pre CV Impl =
  (\<forall> tr s . ((tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>T{x' . Pre x'}) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV) :t [[Spec]]Tf [[ProcNames]]Tfix)
"

text {* Again, sequential variants to allow for easier reasoning. *}

definition (in SFsemSigma) spec_impl_refinement_sf_seq_no_conn ::
 "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) stable_failures.failure \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_sf_seq_no_conn Spec ProcNames Pre CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> (sfsem Impl {x' . Pre x'})) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"
                                                     
definition spec_impl_refinement_t_seq_no_conn ::
 "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> (('s, 'e) trace_state \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_t_seq_no_conn Spec ProcNames Pre CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> (densem Impl {x' . Pre x'})) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV) :t [[Spec]]Tf [[ProcNames]]Tfix))
"

paragraph {* Refinement without Conn, predefined Pre *}

text {* The connecting property is left out and the start condition is fixed to the minimal starting condition. *}

definition spec_impl_refinement_sf_no_conn_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_')=F _")
where
"spec_impl_refinement_sf_no_conn_pre_set Spec ProcNames CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x' . PreStd\<^sub>S\<^sub>F x'}) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"

definition spec_impl_refinement_t_no_conn_pre_set :: "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool"
("_ _ <='(_')=T _")
where
"spec_impl_refinement_t_no_conn_pre_set Spec ProcNames CV Impl =
  (\<forall> tr s . ((tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>T{x' . PreStd\<^sub>T x'}) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV) :t [[Spec]]Tf [[ProcNames]]Tfix)
"

text {* Again, for the sequential case. *}
                                                     
definition (in SFsemSigma) spec_impl_refinement_sf_seq_no_conn_pre_set ::
 "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_sf_seq_no_conn_pre_set Spec ProcNames CV Impl =
  ((\<forall> tr s X . ((tr, s, X) \<in> (sfsem Impl {x' . PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q x'})) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[Spec]]Ff [[ProcNames]]Ffix))))
"
                                                     
definition spec_impl_refinement_t_seq_no_conn_pre_set ::
 "('p, 'a) proc \<Rightarrow> ('p, 'a) pnfun \<Rightarrow> ('e \<Rightarrow> 'a) \<Rightarrow> ('s, 'e) structured_code \<Rightarrow> bool"
where
"spec_impl_refinement_t_seq_no_conn_pre_set Spec ProcNames CV Impl =
  ((\<forall> tr s . ((tr, s) \<in> (densem Impl {x' . PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q x'})) \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV) :t [[Spec]]Tf [[ProcNames]]Tfix))
"

paragraph {* Checks: Implicit precondition can be stated explicitly (no Conn) *}

lemma spec_impl_refinement_sf_no_conn_pre_set_alternative:
  "(Spec ProcNames <=(\<lambda>x. PreStd\<^sub>S\<^sub>F x, CV)=F Impl) = (Spec ProcNames <=(CV)=F Impl)"
  by (simp add: spec_impl_refinement_sf_no_conn_def spec_impl_refinement_sf_no_conn_pre_set_def)

lemma spec_impl_refinement_sf_no_conn_pre_set_alternative':
  "(Spec ProcNames <=(\<lambda>(tr, s, X). PreStd\<^sub>S\<^sub>F (tr, s, X), CV)=F Impl) = (Spec ProcNames <=(CV)=F Impl)"
  by (simp add: spec_impl_refinement_sf_no_conn_def spec_impl_refinement_sf_no_conn_pre_set_def)

lemma spec_impl_refinement_t_no_conn_pre_set_alternative:
  "(Spec ProcNames <=(\<lambda>x. PreStd\<^sub>T x, CV)=T Impl) = (Spec ProcNames <=(CV)=T Impl)"
  by (simp add: spec_impl_refinement_t_no_conn_def spec_impl_refinement_t_no_conn_pre_set_def)

lemma spec_impl_refinement_t_no_conn_pre_set_alternative':
  "(Spec ProcNames <=(\<lambda>(tr, s). PreStd\<^sub>T (tr, s), CV)=T Impl) = (Spec ProcNames <=(CV)=T Impl)"
  by (simp add: spec_impl_refinement_t_no_conn_def spec_impl_refinement_t_no_conn_pre_set_def)

lemma (in SFsemSigma) spec_impl_refinement_sf_seq_no_conn_pre_set_alternative:
  "(spec_impl_refinement_sf_seq_no_conn Spec ProcNames (\<lambda>x. PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q x) CV Impl) = (spec_impl_refinement_sf_seq_no_conn_pre_set Spec ProcNames CV Impl)"
  by (simp add: spec_impl_refinement_sf_seq_no_conn_def spec_impl_refinement_sf_seq_no_conn_pre_set_def)

lemma (in SFsemSigma) spec_impl_refinement_sf_seq_no_conn_pre_set_alternative':
  "(spec_impl_refinement_sf_seq_no_conn Spec ProcNames (\<lambda>(tr, s, X). PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q (tr, s, X)) CV Impl) = (spec_impl_refinement_sf_seq_no_conn_pre_set Spec ProcNames CV Impl)"
  by (simp add: spec_impl_refinement_sf_seq_no_conn_def spec_impl_refinement_sf_seq_no_conn_pre_set_def)

lemma spec_impl_refinement_t_seq_no_conn_pre_set_alternative:
  "(spec_impl_refinement_t_seq_no_conn Spec ProcNames (\<lambda>x. PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q x) CV Impl) = (spec_impl_refinement_t_seq_no_conn_pre_set Spec ProcNames CV Impl)"
  by (simp add: spec_impl_refinement_t_seq_no_conn_def spec_impl_refinement_t_seq_no_conn_pre_set_def)

lemma spec_impl_refinement_t_seq_no_conn_pre_set_alternative':
  "(spec_impl_refinement_t_seq_no_conn Spec ProcNames (\<lambda>(tr, s). PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q (tr, s)) CV Impl) = (spec_impl_refinement_t_seq_no_conn_pre_set Spec ProcNames CV Impl)"
  by (simp add: spec_impl_refinement_t_seq_no_conn_def spec_impl_refinement_t_seq_no_conn_pre_set_def)

paragraph {* Checks: Implicit precondition can be stated explicitly (with Conn) *}

lemma spec_impl_refinement_sf_pre_set_alternative:
  "(Spec ProcNames <=(\<lambda>x. PreStd\<^sub>S\<^sub>F x, Conn, CV)=F Impl) = (Spec ProcNames <=(, Conn, CV)=F Impl)"
  by (simp add: spec_impl_refinement_sf_def spec_impl_refinement_sf_pre_set_def)

lemma spec_impl_refinement_sf_pre_set_alternative':
  "(Spec ProcNames <=(\<lambda>(tr, s, X). PreStd\<^sub>S\<^sub>F (tr, s, X), Conn, CV)=F Impl) = (Spec ProcNames <=(, Conn, CV)=F Impl)"
  by (simp add: spec_impl_refinement_sf_def spec_impl_refinement_sf_pre_set_def)

lemma spec_impl_refinement_t_pre_set_alternative:
  "(Spec ProcNames <=(\<lambda>x. PreStd\<^sub>T x, Conn, CV)=T Impl) = (Spec ProcNames <=(, Conn, CV)=T Impl)"
  by (simp add: spec_impl_refinement_t_def spec_impl_refinement_t_pre_set_def)

lemma spec_impl_refinement_t_pre_set_alternative':
  "(Spec ProcNames <=(\<lambda>(tr, s). PreStd\<^sub>T (tr, s), Conn, CV)=T Impl) = (Spec ProcNames <=(, Conn, CV)=T Impl)"
  by (simp add: spec_impl_refinement_t_def spec_impl_refinement_t_pre_set_def)

lemma (in SFsemSigma) spec_impl_refinement_sf_seq_pre_set_alternative:
  "(spec_impl_refinement_sf_seq Spec ProcNames (\<lambda>x. PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q x) Conn CV Impl) = (spec_impl_refinement_sf_seq_pre_set Spec ProcNames Conn CV Impl)"
  by (simp add: spec_impl_refinement_sf_seq_def spec_impl_refinement_sf_seq_pre_set_def)

lemma (in SFsemSigma) spec_impl_refinement_sf_seq_pre_set_alternative':
  "(spec_impl_refinement_sf_seq Spec ProcNames (\<lambda>(tr, s, X). PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q (tr, s, X)) Conn CV Impl) = (spec_impl_refinement_sf_seq_pre_set Spec ProcNames Conn CV Impl)"
  by (simp add: spec_impl_refinement_sf_seq_def spec_impl_refinement_sf_seq_pre_set_def)

lemma spec_impl_refinement_t_seq_pre_set_alternative:
  "(spec_impl_refinement_t_seq Spec ProcNames (\<lambda>x. PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q x) Conn CV Impl) = (spec_impl_refinement_t_seq_pre_set Spec ProcNames Conn CV Impl)"
  by (simp add: spec_impl_refinement_t_seq_def spec_impl_refinement_t_seq_pre_set_def)

lemma spec_impl_refinement_t_seq_pre_set_alternative':
  "(spec_impl_refinement_t_seq Spec ProcNames (\<lambda>(tr, s). PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q (tr, s)) Conn CV Impl) = (spec_impl_refinement_t_seq_pre_set Spec ProcNames Conn CV Impl)"
  by (simp add: spec_impl_refinement_t_seq_def spec_impl_refinement_t_seq_pre_set_def)

paragraph {* Refinement with @{term Conn} implies refinement without @{term Conn} *}
  
lemma spec_impl_refinement_sf_impl_no_conn:
  assumes "Spec ProcNames <=(Pre, Conn, CV)=F Impl"
  shows "Spec ProcNames <=(Pre, CV)=F Impl"
  using assms
  unfolding spec_impl_refinement_sf_no_conn_def spec_impl_refinement_sf_def
by (simp add: CUCtoCSPStableFailure_def)

lemma spec_impl_refinement_t_impl_no_conn:
  assumes "Spec ProcNames <=(Pre, Conn, CV)=T Impl"
  shows "Spec ProcNames <=(Pre, CV)=T Impl"
  using assms
  unfolding spec_impl_refinement_t_no_conn_def spec_impl_refinement_t_def
by (simp add: CUCtoCSPStableFailure_def)

lemma spec_impl_refinement_sf_seq_impl_no_conn:
  fixes lis :: "('s, 'e)labeled_instruction_set" and \<Sigma> :: "'e set" and Pre :: "('s, 'e) stable_failures.failure \<Rightarrow> bool"
  assumes "SFsemSigma.spec_impl_refinement_sf_seq \<Sigma> Spec ProcNames Pre Conn CV Impl" and pre: "SFsemSigma lis"
  shows "SFsemSigma.spec_impl_refinement_sf_seq_no_conn \<Sigma> Spec ProcNames Pre CV Impl"
  using assms
  unfolding 
     SFsemSigma.spec_impl_refinement_sf_seq_def[OF pre]
     SFsemSigma.spec_impl_refinement_sf_seq_no_conn_def[OF pre]
  by blast

lemma spec_impl_refinement_t_seq_impl_no_conn:
  fixes lis :: "('s, 'e)labeled_instruction_set" and Pre :: "('s, 'e) trace_state \<Rightarrow> bool"
  assumes "spec_impl_refinement_t_seq Spec ProcNames Pre Conn CV Impl"
  shows "spec_impl_refinement_t_seq_no_conn Spec ProcNames Pre CV Impl"
  using assms
  unfolding 
     spec_impl_refinement_t_seq_def
     spec_impl_refinement_t_seq_no_conn_def
  by blast

lemma spec_impl_refinement_sf_impl_no_conn_pre_set:
  assumes "Spec ProcNames <=(, Conn, CV)=F Impl"
  shows "Spec ProcNames <=(CV)=F Impl"
  using assms
  unfolding spec_impl_refinement_sf_pre_set_def spec_impl_refinement_sf_no_conn_pre_set_def
by (simp add: CUCtoCSPStableFailure_def)

lemma spec_impl_refinement_t_impl_no_conn_pre_set:
  assumes "Spec ProcNames <=(, Conn, CV)=T Impl"
  shows "Spec ProcNames <=(CV)=T Impl"
  using assms
  unfolding spec_impl_refinement_t_pre_set_def spec_impl_refinement_t_no_conn_pre_set_def
by (simp add: CUCtoCSPStableFailure_def)

lemma spec_impl_refinement_sf_seq_impl_no_conn_pre_set:
  fixes
    lis :: "('s, 'e)labeled_instruction_set" and
    \<Sigma> :: "'e set" and
    Pre :: "('s, 'e) stable_failures.failure \<Rightarrow> bool" and
    Impl :: "('s, 'e) structured_code"
  assumes
    a:"SFsemSigma.spec_impl_refinement_sf_seq_pre_set \<Sigma> Spec ProcNames Conn CV Impl" and
    b:"SFsemSigma lis"
  shows "SFsemSigma.spec_impl_refinement_sf_seq_no_conn_pre_set \<Sigma> Spec ProcNames CV Impl"
unfolding "SFsemSigma.spec_impl_refinement_sf_seq_no_conn_pre_set_def"[OF b, of \<Sigma> Spec ProcNames CV Impl]
using a[unfolded "SFsemSigma.spec_impl_refinement_sf_seq_pre_set_def"[OF b]]
by blast

lemma spec_impl_refinement_t_seq_impl_no_conn_pre_set:
  fixes
    Pre :: "('s, 'e) trace_state \<Rightarrow> bool" and
    Impl :: "('s, 'e) structured_code"
  assumes
    a:"spec_impl_refinement_t_seq_pre_set Spec ProcNames Conn CV Impl"
  shows "spec_impl_refinement_t_seq_no_conn_pre_set Spec ProcNames CV Impl"
unfolding "spec_impl_refinement_t_seq_no_conn_pre_set_def"[of Spec ProcNames CV Impl]
using a[unfolded "spec_impl_refinement_t_seq_pre_set_def"]
by blast

paragraph {* Checks: The sequentiel helper refinements imply their parallel counterparts *}

text {* Unfolding the short definition into a longer one. *}

lemma seq_refinement_imp_par_refinement:
  assumes
    a:"SFsemSigma.spec_impl_refinement_sf_seq \<Sigma> Spec ProcNames Pre Conn CV Impl" and
    b:"SFsemSigma (\<union>\<^sub>s\<^sub>c Impl)"
  shows "Spec ProcNames <=(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre, Conn, CV)=F (seqn (Impl, \<Sigma>))"
proof -
  show ?thesis
  unfolding spec_impl_refinement_sf_def
    unfolding seq_sf_refinement_set_eq[symmetric]
    using a[unfolded SFsemSigma.spec_impl_refinement_sf_seq_def[OF b]]
    unfolding CUCtoCSPStableFailure_def snd_conv fst_conv
    by fastforce
qed

lemma seq_refinement_imp_par_refinement_no_conn:
  assumes
    a:"SFsemSigma.spec_impl_refinement_sf_seq_no_conn \<Sigma> Spec ProcNames Pre CV Impl" and
    b:"SFsemSigma (\<union>\<^sub>s\<^sub>c Impl)"
  shows "Spec ProcNames <=(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre, CV)=F (seqn (Impl, \<Sigma>))"
proof -
  show ?thesis
    unfolding spec_impl_refinement_sf_no_conn_def
    unfolding seq_sf_refinement_set_eq[symmetric]
    using a[unfolded SFsemSigma.spec_impl_refinement_sf_seq_no_conn_def[OF b]]
    unfolding CUCtoCSPStableFailure_def
    by fastforce
qed

lemma seq_refinement_imp_par_refinement_no_conn_pre_set:
  assumes
    a:"SFsemSigma.spec_impl_refinement_sf_seq_no_conn_pre_set \<Sigma> Spec ProcNames CV Impl" and
    wff: "wff\<^sub>p (seqn (Impl, \<Sigma>))"
  shows "Spec ProcNames <=(CV)=F (seqn (Impl, \<Sigma>))"
proof -
  have ax: "SFsemSigma (\<union>\<^sub>s\<^sub>c Impl)" using wff unfolding SFsemSigma_def wff\<^sub>p_def leafn1_property.simps by simp
  show ?thesis
    unfolding spec_impl_refinement_sf_no_conn_pre_set_def
    using seq_sf_refinement_set_eq_pre_set[OF wff]
    using a[unfolded SFsemSigma.spec_impl_refinement_sf_seq_no_conn_pre_set_def[OF ax]]
    by blast
qed

lemma seq_refinement_imp_par_refinement_pre_set:
  assumes
    a:"SFsemSigma.spec_impl_refinement_sf_seq_pre_set \<Sigma> Spec ProcNames Conn CV Impl" and
    wff: "wff\<^sub>p (seqn (Impl, \<Sigma>))"
  shows "Spec ProcNames <=(, Conn, CV)=F (seqn (Impl, \<Sigma>))"
unfolding spec_impl_refinement_sf_pre_set_def
  using a[unfolded "SFsemSigma.spec_impl_refinement_sf_seq_pre_set_def"[OF wff[unfolded wff\<^sub>p_def leafn1_property.simps SFsemSigma_def[symmetric], simplified, THEN cj2_2]]]
  unfolding CUCtoCSPStableFailure_def snd_conv fst_conv sfsemp_seq failureSetToParallelFailureSet_def PreStd_conv
  by fast

subsection {* CUC-CSP Refinement is composable. *}

lemma refinement_parallel_composition_without_conn:
  fixes Impl1 Impl2 :: "('s, 'e) structured_code_parallel"
  assumes
    inImpl: "(tr, s, X) \<in> \<lbrakk>(parn Impl1 (\<alpha>, \<beta>) Impl2)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x' . Pre x'}" and
    a: "Spec1 ProcNames <=(Pre1, CV)=F Impl1" and
    b: "Spec2 ProcNames <=(Pre2, CV)=F Impl2" and
    c: "\<forall> tr s X . treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s, X) \<longrightarrow> (Pre1 (tr, ATree_l s, X) \<and> Pre2 (tr, ATree_r s, X))" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2)]]Ff [[ProcNames]]Ffix))"
using a b c inImpl
unfolding spec_impl_refinement_sf_no_conn_def
unfolding unfoldFtofailues
unfolding SteveSchneiderAlphaParFailuresEqualToCSPVersion2[symmetric]
unfolding sfsemp_par CUCtoCSPStableFailure_def CUCtoCSPrefusal_def
proof (clarsimp)
  fix X1 X2 X' t1 t2 s1 s2
  assume
    aa1: "\<forall>tr s X. (tr, s, X) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre1 x'} \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, {x. \<exists>a. a \<in> X \<and> x = Ev (CV a)}) :f failures Spec1 [[ProcNames]]Ffix" and
    aa2: "\<forall>tr s X. (tr, s, X) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. Pre2 x'} \<longrightarrow> (CV\<^sub>t\<^sub>r tr CV, {x. \<exists>a. a \<in> X \<and> x = Ev (CV a)}) :f failures Spec2 [[ProcNames]]Ffix" and
    aa3: "\<forall>tr s. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> (\<forall>X. Pre (tr, s, X) \<longrightarrow> (Pre1 (tr, ATree_l s, X) \<and> Pre2 (tr, ATree_r s, X)))" and
    Pre: "Pre ([], par s1 s2, X')" and
    aa6: "N\<^sub>p s1" and
    aa7: "N\<^sub>p s2" and
    aa8: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    aa9: "(tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
    aa10: "(tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" and
    aa11: "set tr \<subseteq> \<alpha> \<union> \<beta>"
  have tm1: "treen_mirror s1 Impl1" using aa9 sfsemp_predecessor wff1 by fastforce
  have tm2: "treen_mirror s2 Impl2" using aa10 sfsemp_predecessor wff2 by fastforce
  have tm: "treen_mirror (par s1 s2) (parn Impl1 (\<alpha>, \<beta>) Impl2)" using tm1 tm2 by simp
  have aa4: "Pre1 ([], s1, X')" using tm aa3 Pre by fastforce
  have aa5: "Pre2 ([], s2, X')" using tm aa3 Pre by fastforce
  let ?G = "(CV\<^sub>t\<^sub>r tr CV, {x. \<exists>x'. x' \<in> X \<and> x = Ev (CV x')})
       :f {(tr, X).
           \<exists>X\<^sub>1 X\<^sub>2. X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                     (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and> (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}f"
  have G1: "CV\<^sub>X X CV \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = CV\<^sub>X X1 CV \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> CV\<^sub>X X2 CV \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
    unfolding CUCtoCSPrefusal_def tickify_failureset_def
    using aa8 inj
  proof (auto)
    show "\<And>x' xb. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> x' \<in> X \<Longrightarrow> CV x' = CV xb \<Longrightarrow> xb \<in> \<alpha> \<Longrightarrow> \<forall>x'. x' \<in> X2 \<longrightarrow> CV xb \<noteq> CV x' \<Longrightarrow> \<exists>x'. x' \<in> X1 \<and> CV xb = CV x'"
    proof -
      fix x' :: 'e and xb :: 'e
      assume a1: "x' \<in> X"
      assume a2: "xb \<in> \<alpha>"
      assume a3: "inj CV"
      assume a4: "CV x' = CV xb"
      assume a5: "\<forall>x'. x' \<in> X2 \<longrightarrow> CV xb \<noteq> CV x'"
      assume a6: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>"
      have "xb = x'"
        using a4 a3 by (simp add: inj_on_def)
      then show "\<exists>e. e \<in> X1 \<and> CV xb = CV e"
        using a6 a5 a2 a1 by blast
    qed
  next
    show "\<And>x' xb. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> x' \<in> X \<Longrightarrow> CV x' = CV xb \<Longrightarrow> xb \<in> \<alpha> \<Longrightarrow> Ev (CV xb) \<notin> Ev ` CV ` \<beta> \<Longrightarrow> \<exists>x'. x' \<in> X1 \<and> CV xb = CV x'"
      by (metis Int_iff Un_iff aa8 event_in_channel_map_inj image_eqI)
  next
    show "\<And>x' xb. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> x' \<in> X \<Longrightarrow> CV x' = CV xb \<Longrightarrow> xb \<in> \<beta> \<Longrightarrow> \<forall>x'. x' \<in> X2 \<longrightarrow> CV xb \<noteq> CV x' \<Longrightarrow> \<exists>x'. x' \<in> X1 \<and> CV xb = CV x'"
      by (metis Int_iff Un_iff aa8 event_in_channel_map_inj)
  next
    show "\<And>x' xb. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> x' \<in> X \<Longrightarrow> CV x' = CV xb \<Longrightarrow> xb \<in> \<beta> \<Longrightarrow> \<forall>x'. x' \<in> X2 \<longrightarrow> CV xb \<noteq> CV x' \<Longrightarrow> Ev (CV xb) \<in> Ev ` CV ` \<alpha>"
      by (metis Int_iff Un_iff aa8 event_in_channel_map_inj image_eqI)
  next
    show "\<And>xb x'. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> xb \<in> \<alpha> \<Longrightarrow> x' \<in> X1 \<Longrightarrow> CV xb = CV x' \<Longrightarrow> \<exists>x'a. x'a \<in> X \<and> CV x' = CV x'a"
      by (metis Int_iff UNIV_I Un_iff inj_on_def)
  next
    show "\<And>xb x'. X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<Longrightarrow> inj CV \<Longrightarrow> xb \<in> \<beta> \<Longrightarrow> x' \<in> X2 \<Longrightarrow> CV xb = CV x' \<Longrightarrow> \<exists>x'a. x'a \<in> X \<and> CV x' = CV x'a"
      by (metis Int_iff UNIV_I Un_iff inj_on_def)
  qed
  have hlpr: "\<And> X . CV\<^sub>X X CV = {x. \<exists>a. a \<in> X \<and> x = Ev (CV a)}"
    using CUCtoCSPrefusal_def by auto
  have trRestrA: "CV\<^sub>t\<^sub>r (tr \<restriction> \<alpha>) CV = CV\<^sub>t\<^sub>r tr CV rest-tr (CV ` \<alpha>)"
    by (simp add: rest_tr_cv_distr[OF inj])
  have trRestrB: "CV\<^sub>t\<^sub>r (tr \<restriction> \<beta>) CV = CV\<^sub>t\<^sub>r tr CV rest-tr (CV ` \<beta>)"
    by (simp add: rest_tr_cv_distr[OF inj])
  have G2: "(CV\<^sub>t\<^sub>r tr \<restriction> \<alpha> CV, CV\<^sub>X X1 CV) :f failures Spec1 [[ProcNames]]Ffix"
    unfolding hlpr
  proof -
    have "([], s1, X') \<in> Collect Pre1"
      using aa4 by blast
    then show "(CV\<^sub>t\<^sub>r tr \<restriction> \<alpha> CV, {x. \<exists>a. a \<in> X1 \<and> x = Ev (CV a)}) :f failures Spec1 [[ProcNames]]Ffix"
      using aa1 aa9 sfsemp_monotone_single wff1 by blast
  qed
  have G3: "(CV\<^sub>t\<^sub>r tr \<restriction> \<beta> CV, CV\<^sub>X X2 CV) :f failures Spec2 [[ProcNames]]Ffix"
    unfolding hlpr
  proof -
    have "([], s2, X') \<in> Collect Pre2"
      using aa5 by blast
    then show "(CV\<^sub>t\<^sub>r tr \<restriction> \<beta> CV, {x. \<exists>a. a \<in> X2 \<and> x = Ev (CV a)}) :f failures Spec2 [[ProcNames]]Ffix"
      using aa2 aa10 sfsemp_monotone_single wff2 by blast
  qed
  have G4: "sett (CV\<^sub>t\<^sub>r tr CV) \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
    using aa11 "CUCtoCSPtrace.simps"
    unfolding tickify_failureset_def
  proof (induct tr, auto) qed
  have "(CV\<^sub>t\<^sub>r tr CV, {x. \<exists>a. a \<in> X \<and> x = Ev (CV a)})
       : {(tr, X).
           \<exists>X\<^sub>1 X\<^sub>2. X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                     (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and> (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}"
    using G1 G2 G3 G4 inj
    unfolding hlpr[symmetric] rest_tr_cv_distr
    using case_prodI mem_Collect_eq trRestrA trRestrB by auto
  moreover have "HC_F2 {(tr, X).
           \<exists>X\<^sub>1 X\<^sub>2. X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                     (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and> (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}"
    unfolding HC_F2_def
  proof (clarsimp)
    fix s X Y X\<^sub>1 X\<^sub>2
    assume
      ba1: "Y \<subseteq> X" and
      ba2: "X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k" and
      ba3: "(s rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix" and
      ba4: "(s rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix" and
      ba5: "sett s \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
    obtain Xr Xs where 
        bo3: "Xr = X\<^sub>1 - (X - Y)" and
        bo4: "Xs = X\<^sub>2 - (X - Y)"
        by blast
      have b1: "Y \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = Xr \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> Xs \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
        using ba1 ba2 bo3 bo4
        by auto
      have b4: "(s rest-tr CV ` \<alpha>, Xr) :f failures Spec1 [[ProcNames]]Ffix"
        using bo3 ba3 memF_F2
        by blast
      have b5: "(s rest-tr CV ` \<beta>, Xs) :f failures Spec2 [[ProcNames]]Ffix"
        using bo4 ba4 memF_F2
        by blast
    show "\<exists>X\<^sub>1 X\<^sub>2. Y \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and> (s rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and> (s rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix"
      using b1 b4 b5 by blast
  qed
  ultimately show ?G
    by (simp add: CollectF_open_memF setF_def)
qed

lemma refinement_parallel_composition_without_conn_Pre_set:
  assumes
    inImpl: "(tr, s, X) \<in> \<lbrakk>(parn Impl1 (\<alpha>, \<beta>) Impl2)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x . PreStd\<^sub>S\<^sub>F x}" and
    a: "Spec1 ProcNames <=(CV)=F Impl1" and
    b: "Spec2 ProcNames <=(CV)=F Impl2" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f (sndF ([[(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2)]]Ff [[ProcNames]]Ffix))"
proof -
  have G4: "\<forall>tr s X. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow>
             (case (tr, s, X) of (tr, s, X) \<Rightarrow> PreStd\<^sub>S\<^sub>F (tr, s, X)) \<longrightarrow> (PreStd\<^sub>S\<^sub>F (tr, ATree_l s, X) \<and> PreStd\<^sub>S\<^sub>F (tr, ATree_r s, X))"
    unfolding PreStd\<^sub>S\<^sub>F_def
  proof (clarsimp)
    fix s tr
    assume
      tm: "treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2)" and
      pcs: "AllPC\<^sub>S\<^sub>F s (Suc 0)" and N: "N\<^sub>p s"
    obtain s1 s2 where s: "s = par s1 s2"
      by (meson tm treen_mirror_unfold_parr)
    show "AllPC\<^sub>S\<^sub>F (ATree_l s) (Suc 0) \<and> N\<^sub>p (ATree_l s) \<and> AllPC\<^sub>S\<^sub>F (ATree_r s) (Suc 0) \<and> N\<^sub>p (ATree_r s)"
      using pcs N
      unfolding s ATree_l_def ATree_r_def AllPC\<^sub>S\<^sub>F_def "treen_mirror" leaf1_property.simps N\<^sub>p_def
      by auto
  qed
  have G1: "(tr, s, X) \<in> \<lbrakk>parn Impl1 (\<alpha>, \<beta>) Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. case x' of (tr, s, X) \<Rightarrow> PreStd\<^sub>S\<^sub>F (tr, s, X)}"
    using inImpl by auto
  have G2: "Spec1 ProcNames <=(PreStd\<^sub>S\<^sub>F, CV)=F Impl1"
    using a spec_impl_refinement_sf_no_conn_pre_set_alternative' by auto
  have G3: "Spec2 ProcNames <=(PreStd\<^sub>S\<^sub>F, CV)=F Impl2"
    using b spec_impl_refinement_sf_no_conn_pre_set_alternative' by auto
  from refinement_parallel_composition_without_conn[OF G1 G2 G3 G4 wff1 wff2 inj] inImpl PreStd_prod
    show ?thesis 
    unfolding PreStd\<^sub>S\<^sub>F_def by fast
qed

text {* We only have to show the stable failure refinement for the sequential parts of the program.
If we take two components in CUC which refine two other components in CSP, we can combine them via the
parallel composition, and the result is still a stable failure refinement.
*}

theorem refinement_parallel_composition_without_conn':
  assumes
    a: "Spec1 ProcNames <=(Pre1, CV)=F Impl1" and
    b: "Spec2 ProcNames <=(Pre2, CV)=F Impl2" and
    c: "\<forall> tr s X . treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s, X) \<longrightarrow> (Pre1 (tr, ATree_l s, X) \<and> Pre2 (tr, ATree_r s, X))" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2) ProcNames <=(Pre, CV)=F (parn Impl1 (\<alpha>, \<beta>) Impl2)"
  unfolding spec_impl_refinement_sf_no_conn_def
  using refinement_parallel_composition_without_conn[OF _ a b c wff1 wff2 inj]
  unfolding sfsemp_starting_set_can_contain_malformed_states
by blast
    
theorem refinement_parallel_composition_without_conn_Pre_set':
  assumes
    a: "Spec1 ProcNames <=(CV)=F Impl1" and
    b: "Spec2 ProcNames <=(CV)=F Impl2" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2) ProcNames <=(CV)=F (parn Impl1 (\<alpha>, \<beta>) Impl2)"
unfolding spec_impl_refinement_sf_no_conn_pre_set_def
using refinement_parallel_composition_without_conn_Pre_set[OF _ a b wff1 wff2 inj]
by blast

text {* We can make the same also with Pre and Conn, which becomes more complex.
But this lemma is superceded by the previous one, as we do not need to know a connecting property. *}

lemma refinement_parallel_composition:
  assumes
    a: "Spec1 ProcNames <=(Pre1, Conn1, CV)=F Impl1" and
    b: "Spec2 ProcNames <=(Pre2, Conn2, CV)=F Impl2" and
    c: "\<forall> tr s X . treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s, X) \<longrightarrow> (Pre1 (tr, ATree_l s, X) \<and> Pre2 (tr, ATree_r s, X))" and
    d: "Conn = (\<lambda> (tr, X) . 
        \<exists> X1 X2 . X = X1 \<union> X2 \<and>
        (* Stuff like X \<inter> (\<alpha> - \<beta>) does not work, as in Conn\<rightarrow>Spec we cannot fulfill spec. *)
        (X1 \<union> X2) \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = (X1) \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> (X2) \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
        Conn1 (tr rest-tr CV ` \<alpha>, X1) \<and> Conn2 (tr rest-tr CV ` \<beta>, X2) \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k)" and
    e: "\<And> tr X X' . Conn1 (tr, X) \<and> (X' \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<subseteq> (X \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<longrightarrow> Conn1 (tr, X')" and
    f: "\<And> tr X X' . Conn2 (tr, X) \<and> (X' \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<subseteq> (X \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<longrightarrow> Conn2 (tr, X')" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(Spec1 |[CV ` \<alpha>,CV ` \<beta>]| Spec2) ProcNames <=(Pre, Conn, CV)=F (parn Impl1 (\<alpha>, \<beta>) Impl2)"
using a b c d 
unfolding spec_impl_refinement_sf_def CSP_F_domain.sndF_semFf
unfolding SteveSchneiderAlphaParFailuresEqualToCSPVersion2[symmetric]
unfolding sfsemp_par CUCtoCSPStableFailure_def CUCtoCSPrefusal_def
proof (clarsimp)
  assume
    Pre_imp: "\<forall>tr s. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> (\<forall>X. Pre (tr, s, X) \<longrightarrow> Pre1 (tr, ATree_l s, X) \<and> Pre2 (tr, ATree_r s, X))" and
    aa1: "\<forall>tr s X. (tr, s, X) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>S\<^sub>FCollect Pre1 \<longrightarrow> Conn1 (CV\<^sub>t\<^sub>r tr CV, {x. \<exists>x'. x' \<in> X \<and> x = Ev (CV x')})" and
    aa2: "\<forall>tr X. Conn1 (tr, X) \<longrightarrow> (tr, X) :f failures Spec1 [[ProcNames]]Ffix" and
    aa3: "\<forall>tr s X. (tr, s, X) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>FCollect Pre2 \<longrightarrow> Conn2 (CV\<^sub>t\<^sub>r tr CV, {x. \<exists>x'. x' \<in> X \<and> x = Ev (CV x')})" and
    aa4: "\<forall>tr X. Conn2 (tr, X) \<longrightarrow> (tr, X) :f failures Spec2 [[ProcNames]]Ffix"
  have "\<forall>tr X. (\<exists>X1 X2. X = X1 \<union> X2 \<and>
                     (X1 \<union> X2) \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                     Conn1 (tr rest-tr CV ` \<alpha>, X1) \<and> Conn2 (tr rest-tr CV ` \<beta>, X2) \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k) \<longrightarrow>
            (tr, X)
            :f {(tr, X).
                \<exists>X\<^sub>1 X\<^sub>2. X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                          (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and>
                          (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}f" (is ?G2)
  proof (clarsimp)
    fix tr X1 X2
    assume
      ba4: "(X1 \<union> X2) \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k" and
      ba1: "Conn1 (tr rest-tr CV ` \<alpha>, X1)" and
      ba2: "Conn2 (tr rest-tr CV ` \<beta>, X2)" and
      ba3: "sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
    have b1: "(tr rest-tr CV ` \<alpha>, X1) :f failures Spec1 [[ProcNames]]Ffix"
      using aa2 ba1
      by blast
    moreover
    have b2: "(tr rest-tr CV ` \<beta>, X2) :f failures Spec2 [[ProcNames]]Ffix"
      using aa4 ba2
      by blast
    ultimately
    have "(tr, X1 \<union> X2) \<in> {(tr, X). \<exists>X\<^sub>1 X\<^sub>2.
                   X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                   (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and>
                   (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}"
      using ba4 ba3
      unfolding tickify_failureset_def by blast
    moreover
    have "HC_F2 {(tr, X). \<exists>X\<^sub>1 X\<^sub>2.
                   X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                   (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and>
                   (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}"
      unfolding HC_F2_def
    proof (clarsimp)
      fix s X Y X\<^sub>1 X\<^sub>2
      assume
        ca1: "Y \<subseteq> X" and
        ca2: "X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k" and
        ca3: "(s rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix" and
        ca4: "(s rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix" and
        ca5: "sett s \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      obtain Xr Xs where 
        co3: "Xr = X\<^sub>1 - (X - Y)" and
        co4: "Xs = X\<^sub>2 - (X - Y)"
        by blast
      have c1: "Y \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = Xr \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> Xs \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
        using ca1 ca2 co3 co4
        by auto
      have c4: "(s rest-tr CV ` \<alpha>, Xr) :f failures Spec1 [[ProcNames]]Ffix"
        using co3 ca3 memF_F2
        by blast
      have c5: "(s rest-tr CV ` \<beta>, Xs) :f failures Spec2 [[ProcNames]]Ffix"
        using co4 ca4 memF_F2
        by blast
      show "\<exists>X\<^sub>1 X\<^sub>2.
          Y \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
          (s rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and> (s rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix"
        using c1 c4 c5
        by blast
    qed
    ultimately
    show  "(tr, X1 \<union> X2) :f {(tr, X). \<exists>X\<^sub>1 X\<^sub>2.
                   X \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X\<^sub>1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X\<^sub>2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                   (tr rest-tr CV ` \<alpha>, X\<^sub>1) :f failures Spec1 [[ProcNames]]Ffix \<and>
                   (tr rest-tr CV ` \<beta>, X\<^sub>2) :f failures Spec2 [[ProcNames]]Ffix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}f"
      by (simp add: CollectF_open_memF setF_def) 
  qed
  moreover
  have "\<forall>tr s X. (\<exists>X1 X2 X' t1 t2 s1 s2.
                  Pre ([], par s1 s2, X') \<and>
                  N\<^sub>p s1 \<and>
                  N\<^sub>p s2 \<and>
                  X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and>
                  (tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and> (tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')} \<and> set tr \<subseteq> \<alpha> \<union> \<beta> \<and> s = par t1 t2) \<longrightarrow>
              (\<exists>X1 X2. {x. \<exists>x'. x' \<in> X \<and> x = Ev (CV x')} = X1 \<union> X2 \<and>
                       (X1 \<union> X2) \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
                       Conn1 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha>, X1) \<and> Conn2 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<beta>, X2) \<and> sett CV\<^sub>t\<^sub>r tr CV \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k)" (is ?G1)
  proof (clarsimp)
    fix tr X X1 X2 X' t1 t2 s1 s2
    assume
      Pre: "Pre ([], par s1 s2, X')" and
      ba3: "N\<^sub>p s1" and
      ba4: "N\<^sub>p s2" and
      ba5: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
      ba6: "(tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
      ba7: "(tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" and
      ba8: "set tr \<subseteq> \<alpha> \<union> \<beta>"
    have ba9: "treen_mirror s1 Impl1" using ba6 sfsemp_predecessor wff1 by fastforce
    have ba10: "treen_mirror s2 Impl2" using ba7 sfsemp_predecessor wff2 by fastforce
    have ba1: "Pre1 ([], s1, X')" using Pre Pre_imp ba9 ba10 treen_mirror_inherit by force
    have ba2: "Pre2 ([], s2, X')" using Pre Pre_imp ba9 ba10 treen_mirror_inherit by force
    have b0_1: "([], s1, X') \<in> {x'. Pre1 x'}"
      using ba9 ba1 by simp
    have b0_2: "([], s2, X') \<in> {x'. Pre2 x'}"
      using ba10 ba2 by simp
    have G1: "sett CV\<^sub>t\<^sub>r tr CV \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      using ba8
      unfolding tickify_failureset_def
    proof (induct tr, auto) qed
    have b1: "Conn1 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha>, {x. \<exists>a. a \<in> X1 \<and> x = Ev (CV a)})"
      using aa1 sfsemp_monotone_single[OF ba6 b0_1 wff1]
      unfolding rest_tr_cv_distr[OF inj, symmetric] by fastforce
    have b2: "Conn2 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<beta>, {x. \<exists>a. a \<in> X2 \<and> x = Ev (CV a)})"
      using aa3 sfsemp_monotone_single[OF ba7 b0_2 wff2]
      unfolding rest_tr_cv_distr[OF inj, symmetric] by fastforce
    obtain X1' where X1': "X1' = {x. \<exists>a. a \<in> ((X1 \<inter> \<alpha>) \<union> (X - (\<alpha> \<union> \<beta>))) \<and> (x = Ev (CV a))}"
      by blast
    obtain X2' where X2': "X2' = {x. \<exists>a. a \<in> ((X2 \<inter> \<beta>) \<union> (X - (\<alpha> \<union> \<beta>))) \<and> (x = Ev (CV a))}"
      by blast 
    have "X1' \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<subseteq> {x. \<exists>a. a \<in> X1 \<and> x = Ev (CV a)} \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      using X1' unfolding tickify_failureset_def
    proof (auto)
      show "\<And>xb a. X1' = {x. \<exists>a. (a \<in> X1 \<and> a \<in> \<alpha> \<or> a \<in> X \<and> a \<notin> \<alpha> \<and> a \<notin> \<beta>) \<and> x = Ev (CV a)} \<Longrightarrow>
            xb \<in> \<alpha> \<Longrightarrow> CV xb = CV a \<Longrightarrow> a \<in> X \<Longrightarrow> a \<notin> \<alpha> \<Longrightarrow> a \<notin> \<beta> \<Longrightarrow> \<exists>aa. aa \<in> X1 \<and> CV a = CV aa"
        by (metis (no_types, hide_lams) event_in_channel_map_inj inj)
    qed
    hence G4: "Conn1 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha>, X1')"
      using b1 e
      by blast 
    have "X2' \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<subseteq> {x. \<exists>a. a \<in> X2 \<and> x = Ev (CV a)} \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      using X2' unfolding tickify_failureset_def
    proof (auto)
      show "\<And>xb a. X2' = {x. \<exists>a. (a \<in> X2 \<and> a \<in> \<beta> \<or> a \<in> X \<and> a \<notin> \<alpha> \<and> a \<notin> \<beta>) \<and> x = Ev (CV a)} \<Longrightarrow>
            xb \<in> \<beta> \<Longrightarrow> CV xb = CV a \<Longrightarrow> a \<in> X \<Longrightarrow> a \<notin> \<alpha> \<Longrightarrow> a \<notin> \<beta> \<Longrightarrow> \<exists>aa. aa \<in> X2 \<and> CV a = CV aa"
        by (metis (no_types, hide_lams) event_in_channel_map_inj inj)
    qed
    hence G5: "Conn2 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<beta>, X2')"
      using b2 f
      by blast
    have G2: "{x. \<exists>a. a \<in> X \<and> x = Ev (CV a)} = X1' \<union> X2'"
      unfolding X1' X2' using ba5 by blast
    have G3: "(X1' \<union> X2') \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X1' \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X2' \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      unfolding tickify_failureset_def X1' X2'
      by auto
    show 
      "\<exists>X1 X2. {x. \<exists>a. a \<in> X \<and> x = Ev (CV a)} = X1 \<union> X2 \<and>
               (X1 \<union> X2) \<inter> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k = X1 \<inter> (CV ` \<alpha>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<union> X2 \<inter> (CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k \<and>
               Conn1 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<alpha>, X1) \<and> Conn2 (CV\<^sub>t\<^sub>r tr CV rest-tr CV ` \<beta>, X2) \<and> sett CV\<^sub>t\<^sub>r tr CV \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
      using G1 G2 G3 G4 G5
      unfolding tickify_failureset_def[symmetric]
      by blast
  qed
  ultimately
  show "?G1 \<and> ?G2"
    by fast
qed

lemma refinement_parallel_composition_without_conn_t:
  fixes Impl1 Impl2 :: "('s, 'e) structured_code_parallel"
  assumes
    inImpl: "(tr, s) \<in> \<lbrakk>(parn Impl1 (\<alpha>, \<beta>) Impl2)\<rbrakk>\<^sub>p\<^sub>T{x' . Pre x'}" and
    a: "Spec1 ProcNames <=(Pre1, CV)=T Impl1" and
    b: "Spec2 ProcNames <=(Pre2, CV)=T Impl2" and
    c: "\<forall> tr s . treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s) \<longrightarrow> (Pre1 (tr, ATree_l s) \<and> Pre2 (tr, ATree_r s))" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(CV\<^sub>t\<^sub>r tr CV) :t [[(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2)]]Tf [[ProcNames]]Tfix"
using a b c inImpl
unfolding spec_impl_refinement_t_no_conn_def
unfolding unfoldTtotraces               
unfolding SteveSchneiderAlphaParTracesEqualToCSPVersion2[symmetric]
unfolding densemp_par
proof (clarsimp)
  fix t1 t2 s1 s2
  assume
    aa1: "\<forall>tr. (\<exists>s. (tr, s) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>TCollect Pre1) \<longrightarrow> CV\<^sub>t\<^sub>r tr CV :t traces Spec1 [[ProcNames]]Tfix" and
    aa2: "\<forall>tr. (\<exists>s. (tr, s) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>TCollect Pre2) \<longrightarrow> CV\<^sub>t\<^sub>r tr CV :t traces Spec2 [[ProcNames]]Tfix" and
    aa3: "\<forall>tr s. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s) \<longrightarrow> Pre1 (tr, ATree_l s) \<and> Pre2 (tr, ATree_r s)" and
    Pre: "Pre ([], par s1 s2)" and
    aa9: "(tr \<restriction> \<alpha>, t1) \<in> \<lbrakk>Impl1\<rbrakk>\<^sub>p\<^sub>T{([], s1)}" and
    aa10: "(tr \<restriction> \<beta>, t2) \<in> \<lbrakk>Impl2\<rbrakk>\<^sub>p\<^sub>T{([], s2)}" and
    aa11: "set tr \<subseteq> \<alpha> \<union> \<beta>"
  have tm1: "treen_mirror s1 Impl1" using aa9 densemp_predecessor wff1 by fastforce
  have tm2: "treen_mirror s2 Impl2" using aa10 densemp_predecessor wff2 by fastforce
  have tm: "treen_mirror (par s1 s2) (parn Impl1 (\<alpha>, \<beta>) Impl2)" using tm1 tm2 by simp
  have aa4: "Pre1 ([], s1)" using tm aa3 Pre by fastforce
  have aa5: "Pre2 ([], s2)" using tm aa3 Pre by fastforce
  let ?G = "CV\<^sub>t\<^sub>r tr CV :t {tr. tr rest-tr CV ` \<alpha> :t traces Spec1 [[ProcNames]]Tfix \<and> tr rest-tr CV ` \<beta> :t traces Spec2 [[ProcNames]]Tfix \<and> sett tr \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k}t"
  have trRestrA: "CV\<^sub>t\<^sub>r (tr \<restriction> \<alpha>) CV = CV\<^sub>t\<^sub>r tr CV rest-tr (CV ` \<alpha>)"
    by (simp add: rest_tr_cv_distr[OF inj])
  have trRestrB: "CV\<^sub>t\<^sub>r (tr \<restriction> \<beta>) CV = CV\<^sub>t\<^sub>r tr CV rest-tr (CV ` \<beta>)"
    by (simp add: rest_tr_cv_distr[OF inj])
  have G2: "(CV\<^sub>t\<^sub>r tr \<restriction> \<alpha> CV) :t traces Spec1 [[ProcNames]]Tfix"
  proof -
    have "([], s1) \<in> Collect Pre1"
      using aa4 by blast
    then show "(CV\<^sub>t\<^sub>r tr \<restriction> \<alpha> CV) :t traces Spec1 [[ProcNames]]Tfix"
      using aa1 aa9 densemp_monotone_single wff1 by blast
  qed
  have G3: "(CV\<^sub>t\<^sub>r tr \<restriction> \<beta> CV) :t traces Spec2 [[ProcNames]]Tfix"
  proof -
    have "([], s2) \<in> Collect Pre2"
      using aa5 by blast
    then show "(CV\<^sub>t\<^sub>r tr \<restriction> \<beta> CV) :t traces Spec2 [[ProcNames]]Tfix"
      using aa2 aa10 densemp_monotone_single wff2 by blast
  qed
  have G4: "sett (CV\<^sub>t\<^sub>r tr CV) \<subseteq> (CV ` \<alpha> \<union> CV ` \<beta>)\<^sub>t\<^sub>i\<^sub>c\<^sub>k"
    using aa11 "CUCtoCSPtrace.simps"
    unfolding tickify_failureset_def
  proof (induct tr, auto) qed
  have "CV\<^sub>t\<^sub>r tr CV :t traces (Spec1 |[CV ` \<alpha>,CV ` \<beta>]| Spec2) [[ProcNames]]Tfix"
    by (metis (no_types) G2 G3 G4 in_traces_Alpha_parallel tickify_failureset_def trRestrA trRestrB)
  then show ?G
    by (simp add: tickify_failureset_def traces_Alpha_parallel)
qed

text {* We only have to show the stable failure refinement for the sequential parts of the program.
If we take two components in CUC which refine two other components in CSP, we can combine them via the
parallel composition, and the result is still a stable failure refinement.
*}

theorem refinement_parallel_composition_without_conn'_t:
  assumes
    a: "Spec1 ProcNames <=(Pre1, CV)=T Impl1" and
    b: "Spec2 ProcNames <=(Pre2, CV)=T Impl2" and
    c: "\<forall> tr s . treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> Pre (tr, s) \<longrightarrow> (Pre1 (tr, ATree_l s) \<and> Pre2 (tr, ATree_r s))" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2) ProcNames <=(Pre, CV)=T (parn Impl1 (\<alpha>, \<beta>) Impl2)"
  unfolding spec_impl_refinement_t_no_conn_def
  using refinement_parallel_composition_without_conn_t[OF _ a b c wff1 wff2 inj]
  unfolding densemp_starting_set_can_contain_malformed_states
by blast
    
theorem refinement_parallel_composition_without_conn_Pre_set'_t:
  assumes
    a: "Spec1 ProcNames <=(CV)=T Impl1" and
    b: "Spec2 ProcNames <=(CV)=T Impl2" and
    wff1: "wff\<^sub>p Impl1" and
    wff2: "wff\<^sub>p Impl2" and
    inj: "inj CV"
  shows
    "(Spec1 |[CV ` \<alpha>, CV ` \<beta>]| Spec2) ProcNames <=(CV)=T (parn Impl1 (\<alpha>, \<beta>) Impl2)"
unfolding spec_impl_refinement_sf_no_conn_pre_set_def
using refinement_parallel_composition_without_conn'_t[of Spec1 ProcNames "\<lambda>x. PreStd\<^sub>T x" CV Impl1 Spec2 "\<lambda>x. PreStd\<^sub>T x" Impl2 \<alpha> \<beta> "\<lambda>x. PreStd\<^sub>T x"]
unfolding spec_impl_refinement_t_no_conn_pre_set_alternative
proof -
  assume
    pre: "Spec1 ProcNames <=(CV)=T Impl1 \<Longrightarrow>
     Spec2 ProcNames <=(CV)=T Impl2 \<Longrightarrow>
     \<forall>tr s. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> PreStd\<^sub>T (tr, s) \<longrightarrow> PreStd\<^sub>T (tr, ATree_l s) \<and> PreStd\<^sub>T (tr, ATree_r s) \<Longrightarrow>
     wff\<^sub>p Impl1 \<Longrightarrow> wff\<^sub>p Impl2 \<Longrightarrow> inj CV \<Longrightarrow> Spec1 |[CV ` \<alpha>,CV ` \<beta>]| Spec2 ProcNames <=(CV)=T parn Impl1 (\<alpha>, \<beta>) Impl2"
  have cond: "\<forall>tr s. treen_mirror s (parn Impl1 (\<alpha>, \<beta>) Impl2) \<longrightarrow> PreStd\<^sub>T (tr, s) \<longrightarrow> PreStd\<^sub>T (tr, ATree_l s) \<and> PreStd\<^sub>T (tr, ATree_r s)"
    unfolding PreStd\<^sub>T_def using AllPC\<^sub>T_inherit
    by (metis ATree.collapse(2) ATree.discI(2) case_prodD case_prodI treen_mirror_unfold_parr)
  show "(Spec1 |[CV ` \<alpha>,CV ` \<beta>]| Spec2) ProcNames <=(CV)=T (parn Impl1 (\<alpha>, \<beta>) Impl2)"
    using pre[OF a b cond wff1 wff2 inj] .
qed

text{* If we have a SF refinement, and each starting instruction on the code has a successor,
then we have also a T refinement.
One thing to note: There is no real condition for the starting states: The existence of a successor
already implies the existence of a wellformed starting state (i.e. matching tree structure and empty
trace).
*}

theorem FailureRefinement_Implied_TracesRefinement_If_NoDivergency_with_pre_with_conn:
  fixes
    Impl :: "('s, 'e) structured_code_parallel"
  assumes
    --"We assume to already have an SF refinement between CUC and CSP."
    sf_refinement: "Spec ProcFun <=(PreF, ConnF, CV)=F Impl" and
    --"The same starting states for the CUC code have to be used in the SF and in the T part."
    PreF: "\<forall> tr s X . treen_mirror s Impl \<longrightarrow> (PreT (tr, (PS2PTS s)) \<and> N\<^sub>p s) \<longrightarrow> PreF (tr, s, X)" and
    --"Also the connecting property has to match."
    ConnF: "ConnF = (\<lambda> (tr, X) . ConnT tr)" and
    wffImpl: "wff\<^sub>p Impl" and
    --"For the sequential leafs, we always want to have successors."
    successors: "leafn1_property (\<lambda> (c, \<Sigma>) . \<forall> tr s X . SFsemSigma.sfsem \<Sigma> c {(tr, s, X)} \<noteq> {}) op \<and> Impl" and
    (* The Process name function has to be passed, and in CMSmode it has to be guarded. *)
    CSP_assumpt: "ProcFun = PNfun \<and> (FPmode = CMSmode \<longrightarrow> guardedfun ProcFun)"
  shows
    "Spec ProcFun <=(PreT, ConnT, CV)=T Impl"
unfolding spec_impl_refinement_t_def
proof(rule ccontr, auto)
  fix tr s
  assume
    aa1: "\<not> ConnT CV\<^sub>t\<^sub>r tr CV" and
    aa2: "(tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>TCollect PreT"
  have "\<exists> s' X . (tr, s', X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>FCollect PreF"
    using tr_of_T_will_be_in_SF_parallel[OF aa2 PreF wffImpl successors] by fast
  hence "\<exists> s' X .ConnF CV\<^sub>f (tr, s', X) CV"
    using sf_refinement[unfolded spec_impl_refinement_sf_def]
    by (simp add: CUCtoCSPStableFailure_def ConnF)
  thus False
    unfolding ConnF CUCtoCSPStableFailure_def
    using aa1
    by force
next
  fix tr
  assume
    aa1: "ConnT tr" and
    aa2: "tr ~:t [[Spec]]Tf [[ProcFun]]Tfix"
  have hlpr: "[[ProcFun]]Ffix = MF"
    by (simp add: CSP_assumpt MF_def)
  have hlpr2: "[[ProcFun]]Tfix = MT"
    by (simp add: CSP_assumpt MT_def)
  from sf_refinement[unfolded spec_impl_refinement_sf_def ConnF] aa1
    have "\<forall> X . (tr, X) :f sndF ([[Spec]]Ff [[ProcFun]]Ffix)"
    by blast
  hence tr_in_fstF: "tr :t fstF ([[Spec]]Ff [[ProcFun]]Ffix)"
    using pairF_domF_T2 by blast
  thus "False"
    unfolding hlpr semF_def[symmetric] fstF_semF
    using aa2[unfolded semT_def semTf_def hlpr2]
    using CSP_assumpt CPOmode_or_CMSmode_or_MIXmode_lm fstF_MF_MT
    by metis
qed

theorem FailureRefinement_Implied_TracesRefinement_If_NoDivergency:
  fixes
    Impl :: "('s, 'e) structured_code_parallel"
  assumes
    --"We assume to already have an SF refinement between CUC and CSP."
    sf_refinement: "Spec ProcFun <=(PreF, CV)=F Impl" and
    --"The same starting states for the CUC code have to be used in the SF and in the T part."
    PreF: "\<forall> tr s X . treen_mirror s Impl \<longrightarrow> (PreT (tr, (PS2PTS s)) \<and> N\<^sub>p s) \<longrightarrow> PreF (tr, s, X)" and
    wffImpl: "wff\<^sub>p Impl" and
    --"For the sequential leafs, we always want to have successors."
    successors: "leafn1_property (\<lambda> (c, \<Sigma>) . \<forall> tr s X . SFsemSigma.sfsem \<Sigma> c {(tr, s, X)} \<noteq> {}) op \<and> Impl" and
    (* The Process name function has to be passed, and in CMSmode it has to be guarded. *)
    CSP_assumpt: "ProcFun = PNfun \<and> (FPmode = CMSmode \<longrightarrow> guardedfun ProcFun)"
  shows
    "Spec ProcFun <=(PreT, CV)=T Impl"
unfolding spec_impl_refinement_t_no_conn_def
proof(rule ccontr, auto)
  fix tr s
  assume
    aa1: "(CV\<^sub>t\<^sub>r tr CV) ~:t [[Spec]]Tf [[ProcFun]]Tfix" and
    aa2: "(tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>TCollect PreT"
  have "\<exists> s' X . (tr, s', X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. PreF x'}"
    using tr_of_T_will_be_in_SF_parallel[OF aa2 PreF wffImpl successors]
    .
  hence maybeGoal: "\<exists> X .(CV\<^sub>t\<^sub>r tr CV, CV\<^sub>X X CV) :f sndF ([[Spec]]Ff [[ProcFun]]Ffix)"
    using sf_refinement[unfolded spec_impl_refinement_sf_no_conn_def]
    by fast
  have hlpr: "[[ProcFun]]Ffix = MF"
    by (simp add: CSP_assumpt MF_def)
  have hlpr2: "[[ProcFun]]Tfix = MT"
    by (simp add: CSP_assumpt MT_def)
  show False
    using maybeGoal
    using aa1
    unfolding hlpr semF_def[symmetric] fstF_semF
    using CSP_assumpt CPOmode_or_CMSmode_or_MIXmode_lm fstF_MF_MT
    by (metis MT_def fstF_semF pairF_domF_T2 semTf_def)
qed

end
