section{* Denotational Semantics *}

theory denotational
imports small_step
begin

subsection {* The Semantics *}

text {* The actual denotational semantics, which is the heart of our approach and takes multiple instructions as the fixpoint of them.
do: execute the non-deterministic function f and use its result as the successor state's R and increase the instruction pointer by one.
cbr: Execute b with the current state's R and take the result as an answer to whether jump to PC m or n. R stays the same.
comm: Extend the trace by some communicated event e, use it to (deterministically) generate the successor state's R and increase the instruction pointer.
*}

fun densem :: "('s, 'e) structured_code \<Rightarrow> ('s, 'e) trace_state set \<Rightarrow> ('s, 'e) trace_state set" ("\<lbrakk>_\<rbrakk>(_)"  [700, 700] 999)
and func :: "('s, 'e) structured_code \<Rightarrow> (('s, 'e) trace_state set \<Rightarrow> ('s, 'e)trace_state set) \<Rightarrow> ('s, 'e) trace_state set \<Rightarrow> ('s, 'e) trace_state set"
where
ddof: "densem (one l (do f)) = (\<lambda>S. S \<union> {(tr, t). (\<exists>s x. (tr, s) \<in> S \<and> x \<in> f (R s) \<and> PC s = l \<and> 
                    t = s\<lparr> R:=x, PC:=PC s + 1\<rparr>)})"
|
dcbr: "densem (one l (cbr b m n)) = (\<lambda>S.  S \<union> {(tr, t). (\<exists>s.  (tr, s) \<in> S \<and> PC s = l \<and>  ((b (R s) \<and>
    t=s\<lparr>PC := m\<rparr>) \<or> (\<not>b (R s) \<and> t = s\<lparr>PC := n\<rparr>)))})"
|
dcomm: "densem (one l (comm ef f)) = (\<lambda>S. S \<union> {(tr', t). (\<exists>tr s e. (tr, s) \<in> S \<and> e \<in> ef (R s) \<and> PC s = l \<and> 
                    tr' = tr@[e] \<and> t = s\<lparr> R := f (R s) e, PC:=PC s + 1\<rparr>)})"
|
dseq: "densem (sc_1 \<oplus> sc_2) = ccpo_class.fixp (func (sc_1 \<oplus> sc_2))"
|
func: "func (sc_1 \<oplus> sc_2) f = (\<lambda>S. f(\<lbrakk>sc_1\<rbrakk>(S)) \<union> f(\<lbrakk>sc_2\<rbrakk>(S)) \<union> S)"

definition densem_refinement :: "('s, 'e) trace_state set \<Rightarrow> ('s, 'e) trace_state set \<Rightarrow> bool"
("_ \<subseteq>\<^sub>T _") where
"(S \<subseteq>\<^sub>T S') = (\<forall> tr s . (tr, s) \<in> S \<longrightarrow> (\<exists> s' . (tr, s') \<in> S'))"

paragraph {* Some fixpoint related lemmas. *}

text{* A small extension to the general @{text "ccpo_class.fixp_induct"} can be made, as func is monotone. *}

lemma func_monotone[simp]: "monotone (op \<le>) (op \<le>) (func (sc_1 \<oplus> sc_2))"
apply(rule monotoneI) apply simp
unfolding le_fun_def by blast

lemma fixp_induct_t:
  "ccpo.admissible Sup op \<le> P \<Longrightarrow> P (Sup {}) \<Longrightarrow> (\<And>x. P x \<Longrightarrow> P (func (c1 \<oplus> c2) x)) \<Longrightarrow> P (ccpo_class.fixp (func (c1 \<oplus> c2)))"
  using "ccpo_class.fixp_induct"[of _ "func (_ \<oplus> _)"] func_monotone
by blast

text{* Convenient unfolds *}

lemma dseq_unfold: "\<lbrakk>(c1 \<oplus> c2)\<rbrakk> S = func (c1 \<oplus> c2) (ccpo_class.fixp (func (c1 \<oplus> c2))) S"
  unfolding dseq
  using fixp_unfold[OF func_monotone, of c1 c2]
  by simp
  
lemma dseq_unfold2: "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk> (\<lbrakk>c1\<rbrakk>S) \<union> \<lbrakk>(c1 \<oplus> c2)\<rbrakk> (\<lbrakk>c2\<rbrakk>S) \<union> S"
  unfolding dseq
  using fixp_unfold[OF func_monotone, unfolded func, of c1 c2]
  using dseq_unfold by auto
  
paragraph {* Semantic lemmas *}

text{* densem is commutative *}

theorem oplus_commutative: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>(S) = \<lbrakk>(code2 \<oplus> code1)\<rbrakk>(S)"
unfolding dseq
proof -
  have hlpr: "\<And> f g S . \<forall> S . f S = g S \<Longrightarrow> ccpo_class.fixp f S = ccpo_class.fixp g S"
    by presburger
  have eq: "\<forall>S. func (code1 \<oplus> code2) S = func (code2 \<oplus> code1) S" by (simp add: Un_assoc sup.left_commute)
  from hlpr[OF eq, of S]
  show "ccpo_class.fixp (func (code1 \<oplus> code2)) S = ccpo_class.fixp (func (code2 \<oplus> code1)) S"
    .
qed

text {* Imagine in the following @{text "g=id"}, then taking the union over the start states is distributive over densem. *}

lemma densem_big_union_additive: "\<lbrakk>c\<rbrakk>(\<Union>f\<in>PS . g f) = (\<Union>f\<in>PS. \<lbrakk>c\<rbrakk>(g f))"
proof (induct c arbitrary: PS g)
  case (one l i)
  show ?case
  proof (cases i, simp_all, blast+) qed
next
  fix c1 c2 :: "('b, 'a) structured_code" and PS :: "'c set" and g :: "'c \<Rightarrow> ('b, 'a) trace_state set"
  assume
    IH1: "\<And>(PS:: 'c set) (g::'c \<Rightarrow> ('b, 'a) trace_state set). \<lbrakk>c1\<rbrakk>(\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. \<lbrakk>c1\<rbrakk>g f)" and
    IH2: "\<And>(PS:: 'c set) (g::'c \<Rightarrow> ('b, 'a) trace_state set). \<lbrakk>c2\<rbrakk>(\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. \<lbrakk>c2\<rbrakk>g f)"
  have "ccpo_class.fixp (func (c1 \<oplus> c2)) (\<Union>f\<in>PS. g f) = (\<Union>f\<in>PS. ccpo_class.fixp (func (c1 \<oplus> c2)) (g f))"
  proof (induct arbitrary: PS g rule: "fixp_induct_t")
  next
    show "ccpo.admissible Sup op \<le> (\<lambda>a. \<forall>x xa. a (UNION x xa) = (\<Union>f\<in>x. a (xa f)))"
      unfolding "ccpo.admissible_def"
    proof (clarsimp)
      fix A :: "('d set \<Rightarrow> 'e set) set" and PS :: "'f set" and g :: "'f \<Rightarrow> 'd set"
      assume "\<forall>a\<in>A. \<forall>PS (g :: 'f \<Rightarrow> 'd set). a (\<Union>s\<in>PS. g s) = (\<Union>s\<in>PS. a (g s))"
      hence "(\<Union>a\<in>A. a (\<Union>s\<in>PS. g s)) = (\<Union>a\<in>A. (\<Union>s\<in>PS. a (g s)))"
        by auto
      thus "(\<Union>a\<in>A. a (\<Union>s\<in>PS. g s)) = (\<Union>s\<in>PS. (\<Union>a\<in>A. a (g s)))"
        unfolding UNION_eq
        by fastforce
    qed
  next
    fix x :: "('b, 'a) trace_state set \<Rightarrow> ('b, 'a) trace_state set" and PS :: "'c set" and g :: "'c \<Rightarrow> ('b, 'a) trace_state set"
    assume step: "\<And>(PS::'c set) g. x (\<Union>s\<in>PS. g s) = (\<Union>s\<in>PS. x (g s))"
    have "x (\<lbrakk>c1\<rbrakk>(\<Union>s\<in>PS. g s)) = x (\<Union>s\<in>PS. \<lbrakk>c1\<rbrakk>g s)"
        using IH1[of g] by simp
    hence "x (\<lbrakk>c1\<rbrakk>(\<Union>s\<in>PS. g s)) = (\<Union>s\<in>PS. x (\<lbrakk>c1\<rbrakk>g s))"
      using step[of "\<lambda>s. \<lbrakk>c1\<rbrakk>g s"] by presburger
    moreover
    have "x (\<lbrakk>c2\<rbrakk>(\<Union>s\<in>PS. g s)) = x (\<Union>s\<in>PS. \<lbrakk>c2\<rbrakk>g s)"
      using IH2[of g] by simp
    hence "x (\<lbrakk>c2\<rbrakk>(\<Union>s\<in>PS. g s)) = (\<Union>s\<in>PS. x (\<lbrakk>c2\<rbrakk>g s))"
      using step[of "\<lambda>s. \<lbrakk>c2\<rbrakk>g s"] by presburger
    ultimately
    show "func (c1 \<oplus> c2) x (UNION PS g) = (\<Union>f\<in>PS. func (c1 \<oplus> c2) x (g f))"
      unfolding func
      by auto
  qed auto
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>UNION PS g = (\<Union>f\<in>PS. \<lbrakk>(c1 \<oplus> c2)\<rbrakk>g f)"
    by auto
qed

text {* A convenient special case of the above for only two sets. *}

lemma densem_union_additive: "\<lbrakk>c\<rbrakk>(A \<union> B) =  \<lbrakk>c\<rbrakk>(A) \<union> \<lbrakk>c\<rbrakk>(B)"
  using densem_big_union_additive[of c _ "{A, B}"]
  by fastforce

lemma densem_super_id: "S \<subseteq> \<lbrakk>code\<rbrakk>(S)"
proof (induct code)
  fix l i
  show "S \<subseteq> \<lbrakk>one l i\<rbrakk>S"
  proof (cases i, auto) qed
next
  fix c1 c2
  assume
    IH1: "S \<subseteq> \<lbrakk>c1\<rbrakk>S"and
    IH2: "S \<subseteq> \<lbrakk>c2\<rbrakk>S"
  show "S \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
    unfolding dseq_unfold
  proof (induct rule: fixp_induct_t, auto simp add: admissible_True) qed
qed

lemma densem_super_id_element: "s \<in> S \<Longrightarrow> s \<in> \<lbrakk>code\<rbrakk>(S)"
apply(insert densem_super_id)
by auto

text{* Code can only change S if there is an s in S with a PC in the domain of the Code.
Note, that this is no bi-implication, as all resulting states already could be part of S and thus nothing new will be added.
*}

lemma pc_does_not_point_to:
  assumes "(\<forall>s \<in> S. PC (snd s) \<notin> dom c)"
  shows "\<lbrakk>c\<rbrakk>S = S"
using assms
proof (induct c arbitrary: S)
  fix l i S
  show "\<forall>s\<in>S. PC (snd s) \<notin> definitions.dom (one l i) \<Longrightarrow> \<lbrakk>one l i\<rbrakk>S = S"
  proof (induct i, auto) qed
next
  fix c1 c2 :: "('s, 'e) structured_code" and S :: "('s, 'e) trace_state set"
  assume
    IH1: "\<And>S. \<forall>s\<in>S. PC (snd s) \<notin> definitions.dom c1 \<Longrightarrow> \<lbrakk>c1\<rbrakk>S = S" and
    IH2: "\<And>S. \<forall>s\<in>S. PC (snd s) \<notin> definitions.dom c2 \<Longrightarrow> \<lbrakk>c2\<rbrakk>S = S" and
    pcs_outside: "\<forall>s\<in>S. PC (snd s) \<notin> definitions.dom (c1 \<oplus> c2)"
  show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = S"
    using pcs_outside
    unfolding dseq_unfold
  proof (induct arbitrary: S rule: fixp_induct_t)
    show "\<And>S . \<forall>s\<in>S. PC (snd s) \<notin> definitions.dom (c1 \<oplus> c2) \<Longrightarrow> func (c1 \<oplus> c2) (Sup {}) S = S"
      unfolding "func.simps" by simp
  next
    fix x and S :: "('s, 'e) trace_state set"
    assume
      step: "\<And>S. \<forall>s\<in>S. PC (snd s) \<notin> definitions.dom (c1 \<oplus> c2) \<Longrightarrow> func (c1 \<oplus> c2) x S = S" and
      pcs_outside: "\<forall>s\<in>S. PC (snd s) \<notin> definitions.dom (c1 \<oplus> c2)"
    have eq1: "\<lbrakk>c1\<rbrakk>S = S"
      using IH1[of S] pcs_outside by simp
    have eq2: "\<lbrakk>c2\<rbrakk>S = S"
      using IH2[of S] pcs_outside by simp
    have eq: "x (\<lbrakk>c1\<rbrakk>\<lbrakk>c1\<rbrakk>S) \<union> x (\<lbrakk>c2\<rbrakk>\<lbrakk>c1\<rbrakk>S) \<union> x (\<lbrakk>c1\<rbrakk>\<lbrakk>c2\<rbrakk>S) \<union> x (\<lbrakk>c2\<rbrakk>\<lbrakk>c2\<rbrakk>S) \<union> S = x (\<lbrakk>c1\<rbrakk>S) \<union> x (\<lbrakk>c2\<rbrakk>S) \<union> S"
      using eq1 eq2 by simp
    also have "... = S" using step[of S, OF pcs_outside] unfolding func by simp
    ultimately show "func (c1 \<oplus> c2) (func (c1 \<oplus> c2) x) S = S"
      unfolding func
      using eq1 eq2
      by simp
  qed (auto simp add: ccpo.admissible_def)
qed

lemma pc_does_not_point_toElement: "PC (snd s) \<notin> dom code \<Longrightarrow>  \<lbrakk>code\<rbrakk>{s} = {s}"
by(metis pc_does_not_point_to singletonD)

paragraph {* Monotonicity *}

lemma densem_monotone: "A \<subseteq> B \<Longrightarrow> \<lbrakk>code\<rbrakk>(A) \<subseteq> \<lbrakk>code\<rbrakk>(B)"
by(metis (full_types) densem_union_additive le_iff_sup)

lemma densem_subset_single:
  assumes "x' \<in> S" and "x \<in> \<lbrakk>c\<rbrakk>{x'}"
  shows "x \<in> \<lbrakk>c\<rbrakk>S"
using densem_monotone[of "{x'}" S c] assms(1) assms(2) by blast

paragraph {* Predecessor Knowledge *}

text{* For each state in Code S there is a densem_predecessor in S (might be itself) *}

lemma densem_predecessor:
  assumes
    a: "s \<in> \<lbrakk>c\<rbrakk>S"
  shows "\<exists> s' \<in> S . s \<in> \<lbrakk>c\<rbrakk>{s'} \<and> (\<not> (PC (snd s') \<in> dom c) \<longrightarrow> s = s')"
proof -
  have "\<exists> s' \<in> S . s \<in> \<lbrakk>c\<rbrakk>{s'}"
  proof (rule ccontr, simp)
    --"We now can show via the big union, that s also could not be created with @{term S} as start."
    assume a1: "\<forall> s' \<in> S. s \<notin> \<lbrakk>c\<rbrakk>{s'}"
    from a1 have "s \<notin> (\<Union>x'\<in>S . \<lbrakk>c\<rbrakk>{x'})"
      by auto
    moreover
    have "(\<Union>x'\<in>S . \<lbrakk>c\<rbrakk>{x'}) = \<lbrakk>c\<rbrakk>(\<Union>x'\<in>S . {x'})"
      using densem_big_union_additive[of c _ S, symmetric]
      by presburger
    moreover
    have "(\<Union>x'\<in>S . {x'}) = S"
      by simp
    ultimately
    have "s \<notin> \<lbrakk>c\<rbrakk>S"
      by simp
    thus "False"
      using a
      by auto
  qed
  thus ?thesis
    using pc_does_not_point_to[of _ c] a by auto
qed

text {* We cannot generate states from nothing. *}
    
lemma code_empty: "\<lbrakk>c\<rbrakk>{}={}" using densem_predecessor by fast

paragraph {* Code absorption - multiple code applications *}

lemma densem_idem:
  fixes S :: "('s, 'e) trace_state set"
  shows "\<lbrakk>code\<rbrakk>(S) = \<lbrakk>code\<rbrakk>\<lbrakk>code\<rbrakk>(S)"
proof (induct code arbitrary: S)
  fix l i and S :: "('e list \<times> 's state) set"
  show "\<lbrakk>one l i\<rbrakk>S = \<lbrakk>one l i\<rbrakk>\<lbrakk>one l i\<rbrakk>S"
  proof (induct i, auto) qed
next
  fix c1 c2 and S :: "('s, 'e) trace_state set"
  assume
    IH1: "\<And>(S::('s, 'e) trace_state set) . \<lbrakk>c1\<rbrakk>S = \<lbrakk>c1\<rbrakk>\<lbrakk>c1\<rbrakk>S" and
    IH2: "\<And>(S::('s, 'e) trace_state set) . \<lbrakk>c2\<rbrakk>S = \<lbrakk>c2\<rbrakk>\<lbrakk>c2\<rbrakk>S"
  have "((densem (c1 \<oplus> c2)) S) \<supseteq>
    (densem (c1 \<oplus> c2)) ((ccpo_class.fixp (func (c1 \<oplus> c2))) S)"
  proof (induct arbitrary: S rule: fixp_induct_t)
    show "\<And>S . \<lbrakk>(c1 \<oplus> c2)\<rbrakk>Sup {} S \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" using code_empty[of "(c1 \<oplus> c2)"] by simp
  next
    show "ccpo.admissible Sup op \<le> (\<lambda>a. \<forall> S. \<lbrakk>(c1 \<oplus> c2)\<rbrakk>a S \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S)"
      unfolding "ccpo.admissible_def" 
    proof (clarsimp)
      fix A S a b
      assume
        subs: "\<forall>a\<in>A. \<forall> S . ccpo_class.fixp (func (c1 \<oplus> c2)) (a S) \<subseteq> ccpo_class.fixp (func (c1 \<oplus> c2)) S" and
        in_fixp: "(a, b) \<in> ccpo_class.fixp (func (c1 \<oplus> c2)) (\<Union>f\<in>A. f S)"
      have "ccpo_class.fixp (func (c1 \<oplus> c2)) (\<Union>f\<in>A. f S) = (\<Union> a \<in> A . ccpo_class.fixp (func (c1 \<oplus> c2)) (a S))"
        using densem_big_union_additive[of "c1 \<oplus> c2", unfolded dseq] .
      moreover have "(\<Union> a \<in> A . ccpo_class.fixp (func (c1 \<oplus> c2)) (a S)) \<subseteq> ccpo_class.fixp (func (c1 \<oplus> c2)) S"
        using subs by fast
      ultimately show "(a, b) \<in> ccpo_class.fixp (func (c1 \<oplus> c2)) S" using in_fixp by blast
    qed
  next
    fix x S
    assume
      step: "\<And>S. \<lbrakk>(c1 \<oplus> c2)\<rbrakk>x S \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
    show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>func (c1 \<oplus> c2) x S \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
      unfolding func
      using step[of "\<lbrakk>c1\<rbrakk>S"] step[of "\<lbrakk>c2\<rbrakk>S"]
      using dseq_unfold2[of c1 c2 S]
      unfolding  densem_union_additive[of "c1 \<oplus> c2"] by blast
  qed
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
    unfolding dseq[symmetric]
    using densem_super_id[of "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>S" "c1 \<oplus> c2"]
    by blast
qed

lemma densem_transitive: 
  assumes "S' \<subseteq> \<lbrakk>code\<rbrakk>(S)"
          "T \<subseteq> \<lbrakk>code\<rbrakk>(S')"
  shows   "T \<subseteq> \<lbrakk>code\<rbrakk>(S)"
proof -
  from assms have "T \<subseteq> \<lbrakk>code\<rbrakk>\<lbrakk>code\<rbrakk>(S)" using densem_monotone order_subst1 by blast
  thus ?thesis using densem_idem by auto
qed

lemma densem_transitive_element:
  assumes "s' \<in> \<lbrakk>code\<rbrakk>({s})"
          "t \<in> \<lbrakk>code\<rbrakk>({s'})"
  shows   "t \<in> \<lbrakk>code\<rbrakk>({s})"
  using assms densem_transitive[of "{s'}" code "{s}" "{t}"]
proof - qed blast

text {* As code is taken as a fixpoint it doesnt matter whether one uses code once or twice (or even more often for that matter) *}

context
begin

private lemma densem_partial_result_inclusion_1: "\<lbrakk>c1\<rbrakk>(S) \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>(S)"
by (metis denotational.dseq densem_super_id dseq_unfold func.simps le_supI1)

private lemma densem_partial_result_inclusion_2: "\<lbrakk>c2\<rbrakk>(S) \<subseteq> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>(S)"
  unfolding oplus_commutative[of c1 c2 S]
  using densem_partial_result_inclusion_1
by blast

private lemma "\<lbrakk>code1\<rbrakk>\<lbrakk>(code1 \<oplus> code2)\<rbrakk> S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk> S"
by(metis densem_idem densem_partial_result_inclusion_1 densem_super_id subset_antisym)

lemma code_absorbtion_right1: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code1\<rbrakk> S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk> S"
by (metis densem_idem densem_partial_result_inclusion_1 densem_super_id densem_union_additive sup.absorb1 sup_absorb2)

lemma codeReduction1MalAnders2: "\<lbrakk>code2\<rbrakk>\<lbrakk>(code1 \<oplus> code2)\<rbrakk> S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk> S"
by(metis densem_idem densem_partial_result_inclusion_2 densem_super_id subset_antisym)

lemma code_absorbtion_right2: "\<lbrakk>(code1 \<oplus> code2)\<rbrakk>\<lbrakk>code2\<rbrakk> S = \<lbrakk>(code1 \<oplus> code2)\<rbrakk> S"
by (metis densem_idem densem_partial_result_inclusion_2 densem_super_id densem_union_additive sup.absorb1 sup_absorb2)

lemma densem_code_absorption_right_one:
  assumes "{(l, i)} \<subseteq> (\<union>\<^sub>s\<^sub>c c)"
  shows "\<lbrakk>c\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>c\<rbrakk>S"
  using assms
proof (induct c arbitrary: S, clarsimp)
  fix S
  show "\<lbrakk>(one l i)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(one l i)\<rbrakk>S"
    using densem_idem[of "one l i" S, symmetric]
    .
next
  fix
    c1 c2 S
  assume
    IH1: "\<And>S. {(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c1 \<Longrightarrow> \<lbrakk>c1\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>c1\<rbrakk>S" and
    IH2: "\<And>S. {(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c2 \<Longrightarrow> \<lbrakk>c2\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>c2\<rbrakk>S" and
    subs: "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)"
  consider (inc1) "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c1" | (inc2) "{(l, i)} \<subseteq> \<union>\<^sub>s\<^sub>c c2"
    using subs by fastforce
  thus "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
  proof cases
    case inc1
    have b1:  "\<And>S . \<lbrakk>c1\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>c1\<rbrakk>S"
      using IH1[OF inc1] .
    show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
      using code_absorbtion_right1 b1
      by metis
  next
    case inc2
    have b1:  "\<And>S . \<lbrakk>c2\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>c2\<rbrakk>S"
      using IH2[OF inc2] .
    show "\<lbrakk>(c1 \<oplus> c2)\<rbrakk>\<lbrakk>(one l i)\<rbrakk>S = \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
      using code_absorbtion_right2 b1
      by metis
  qed
qed

lemma densem_code_absorption_left_one: 
  assumes "\<union>\<^sub>s\<^sub>c (one l i) \<subseteq> \<union>\<^sub>s\<^sub>c c"
  shows "\<lbrakk>(one l i)\<rbrakk>\<lbrakk>c\<rbrakk>S = \<lbrakk>c\<rbrakk>S"
  using assms
by (metis densem_idem densem_super_id densem_code_absorption_right_one sc_inj.one subset_antisym)

lemma densem_partial_result_inclusion_element1: "t \<in> \<lbrakk>sc_1\<rbrakk>(S) \<Longrightarrow> t \<in> \<lbrakk>(sc_1 \<oplus> sc_2)\<rbrakk>(S)"
  using densem_partial_result_inclusion_1
by auto

lemma densem_partial_result_inclusion_element2: "t \<in> \<lbrakk>sc_2\<rbrakk>(S) \<Longrightarrow> t \<in> \<lbrakk>(sc_1 \<oplus> sc_2)\<rbrakk>(S)"
  using densem_partial_result_inclusion_2
by auto

end

subsection {* Connecting T with multistep and other theorems. *}

context liSet
begin

text {* A smallstep can be directly translated into a single instruction execution. *}

lemma Smallstep_implies_densem: "((trs, s), trt, t) \<in> Smallstep \<Longrightarrow>
       \<exists> instr . (PC s, instr) \<in> lis \<and> (trt, t) \<in> \<lbrakk>(one (PC s) instr)\<rbrakk>{(trs, s)}"
  unfolding Stos
proof (induct rule: smallstep.induct, fastforce+) qed

text {* A single instruction execution can be directly translated into a smallstep. *}

lemma single_densem_implies_Smallstep:
  assumes 
    ind: "t \<in> \<lbrakk>(one l instr)\<rbrakk>S" and
    subs: "{(l, instr)} \<subseteq> lis" and
    notInS: "t \<notin> S"
  shows "\<exists> s. s\<in>S \<and> (s, t) \<in> Smallstep \<and> PC (snd s) = l"
proof -
  obtain s where inds: "t \<in> \<lbrakk>one l instr\<rbrakk>{s}" and inS: "s \<in> S" using ind densem_predecessor
    by blast
  have neq: "t \<noteq> s" using notInS inS by fast
  obtain tr' ts tr ss where t: "t = (tr', ts)" and s: "s = (tr, ss)" by fastforce
  have pcsseql: "PC ss = l" using neq inds unfolding s proof (induct instr, auto) qed
  have "((tr, ss), (tr', ts)) \<in> Smallstep" (* using inds[unfolded t s] *)
    unfolding Stos
  using inds subs unfolding s t
  proof (induct instr)
    case (do f)
    from sdof[of l f lis _ ss ts tr, unfolded Suc_eq_plus1[symmetric]] do[simplified] neq s t
    show "((tr, ss), tr', ts) \<in> smallstep lis" by blast
  next
    case (cbr b m n)
    from scbrf[of l b m n lis ss ts tr] scbrt[of l b m n lis ss ts tr] cbr[simplified] neq
    show "((tr, ss), tr', ts) \<in> smallstep lis" unfolding Suc_eq_plus1 s t by fast
  next
    case (comm ef f)
    from scomm[of l ef f lis _ ss tr' tr ts] comm[simplified] neq
    show "((tr, ss), tr', ts) \<in> smallstep lis" unfolding Suc_eq_plus1 s t by blast
  qed
  thus ?thesis using s t pcsseql inS by auto
qed
 
text {* A @{term multistep_k} can be translated into traces semantics. *}

theorem multistep_k_implies_densem:
  assumes
    subs: "\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis" and
    ms_k: "((trs, s), \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k) \<in> multistep_k"
  shows "(trt,t) \<in> \<lbrakk>strInstr\<rbrakk>({(trs,s)})"
  using ms_k
proof (induct k arbitrary: trs s rule: nat_induct)
  case 0
  show ?case
    using multistep_k_0_same[OF 0] densem_super_id by blast
next
  case (Suc k)
  from multistep_k_split[OF Suc(2), unfolded Stos]
  obtain s' trs' where
    smallstep: "((trs, s), trs', s') \<in> smallstep lis" and
    ms_k: "((trs', s'), \<union>\<^sub>s\<^sub>c strInstr, (trt, t), k) \<in> multistep_k"
    by blast
  from Suc(1)[OF ms_k] have base: "(trt, t) \<in> \<lbrakk>strInstr\<rbrakk>{(trs', s')}" .
  obtain instr where inlis: "(PC s, instr) \<in> lis" and ex: "(trs', s') \<in> \<lbrakk>one (PC s) instr\<rbrakk>{(trs, s)}"
    using  smallstep Stos Smallstep_implies_densem
    using smallstep_rev[OF smallstep] by blast
  hence inStrInstr: "{(PC s, instr)} \<subseteq> \<union>\<^sub>s\<^sub>c strInstr"
    using indomSC_element inlis liSet.multistep_k_suc liSet_axioms Suc subs by blast
  hence "(trt, t) \<in> \<lbrakk>strInstr\<rbrakk>\<lbrakk>one (PC s) instr\<rbrakk>{(trs, s)}" using base densem_subset_single ex
    by blast
  thus ?case
    using densem_code_absorption_right_one[OF inStrInstr] by blast
qed
 
text {* An @{term exec_k} step can be translated into traces semantics. *}

lemma reflection_of_stuck_reductions_k:
  assumes
    subs: " (\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis)" and
    exec_k: "(s, \<union>\<^sub>s\<^sub>c strInstr, t, k) \<in> Exec_k"
  shows "t \<in> \<lbrakk>strInstr\<rbrakk>({s})"
using exec_k
proof (induct k arbitrary: s)
  case 0
  obtain tr ss tr' ss' where
    s: "s = (tr, ss)" and
    t: "t = (tr', ss')" by fastforce
  have seqt: "s = t"
    using Exec_k_0_same[OF 0[unfolded s t]]
    unfolding s t by simp
  thus ?case
    using densem_super_id by fastforce
next
  case (Suc k)
  obtain tr ss tr' ss' where
    s: "s = (tr, ss)" and
    t: "t = (tr', ss')" by fastforce
  from Exec_k_split[OF Suc(2)[unfolded s t], unfolded Stos]
  obtain s' trs' where
    smallstep: "((tr, ss), trs', s') \<in> smallstep lis" and
    exec_k: "((trs', s'), \<union>\<^sub>s\<^sub>c strInstr, (tr', ss'), k) \<in> Exec_k" and
    neq0: "0 < k \<longrightarrow> indom (PC s') (\<union>\<^sub>s\<^sub>c strInstr)" and
    eq0: "k = 0 \<longrightarrow> \<not> indom (PC s') (\<union>\<^sub>s\<^sub>c strInstr)"
    by blast
  from Suc(1)[unfolded t, OF exec_k] have base: "(tr', ss') \<in> \<lbrakk>strInstr\<rbrakk>{(trs', s')}" .
  obtain instr where inlis: "(PC ss, instr) \<in> lis" and exd: "(trs', s') \<in> \<lbrakk>one (PC ss) instr\<rbrakk>{(tr, ss)}"
    using Smallstep_implies_densem[OF smallstep[unfolded Stos[symmetric]]]
    by blast
  have inStrInstr: "{(PC ss, instr)} \<subseteq> \<union>\<^sub>s\<^sub>c strInstr"
    using Suc(2) indomSC_element inlis liSet.Exec_k_suc liSet_axioms s subs t by fastforce
  have "(tr', ss') \<in> \<lbrakk>strInstr\<rbrakk>\<lbrakk>one (PC ss) instr\<rbrakk>{(tr, ss)}" using base densem_subset_single exd
    by blast
  thus ?case
    unfolding t s
    using densem_code_absorption_right_one[OF inStrInstr] by blast
qed

text{* densem equals multistep and Exec *}

theorem densem_implies_multistep: "(\<union>\<^sub>s\<^sub>c c \<subseteq> lis) \<Longrightarrow> 
t \<in> \<lbrakk>c\<rbrakk>(S) \<Longrightarrow>  (\<exists>s\<in>S.(s, \<union>\<^sub>s\<^sub>c c, t) \<in> multistep)"
proof (induct c arbitrary: t S)
  case (one l i)
  {
    assume "t \<in> S"
    hence ?case
      using Mterm one
      by (metis old.prod.exhaust)
  }
  moreover { 
    assume tNotInS: "t \<notin> S"
    from single_densem_implies_Smallstep[OF _ _ tNotInS] one Mseq
      have ?case
      by (metis insert_subset multistep.Mterm order_refl prod.collapse sc_inj.simps(1))
  }
  ultimately show ?case by blast
next
  fix c1 c2 :: "('s, 'e) structured_code" and t S
  assume
    IH1: "\<And> t S . \<union>\<^sub>s\<^sub>c c1 \<subseteq> lis \<Longrightarrow> t \<in> \<lbrakk>c1\<rbrakk>S \<Longrightarrow> \<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c c1, t) \<in> multistep" and
    IH2: "\<And> t S . \<union>\<^sub>s\<^sub>c c2 \<subseteq> lis \<Longrightarrow> t \<in> \<lbrakk>c2\<rbrakk>S \<Longrightarrow> \<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c c2, t) \<in> multistep" and
    subLis: "\<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<subseteq> lis" and
    tInD: "t \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
  show "\<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep"
    using tInD unfolding dseq
  proof (induct arbitrary: S rule: fixp_induct_t)
    show "ccpo.admissible Sup op \<le> (\<lambda>a. \<forall> S . t \<in> a S \<longrightarrow> (\<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep))"
      unfolding "ccpo.admissible_def"
      by auto
  next
    fix x :: "('s, 'e) trace_state set \<Rightarrow> ('s, 'e) trace_state set" and S
    assume
      step: "\<And> S . t \<in> x S \<Longrightarrow> \<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep" and
      tInF: "t \<in> func (c1 \<oplus> c2) x S"
    consider (tinc1) "t \<in> x (\<lbrakk>c1\<rbrakk>S)" | (tinc2) "t \<in> x (\<lbrakk>c2\<rbrakk>S)" | (tinS) "t \<in> S"
      using tInF[unfolded func] by blast
    thus "\<exists>s\<in>S. (s, \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep"
    proof cases
      case tinS
      thus ?thesis
        using Mterm subLis
        by (metis old.prod.exhaust)
    next
      case tinc1
      obtain t' where
        t' : "t' \<in>\<lbrakk>c1\<rbrakk>S" and
        t'2: "(t', \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep"
        using step[of "\<lbrakk>c1\<rbrakk>S", OF tinc1]
        by blast
      obtain t'' where
        t'': "t''\<in>S" and
        t''2: "(t'', \<union>\<^sub>s\<^sub>c c1, t') \<in> multistep"
        using IH1[OF _ t'] subLis[unfolded sc_inj.simps]
        by blast
      have eq: "\<union>\<^sub>s\<^sub>c c1 \<union> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) = \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" by simp
      show ?thesis using t'2 t'' t''2
        using multistep_seq[OF t''2 t'2]
        unfolding eq
        by blast
    next
      case tinc2
      obtain t' where
        t' : "t' \<in>\<lbrakk>c2\<rbrakk>S" and
        t'2: "(t', \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2), t) \<in> multistep"
        using step[of "\<lbrakk>c2\<rbrakk>S", OF tinc2]
        by blast
      obtain t'' where
        t'': "t''\<in>S" and
        t''2: "(t'', \<union>\<^sub>s\<^sub>c c2, t') \<in> multistep"
        using IH2[OF _ t'] subLis[unfolded sc_inj.simps]
        by blast
      have eq: "\<union>\<^sub>s\<^sub>c c2 \<union> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) = \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" by auto
      show ?thesis using t'2 t'' t''2
        using multistep_seq[OF t''2 t'2]
        unfolding eq
        by blast
    qed
  qed auto
qed

theorem reflection_of_stuck_reductions  : " (\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> Exec \<Longrightarrow> t \<in> \<lbrakk>strInstr\<rbrakk>({s})"
  using
    Exec_implies_Exec_k[of "fst s" "snd s" "\<union>\<^sub>s\<^sub>c strInstr" "fst t" "snd t"]
    reflection_of_stuck_reductions_k[of strInstr s t]
  by auto

theorem Exec_implies_densem : " (\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> s \<in> S \<Longrightarrow> (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> Exec \<Longrightarrow> t \<in> \<lbrakk>strInstr\<rbrakk>(S)"
using densem_subset_single reflection_of_stuck_reductions by blast

theorem Reflection_of_stuck_reductions  : " (\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> multistep \<Longrightarrow> t \<in> \<lbrakk>strInstr\<rbrakk>({s})"
by (metis liSet.multistep_k_implies_densem liSet_axioms multistep_implies_multistep_k prod.collapse)

theorem multistep_implies_densem : " (\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> s \<in> S \<Longrightarrow> (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> multistep \<Longrightarrow> t \<in> \<lbrakk>strInstr\<rbrakk>(S)"
  using Reflection_of_stuck_reductions densem_subset_single by blast

theorem densem_implies_Exec_single: "(\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> t \<in> \<lbrakk>strInstr\<rbrakk>({s}) \<Longrightarrow> \<not> (PC (snd t)) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstr) \<Longrightarrow> (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> Exec"
using densem_implies_multistep[of strInstr t "{s}", simplified]
using notIndomIsStuck[of "fst s" "snd s" "\<union>\<^sub>s\<^sub>c strInstr" "fst t" "snd t"] by simp

theorem densem_implies_multistep_single: "(\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow> 
t \<in> \<lbrakk>strInstr\<rbrakk>({s}) \<Longrightarrow>  (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> multistep"
by (metis densem_implies_multistep singletonD)

theorem densem_implies_Exec: "(\<union>\<^sub>s\<^sub>c strInstr \<subseteq> lis) \<Longrightarrow>
t \<in> \<lbrakk>strInstr\<rbrakk>(S) \<Longrightarrow> \<not> (PC (snd t)) \<in>\<^sub>l\<^sub>i\<^sub>s (\<union>\<^sub>s\<^sub>c strInstr) \<Longrightarrow> (\<exists>s \<in> S. (s, \<union>\<^sub>s\<^sub>c strInstr, t) \<in> Exec)"
by(metis densem_implies_multistep notIndomIsStuck prod.collapse)

paragraph {* Associativity *}

theorem oplus_associative:
  assumes subs:"(\<union>\<^sub>s\<^sub>c ((code1 \<oplus> code2) \<oplus> code3) \<subseteq> lis)"
  shows "\<lbrakk>(code1 \<oplus> (code2 \<oplus> code3))\<rbrakk>(S) = \<lbrakk>((code1 \<oplus> code2) \<oplus> code3)\<rbrakk>(S)"
proof -
  have eq: "\<union>\<^sub>s\<^sub>c ((code1 \<oplus> code2) \<oplus> code3) = \<union>\<^sub>s\<^sub>c (code1 \<oplus> (code2 \<oplus> code3))" by force
  {
    fix t
    assume ind: "t \<in> \<lbrakk>((code1 \<oplus> code2) \<oplus> code3)\<rbrakk>(S)"
    have "t \<in> \<lbrakk>(code1 \<oplus> (code2 \<oplus> code3))\<rbrakk>(S)"
      using densem_implies_multistep[OF subs ind, unfolded eq] multistep_implies_densem[OF subs[unfolded eq], of _ S t]
      by blast
  } moreover {
    fix t
    assume ind: "t \<in> \<lbrakk>(code1 \<oplus> (code2 \<oplus> code3))\<rbrakk>(S)"
    have "t \<in> \<lbrakk>((code1 \<oplus> code2) \<oplus> code3)\<rbrakk>(S)"
      using densem_implies_multistep[OF subs[unfolded eq] ind, unfolded eq[symmetric]] multistep_implies_densem[OF subs, of _ S t]
      by blast
  } ultimately show ?thesis by blast
qed

paragraph {* Trace related lemmas *}
  
text {* A successor state keeps the trace of the starting state as a prefix. *}
  
lemma densem_tr_growing:
  assumes
    "(tr, s) \<in> \<lbrakk>c\<rbrakk>{(tr', s')}" and
    "\<union>\<^sub>s\<^sub>c c \<subseteq> lis"
  shows "\<exists> l . tr = tr' @ l"
proof-
  have ms: "((tr', s'), \<union>\<^sub>s\<^sub>c c, (tr, s)) \<in> multistep"
    by (simp add: assms densem_implies_multistep_single)
  show ?thesis
    using multistep_implies_multistep_k[OF ms]
      multistep_k_growing_trace[of tr' s' "\<union>\<^sub>s\<^sub>c c" tr s]
    by blast
qed

text {* If trace @{term tr} is not in the semantics, but some of its prefixes @{term tr'},
then there is an intermediate prefix which still is contained, whereas the next longer prefix is not. *}

lemma densem_intermediate_trace_nonacceptance:
  assumes
    --"Some trace is not contained in the denotational semantics..."
    "\<forall>s'. (tr, s') \<notin> \<lbrakk>c\<rbrakk> S" and
    "tr = tr' @ l" and
    --"But one of its prefixes is."
    "(tr', s) \<in> \<lbrakk>c\<rbrakk> S"
  --"This means, we can find some longest prefix of @{term tr}, such that it is still contained."
  shows "\<exists> r a l s . (tr' @ r, s) \<in> \<lbrakk>c\<rbrakk> S \<and>
        (tr' @ r @ [a], s') \<notin> \<lbrakk>c\<rbrakk> S \<and>
        (tr' @ r @ [a] @ l = tr)"
using assms
proof (induct "rev l" arbitrary: l tr, auto)
  fix a x la
  assume 
    IH: "\<And>l tr. x = rev l \<Longrightarrow>
                \<forall>s'. (tr' @ l, s') \<notin> \<lbrakk>c\<rbrakk>S \<Longrightarrow>
                tr = tr' @ l \<Longrightarrow>
                \<exists>r. (\<exists>s. (tr' @ r, s) \<in> \<lbrakk>c\<rbrakk>S) \<and> (\<exists>a. ((tr' @ r @ [a], s') \<notin> \<lbrakk>c\<rbrakk>S) \<and> (\<exists>la. r @ a # la = l))" and
    ax: "a # x = rev la" and
    not_contained: "\<forall>s'. (tr' @ la, s') \<notin> \<lbrakk>c\<rbrakk>S"
  have l_def: "la = (rev x) @ [a]"
    using ax
    by (metis rev_eq_Cons_iff)
  consider (withIH) "\<forall> s' . (tr' @ (rev x), s') \<notin> \<lbrakk>c\<rbrakk>S" | (directFin) "\<exists> s . (tr' @ (rev x), s) \<in> \<lbrakk>c\<rbrakk>S"
    by blast
  thus "\<exists>r. (\<exists>s. (tr' @ r, s) \<in> \<lbrakk>c\<rbrakk>S) \<and> (\<exists>a. ((tr' @ r @ [a], s') \<notin> \<lbrakk>c\<rbrakk>S) \<and> (\<exists>l. r @ a # l = la))"
  proof (cases)
    case directFin
    obtain s'' where s'': "(tr' @ rev x, s'') \<in> \<lbrakk>c\<rbrakk>S" using directFin by blast
    moreover
    have "(tr' @ (rev x) @ [a], s') \<notin> \<lbrakk>c\<rbrakk>S"
      using not_contained[unfolded l_def] 
      by blast
    moreover
    have "(rev x) @ a # [] = la"
      using l_def by blast
    ultimately
    show ?thesis
      by blast
  next
    case withIH
    obtain l where l_def: "x = rev l"
      by (metis rev_rev_ident)
    have l_def': "l = rev x" using l_def by simp
    obtain r s'' a la' where
      "(tr' @ r, s'') \<in> \<lbrakk>c\<rbrakk>S" and
      "(tr' @ r @ [a], s') \<notin> \<lbrakk>c\<rbrakk>S" and
      l: "l = r @ a # la'"
      using IH[OF l_def withIH[unfolded l_def'[symmetric]]]
      by metis
    moreover
    have "\<exists>l. r @ a # l = la"
      using l ax l_def l_def'
      by (metis append.simps(2) append_assoc rev.simps(2) rev_rev_ident)
    ultimately
    show ?thesis
      by blast
  qed
qed

end --"context liSet"

subsection {* T instruction path execution *}

text {* The concept behind this part is equivalent to the one defined in the stable failures semantics.
Please refer to the documentation there. *}

fun densem_exec_b :: "(label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) trace_state set \<Rightarrow> ('s, 'e) trace_state set" where
  "densem_exec_b [] S = S" |
  "densem_exec_b ((l, i) # lil) S = densem_exec_b lil (densem (one l i) S)"

lemma densem_exec_b_unfold:
  "densem_exec_b (lil @ [(l, i)]) S = (densem (one l i) (densem_exec_b (lil) S))"
proof (induct lil arbitrary: S l i)
  case (Cons li lil)
  have "densem_exec_b ((li # lil) @ [(l, i)]) S = densem_exec_b ((lil) @ [(l, i)]) (densem (one (fst li) (snd li)) S)"
    by (metis append_Cons densem_exec_b.simps(2) prod.collapse)
  also have "... = \<lbrakk>(one l i)\<rbrakk>densem_exec_b lil (densem (one (fst li) (snd li)) S)"
    using Cons
    by auto
  finally show "densem_exec_b ((li # lil) @ [(l, i)]) S = \<lbrakk>(one l i)\<rbrakk>densem_exec_b (li # lil) S"
    by (metis densem_exec_b.simps(2) prod.collapse)
qed auto
  
lemma densem_exec_b_unfold2:
  "densem_exec_b (l @ r) S = (densem_exec_b (r) (densem_exec_b (l) S))"
proof (induct l arbitrary: S)
  case (Cons a l)
  obtain lbl i where a: "a = (lbl, i)" by fastforce
  show "densem_exec_b ((a # l) @ r) S = densem_exec_b r (densem_exec_b (a # l) S)"
    using densem_exec_b.simps(2)[of lbl i l S] Cons[of "\<lbrakk>one lbl i\<rbrakk>S"]
    unfolding a
    by simp
qed auto

definition "densem_exec lil S \<equiv> densem_exec_b (rev lil) S"

lemma densem_exec_empty [simp]: "densem_exec [] S = S"
  using densem_exec_def[of "[]" S]
  by auto

lemma densem_exec_unfold:
  "densem_exec ((l, i) # lil) S = (densem (one l i) (densem_exec (lil) S))"
using densem_exec_b_unfold densem_exec_def
  by (metis rev.simps(2))

lemma densem_exec_pred1: 
  assumes "(tr, s) \<in> densem_exec ((l, i) # lil) S"
  shows "\<exists> tr' s' . (tr', s') \<in> densem_exec lil S \<and> (tr, s) \<in> \<lbrakk>one l i\<rbrakk>{(tr', s')}"
  using densem_exec_unfold[of l i lil S] assms densem_predecessor[of "(tr, s)" "(one l i)" "densem_exec lil S"]
  by fast

lemma densem_exec_additive:
  "densem_exec lil (P \<union> Q) = densem_exec lil P \<union> densem_exec lil Q"
proof (induct lil, simp)
  case (Cons li lil)
  obtain l i where li: "li = (l, i)" by fastforce
  show "densem_exec (li # lil) (P \<union> Q) = densem_exec (li # lil) P \<union> densem_exec (li # lil) Q"
    unfolding li densem_exec_unfold densem_union_additive[of "one l i" "densem_exec lil P" "densem_exec lil Q", symmetric] Cons
    ..
qed

lemma densem_exec_pred2:
  assumes "(tr, s) \<in> densem_exec (lil) S"
  shows "\<exists> tr' s' . (tr, s) \<in> densem_exec (lil) {(tr', s')} \<and> (tr', s') \<in> S"
using assms
proof(induct lil arbitrary: tr s, simp)
  case (Cons li lil)
  obtain l i where li: "li = (l, i)" by fastforce
  show "(tr, s) \<in> densem_exec (li # lil) S \<Longrightarrow> \<exists>tr' s'. (tr, s) \<in> densem_exec (li # lil) {(tr', s')} \<and> (tr', s') \<in> S"
    unfolding li densem_exec_unfold
    using densem_predecessor[of "(tr, s)" "one l i" "densem_exec lil S"] Cons densem_subset_single[of _ _ "(tr, s)" "one l i"]
    by fast
qed

lemma densem_exec_unfold2:
  "densem_exec (l @ r) S = (densem_exec (l) (densem_exec (r) S))"
by (simp add: densem_exec_b_unfold2 densem_exec_def)

lemma densem_exec_unfold_pred':
  assumes 
    a: "(tr, s) \<in> densem_exec (l @ r) S"
  shows
    "\<exists> tr' s' . (tr, s) \<in> densem_exec (l) {(tr', s')} \<and> (tr', s') \<in> densem_exec (r) S"
using densem_exec_pred2[of tr s l "densem_exec r S", OF a[unfolded densem_exec_unfold2]] .

lemma densem_exec_unfold_pred:
  assumes 
    "x \<in> densem_exec (l @ r) S"
  shows
    "\<exists> x' . x \<in> densem_exec (l) {x'} \<and> x' \<in> densem_exec (r) S"
using densem_exec_pred2[of _ _ l "densem_exec r S"] assms[unfolded densem_exec_unfold2]
  by (metis prod.collapse)

lemma densem_exec_fold:
  assumes
    a: "x \<in> densem_exec l {x'}" and
    b: "x' \<in> densem_exec r S"
  shows "x \<in> densem_exec (l @ r) S"
unfolding densem_exec_unfold2[of l r S]
using a b  densem_exec_additive[of l "{x'}"]
by (metis (no_types, lifting) UnCI insert_Diff insert_is_Un)

lemma densem_exec1: 
  assumes
    "(tr, s) \<in> densem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> (tr, s) \<in> densem_exec lil S"
using assms
proof(induct c arbitrary: S tr s)
  case (one l i)
  show ?case
    using one
    by (metis densem_exec_empty append_Nil empty_set inf_sup_ord(4) list.simps(15) sc_inj.one set_append densem_exec_unfold)
next
  fix c1 c2 :: "('b, 'a) structured_code" and S tr s
  assume
    IH1: "\<And> S tr s . (tr, s) \<in> \<lbrakk>c1\<rbrakk>S \<Longrightarrow>
        \<exists>lil.
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c1 \<and>
           (tr, s) \<in> densem_exec lil S" and
    IH2: "\<And> S tr s . (tr, s) \<in> \<lbrakk>c2\<rbrakk>S \<Longrightarrow>
        \<exists>lil.
           set lil \<subseteq> \<union>\<^sub>s\<^sub>c c2 \<and>
           (tr, s) \<in> densem_exec lil S" and
    inD: "(tr, s) \<in> \<lbrakk>(c1 \<oplus> c2)\<rbrakk>S"
  show "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s) \<in> densem_exec lil S"
    using inD
    unfolding dseq
  proof (induct arbitrary: S rule: fixp_induct_t)
    show "ccpo.admissible Sup op \<le> (\<lambda>a. \<forall>x. (tr, s) \<in> a x \<longrightarrow> (\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s) \<in> densem_exec lil x))"
      unfolding "ccpo.admissible_def" chain_def
      by auto
  next
    fix x S
    assume
      IH: "\<And>S. (tr, s) \<in> x S \<Longrightarrow>
             \<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s) \<in> densem_exec lil S" and
      "(tr, s) \<in> func (c1 \<oplus> c2) x S "
    then consider (inc1) "(tr, s) \<in> x (\<lbrakk>c1\<rbrakk>S)" | (inc2) "(tr, s) \<in> x (\<lbrakk>c2\<rbrakk>S)" | (inS) "(tr, s) \<in> S"
      unfolding func by blast
    thus "\<exists>lil. set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s) \<in> densem_exec lil S"
    proof cases
      case inc1
      obtain lil where
        lilSubs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
        ex_lil: "(tr, s) \<in> densem_exec lil (\<lbrakk>c1\<rbrakk>S)"
        using IH[of "\<lbrakk>c1\<rbrakk>S"] inc1
        by blast
      then obtain tr' s' where
        ex_lil_pred: "(tr, s) \<in> densem_exec lil ({(tr', s')})" and
        inD: "(tr', s') \<in> \<lbrakk>c1\<rbrakk>S"
        using densem_exec_pred2[of tr s lil "\<lbrakk>c1\<rbrakk>S"]
        by fast
      then obtain lil' where
        subs: "set (lil @ lil') \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
        ex_lil': "(tr', s') \<in> densem_exec lil' S"
        using IH1[of tr' s' S] lilSubs
        by auto
      show ?thesis
        using densem_exec_fold[OF ex_lil_pred ex_lil'] subs
        by blast
    next
      case inc2
      obtain lil where
        lilSubs: "set lil \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
        ex_lil: "(tr, s) \<in> densem_exec lil (\<lbrakk>c2\<rbrakk>S)"
        using IH[of "\<lbrakk>c2\<rbrakk>S"] inc2
        by blast
      then obtain tr' s' where
        ex_lil_pred: "(tr, s) \<in> densem_exec lil ({(tr', s')})" and
        inD: "(tr', s') \<in> \<lbrakk>c2\<rbrakk>S"
        using densem_exec_pred2[of tr s lil "\<lbrakk>c2\<rbrakk>S"]
        by fast
      then obtain lil' where
        subs: "set (lil @ lil') \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2)" and
        ex_lil': "(tr', s') \<in> densem_exec lil' S"
        using IH2[of tr' s' S] lilSubs
        by auto
      show ?thesis
        using densem_exec_fold[OF ex_lil_pred ex_lil'] subs
        by blast
    next
      case inS
      hence "set [] \<subseteq> \<union>\<^sub>s\<^sub>c (c1 \<oplus> c2) \<and> (tr, s) \<in> densem_exec [] S"
        unfolding densem_exec_def
        by auto
      thus ?thesis
        by blast
    qed
  qed simp
qed

lemma densem_exec1': 
  assumes
    a: "x \<in> densem c S"
  shows "\<exists> lil . set lil \<subseteq> (\<union>\<^sub>s\<^sub>c c) \<and> x \<in> densem_exec lil S"
proof -
  obtain tr s where
    ao1: "x = (tr, s)"
    by fastforce
  show ?thesis
    unfolding ao1
    using densem_exec1[OF a[unfolded ao1]]
    .
qed

lemma densem_exec2: 
  assumes 
    subs: "set (lil) \<subseteq> (\<union>\<^sub>s\<^sub>c c)" and
    ex_lil: "(tr, s) \<in> densem_exec lil S"
  shows "(tr, s) \<in> densem c S"
proof (cases "(tr, s) \<in> S")
  case True
  thus ?thesis
    by (simp add: densem_super_id_element)
next
  case False
  have "lil \<noteq> []" using False ex_lil by fastforce
  then obtain l i right where lil: "lil = (l, i) # right" by (meson densem_exec_b.elims)
  then obtain right l i tr' s' where
    lil: "lil = [(l, i)] @ right" and
    d_one: "(tr, s) \<in> \<lbrakk>(one l i)\<rbrakk>{(tr', s')}" and
    ex_right: "(tr', s') \<in> densem_exec right S"
    using densem_exec_pred1 ex_lil by fastforce
  have "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<lbrakk>(one l i)\<rbrakk>{(tr', s')}"
    using densem_super_id_element [of "(tr, s)" "\<lbrakk>(one l i)\<rbrakk>{(tr', s')}" c, OF d_one] .
  hence "(tr, s) \<in> \<lbrakk>c\<rbrakk>{(tr', s')}"
    using densem_code_absorption_right_one[of l i c "{(tr', s')}"] assms lil
    by auto
  moreover
  have "set right \<subseteq> (\<union>\<^sub>s\<^sub>c c)"
    using assms lil
    by auto
  ultimately
  show "(tr, s) \<in> \<lbrakk>c\<rbrakk>S"
    using ex_right
  proof (induct right arbitrary: tr' s')
    case Nil
    hence "\<lbrakk>c\<rbrakk>{(tr', s')} \<subseteq> \<lbrakk>c\<rbrakk>S"
      by (metis (no_types) densem_union_additive Un_insert_right empty_subsetI insert_absorb densem_exec_empty sup.order_iff)
    then show ?case
      using Nil by blast
  next
    fix li' right tr' s' X'
    assume
      aa1: "\<And>tr' s'.
           (tr, s) \<in> \<lbrakk>c\<rbrakk>{(tr', s')} \<Longrightarrow>
           set right \<subseteq> \<union>\<^sub>s\<^sub>c c \<Longrightarrow>
           (tr', s') \<in> densem_exec right S \<Longrightarrow>
           (tr, s) \<in> \<lbrakk>c\<rbrakk>S" and
      aa2: "set (li' # right) \<subseteq> \<union>\<^sub>s\<^sub>c c" and
      aa3: "(tr, s) \<in> \<lbrakk>c\<rbrakk>{(tr', s')}" and
      aa4: "(tr', s') \<in> densem_exec (li' # right) S"
    obtain l' i' where ao1: "li' = (l', i')" by fastforce
    from aa4 have "(tr', s') \<in> \<lbrakk>(one l' i')\<rbrakk>densem_exec right S"
      using densem_exec_unfold[of l' i' right S] ao1
      by blast
    then obtain tr'' s'' where
      ao2: "(tr'', s'') \<in> densem_exec right S \<and> (tr', s') \<in> \<lbrakk>(one l' i')\<rbrakk>{(tr'', s'')}"
      using densem_predecessor[of "(tr', s')" "one l' i'" "densem_exec right S"] by fast
    from ao2 have "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<lbrakk>(one l' i')\<rbrakk>{(tr'', s'')}"
      by (metis (no_types, lifting) Set.set_insert UnI1 aa3 insert_is_Un densem_union_additive)
    moreover
    from aa2 have "set (right) \<subseteq> \<union>\<^sub>s\<^sub>c c \<and> {li'} \<subseteq> \<union>\<^sub>s\<^sub>c c"
      by auto
    moreover
    have "(tr, s) \<in> \<lbrakk>c\<rbrakk>{(tr'', s'')}"
      using aa2 ao1 calculation
      by (simp add: densem_code_absorption_right_one)
    ultimately show "(tr, s) \<in> \<lbrakk>c\<rbrakk>S"
      using aa1 ao2 by auto
  qed
qed

lemma sfsem_exec2': 
  assumes 
    a: "set (lil) \<subseteq> (\<union>\<^sub>s\<^sub>c c)" and
    b: "x \<in> densem_exec lil S"
  shows "x \<in> densem c S"
proof -
  obtain tr s where
    ao1: "x = (tr, s)"
    by fastforce
  have a1: "snd (tr, s) = s"
    by auto
  show ?thesis
    using assms
    unfolding ao1 a1
    using densem_exec2[of lil c tr s S]
    by blast
qed

fun densem_exec_path_is_critical :: "('e trace \<times> 's state) \<Rightarrow> (label \<times> ('s, 'e) instr) list \<Rightarrow> ('s, 'e) trace_state set \<Rightarrow> bool" where
  "densem_exec_path_is_critical x [] S = (x \<in> S)" |
  "densem_exec_path_is_critical x ((l, i) # lil) S = (x \<notin> densem_exec lil S \<and> (\<exists> x' . x \<noteq> x' \<and> x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> densem_exec_path_is_critical x' lil S))"

lemma list_append_singleton:
  fixes x :: 'a and xs :: "'a list"
  shows "[x] @ xs = x # xs"
by simp

lemma densem_exec_critical_path:
  assumes
    a: "x \<in> densem_exec lil S"
  shows "\<exists> lil' . set lil' \<subseteq> set lil \<and> length lil' \<le> length lil \<and> densem_exec_path_is_critical x lil' S"
  using a
proof (induct lil arbitrary: x rule: length_induct)
  fix lil x
  assume
    aa1: "\<forall>lil'. length lil' < length lil \<longrightarrow>
            (\<forall>x. x \<in> densem_exec lil' S \<longrightarrow>
                 (\<exists>lil''.
                     set lil'' \<subseteq> set lil' \<and>
                     length lil'' \<le> length lil' \<and>
                     densem_exec_path_is_critical x
                      lil'' S))" and
    aa2: "x \<in> densem_exec lil S"
  show "\<exists>lil'.
          set lil' \<subseteq> set lil \<and>
          length lil' \<le> length lil \<and>
          densem_exec_path_is_critical x lil' S" (is ?A)
  proof (cases "lil = []")
    assume ba1: "lil = []"
    thus ?thesis
      using aa2[unfolded ba1, unfolded densem_exec_empty]
      by simp
  next
    assume ba1: "lil \<noteq> []"
    obtain l i right where
      bo1: "lil = [(l, i)] @ right"
      using ba1
      by (metis append_Cons append_Nil neq_Nil_conv prod.collapse)
    obtain x' where
      bo2: "x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> x' \<in> densem_exec right S"
      using densem_predecessor[of "x" "one l i" "densem_exec right S", OF aa2[unfolded bo1, unfolded list_append_singleton, unfolded densem_exec_unfold]]
      by blast
    have b1: "length right < length lil \<and> set right \<subseteq> set lil"
      using bo1
      by auto
    obtain rightC where
      bo3: "set rightC \<subseteq> set right \<and> length rightC \<le> length right \<and> densem_exec_path_is_critical x' rightC S"
      using aa1 b1 bo2
      by blast
    consider (ex_right) "x \<in> densem_exec right S" | (ex_rightC) "x \<in> densem_exec rightC S" | (no_ex) "x \<notin> densem_exec right S \<and> x \<notin> densem_exec rightC S"
      by blast
    thus ?thesis
    proof cases
      case ex_right
      thus ?thesis
        by (meson aa1 b1 dual_order.strict_implies_order order_trans)
    next
      case ex_rightC
      thus ?thesis
        by (meson aa1 b1 bo3 dual_order.strict_trans2 order.strict_implies_order order_trans)
    next
      case no_ex
      hence d1: "x \<noteq> x'"
        using bo2
        by blast
      have "(x \<noteq> x' \<and> x \<in> \<lbrakk>one l i\<rbrakk>{x'} \<and> densem_exec_path_is_critical x' rightC S)"
        using bo3 d1 bo2
        by simp
      hence "densem_exec_path_is_critical x ((l, i) # rightC) S"
        using densem_exec_path_is_critical.simps(2)[symmetric, of x rightC S l i] no_ex
        by blast
      moreover
      have "set ((l, i) # rightC) \<subseteq> set lil \<and> length ((l, i) # rightC) \<le> length lil"
        using bo1 bo3
        by auto
      ultimately
      show ?thesis
        by blast
    qed
  qed
qed

lemma densem_exec_critical_path_unfold_single:
  assumes
    "densem_exec_path_is_critical x (li # lil) S"
  shows
    "\<exists> x' . densem_exec_path_is_critical x [li] {x'} \<and> densem_exec_path_is_critical x' lil S"
proof -
  obtain l i where
    ao1: "li = (l, i)"
    by fastforce
  from assms show ?thesis
    unfolding densem_exec_path_is_critical.simps ao1
    using densem_exec_empty
    by blast
qed

lemma densem_exec_is_critical_is_exec:
  assumes "densem_exec_path_is_critical x lil S"
  shows "x \<in> densem_exec lil S"
using assms
proof (induct lil arbitrary: x)
  show "\<And>x. densem_exec_path_is_critical x [] S \<Longrightarrow>
         x \<in> densem_exec [] S"
    unfolding densem_exec_empty densem_exec_path_is_critical.simps .
next
  fix a lil x
  assume
    aa1: "\<And>x. densem_exec_path_is_critical x lil S \<Longrightarrow>
             x \<in> densem_exec lil S" and
    aa2: "densem_exec_path_is_critical x (a # lil) S"
  obtain l i where
    ao1: "a = (l, i)"
    by fastforce
  obtain x' where
    ao2: "densem_exec_path_is_critical x [a] {x'} \<and>
       densem_exec_path_is_critical x' lil S"
    using densem_exec_critical_path_unfold_single[OF aa2]
    by blast
  have a1: "x' \<in> densem_exec lil S"
    using aa1[OF ao2[THEN conjunct2]]
    .
  have a2: "x \<in> \<lbrakk>(one l i)\<rbrakk>{x'}"
    using ao2[THEN conjunct1, unfolded ao1 densem_exec_path_is_critical.simps]
    by blast
  have a3: "x \<in> \<lbrakk>(one l i)\<rbrakk>densem_exec lil S"
    using densem_subset_single[OF a1 a2] .
  show "x \<in> densem_exec (a # lil) S"
    using a3[unfolded densem_exec_unfold[symmetric] ao1[symmetric]]
    .
qed

lemma densem_exec_critical_path_unfold:
  assumes
    "densem_exec_path_is_critical x (l @ r) S"
  shows
    "\<exists> x' . densem_exec_path_is_critical x l {x'} \<and> densem_exec_path_is_critical x' r S"
  using assms
proof (induct l arbitrary: x, simp)
  fix a l x
  assume
    aa1: "\<And>x. densem_exec_path_is_critical x (l @ r) S \<Longrightarrow>
             \<exists>x'. densem_exec_path_is_critical x l {x'} 
                 \<and> densem_exec_path_is_critical x' r S" and
    aa2: "densem_exec_path_is_critical x ((a # l) @ r) S"
  from aa2 have aa2': "densem_exec_path_is_critical x (a # (l @ r)) S"
    by simp
  obtain x' where
    ao1: "densem_exec_path_is_critical x [a] {x'} \<and>
       densem_exec_path_is_critical x' (l @ r) S"
    using densem_exec_critical_path_unfold_single[OF aa2']
    by blast
  obtain x'' where
    ao2: "densem_exec_path_is_critical x' l {x''} \<and>
     densem_exec_path_is_critical x'' r S"
    using aa1[OF ao1[THEN conjunct2]]
    by blast
  obtain lbl i where
    ao3: "a = (lbl, i)"
    by fastforce
  have a1: "x \<notin> densem_exec (l @ r) S"
    using aa2'[unfolded ao3 densem_exec_path_is_critical.simps, THEN conjunct1]
    .
  hence "x \<notin> densem_exec l (densem_exec r S)"
    unfolding densem_exec_unfold2[of l r S] .
  hence a2: "\<forall> x' \<in> densem_exec r S . x \<notin> densem_exec l {x'}"
    using densem_exec_additive[of l] UnCI[of x "densem_exec l _"] insert_is_Un mk_disjoint_insert
    by (metis (no_types, lifting))
  hence a3: "x \<notin> densem_exec l {x''}"
    using densem_exec_is_critical_is_exec[OF ao2[THEN conjunct2]]
    by blast
  have a4: "x \<noteq> x' \<and> x \<in> \<lbrakk>(one lbl i)\<rbrakk>{x'} \<and> densem_exec_path_is_critical x' l {x''}"
    using ao1[unfolded ao3 densem_exec_path_is_critical.simps densem_exec_empty]
          ao2[THEN conjunct1]
    by blast
  have a5: "densem_exec_path_is_critical x (a # l) {x''}"
    unfolding ao3 densem_exec_path_is_critical.simps
    using a3 a4
    by blast
  show "\<exists>x'. densem_exec_path_is_critical x (a # l) {x'} \<and>
             densem_exec_path_is_critical x' r S"
    using a5 ao2[THEN conjunct2]
    by blast
qed

end