section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory "denotational_parallel"
imports denotational parallel_base trace_restriction
begin

subsection {* Preparations *}

paragraph {* Datatypes *}

text {* First we need to introduce datatypes to handle the tree structure of parallel code.  *}

type_synonym ('s)parallelState = "'s state ATree"

type_synonym ('s, 'e) parallelTraceState = "'e definitions.trace \<times> 's parallelState"

text {* For every sequential state, we want it to be normal. *}

definition AllPC\<^sub>T :: "'s parallelState \<Rightarrow> nat \<Rightarrow> bool" where
"AllPC\<^sub>T ss lbl = leaf1_property (\<lambda> s . PC s = lbl) op \<and> ss"

lemma AllPC\<^sub>T_inherit: "AllPC\<^sub>T (par sl sr) lbl = (AllPC\<^sub>T sl lbl \<and> AllPC\<^sub>T sr lbl)"
  using leaf1_property_inherit AllPC\<^sub>T_def by blast

paragraph {* Transformations between the two parallel state worlds. *}

fun parallelTraceStateToTraceState :: "('s, 'e) parallelTraceState \<Rightarrow> ('s, 'e) trace_state set" where
"parallelTraceStateToTraceState (tr, seq s) = {(tr, s)}" |
"parallelTraceStateToTraceState (tr, par _ _) = {}"

definition parallelTraceStateSetToTraceStateSet :: "('s, 'e) parallelTraceState set \<Rightarrow> ('s, 'e) trace_state set" where
"parallelTraceStateSetToTraceStateSet S = {x . \<exists> x' . x' \<in> S \<and> x \<in> parallelTraceStateToTraceState x'}"

lemma parallelTraceStateToTraceState_ex_seq: "(tr, s) \<in> parallelTraceStateToTraceState (tr', s') \<Longrightarrow> (tr, seq s) = (tr', s')"
  by (metis ATree.exhaust empty_iff fst_conv insert_iff parallelTraceStateToTraceState.simps(1) parallelTraceStateToTraceState.simps(2) snd_conv)

lemma parallelTraceStateSingletonToTraceStateSingleton:
 "parallelTraceStateSetToTraceStateSet {(tr, s)} \<noteq> {} \<Longrightarrow> \<exists> s' . {(tr, s')} = parallelTraceStateSetToTraceStateSet {(tr, s)} \<and> s = seq s'"
  unfolding parallelTraceStateSetToTraceStateSet_def
proof (induct s, auto) qed

definition trace_state_set_to_parallel_trace_state_set :: "('s, 'e) trace_state set \<Rightarrow> ('s, 'e) parallelTraceState set" where
"trace_state_set_to_parallel_trace_state_set S = {(tr, s) . \<exists> s' . (tr, s') \<in> S \<and> s = seq s'}"

subsection {* The parallel semantics *}

fun densem\<^sub>p :: "('s, 'e) structured_code_parallel \<Rightarrow>
              ('s, 'e) parallelTraceState set \<Rightarrow>
              ('s, 'e) parallelTraceState set"(*<*)("\<lbrakk>_\<rbrakk>\<^sub>p\<^sub>T(_)")(*>*) where
densemp_par: "(\<lbrakk>parn c1 (\<alpha>, \<beta>) c2\<rbrakk>\<^sub>p\<^sub>T (S)) = {(tr, t12) .
  \<exists> t1 t2 s1 s2 s12. ([], s12) \<in> S \<and>
  (tr \<restriction> \<alpha>, t1) \<in> densem\<^sub>p c1 {([], s1)} \<and>
  (tr \<restriction> \<beta>, t2) \<in> densem\<^sub>p c2 {([], s2)} \<and>
  set tr \<subseteq> (\<alpha> \<union> \<beta>) \<and>
  t12 = par t1 t2 \<and>
  s12 = par s1 s2
}" |
densemp_seq: "(\<lbrakk>seqn (c, _)\<rbrakk>\<^sub>p\<^sub>T(S)) = trace_state_set_to_parallel_trace_state_set (densem c (parallelTraceStateSetToTraceStateSet S))"

lemma sensemp_seq_unfold: "((tr, seq s) \<in> (\<lbrakk>seqn (c, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>T{(tr', seq s')})) = ((tr, s) \<in> (densem c {(tr',  s')}))"
  unfolding densemp_seq trace_state_set_to_parallel_trace_state_set_def parallelTraceStateSetToTraceStateSet_def
  using parallelTraceStateToTraceState.simps
  by simp

text {* We can add arbitrary many other states to the besides the known predecessor state. *}

lemma densemp_monotone_single:
  fixes
    X X' :: "'e definitions.event set" and
    s s' :: "'s parallelState" and
    tr tr' :: "'e list"
  assumes
    a: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T({(tr', s')})" and
    b: "(tr', s') \<in> S" and
    wff: "wff\<^sub>p c"
  shows "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T(S)"
using a b wff
proof (induct c rule: structured_code_parallel_induct, simp_all add: wff\<^sub>p_def)
  fix x :: "('s, 'e) structured_code"
  assume
    aa1: "(tr, s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>x\<rbrakk>parallelTraceStateSetToTraceStateSet {(tr', s')})" and
    aa2: "(tr', s') \<in> S" and
    wff: "wff\<^sub>s\<^sub>c x \<and> liSet (\<union>\<^sub>s\<^sub>c x)"
  have subs: "parallelTraceStateSetToTraceStateSet {(tr', s')} \<subseteq> parallelTraceStateSetToTraceStateSet S"
    unfolding parallelTraceStateSetToTraceStateSet_def
    using aa2
    by blast
  moreover have "\<forall> S T . S \<subseteq> T \<longrightarrow> trace_state_set_to_parallel_trace_state_set S \<subseteq> trace_state_set_to_parallel_trace_state_set T"
    unfolding trace_state_set_to_parallel_trace_state_set_def
    by fast
  ultimately show "(tr, s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>x\<rbrakk>parallelTraceStateSetToTraceStateSet S)"
    using aa1 densem_monotone[OF subs]
    by blast
qed blast

lemma densemp_predecessor:
  fixes
    X X' :: "'e definitions.event set" and
    s s' :: "'s parallelState"
  assumes
    a: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T(S)" and
    wff: "wff\<^sub>p c"
  shows "\<exists> tr' s'. (tr', s') \<in> S \<and> (tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T({(tr', s')}) \<and> treen_mirror s' c \<and> treen_mirror s c \<and> tree_mirror s s'"
using a wff
proof (induct c arbitrary: tr s S rule: structured_code_parallel_induct, simp_all add: wff\<^sub>p_def)
  fix ca tr s S
  assume
    inside: "(tr, s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>ca\<rbrakk>parallelTraceStateSetToTraceStateSet S)" and
    wff: "wff\<^sub>s\<^sub>c ca \<and> liSet (\<union>\<^sub>s\<^sub>c ca)"
  obtain s' where
    inside: "(tr, s') \<in> densem ca {x. \<exists>x'. x' \<in> S \<and> x \<in> parallelTraceStateToTraceState x'}" and
    s: "s = ATree.seq s'"
    using inside
    unfolding trace_state_set_to_parallel_trace_state_set_def parallelTraceStateSetToTraceStateSet_def
    by fast
  obtain trp sp where
    inside_single: "(tr, s') \<in> densem ca {(trp, sp)}" and
    inS: "(trp, sp) \<in> {x. \<exists>x'. x' \<in> S \<and> x \<in> parallelTraceStateToTraceState x'}"
    using densem_predecessor[OF inside] by fastforce
  then obtain trs ss where
    p1: "(trp, sp) \<in> {x. \<exists>x'. x' \<in> {(trs, ss)} \<and> x \<in> parallelTraceStateToTraceState x'}" and
    pInS: "(trs, ss) \<in> S"
    using inS by auto
  have inside: "(tr, s') \<in> densem ca {x. \<exists>x'. x' \<in> {(trs, ss)} \<and> x \<in> parallelTraceStateToTraceState x'}"
    using p1 inside_single
    by (simp add: densem_subset_single)
  have "\<exists>s. ss = ATree.seq s"
    using p1
    by (metis (full_types) empty_iff parallelTraceStateSetToTraceStateSet_def parallelTraceStateSingletonToTraceStateSingleton)
  thus "\<exists>tr' s'. (tr', s') \<in> S \<and> (tr, s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>ca\<rbrakk>parallelTraceStateSetToTraceStateSet {(tr', s')}) \<and> (\<exists>s. s' = ATree.seq s) \<and> (\<exists>s'. s = ATree.seq s') \<and> tree_mirror s s'"
    using inside pInS p1 s
    unfolding trace_state_set_to_parallel_trace_state_set_def parallelTraceStateSetToTraceStateSet_def
    by auto
next
  fix cl \<alpha> \<beta> cr tra sa Xa Sa
  assume
    IH1: "\<And>tr s S. (tr, s) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>TS \<Longrightarrow> \<exists>tr' s'. (tr', s') \<in> S \<and> (tr, s) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{(tr', s')} \<and> treen_mirror s' cl \<and> treen_mirror s cl \<and> tree_mirror s s'" and
    IH2: "\<And>tr s S. (tr, s) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>TS \<Longrightarrow> \<exists>tr' s'. (tr', s') \<in> S \<and> (tr, s) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{(tr', s')} \<and> treen_mirror s' cr \<and> treen_mirror s cr \<and> tree_mirror s s'" and
    base: "\<exists>t1 t2 s1 s2. ([], par s1 s2) \<in> Sa \<and> (tra \<restriction> \<alpha>, t1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{([], s1)} \<and> (tra \<restriction> \<beta>, t2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{([], s2)} \<and> set tra \<subseteq> \<alpha> \<union> \<beta> \<and> sa = par t1 t2"
  obtain t1 t2 s1 s2 where
    "([], par s1 s2) \<in> Sa" and
    ex_l: "(tra \<restriction> \<alpha>, t1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{([], s1)}" and
    ex_r: "(tra \<restriction> \<beta>, t2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{([], s2)}" and
    settra: "set tra \<subseteq> \<alpha> \<union> \<beta>" and
    sa: "sa = par t1 t2"
    using base by blast
  moreover
  have "treen_mirror (par s1 s2) (parn cl (\<alpha>, \<beta>) cr) \<and> treen_mirror (sa) (parn cl (\<alpha>, \<beta>) cr) \<and> tree_mirror (sa) (par s1 s2)"
    using IH1[OF ex_l, simplified] IH2[OF ex_r, simplified] sa
    by auto
  ultimately
  show "\<exists>s'. ([], s') \<in> Sa \<and>
            (\<exists>t1 t2 s1. (tra \<restriction> \<alpha>, t1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{([], s1)} \<and> (\<exists>s2. (tra \<restriction> \<beta>, t2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{([], s2)} \<and> set tra \<subseteq> \<alpha> \<union> \<beta> \<and> sa = par t1 t2 \<and> s' = par s1 s2)) \<and>
            treen_mirror s' (parn cl (\<alpha>, \<beta>) cr) \<and> treen_mirror sa (parn cl (\<alpha>, \<beta>) cr) \<and> tree_mirror sa s'"
    by blast
qed

lemma densemp_monotone:
  assumes subs: "S \<subseteq> T" and wff: "wff\<^sub>p c"
  shows "\<lbrakk>c\<rbrakk>\<^sub>p\<^sub>TS \<subseteq> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>TT"
proof
  fix x
  assume inside: "x \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>TS"
  obtain tr s where x: "x = (tr, s)" by fastforce
  obtain tr' s' where
    ex: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{(tr', s')}" and
    inS: "(tr', s') \<in> S"
    using densemp_predecessor[OF inside[unfolded x] wff]
    by blast
  have inT: "(tr', s') \<in> T" using subs inS by blast
  thus "x \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>TT"
    using densemp_monotone_single ex wff
    unfolding x
    by blast
qed

lemma densemp_starting_set_can_contain_malformed_states:
  fixes
    c :: "('s, 'e) structured_code_parallel"
  assumes
    wff: "wff\<^sub>p c"
  shows "\<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x' \<and> treen_mirror (snd x') c} = \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x'}"
proof -
  {
    fix tr s
    assume inside: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x' \<and> treen_mirror (snd x') c}"
    hence "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x'}"
      using densemp_monotone[of "{x'. Pre x' \<and> treen_mirror (snd x') c}" "{x'. Pre x'}" c] wff
      by blast
  } moreover {
    fix tr s X
    assume inside: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x'}"
    obtain tr' s' where
      ex: "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{(tr', s')}" and
      inS: "(tr', s') \<in> {x'. Pre x'}" and
      tm: "treen_mirror s' c"
      using densemp_predecessor[OF inside wff]
      by blast
    hence start: "(tr', s') \<in> {x'. Pre x' \<and> treen_mirror (snd x') c}"
      by simp
    have "(tr, s) \<in> \<lbrakk>c\<rbrakk>\<^sub>p\<^sub>T{x'. Pre x' \<and> treen_mirror (snd x') c}"
      using densemp_monotone_single[OF ex start wff] .
  } ultimately show ?thesis by fast
qed

paragraph {* Preconditions *}

text {* The precondition for start states - we want them all to start at a welldefined pc, be normal
and have an empty trace. *}

definition "PreStd\<^sub>T\<^sub>s\<^sub>e\<^sub>q x = (case x of (tr, s) \<Rightarrow> PC s = 1 \<and> tr = [])"
definition "PreStd\<^sub>T x = (case x of (tr, s) \<Rightarrow> AllPC\<^sub>T s 1 \<and> tr = [])"

text {* Converts a predicate over a non-parallel stable failure to a parallel stable failure, considering only the cases where
the state is a single state.*}

definition "CV\<^sub>P\<^sub>r\<^sub>e\<^sub>T Pre = (\<lambda> x . case x of (tr, seq s') \<Rightarrow> Pre (tr, s') | _ \<Rightarrow> False)"

end
