section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory parallel_base
imports small_step tree static_code_analysis
begin

text {* First we need to introduce datatypes to handle the tree structure of parallel code.  *}

type_synonym ('s, 'e)structured_code_parallel = "(('s, 'e)structured_code \<times> 'e definitions.event set, 'e definitions.event set \<times> 'e definitions.event set) ATreeNode"

text {* A slightly modified induction scheme with already expanded variables. *}

lemma structured_code_parallel_induct: "(\<And> c \<Sigma> . P (seqn (c, \<Sigma>))) \<Longrightarrow>
(\<And>cl \<alpha> \<beta> cr. P cl \<Longrightarrow> P cr \<Longrightarrow> P (parn cl (\<alpha>, \<beta>) cr)) \<Longrightarrow>
P ATreeNode"
proof (induct ATreeNode, auto) qed

paragraph {* Functions describing state and code characteristics *}

text {* In here we gather all labeled instructions, thus losing the tree structure of the parallel code. *}

definition scp_inj :: " ('s, 'e)structured_code_parallel \<Rightarrow> ('s, 'e)labeled_instruction_set"  ("\<union>\<^sub>p\<^sub>s\<^sub>c") where
"scp_inj cc = (ATreeNode_mapfoldl cc (\<lambda> cle . \<union>\<^sub>s\<^sub>c (fst cle)) (\<lambda> l r n . l \<union> r))"

text {* We only do care for wellformedness in the sequential part, as parallel code can
 mean executing the very same code in parallel, which would not work with a definition
 equal to @{term "wff\<^sub>s\<^sub>c"}. Also, we include correctness of the lis in here. *}
definition wff\<^sub>p :: "('s, 'e) structured_code_parallel \<Rightarrow> bool" where
"(wff\<^sub>p cc) = (leafn1_property (\<lambda> (c, _) . wff\<^sub>s\<^sub>c c \<and> liSet (\<union>\<^sub>s\<^sub>c c)) op \<and> cc)"

lemma wff\<^sub>p_inherit: "wff\<^sub>p (parn cl \<alpha>\<beta> cr) = (wff\<^sub>p cl \<and> wff\<^sub>p cr)"
  using leafn1_property_inherit[of wff\<^sub>p] wff\<^sub>p_def by blast

definition wff_p_sigma :: "('s, 'e) structured_code_parallel \<Rightarrow> 'e set \<Rightarrow> bool" where
"wff_p_sigma cc \<Sigma> = (leafn1_property (\<lambda> (_, \<Sigma>') . \<Sigma> = \<Sigma>') op \<and> cc \<and> node1_property (\<lambda> (\<alpha>, \<beta>) . \<alpha> \<subseteq> \<Sigma> \<and> \<beta> \<subseteq> \<Sigma>) cc)"

lemma wff_p_sigma_inherit: "wff_p_sigma (parn l (\<alpha>, \<beta>) r) \<Sigma> \<Longrightarrow> (wff_p_sigma l \<Sigma> \<and> wff_p_sigma r \<Sigma>)"
  unfolding wff_p_sigma_def by auto

text {* The refusal sets of a child are always a subset of the respective interface's refusal set. *}
definition wff_p_refusal_subsets :: "('s, 'e) structured_code_parallel \<Rightarrow> bool" where
"wff_p_refusal_subsets cc = (\<forall> lrns lf . (lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N cc \<longrightarrow>
  fst (foldr (\<lambda> (lr, (\<alpha>, \<beta>)) (tf, child) . (tf \<and> (lr = Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) lrns (True, {})))"

lemma wff_p_refusal_subsets_inherit: "wff_p_refusal_subsets (parn l n r) \<Longrightarrow> (wff_p_refusal_subsets l \<and> wff_p_refusal_subsets r)"
  unfolding wff_p_refusal_subsets_def
proof (auto)
  fix lrns a b
  assume
    IH: "\<forall>lrns. (\<exists>a b. (lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r)) \<longrightarrow>
                          fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) lrns (True, {}))" and
    inl: "(lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l"
  obtain \<alpha> \<beta> where n: "n = (\<alpha>, \<beta>)" by fastforce
  have "((Left, n) # lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r)"
    by (simp add: ATreeNode_gather_left inl)
  hence g: "fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) ((Left, n) # lrns) (True, {}))"
    using IH by blast
  show "fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) lrns (True, {}))"
    using g[unfolded n foldr.simps, simplified]
    by (metis (no_types, lifting) case_prod_conv prod.collapse prod.inject)
next
  fix lrns a b
  assume
    IH: "\<forall>lrns. (\<exists>a b. (lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r)) \<longrightarrow>
                          fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) lrns (True, {}))" and
    inr: "(lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r"
  obtain \<alpha> \<beta> where n: "n = (\<alpha>, \<beta>)" by fastforce
  have "((Right, n) # lrns, a, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r)"
    by (simp add: ATreeNode_gather_right inr)
  hence g: "fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) ((Right, n) # lrns) (True, {}))"
    using IH by blast
  show "fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) lrns (True, {}))"
    using g[unfolded n foldr.simps, simplified]
    by (metis (no_types, lifting) case_prod_conv prod.collapse prod.inject)
qed

lemma (* In general auto provable. *)
  assumes cc: "cc = parn (parn (seqn a) ({1,2},{2,3}) (parn (seqn b) ({2}, {3}) (seqn c))) ({1,2,3}, {2,3,4}) (parn (seqn a) ({2,3},{3,4}) (parn (seqn b) ({3}, {4}) (seqn c)))"
  shows "wff_p_refusal_subsets cc"
unfolding cc wff_p_refusal_subsets_def ATreeNode_gather_def ATreeNode_gather_hlpr.simps append_Nil by fastforce

lemma wff_p_refusal_subsets_useable:
  assumes
    subs: "fst (foldr (\<lambda> (lr, (\<alpha>, \<beta>)) (tf, child) . (tf \<and> (lr = Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) (l @ r) (True, {}))" and
    parent: "(lr, \<alpha>, \<beta>) \<in> set l" and
    child: "(lr', \<alpha>', \<beta>') \<in> set r"
  shows "case lr of Left \<Rightarrow> (\<alpha>' \<union> \<beta>') \<subseteq> \<alpha>
                  | Right \<Rightarrow> (\<alpha>' \<union> \<beta>') \<subseteq> \<beta>"
    using assms
proof (induct "length (l @ r)" arbitrary: lr \<alpha> \<beta> l r)
  case 0
  thus ?case by simp
next
  case (Suc x)
  obtain lru au bu where a: "hd (l @ r) = (lru, au, bu)" using prod_cases3 by blast
  define fr where fr: "fr = foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) (tl (l @ r)) (True, {})"
  define lt where lt: "lt = (l @ r)"
  have eq: "foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) (tl (l @ r)) (True, {}) =
    (fst fr,
     snd fr)" using fr by simp
  have eq2: "(fst (case (fst fr, snd fr) of (tf, child) \<Rightarrow> (tf \<and> (lru = NodeAccessor.Left \<longrightarrow> child \<subseteq> au) \<and> (lru = NodeAccessor.Right \<longrightarrow> child \<subseteq> bu), au \<union> bu))) =
             (fst (((fst fr) \<and> (lru = NodeAccessor.Left \<longrightarrow> (snd fr) \<subseteq> au) \<and> (lru = NodeAccessor.Right \<longrightarrow> (snd fr) \<subseteq> bu), au \<union> bu)))" by fast
  have eq3: "(l @ r) = (hd lt) # (tl lt)" using Suc(2) lt
    by (metis length_0_conv list.exhaust_sel nat.simps(3))
  have g: "fst (foldr (\<lambda>a. case a of (lr, a) \<Rightarrow> case a of (\<alpha>, \<beta>) \<Rightarrow> \<lambda>a. case a of (tf, child) \<Rightarrow> (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) (tl (l @ r)) (True, {}))"
    using Suc.prems(1)
    unfolding foldr.simps List.append.simps(2) eq3
    unfolding o_apply prod.collapse prod.inject prod.simps(2) a lt
    unfolding eq eq2 fst_conv 
    unfolding fr
    by auto
  obtain hd_l tl_l where l: "l = hd_l # tl_l" using Suc(4)
    by (meson list.set_cases)
  have tl_eq: "tl (l @ r) = tl_l @ r" unfolding l by simp
  have tl_eq2: "tl l = tl_l" using l by simp
  consider (in_tl) "(lr, \<alpha>, \<beta>) \<in> set (tl_l)" | (eq_hd) "(lr, \<alpha>, \<beta>) = hd (l)" using Suc(2) Suc(4) unfolding tl_eq2[symmetric]
    by (metis Nitpick.size_list_simp(2) hd_Cons_tl length_pos_if_in_set neq0_conv set_ConsD)
  thus ?case
  proof cases
    case in_tl
    show ?thesis
      using Suc(1)[OF _ g[unfolded tl_eq] in_tl Suc(5)] Suc(2) l
      by auto
  next
    case eq_hd
    hence hd_l: "hd_l = (lr, \<alpha>, \<beta>)"
      by (simp add: l)
    obtain lru au bu where a: "hd (tl_l @ r) = (lru, au, bu)" using prod_cases3 by blast
    obtain r' where r'eq: "tl_l @ r = (lru, au, bu) # r'"
      by (metis Nil_is_append_conv Suc.prems(3) a length_pos_if_in_set less_numeral_extra(3) list.exhaust_sel list.size(3))
    hence r'eq2: "tl_l @ r = [(lru, au, bu)] @ r'" by auto
    have x: "x = length ([(lru, au, bu)] @ r')"
      by (metis Suc.hyps(2) eq3 length_Cons lt nat.inject r'eq2 tl_eq)
    have g2: "fst (foldr (\<lambda>(lr, \<alpha>, \<beta>) (tf, child). (tf \<and> (lr = NodeAccessor.Left \<longrightarrow> child \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> child \<subseteq> \<beta>), \<alpha> \<union> \<beta>)) ([(lru, au, bu)] @ r') (True, {}))"
      using g[unfolded tl_eq r'eq] by force
    consider (in_r') "(lr', \<alpha>', \<beta>') \<in> set r'" | (not_in_r') "(lr', \<alpha>', \<beta>') \<notin> set r'" by blast
    thus ?thesis
    proof cases
      case in_r'
      have "case lru of NodeAccessor.Left \<Rightarrow> \<alpha>' \<union> \<beta>' \<subseteq> au | NodeAccessor.Right \<Rightarrow> \<alpha>' \<union> \<beta>' \<subseteq> bu"
        using Suc.hyps(1)[OF x g2, of lru au bu, simplified] Suc(5)
        by (metis (no_types, lifting) Suc.hyps(1) a g2 hd_append hd_in_set r'eq2 in_r' x)
      hence "\<alpha>' \<union> \<beta>' \<subseteq> au \<union> bu" using NodeAccessor.exhaust
        by (metis (full_types) NodeAccessor.simps(3) NodeAccessor.simps(4) sup.coboundedI1 sup.coboundedI2)
      moreover have "(lr = NodeAccessor.Left \<longrightarrow> au \<union> bu \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> au \<union> bu \<subseteq> \<beta>)"
        using Suc(3)[unfolded l hd_l]
        unfolding foldr.simps List.append.simps(2) eq3
        unfolding o_apply prod.collapse prod.inject prod.simps(2) a lt
        unfolding eq eq2 fst_conv 
        unfolding fr r'eq
        unfolding foldr.simps List.append.simps(2) eq3
        unfolding o_apply prod.collapse prod.inject prod.simps(2) a lt
        unfolding eq eq2 fst_conv
        by (simp add: case_prod_unfold)
      ultimately show ?thesis using NodeAccessor.exhaust
        by (metis (mono_tags, hide_lams) NodeAccessor.simps(3) NodeAccessor.simps(4) subset_trans)
    next
      case not_in_r'
      have "(lru, au, bu) = (lr', \<alpha>', \<beta>')" 
      proof -
        have f1: "\<forall>p ps. (p::NodeAccessor \<times> 'a set \<times> 'a set) \<notin> set ps \<or> (\<exists>psa. ps = p # psa) \<or> (\<exists>pa psa. ps = pa # psa \<and> p \<in> set psa)"
          by (meson list.set_cases)
        have "\<And>ps psa. (lr', \<alpha>', \<beta>') \<in> set ps \<or> psa @ r \<noteq> ps"
          using Suc.prems(3) by force
        then show ?thesis
          using f1 by (metis (no_types) list.sel(1) list.sel(3) not_in_r' r'eq)
      qed
      moreover have "(lr = NodeAccessor.Left \<longrightarrow> au \<union> bu \<subseteq> \<alpha>) \<and> (lr = NodeAccessor.Right \<longrightarrow> au \<union> bu \<subseteq> \<beta>)"
        using Suc(3)[unfolded l hd_l]
        unfolding foldr.simps List.append.simps(2) eq3
        unfolding o_apply prod.collapse prod.inject prod.simps(2) a lt
        unfolding eq eq2 fst_conv 
        unfolding fr r'eq
        unfolding foldr.simps List.append.simps(2) eq3
        unfolding o_apply prod.collapse prod.inject prod.simps(2) a lt
        unfolding eq eq2 fst_conv
        by (simp add: case_prod_unfold)
      ultimately show ?thesis using NodeAccessor.exhaust
        by (metis (mono_tags, lifting) NodeAccessor.simps(3) NodeAccessor.simps(4) Pair_inject)
    qed
  qed
qed

end
