section {* Parallel SF semantics *}

text{* This section introduces parallel concepts of the stable failure semantics and ultimately exercises it on a parallel buffer example. *}

theory "stable_failures_traces_conformance_parallel"
imports "stable_failures_traces_conformant" stable_failures_parallel denotational_parallel
begin

subsection {* Preparations *}

paragraph {* Transformations between the two parallel state worlds. *}

definition PS2PTS :: "'a stable_failures_parallel.parallelState \<Rightarrow> 'a denotational_parallel.parallelState" where
"PS2PTS tree = ATree_map tree (\<lambda> s . NCS s)"

definition PTS2PS :: "'a denotational_parallel.parallelState \<Rightarrow> 'a stable_failures_parallel.parallelState" where
"PTS2PS tree = ATree_map tree (\<lambda> s . stable_failures.normal s)"

lemma NpPTS2PS: "N\<^sub>p (PTS2PS s)" unfolding N\<^sub>p_def PTS2PS_def proof (induct s, auto) qed

lemma PS2PTS2PS_id: "PS2PTS (PTS2PS s) = s" unfolding PS2PTS_def PTS2PS_def proof (induct s, auto) qed
lemma Np_PTS2PS2PTS_id: "N\<^sub>p s \<Longrightarrow> PTS2PS (PS2PTS s) = s" unfolding PS2PTS_def PTS2PS_def N\<^sub>p_def
apply (induct s, auto)
using Nnormal by fastforce

definition parallel_failure_set_to_parallel_trace_state_set :: "('s, 'e) parallelFailure set \<Rightarrow> ('s, 'e) parallelTraceState set" where
"parallel_failure_set_to_parallel_trace_state_set S = {(tr, s) . \<exists> s' X . (tr, s', X) \<in> S \<and> s = PS2PTS s'}"

definition parallel_trace_state_set_to_parallel_failure_set :: "('s, 'e) parallelTraceState set \<Rightarrow> 'e set \<Rightarrow> ('s, 'e) parallelFailure set" where
"parallel_trace_state_set_to_parallel_failure_set S \<Sigma> = {(tr, s, X) . \<exists> s' . (tr, s') \<in> S \<and> s = PTS2PS s' \<and> X \<subseteq> \<Sigma>}"

lemma state_set_conversion_id1: "failureSetToTraceStateSet (parallelFailureSetToFailureSet S) = parallelTraceStateSetToTraceStateSet (parallel_failure_set_to_parallel_trace_state_set S)"
  unfolding failureSetToTraceStateSet_def parallelFailureSetToFailureSet_def parallelTraceStateSetToTraceStateSet_def parallel_failure_set_to_parallel_trace_state_set_def
proof auto
  fix tr' X s' tr s X'
  assume
    inS: "(tr, s, X') \<in> S" and
    fail: "(tr', s', X) \<in> parallelFailureToFailure (tr, s, X')"
  have s: "s = seq s'"
    using fail parallelFailureToFailure_ex_seq
    by blast
  have G1: "(tr', NCS s') \<in> parallelTraceStateToTraceState (tr, PS2PTS s)"
    using fail
    unfolding s PS2PTS_def ATree_map.simps parallelTraceStateToTraceState.simps parallelFailureToFailure.simps
    by simp
  thus "\<exists>tr b. (\<exists>s. (\<exists>X. (tr, s, X) \<in> S) \<and> b = PS2PTS s) \<and> (tr', NCS s') \<in> parallelTraceStateToTraceState (tr, b)"
    using inS by blast
next
  fix tr s tr' s' X
  assume
    trace: "(tr, s) \<in> parallelTraceStateToTraceState (tr', PS2PTS s')" and
    inS: "(tr', s', X) \<in> S"
  have map_s': "PS2PTS s' = seq s"
    using trace parallelTraceStateToTraceState_ex_seq by fastforce
  have "\<exists> s's. s' = seq s's"
  proof (rule ccontr)
    assume "\<nexists>s's. s' = ATree.seq s's"
    then obtain s1 s2 where s'': "s' = par s1 s2" using ATree.exhaust by metis
    from map_s' show False
      unfolding s'' PS2PTS_def ATree_map.simps by simp
  qed
  then obtain s's where s': "s' = seq s's" by blast
  have "(tr, s's, X) \<in> parallelFailureToFailure (tr', s', X) \<and> s = NCS s's"
    using trace
    unfolding s' parallelFailureToFailure.simps PS2PTS_def ATree_map.simps parallelTraceStateToTraceState.simps
    by blast
  thus "\<exists>X su. (\<exists>tr' s' X'. (tr', s', X') \<in> S \<and> (tr, su, X) \<in> parallelFailureToFailure (tr', s', X')) \<and> s = NCS su"
    using inS by blast
qed

definition no_pc_in_dom :: "('s)denotational_parallel.parallelState \<Rightarrow> ('s, 'e) structured_code_parallel \<Rightarrow> bool" where
  "(no_pc_in_dom s c) = (leafn_property (\<lambda> (s::'s state) (c, _) . \<not> (PC s \<in> dom c)) op \<and> s c)"

subsection {* SF1-4 for parallel *}

text {* Each state in the SF semantics also is in the traces semantics. *}

lemma SF1_parallel:
  fixes code :: "('s, 'e) structured_code_parallel"
  assumes "(tr, s, X) \<in> sfsem\<^sub>p code S" and "wff\<^sub>p code"
  shows "(tr, PS2PTS s) \<in> densem\<^sub>p code (parallel_failure_set_to_parallel_trace_state_set S)"
  using assms
proof (induct code arbitrary: tr s X S, auto simp add: wff\<^sub>p_inherit wff\<^sub>p_def)
  fix \<Sigma> tr s X S and c :: "('s, 'e) structured_code"
  assume
    insf: "(tr, s, X) \<in> failureSetToParallelFailureSet (SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet S))" and
    wff: "wff\<^sub>s\<^sub>c c" and
    ax: "liSet (\<union>\<^sub>s\<^sub>c c)"
  have SFsemSigma: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using ax unfolding SFsemSigma_def by simp
  obtain s' where
    insf: "(tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet S)" and
    s: "s = ATree.seq s'"
    using insf[unfolded failureSetToParallelFailureSet_def] by blast
  show "(tr, PS2PTS s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>c\<rbrakk>parallelTraceStateSetToTraceStateSet (parallel_failure_set_to_parallel_trace_state_set S))"
    using SFsemSigma.SF1[OF SFsemSigma insf]
    unfolding s PS2PTS_def ATree_map.simps trace_state_set_to_parallel_trace_state_set_def state_set_conversion_id1
    by fast
next
  fix code1 code2 :: "('s, 'e) structured_code_parallel" and S :: "('s, 'e) parallelFailure set" and a b tr X X1 X2 X' t1 t2 s1 s2
  assume
    IH1: "\<And>tr s X S. (tr, s, X) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS \<Longrightarrow> (tr, PS2PTS s) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S" and
    IH2: "\<And>tr s X S. (tr, s, X) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS \<Longrightarrow> (tr, PS2PTS s) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S" and
    wff1: "leafn1_property (\<lambda>(c, ab). wff\<^sub>s\<^sub>c c \<and> liSet (\<union>\<^sub>s\<^sub>c c)) op \<and> code1" and
    wff2: "leafn1_property (\<lambda>(c, ab). wff\<^sub>s\<^sub>c c \<and> liSet (\<union>\<^sub>s\<^sub>c c)) op \<and> code2" and
    inS: "([], par s1 s2, X') \<in> S" and
    N1: "N\<^sub>p s1" and N2: "N\<^sub>p s2" and X: "X \<inter> (a \<union> b) = X1 \<inter> a \<union> X2 \<inter> b" and
    ex1: "(tr \<restriction> a, t1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')}" and
    ex2: "(tr \<restriction> b, t2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}" and
    tr: "set tr \<subseteq> a \<union> b"
  have eq: "PS2PTS (par t1 t2) = par (PS2PTS t1) (PS2PTS t2)"
    by (simp add: PS2PTS_def)
  have eq2: "PS2PTS (par s1 s2) = par (PS2PTS s1) (PS2PTS s2)"
    by (simp add: PS2PTS_def)
  have "([], par (PS2PTS s1) (PS2PTS s2)) \<in> parallel_failure_set_to_parallel_trace_state_set S"
    using inS
    unfolding parallel_failure_set_to_parallel_trace_state_set_def eq2[symmetric] by blast
  thus "\<exists>t1a t2a s1 s2. ([], par s1 s2) \<in> parallel_failure_set_to_parallel_trace_state_set S \<and> (tr \<restriction> a, t1a) \<in> (\<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T{([], s1)}) \<and> (tr \<restriction> b, t2a) \<in> (\<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T{([], s2)}) \<and> PS2PTS (par t1 t2) = par t1a t2a"
    unfolding eq
    using IH1[OF ex1] IH2[OF ex2]
    unfolding parallel_failure_set_to_parallel_trace_state_set_def by fastforce
qed

lemma SF2_parallel:
  fixes X
  assumes exec_code: "(tr, s, X) \<in> sfsem\<^sub>p code S"
      and X'subX: "X' \<subseteq> X"
      and base: "\<forall> tr s X X'. (tr, s, X) \<in> S \<and> X' \<subseteq> X \<longrightarrow> (tr, s, X') \<in> S"
      and wff: "wff\<^sub>p code"
  shows "(tr, s, X') \<in> sfsem\<^sub>p code S"
  using assms
proof (induct code arbitrary: tr s X S X')
  case (seqn x)
  obtain c \<Sigma> where x: "x = (c, \<Sigma>)" by fastforce
  have SFsemSigma: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(4)[unfolded wff\<^sub>p_def x, simplified] unfolding SFsemSigma_def by simp
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(4)[unfolded wff\<^sub>p_def x, simplified] by simp
  obtain s' where
    insf: "(tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> c (parallelFailureSetToFailureSet S)" and
    s: "s = ATree.seq s'"
    using seqn(1)[unfolded x, simplified, unfolded failureSetToParallelFailureSet_def] by blast
  have pre: "\<forall>tr s X X'. (tr, s, X) \<in> parallelFailureSetToFailureSet S \<and> X' \<subseteq> X \<longrightarrow> (tr, s, X') \<in> parallelFailureSetToFailureSet S"
    unfolding parallelFailureSetToFailureSet_def
  proof auto
    fix tr s X X' a aa b
    assume
      ref_subs: "X' \<subseteq> X" and
      inS: "(a, aa, b) \<in> S" and
      in_par: "(tr, s, X) \<in> parallelFailureToFailure (a, aa, b)"
    have eqs: "tr = a \<and> X = b \<and> ATree.seq s = aa" using parallelFailureToFailure_ex_seq[OF in_par] by fast
    then have "(tr, aa, X') \<in> S" using seqn inS ref_subs eqs by blast
    moreover have "(tr, s, X') \<in> parallelFailureToFailure (tr, aa, X')" using parallelFailureToFailure_ex_seq eqs by force
    ultimately show "\<exists>a aa b. (a, aa, b) \<in> S \<and> (tr, s, X') \<in> parallelFailureToFailure (a, aa, b)" by blast
  qed
  show ?case
    using SFsemSigma.SF2[OF SFsemSigma insf seqn(2) pre wff]
    unfolding failureSetToParallelFailureSet_def s x sfsemp_seq
    by blast
next
  case (parn code1 x2 code2)
  obtain a b where x2: "x2 = (a, b)" by fastforce
  have wff12: "wff\<^sub>p code1 \<and> wff\<^sub>p code2" using parn(6)[unfolded wff\<^sub>p_inherit] .
  obtain X1 X2 X'a t1 t2 s1 s2 where
     inS: "([], par s1 s2, X'a) \<in> S" and
     N1: "N\<^sub>p s1" and
     N2: "N\<^sub>p s2" and
     X: "X \<inter> (a \<union> b) = X1 \<inter> a \<union> X2 \<inter> b" and
     insf1: "(tr \<restriction> a, t1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X'a)}" and
     insf2: "(tr \<restriction> b, t2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X'a)}" and
     tr: "set tr \<subseteq> a \<union> b" and
     s: "s = par t1 t2"
    using parn(3)[unfolded x2, simplified] by blast
  have el1: "([], s1, X'a) \<in> {(tr, s, X) . X \<subseteq> X'a \<and> tr = [] \<and> s = s1}" by simp
  have el2: "([], s2, X'a) \<in> {(tr, s, X) . X \<subseteq> X'a \<and> tr = [] \<and> s = s2}" by simp
  have insf1': "(tr \<restriction> a, t1, X1 - (X - X')) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X'a - (X - X'))}" using sfsemp_refusal_set_subtraction[OF insf1] wff12 by simp
  have insf2': "(tr \<restriction> b, t2, X2 - (X - X')) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X'a - (X - X'))}" using sfsemp_refusal_set_subtraction[OF insf2] wff12 by simp
  define X1' where X1': "X1' = (X1) - (X - X')"
  define X2' where X2': "X2' = (X2) - (X - X')"
  have X': "X' \<inter> (a \<union> b) = X1' \<inter> a \<union> X2' \<inter> b"
    unfolding X1' X2' using X parn(4) by auto
  have "([], par s1 s2, X'a - (X - X')) \<in> S" using inS parn(5) by blast
  thus ?case
    using N1 N2 insf1' insf2' X' X1' X2' s tr
    unfolding x2 sfsemp_par by blast
qed


lemma SF3_parallel:
  fixes X and S :: "('s, 'e) parallelFailure set"
  assumes
    ex: "(tr, s, X) \<in> sfsem\<^sub>p code S" and
    base: "\<forall> tr s X X' . (tr, s, X) \<in> S \<and> X \<union> X' \<subseteq> \<Sigma> \<longrightarrow> (tr, s, X \<union> X') \<in> S" and
      --"Sadly we have to add this restriction. In the sequential case we could link the output
         refusal set to the starting refusal set if there was no step and got this guarantee for free.
         But here, if we have two sequential children and none is doing anything (PCs not in dom),
         the starting state's refusal set could contain events not in Sigma. Hence we cannot extend
         it arbitrarily. (Not, that the intersection with the interface sets allows for X to still be
         a subset of Sigma.)"
    base2: "\<forall> tr s X . (tr, s, X) \<in> S \<longrightarrow> X \<subseteq> \<Sigma>" and
    no_trace: "\<forall> t . \<forall> a \<in> X' . (tr@[a], t) \<notin> densem\<^sub>p code (parallel_failure_set_to_parallel_trace_state_set S)" and
    subs: "X \<union> X' \<subseteq> \<Sigma>" and
    wff: "wff\<^sub>p code" and
    wff_sigma: "wff_p_sigma code \<Sigma>" 
    and all_events_in_Sigma: "\<forall>X. X \<subseteq> \<Sigma>"
  shows "(tr, s, X \<union> X') \<in> sfsem\<^sub>p code S"
  using assms
proof (induct code arbitrary: tr s X S X')
  case (seqn x)
  obtain c \<Sigma>' where x: "x = (c, \<Sigma>')" by fastforce
  obtain ss where
    ex: "(tr, ss, X) \<in> SFsemSigma.sfsem \<Sigma>' c (parallelFailureSetToFailureSet S)" and
    s: "s = ATree.seq ss"
    using seqn(1) unfolding sfsemp_seq x failureSetToParallelFailureSet_def
    by blast
  have sf_axiom: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(6)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] SFsemSigma_def by auto
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(6)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] by simp
  have \<Sigma>': "\<Sigma>' = \<Sigma>"
    using seqn(7)[unfolded wff_p_sigma_def leafn1_property.simps x, simplified] ..
  have "(tr, ss, X \<union> X') \<in> SFsemSigma.sfsem \<Sigma>' c (parallelFailureSetToFailureSet S)"
  proof (rule SFsemSigma.SF3[OF sf_axiom ex _ _ _ wff])
    show "\<forall>tr s X X'. (tr, s, X) \<in> parallelFailureSetToFailureSet S \<and> X \<union> X' \<subseteq> \<Sigma>' \<longrightarrow> (tr, s, X \<union> X') \<in> parallelFailureSetToFailureSet S"
      unfolding parallelFailureSetToFailureSet_def
    proof auto
      fix tr s X X' a aa b
      assume
        X_subs_\<Sigma>: "X \<subseteq> \<Sigma>'" and
        X'_subs_\<Sigma>: "X' \<subseteq> \<Sigma>'" and
        inS: "(a, aa, b) \<in> S" and
        transformed: "(tr, s, X) \<in> parallelFailureToFailure (a, aa, b)"
      have aa: "aa = seq s"
        by (metis fst_conv parallelFailureToFailure_ex_seq snd_conv transformed)
      have X: "X = b \<and> tr = a" using transformed aa by simp
      hence *: "(a, aa, X \<union> X') \<in> S" using inS seqn(2) X_subs_\<Sigma> X'_subs_\<Sigma> \<Sigma>' by simp
      have **: "(tr, s, X \<union> X') \<in> parallelFailureToFailure (a, aa, X \<union> X')"
        unfolding aa parallelFailureToFailure.simps using X by simp
      show "\<exists>a aa b. (a, aa, b) \<in> S \<and> (tr, s, X \<union> X') \<in> parallelFailureToFailure (a, aa, b)"
        using * ** by blast
    qed
    have "failureSetToTraceStateSet (parallelFailureSetToFailureSet S) = parallelTraceStateSetToTraceStateSet (parallel_failure_set_to_parallel_trace_state_set S)"
      by (simp add: state_set_conversion_id1)
    show "\<forall>t. \<forall>a\<in>X'. (tr @ [a], t) \<notin> \<lbrakk>c\<rbrakk>failureSetToTraceStateSet (parallelFailureSetToFailureSet S)"
      using seqn(4)[unfolded densemp_seq x state_set_conversion_id1[symmetric] trace_state_set_to_parallel_trace_state_set_def]
      by blast
    show "X \<union> X' \<subseteq> \<Sigma>'"
      using \<Sigma>' seqn(5) by simp
  qed
  thus ?case
    unfolding s sfsemp_seq x failureSetToParallelFailureSet_def
    by blast
next
  case (parn code1 x2 code2)
  obtain \<alpha> \<beta> where x2: "x2 = (\<alpha>, \<beta>)" by fastforce
  obtain X1 X2 Xs s1 s2 s'1 s'2 where
    inS: "([], par s'1 s'2, Xs) \<in> S" and Ns1: "N\<^sub>p s'1" and Ns2: "N\<^sub>p s'2" and X: "X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>" and
    sf1: "(tr \<restriction> \<alpha>, s1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, Xs)}" and
    sf2: "(tr \<restriction> \<beta>, s2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, Xs)}" and tr: "set tr \<subseteq> \<alpha> \<union> \<beta>" and s: "s = par s1 s2"
    using parn(3)[unfolded x2, simplified] by blast
  have Xs_subs_\<Sigma>: "Xs \<subseteq> \<Sigma>" using parn(5) inS inS by blast
  consider (easy) "(X' - X) \<inter> (\<alpha> \<union> \<beta>) = {}" | (element_in_interface) "(X' - X) \<inter> (\<alpha> \<union> \<beta>) \<noteq> {}" by blast
  thus ?case
  proof cases
    case easy
    hence "(X \<union> X') \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta>"
      using X by blast
    thus ?thesis 
      unfolding x2 sfsemp_par
      using inS Ns1 Ns2 X sf1 sf2 tr s by blast
  next
    case element_in_interface
    define X'\<alpha>\<beta> where X'\<alpha>\<beta>: "X'\<alpha>\<beta> = X' \<inter> (\<alpha> \<union> \<beta>)"
    define X'' where X'': "X'' = X' - (\<alpha> \<union> \<beta>)"
    have wff1: "wff\<^sub>p code1" using parn(8) x2 wff\<^sub>p_inherit by blast
    have wff2: "wff\<^sub>p code2" using parn(8) x2 wff\<^sub>p_inherit by blast
    have wffS1: "wff_p_sigma code1 \<Sigma>" using parn(9) x2 wff_p_sigma_inherit by blast
    have wffS2: "wff_p_sigma code2 \<Sigma>" using parn(9) x2 wff_p_sigma_inherit by blast
    have Xs_subs_\<Sigma>: "Xs \<subseteq> \<Sigma>" using inS parn(5) by blast
    have X1_subs_\<Sigma>: "X1 \<subseteq> \<Sigma>" using sf1 Xs_subs_\<Sigma> all_events_in_Sigma wff1 by blast
    have X2_subs_\<Sigma>: "X2 \<subseteq> \<Sigma>" using sf2 Xs_subs_\<Sigma> all_events_in_Sigma wff2 by blast
    define S1 :: "('s, 'e) parallelFailure set" where S1: "S1 = {(trn, sn, Xn) . trn = [] \<and> sn = s'1 \<and> Xs \<subseteq> Xn \<and> ((* Xs = Xs \<or>  *)Xn \<subseteq> \<Sigma>)}"
    define S2 :: "('s, 'e) parallelFailure set" where S2: "S2 = {(trn, sn, Xn) . trn = [] \<and> sn = s'2 \<and> Xs \<subseteq> Xn \<and> ((* Xs = Xs \<or>  *)Xn \<subseteq> \<Sigma>)}"
    have sf1S: "(tr \<restriction> \<alpha>, s1, X1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS1" using sf1 S1 sfsemp_monotone_single Xs_subs_\<Sigma> wff1 by fastforce
    have sf2S: "(tr \<restriction> \<beta>, s2, X2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS2" using sf2 S2 sfsemp_monotone_single Xs_subs_\<Sigma> wff2 by fastforce
    have X'\<alpha>\<beta>_either_refuse: "\<forall> a \<in> X'\<alpha>\<beta> . \<forall> t1 t2 . ((tr @ [a]) \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S1) \<or> ((tr @ [a]) \<restriction> \<beta>, t2) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S2)"
    proof auto
      fix a t1 t2
      assume
        a_in_X'\<alpha>\<beta>: "a \<in> X'\<alpha>\<beta>" and
        in_code1: "((tr @ [a]) \<restriction> \<alpha>, t1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T(parallel_failure_set_to_parallel_trace_state_set S1)" and
        in_code2: "((tr @ [a]) \<restriction> \<beta>, t2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T(parallel_failure_set_to_parallel_trace_state_set S2)"
      have set_eq1: "parallel_failure_set_to_parallel_trace_state_set S1 = {([], PS2PTS s'1)}"
        unfolding parallel_failure_set_to_parallel_trace_state_set_def S1
        using Xs_subs_\<Sigma> by auto
      have set_eq2: "parallel_failure_set_to_parallel_trace_state_set S2 = {([], PS2PTS s'2)}"
        unfolding parallel_failure_set_to_parallel_trace_state_set_def S2
        using Xs_subs_\<Sigma> by auto
      have "(tr @ [a], par t1 t2) \<in> \<lbrakk>parn code1 x2 code2\<rbrakk>\<^sub>p\<^sub>T{([], PS2PTS (par s'1 s'2))}"
        using in_code1 in_code2 a_in_X'\<alpha>\<beta> tr
        unfolding densemp_par x2 set_eq1 set_eq2 PS2PTS_def ATree_map.simps X'\<alpha>\<beta> 
        by auto
      moreover have "([], PS2PTS (par s'1 s'2)) \<in> parallel_failure_set_to_parallel_trace_state_set S"
        unfolding parallel_failure_set_to_parallel_trace_state_set_def using inS by blast
      ultimately have "(tr @ [a], par t1 t2) \<in> \<lbrakk>parn code1 x2 code2\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S"
        using densemp_monotone_single parn(8) by blast
      thus False using parn(6) a_in_X'\<alpha>\<beta> X'\<alpha>\<beta> by blast
    qed
    obtain X'restr_\<alpha> X'restr_\<beta> where
      X'restr_\<alpha>: "\<forall> a \<in> X'restr_\<alpha> . \<forall> t1 . ((tr @ [a]) \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S1)" and
      X'restr_\<beta>: "\<forall> a \<in> X'restr_\<beta> . \<forall> t2 . ((tr @ [a]) \<restriction> \<beta>, t2) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S2)" and
      X'\<alpha>\<beta>_sup: "X'\<alpha>\<beta> = X'restr_\<alpha> \<union> X'restr_\<beta>" (* sadly, Isabelle does not trust me on this property. *)
    proof -
      assume asm: "\<And>X'restr_\<alpha> X'restr_\<beta>.
        \<forall>a\<in>X'restr_\<alpha>. \<forall>t1. ((tr @ [a]) \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S1 \<Longrightarrow>
        \<forall>a\<in>X'restr_\<beta>. \<forall>t2. ((tr @ [a]) \<restriction> \<beta>, t2) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S2 \<Longrightarrow> X'\<alpha>\<beta> = X'restr_\<alpha> \<union> X'restr_\<beta> \<Longrightarrow> thesis"
      obtain a where a: "a \<in> X'\<alpha>\<beta>"
        using X'\<alpha>\<beta> element_in_interface by blast
      define left_state_exists where left_state_exists: "left_state_exists = {a . a \<in> X'\<alpha>\<beta> \<and> (\<exists> t1 . ((tr @ [a]) \<restriction> \<alpha>, t1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S1)}"
      define right_state_exists where right_state_exists: "right_state_exists = {a . a \<in> X'\<alpha>\<beta> \<and> (\<exists> t2 . ((tr @ [a]) \<restriction> \<beta>, t2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S2)}"
      have decomposition: "X'\<alpha>\<beta> = left_state_exists \<union> right_state_exists \<union> (X'\<alpha>\<beta> - left_state_exists) \<union> (X'\<alpha>\<beta> - right_state_exists)"
        unfolding left_state_exists right_state_exists by blast
      define no_state_exists where no_state_exists: "no_state_exists = X'\<alpha>\<beta> - (left_state_exists \<union> right_state_exists)"
      have "\<forall> a \<in> left_state_exists . \<forall> t2 . ((tr @ [a]) \<restriction> \<beta>, t2) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S2)"
        unfolding left_state_exists
        using X'\<alpha>\<beta>_either_refuse by blast
      hence "\<forall>a\<in>left_state_exists \<union> (X'\<alpha>\<beta> - right_state_exists). \<forall>t2. ((tr @ [a]) \<restriction> \<beta>, t2) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S2"
        unfolding left_state_exists right_state_exists by blast
      moreover have "\<forall> a \<in> right_state_exists . \<forall> t1 . ((tr @ [a]) \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S1)"
        unfolding right_state_exists
        using X'\<alpha>\<beta>_either_refuse by blast
      hence "\<forall>a\<in>right_state_exists \<union> (X'\<alpha>\<beta> - left_state_exists). \<forall>t1. ((tr @ [a]) \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>Tparallel_failure_set_to_parallel_trace_state_set S1"
        unfolding left_state_exists right_state_exists by blast
      ultimately show thesis
        using asm[of "right_state_exists \<union> (X'\<alpha>\<beta> - left_state_exists)" "left_state_exists \<union> (X'\<alpha>\<beta> - right_state_exists)"]
        using decomposition
        by blast
    qed
    have X'restr_\<alpha>_subs_\<alpha>: "\<forall> a \<in> X'restr_\<alpha> . a \<in> \<alpha>"
    proof (rule ccontr, auto)
      fix a
      assume in_X'restr_\<alpha>: "a \<in> X'restr_\<alpha>" and not_in_\<alpha>: "a \<notin> \<alpha>"
      have "\<forall> t1 . (tr \<restriction> \<alpha>, t1) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S1)"
        using X'restr_\<alpha> not_in_\<alpha> in_X'restr_\<alpha>
        by (metis restrictTo_append_not_in_alpha)
      thus False
        using SF1_parallel[OF sf1S wff1]
        by blast
    qed
    have X'restr_\<beta>_subs_\<beta>: "\<forall> a \<in> X'restr_\<beta> . a \<in> \<beta>"
    proof (rule ccontr, auto)
      fix a
      assume in_X'restr_\<beta>: "a \<in> X'restr_\<beta>" and not_in_\<beta>: "a \<notin> \<beta>"
      have "\<forall> t1 . (tr \<restriction> \<beta>, t1) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T (parallel_failure_set_to_parallel_trace_state_set S2)"
        using X'restr_\<beta> not_in_\<beta> in_X'restr_\<beta>
        by (metis restrictTo_append_not_in_alpha)
      thus False
        using SF1_parallel[OF sf2S wff2]
        by blast
    qed
    have "([], par s'1 s'2, Xs \<union> X') \<in> S"
      using parn(4) parn(7) Xs_subs_\<Sigma> inS by simp
    moreover {
      have "([], s'1, Xs) \<in> S1" using S1 Xs_subs_\<Sigma> by blast
      moreover have "\<forall>tr s X. (tr, s, X) \<in> S1 \<longrightarrow> X \<subseteq> \<Sigma>" using S1 by blast
      moreover have X1_X'_subs_\<Sigma>: "X1 \<union> X'restr_\<alpha> \<subseteq> \<Sigma>" using X1_subs_\<Sigma> parn(7) X'\<alpha>\<beta> X'\<alpha>\<beta>_sup by blast
      moreover have extended_in_ex1: "(tr \<restriction> \<alpha>, s1, X1 \<union> X'restr_\<alpha>) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS1"
      proof (rule parn(1)[OF sf1S _ _ _ X1_X'_subs_\<Sigma> wff1 wffS1])
        show "\<forall>tr s X X'. (tr, s, X) \<in> S1 \<and> X \<union> X' \<subseteq> \<Sigma> \<longrightarrow> (tr, s, X \<union> X') \<in> S1" using S1 by fastforce
        show "\<forall>tr s X. (tr, s, X) \<in> S1 \<longrightarrow> X \<subseteq> \<Sigma>" using S1 by fastforce
        show "\<forall>t. \<forall>a\<in>X'restr_\<alpha>. (tr \<restriction> \<alpha> @ [a], t) \<notin> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T(parallel_failure_set_to_parallel_trace_state_set S1)"
          using X'restr_\<alpha> X'\<alpha>\<beta>_sup X'\<alpha>\<beta> restrictTo_append_in_alpha
          by (simp add: restrictTo_append_in_alpha X'restr_\<alpha>_subs_\<alpha>)
        show "\<forall>X. X \<subseteq> \<Sigma>" using all_events_in_Sigma by simp
      qed
      obtain XOminous where
        extended_in_ex1: "(tr \<restriction> \<alpha>, s1, X1 \<union> X'restr_\<alpha>) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, XOminous)}" and
        tms1: "treen_mirror s'1 code1 \<and> treen_mirror s1 code1 \<and> tree_mirror s1 s'1"
        using S1 sfsemp_predecessor[OF extended_in_ex1 wff1] by blast
      have "\<exists> XR . X1 \<union> X'restr_\<alpha> \<subseteq> XR \<and> XR \<subseteq> X1 \<union> X'restr_\<alpha> \<union> X' \<and> (tr \<restriction> \<alpha>, s1, XR) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, Xs \<union> X')}"
      proof -
        have "(X1 \<union> X'restr_\<alpha>) \<inter> X'restr_\<alpha> = X'restr_\<alpha>" by auto
        hence *: "(tr \<restriction> \<alpha>, s1, X'restr_\<alpha>) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, XOminous \<inter> X'restr_\<alpha>)}"
          using sfsemp_refusal_set_intersection[
              OF extended_in_ex1  wff1,
              of X'restr_\<alpha>
            ] by argo
        have *: "(tr \<restriction> \<alpha>, s1, X1 \<union> X'restr_\<alpha>) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, Xs \<union> XOminous \<inter> X'restr_\<alpha>)}"
          using sfsemp_refusal_set_merge[
              OF sf1 * wff1
            ] .
        have "\<exists>XR. X1 \<union> X'restr_\<alpha> \<subseteq> XR \<and> XR \<subseteq> X1 \<union> X'restr_\<alpha> \<union> X' \<and> (tr \<restriction> \<alpha>, s1, XR) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'1, Xs \<union> XOminous \<inter> X'restr_\<alpha> \<union> X')}"
          using X1_X'_subs_\<Sigma> parn(7) Xs_subs_\<Sigma>
          using sfsemp_extend_start_refusal_set[OF * _ _ wff1 wffS1, of X']
          by blast
        moreover have "Xs \<union> XOminous \<inter> X'restr_\<alpha> \<union> X' = Xs \<union> X'"
          using X'\<alpha>\<beta> X'\<alpha>\<beta>_sup
          by blast
        ultimately show ?thesis by simp
      qed
    } moreover {
      have "([], s'2, Xs) \<in> S2" using S2 Xs_subs_\<Sigma> by blast
      moreover have "\<forall>tr s X. (tr, s, X) \<in> S2 \<longrightarrow> X \<subseteq> \<Sigma>" using S2 by blast
      moreover have X2_X'_subs_\<Sigma>: "X2 \<union> X'restr_\<beta> \<subseteq> \<Sigma>" using X2_subs_\<Sigma> parn(7) X'\<alpha>\<beta> X'\<alpha>\<beta>_sup by blast
      moreover have extended_in_ex2: "(tr \<restriction> \<beta>, s2, X2 \<union> X'restr_\<beta>) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>FS2"
      proof (rule parn(2)[OF sf2S _ _ _ X2_X'_subs_\<Sigma> wff2 wffS2])
        show "\<forall>tr s X X'. (tr, s, X) \<in> S2 \<and> X \<union> X' \<subseteq> \<Sigma> \<longrightarrow> (tr, s, X \<union> X') \<in> S2" using S2 by fastforce
        show "\<forall>tr s X. (tr, s, X) \<in> S2 \<longrightarrow> X \<subseteq> \<Sigma>" using S2 by fastforce
        show "\<forall>t. \<forall>a\<in>X'restr_\<beta>. (tr \<restriction> \<beta> @ [a], t) \<notin> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T(parallel_failure_set_to_parallel_trace_state_set S2)"
          using X'restr_\<beta> X'\<alpha>\<beta>_sup X'\<alpha>\<beta> restrictTo_append_in_alpha
          by (simp add: restrictTo_append_in_alpha X'restr_\<beta>_subs_\<beta>)
        show "\<forall>X. X \<subseteq> \<Sigma>" using all_events_in_Sigma by simp
      qed
      obtain XOminous where
        extended_in_ex2: "(tr \<restriction> \<beta>, s2, X2 \<union> X'restr_\<beta>) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, XOminous)}" and
        tms2: "treen_mirror s'2 code2 \<and> treen_mirror s2 code2 \<and> tree_mirror s2 s'2"
        using S2 sfsemp_predecessor[OF extended_in_ex2 wff2] by blast
      have "\<exists> XR . X2 \<union> X'restr_\<beta> \<subseteq> XR \<and> XR \<subseteq> X2 \<union> X'restr_\<beta> \<union> X' \<and> (tr \<restriction> \<beta>, s2, XR) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, Xs \<union> X')}"
      proof -
        have "(X2 \<union> X'restr_\<beta>) \<inter> X'restr_\<beta> = X'restr_\<beta>" by auto
        hence *: "(tr \<restriction> \<beta>, s2, X'restr_\<beta>) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, XOminous \<inter> X'restr_\<beta>)}"
          using sfsemp_refusal_set_intersection[
              OF extended_in_ex2  wff2,
              of X'restr_\<beta>
            ] by argo
        have *: "(tr \<restriction> \<beta>, s2, X2 \<union> X'restr_\<beta>) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, Xs \<union> XOminous \<inter> X'restr_\<beta>)}"
          using sfsemp_refusal_set_merge[
              OF sf2 * wff2
            ] .
        have "\<exists>XR. X2 \<union> X'restr_\<beta> \<subseteq> XR \<and> XR \<subseteq> X2 \<union> X'restr_\<beta> \<union> X' \<and> (tr \<restriction> \<beta>, s2, XR) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s'2, Xs \<union> XOminous \<inter> X'restr_\<beta> \<union> X')}"
          using X2_X'_subs_\<Sigma> parn(7) Xs_subs_\<Sigma>
          using sfsemp_extend_start_refusal_set[OF * _ _ wff2 wffS2, of X']
          by blast
        moreover have "Xs \<union> XOminous \<inter> X'restr_\<beta> \<union> X' = Xs \<union> X'"
          using X'\<alpha>\<beta> X'\<alpha>\<beta>_sup
          by blast
        ultimately show ?thesis by simp
      qed
    }
    moreover have "\<forall> XR1 XR2 .
       X2 \<union> X'restr_\<beta> \<subseteq> XR2 \<and> XR2 \<subseteq> X2 \<union> X'restr_\<beta> \<union> X' \<and>
       X1 \<union> X'restr_\<alpha> \<subseteq> XR1 \<and> XR1 \<subseteq> X1 \<union> X'restr_\<alpha> \<union> X' \<longrightarrow>
       (X \<union> X') \<inter> (\<alpha> \<union> \<beta>) = (XR1) \<inter> \<alpha> \<union> (XR2) \<inter> \<beta>"
      using X'\<alpha>\<beta> X'\<alpha>\<beta>_sup X'restr_\<alpha>_subs_\<alpha> X'restr_\<beta>_subs_\<beta> X by blast
    ultimately show ?thesis
      unfolding x2 s sfsemp_par
      using inS Ns1 Ns2 tr X'\<alpha>\<beta> X'\<alpha>\<beta>_sup by fastforce
  qed
qed

lemma SF4_parallel:
  assumes
    densem: "(tr, s) \<in> densem\<^sub>p code S" and
    no_pc_in_dom: "no_pc_in_dom s code" and
    X_sub_\<Sigma>: "X \<subseteq> \<Sigma>" and
    wff: "wff\<^sub>p code" and
    wff_sigma: "wff_p_sigma code \<Sigma>"
  shows "(tr, PTS2PS s, X) \<in> sfsem\<^sub>p code (parallel_trace_state_set_to_parallel_failure_set S \<Sigma>)"
    using densem no_pc_in_dom wff wff_sigma
proof (induct code arbitrary: tr s S)
  case (seqn x)
  obtain c \<Sigma>' where x: "x = (c, \<Sigma>')" by fastforce
  have sf_axiom: "SFsemSigma (\<union>\<^sub>s\<^sub>c c)"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] SFsemSigma_def by auto
  have wff: "wff\<^sub>s\<^sub>c c"
    using seqn(3)[unfolded wff\<^sub>p_def leafn1_property.simps x, simplified] by simp
  obtain ss s' tr'  where
    ex: "(tr, ss) \<in> \<lbrakk>c\<rbrakk>{(tr', s')}" and
    inS: "(tr', s') \<in> parallelTraceStateSetToTraceStateSet S" and
    s: "s = ATree.seq ss"
    using seqn(1)[unfolded densemp_seq x trace_state_set_to_parallel_trace_state_set_def, simplified]
    using densem_predecessor
    by fast
  have \<Sigma>_eq: "\<Sigma> = \<Sigma>'"
    using seqn(4)[unfolded wff_p_sigma_def x leafn1_property.simps, simplified] .
  have pc_not_in_dom: "PC ss \<notin> definitions.dom c"
    using seqn(2)[unfolded x no_pc_in_dom_def leafn_property.simps s, simplified] .
  have ex: "(tr, normal ss, X) \<in> SFsemSigma.sfsem \<Sigma> c (SFsemSigma.trace_state_set_to_failure_set \<Sigma> {(tr', s')})"
    using SFsemSigma.SF4[OF sf_axiom ex pc_not_in_dom X_sub_\<Sigma> wff]
    using \<Sigma>_eq
    by simp
  have subs: "(SFsemSigma.trace_state_set_to_failure_set \<Sigma> {(tr', s')}) \<subseteq> (parallelFailureSetToFailureSet (parallel_trace_state_set_to_parallel_failure_set S \<Sigma>))"
    using inS
    unfolding SFsemSigma.trace_state_set_to_failure_set_def[OF sf_axiom] parallelTraceStateSetToTraceStateSet_def
    unfolding parallelFailureSetToFailureSet_def parallel_trace_state_set_to_parallel_failure_set_def
  proof auto
    fix X tr s
    assume
      inS: "(tr, s) \<in> S" and
      conv: "(tr', s') \<in> parallelTraceStateToTraceState (tr, s)" and
      subs: "X \<subseteq> \<Sigma>"
    moreover have state_eq: "(tr', ATree.seq s') = (tr, s)"
      using parallelTraceStateToTraceState_ex_seq[OF conv] by auto
    moreover have "(tr', normal s', X) \<in> parallelFailureToFailure (tr, seq (normal s'), X)"
      unfolding parallelFailureToFailure.simps
      using state_eq by simp
    moreover have "seq (normal s') = PTS2PS s" using state_eq
      by (metis ATree_map.simps(1) PTS2PS_def Pair_inject)
    ultimately show "\<exists>tr s X'. (\<exists>s'. (tr, s') \<in> S \<and> s = PTS2PS s' \<and> X' \<subseteq> \<Sigma>) \<and> (tr', normal s', X) \<in> parallelFailureToFailure (tr, s, X')"
      by blast
  qed
  have "(tr, normal ss, X) \<in> SFsemSigma.sfsem \<Sigma>' c (parallelFailureSetToFailureSet (parallel_trace_state_set_to_parallel_failure_set S \<Sigma>))"
    using \<Sigma>_eq SFsemSigma.sfsem_monotone[OF sf_axiom subs, of \<Sigma> c] ex
    by blast
  thus ?case
    unfolding sfsemp_seq x failureSetToParallelFailureSet_def s PTS2PS_def
    by fastforce
next
  case (parn code1 \<alpha>\<beta> code2)
  obtain \<alpha> \<beta> where \<alpha>\<beta>: "\<alpha>\<beta> = (\<alpha>, \<beta>)" by fastforce
  have wff1: "wff\<^sub>p code1" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wff2: "wff\<^sub>p code2" using parn(5) \<alpha>\<beta> wff\<^sub>p_inherit by blast
  have wffS1: "wff_p_sigma code1 \<Sigma>" using parn(6) \<alpha>\<beta> wff_p_sigma_inherit by blast
  have wffS2: "wff_p_sigma code2 \<Sigma>" using parn(6) \<alpha>\<beta> wff_p_sigma_inherit by blast
  obtain s1 s2 s1' s2' s' where
    inS: "([], par s1' s2') \<in> S" and
    t1: "(tr \<restriction> \<alpha>, s1) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>T{([], s1')}" and
    t2: "(tr \<restriction> \<beta>, s2) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>T{([], s2')}" and
    tr: "set tr \<subseteq> \<alpha> \<union> \<beta>" and
    s: "s = par s1 s2" and
    s': "s' = par s1' s2'"
    using parn(3)[unfolded \<alpha>\<beta>, simplified]
    by blast
  have no_pc_in_dom1: "no_pc_in_dom s1 code1" using parn(4)[unfolded s]
    unfolding no_pc_in_dom_def
    using leafn_property_inherit by simp
  have no_pc_in_dom2: "no_pc_in_dom s2 code2" using parn(4)[unfolded s]
    unfolding no_pc_in_dom_def
    using leafn_property_inherit by simp
  have PTS2PS_s_unfold: "PTS2PS s = par (PTS2PS s1) (PTS2PS s2)"
    unfolding PTS2PS_def s by simp
  have G1: "(tr \<restriction> \<alpha>, PTS2PS s1, X) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F (parallel_trace_state_set_to_parallel_failure_set {([], s1')} \<Sigma>)"
    using parn(1)[OF t1 no_pc_in_dom1 wff1 wffS1] .
  obtain Xs1 where
    G1: "(tr \<restriction> \<alpha>, PTS2PS s1, X) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F {([], PTS2PS s1', Xs1)}" and
    Xs1_sub_\<Sigma>: "Xs1 \<subseteq> \<Sigma>"
    using sfsemp_predecessor[OF G1 wff1]
    unfolding parallel_trace_state_set_to_parallel_failure_set_def
    by blast
  have "(tr \<restriction> \<alpha>, PTS2PS s1, X) \<in> \<lbrakk>code1\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], PTS2PS s1', X)}"
    using sfsemp_replace_start_refusal_set_by_result_refusal_set[OF G1 X_sub_\<Sigma> wff1 wffS1] .
  moreover have G2: "(tr \<restriction> \<beta>, PTS2PS s2, X) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F (parallel_trace_state_set_to_parallel_failure_set {([], s2')} \<Sigma>)"
    using parn(2)[OF t2 no_pc_in_dom2 wff2 wffS2] .
  obtain Xs2 where
    G2: "(tr \<restriction> \<beta>, PTS2PS s2, X) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F {([], PTS2PS s2', Xs2)}" and
    Xs2_sub_\<Sigma>: "Xs2 \<subseteq> \<Sigma>"
    using sfsemp_predecessor[OF G2 wff2]
    unfolding parallel_trace_state_set_to_parallel_failure_set_def
    by blast
  have "(tr \<restriction> \<beta>, PTS2PS s2, X) \<in> \<lbrakk>code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F {([], PTS2PS s2', X)}"
    using sfsemp_replace_start_refusal_set_by_result_refusal_set[OF G2 X_sub_\<Sigma> wff2 wffS2] .
  moreover have "X \<inter> (\<alpha> \<union> \<beta>) = X \<inter> \<alpha> \<union> X \<inter> \<beta>"
    by blast
  moreover have "([], par (PTS2PS s1') (PTS2PS s2'), X) \<in> parallel_trace_state_set_to_parallel_failure_set {([], s')} \<Sigma>"
    unfolding parallel_trace_state_set_to_parallel_failure_set_def s' PTS2PS_def
    using X_sub_\<Sigma> by simp
  ultimately have "(tr, PTS2PS s, X) \<in> \<lbrakk>parn code1 \<alpha>\<beta> code2\<rbrakk>\<^sub>p\<^sub>S\<^sub>F (parallel_trace_state_set_to_parallel_failure_set {([], s')} \<Sigma>)"
    unfolding sfsemp_par \<alpha>\<beta> PTS2PS_s_unfold
    using NpPTS2PS[of s1'] NpPTS2PS[of s2'] tr s s'
    by blast
  moreover have "(parallel_trace_state_set_to_parallel_failure_set {([], s')} \<Sigma>) \<subseteq> (parallel_trace_state_set_to_parallel_failure_set S \<Sigma>)"
    unfolding parallel_trace_state_set_to_parallel_failure_set_def using inS s'
    by blast
  ultimately show ?case
    using sfsemp_monotone parn(5)
    by blast
qed

subsection {* Connecting Property and Refinement *}

text {* In the following definitions, we always need to carry the process name function as it is used in the
CSP Prover. For its documentation, refer to the CSP Prover package.
In general, we have two basic concepts of the CUC-CSP refinement. One with the user definable connecting property,
the other one without a connecting property. Also, both for SF semantics and T semantics.
A further split is due to the starting property, which in general can be generic. There is also a predefined precondition,
which fixes the starting program counter to be 1, the state to be a normal state, and the trace to be empty. The overall idea is,
that each code probably starts the execution at some program counter, which we simply fixed arbitrarily. Also, the trace should always
be empty in the beginning. Further assumptions concerning the state (i.e. initial values of the registers) can be made
explicit via some initializing instruction(s). Hence, this fixed starting property should be sufficient. Still, we want
to praise genericity and thus allow for arbitrary starting properties.
A third parameter, being required in any case, is the conversion function to relate the CUC event type to the CSP event type.
*}

paragraph {* Preconditions *}

text {* The precondition for start states - we want them all to start at a welldefined pc, be normal
and have an empty trace. *}

lemma Pre_SF_T_Eq:
  shows "PreStd\<^sub>S\<^sub>F = (\<lambda>(tr::'e definitions.event list, s::'s stable_failures_parallel.parallelState, X :: 'e definitions.event set). PreStd\<^sub>T (tr, PS2PTS s) \<and> N\<^sub>p s)"
proof (rule ext)
  fix x :: "('s, 'e) parallelFailure"
  obtain tr s X where x: "x = (tr, s, X)" using prod_cases3 by blast
  show "PreStd\<^sub>S\<^sub>F x = (\<lambda>(tr, s, X). PreStd\<^sub>T (tr, PS2PTS s) \<and> N\<^sub>p s) x"
    unfolding x
  proof (induct s)
    case (seq x)
    show ?case
      unfolding PreStd\<^sub>S\<^sub>F_def PreStd\<^sub>T_def AllPC\<^sub>S\<^sub>F_def AllPC\<^sub>T_def PS2PTS_def
      by (auto simp add: PCNCS)
  next
    case (par s1 s2)
    thus ?case
      unfolding PreStd\<^sub>S\<^sub>F_def N\<^sub>p_def PreStd\<^sub>T_def AllPC\<^sub>S\<^sub>F_def AllPC\<^sub>T_def PS2PTS_def
      by force
  qed
qed

paragraph {* Checks: The sequentiel helper refinements imply their parallel counterparts *}

text {* If we have an NCstate, then it has the same height as another tree of size 1 (being just a single node). *}

lemma parallelState_eq: "parallelState_is_NCstate s = treen_mirror s (seqn (a, b, c))"
  unfolding treen_mirror
  by (metis leafn_property.simps(1) leafn_property_seql parallelState_is_NCstate.elims(2) parallelState_is_NCstate.simps(1))

text {* Unfolding the short definition into a longer one. *}

lemma convert_pre: "(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = (\<lambda> x . PreStd\<^sub>S\<^sub>F x \<and> parallelState_is_NCstate (fst (snd x)))"
  unfolding PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q_def PreStd\<^sub>S\<^sub>F_def CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F_def
proof (auto, rule ext, rule iffI)
  fix x :: "'a list \<times> 'b NCstate ATree \<times> 'c"
  assume aa1: "case x of (tr, seq s', X) \<Rightarrow> NCPC s' = 1 \<and> N s' \<and> tr = []
         | (tr, par parallelState1 parallelState2, X) \<Rightarrow> False"
  obtain tr ss' X where x: "x = (tr, ss', X)" using prod_cases3 by blast
  then consider (s) "\<exists> s' . ss' = seq s'" | (p) "\<exists> cl cr . ss' = par cl cr"
    by (meson parallelState_is_NCstate.cases)
  thus "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
  proof cases
    case s
    then obtain s' where ss': "ss' = seq s'" by blast
    hence "NCPC s' = Suc 0 \<and> N s' \<and> tr = []" using aa1[unfolded x ss', simplified] by simp
    hence "AllPC\<^sub>S\<^sub>F (seq s') (Suc 0) \<and> N\<^sub>p (seq s') \<and> tr = [] \<and> parallelState_is_NCstate (fst (snd x))"
      unfolding x ss' fst_conv snd_conv leaf1_property.simps "parallelState_is_NCstate.simps" "AllPC\<^sub>S\<^sub>F_def" N\<^sub>p_def by simp
    thus "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
      unfolding x ss'
      by simp
  next
    case p
    with aa1 x have False by fastforce
    thus ?thesis by simp
  qed
next
  fix x :: "'a list \<times> 'b NCstate ATree \<times> 'c"
  assume aa1: "(case x of (tr, s, xa) \<Rightarrow> AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = []) \<and> parallelState_is_NCstate (fst (snd x))"
  then obtain tr s X where x: "x = (tr, s, X)" by blast
  with aa1 have G: "AllPC\<^sub>S\<^sub>F s (Suc 0) \<and> N\<^sub>p s \<and> tr = [] \<and> parallelState_is_NCstate (fst (snd x))"
    by fast
  then obtain s' where s: "s = seq s'"
    unfolding fst_conv snd_conv x using "parallelState_is_NCstate.elims"(2)
    by metis
  have "NCPC s' = 1 \<and> N s' \<and> tr = []"
    using G
    unfolding s "AllPC\<^sub>S\<^sub>F_def" N\<^sub>p_def by simp
  thus "case x of (tr, seq s', X) \<Rightarrow> NCPC s' = 1 \<and> N s' \<and> tr = []
         | (tr, par parallelState1  parallelState2, X) \<Rightarrow> False"
    unfolding x s
    by auto
qed

lemma convert_pre2: "(CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = (\<lambda> x . PreStd\<^sub>S\<^sub>F x \<and> treen_mirror (fst (snd x)) (seqn (a, b, c)))"
  by (meson convert_pre parallelState_eq)

lemma convert_pre3: "Collect (CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F PreStd\<^sub>S\<^sub>F\<^sub>s\<^sub>e\<^sub>q) = { x . PreStd\<^sub>S\<^sub>F x \<and> treen_mirror (fst (snd x)) (seqn (a, b, c))}"
  by (meson convert_pre parallelState_eq)

lemma seq_sf_refinement_set_eq:
  "{(tr, s, X) . \<exists> s' . (tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre) \<and> s = seq s'} =
   \<lbrakk>seqn (Impl, \<Sigma>)\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}"
proof -
  have eq: "Collect Pre = (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'})"
    unfolding parallelFailureSetToFailureSet_def CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F_def
  proof (auto)
    fix a aa b
    assume ca1: "Pre (a, aa, b)"
    hence "(case seq aa of seq aa \<Rightarrow> Pre (a, aa, b) | par parallelState1 parallelState2 \<Rightarrow> False) \<and>
               treen_mirror (seq aa) (seqn (Impl, \<Sigma>)) \<and> (a, aa, b) \<in> parallelFailureToFailure (a, seq aa, b)"
      unfolding "parallelFailureToFailure.simps"
      by simp
    thus "\<exists>aaa aaaa ba.
               (case aaaa of seq s' \<Rightarrow> Pre (aaa, s', ba) | par parallelState1 parallelState2 \<Rightarrow> False) \<and>
               (a, aa, b) \<in> parallelFailureToFailure (aaa, aaaa, ba)"
      by blast
  next
    fix a aa b ab ac ba
    assume
      seq: "case ac of ATree.seq s' \<Rightarrow> Pre (ab, s', ba) | par ATree1 ATree2 \<Rightarrow> False" and
      ins: "(a, aa, b) \<in> parallelFailureToFailure (ab, ac, ba)"
    obtain s' where ac: "ac = seq s'" using seq
      by (metis empty_iff ins parallelFailureToFailure.simps(2) parallelState_is_NCstate.cases)
    have eq: "(a, aa, b) = (ab, s', ba)" using ins unfolding ac parallelFailureToFailure.simps
      by simp
    thus "Pre (a, aa, b)"
      using ac seq
      by simp
  qed
  show ?thesis
  proof (auto simp del: treen_mirror_unfold_seqr)
    fix a b s'
    assume aa1: "(a, s', b) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre)"
    
    show "(a, seq s', b)
                \<in> failureSetToParallelFailureSet
                    (SFsemSigma.sfsem \<Sigma> Impl (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}))"
      using aa1[unfolded eq]
      unfolding failureSetToParallelFailureSet_def
      by (metis (mono_tags, lifting) case_prodI eq mem_Collect_eq)
  next
    fix a aa b
    assume aa1: "(a, aa, b) \<in> failureSetToParallelFailureSet
                    (SFsemSigma.sfsem \<Sigma> Impl (parallelFailureSetToFailureSet {x'. CV\<^sub>P\<^sub>r\<^sub>e\<^sub>S\<^sub>F Pre x'}))"
    thus "\<exists>s'. (a, s', b) \<in> SFsemSigma.sfsem \<Sigma> Impl (Collect Pre) \<and> aa = seq s'"
      unfolding eq
      unfolding failureSetToParallelFailureSet_def
      by simp
  qed
qed

text {* A trace in the denotational T world can also be found in the SF world, if we can make sure to always have a successor in the
  sequential code for any start state (related to divergency - we want to prevent disappearing states which are trapped in
  an infinite loop). *}


theorem tr_of_T_will_be_in_SF_parallel:
  fixes
    Impl :: "('s, 'e) structured_code_parallel"
  assumes
    --"We have a trace and some state in the denotational world."
    tr_in_dsem_parallel: "(tr, s) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>T{x'. PreT x'}" and
    --"The same starting states for the CUC code have to be used in the SF and in the T part."
    PreF: "\<forall> tr s X . treen_mirror s Impl \<longrightarrow> (PreT (tr, (PS2PTS s)) \<and> N\<^sub>p s) \<longrightarrow> PreF (tr, s, X)" and
    wffImpl: "wff\<^sub>p Impl" and
    --"For the sequential leafs, we always want to have successors."
    successors: "leafn1_property (\<lambda> (c, \<Sigma>) . \<forall> tr s X . SFsemSigma.sfsem \<Sigma> c {(tr, s, X)} \<noteq> {}) op \<and> Impl"
shows "\<exists> s' X . (tr, s', X) \<in> \<lbrakk>Impl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x'. PreF x'}"
  using tr_in_dsem_parallel PreF successors wffImpl
proof (induct Impl arbitrary: PreF PreT tr s rule: structured_code_parallel_induct, auto)
  fix \<Sigma> :: "'e definitions.event set" and x :: "('s, 'e) structured_code" (* and PreF :: "'e definitions.trace \<times> 's parallelState \<times> 'e definitions.event set \<Rightarrow> bool" *)
    and PreT :: "('s, 'e) parallelTraceState \<Rightarrow> bool" and PreF :: "('s, 'e) parallelFailure \<Rightarrow> bool" and tr s
  assume
    ins: "(tr, s) \<in> trace_state_set_to_parallel_trace_state_set (\<lbrakk>x\<rbrakk>parallelTraceStateSetToTraceStateSet (Collect PreT))" and
    aa3: "\<forall>tr s X. SFsemSigma.sfsem \<Sigma> x {(tr, s, X)} \<noteq> {}" and
    aa5: "wff\<^sub>p (seqn (x, \<Sigma>))" and
    Pre_imp': "\<forall>tr s. (\<exists>sa. s = ATree.seq sa) \<longrightarrow> PreT (tr, PS2PTS s) \<and> N\<^sub>p s \<longrightarrow> (\<forall>X. PreF (tr, s, X))"
  obtain s' where
    trs': "(tr, s') \<in> \<lbrakk>x\<rbrakk>{x'. \<exists>a b. PreT (a, b) \<and> x' \<in> parallelTraceStateToTraceState (a, b)}" and
    s: "s = seq s'"
    using ins[unfolded "densem\<^sub>p.simps" parallelTraceStateSetToTraceStateSet_def, simplified]
    unfolding trace_state_set_to_parallel_trace_state_set_def treen_mirror[symmetric] treen_mirror_unfold_seqr
    by blast
  then obtain tru su where
    u: "(tr, s') \<in> \<lbrakk>x\<rbrakk>{(tru, su)}" and
    u2: "(tru, su) \<in> {x'. \<exists>a b. PreT (a, b) \<and> x' \<in> parallelTraceStateToTraceState (a, b)}"
    using densem_predecessor prod.collapse unfolding treen_mirror_unfold_seqr by fastforce
  then obtain trP sP where
    P:  "(tru, su) \<in> parallelTraceStateToTraceState (trP, sP)" and
    P2: "PreT (trP, sP)" and
    truEqtrP: "tru = trP" and
    sP: "sP = ATree.seq su"
    unfolding treen_mirror_unfold_seqr using parallelTraceStateToTraceState_ex_seq
    by blast
  obtain sP' where sP': "sP' = PTS2PS sP" by simp
  hence sP'2: "PS2PTS sP' = sP" using PS2PTS2PS_id by blast
  have NsP': "N\<^sub>p (sP')" using sP' NpPTS2PS by blast
  have "PreT (tru, PS2PTS sP')" using u2 unfolding sP' PS2PTS2PS_id
    by (simp add: P2 truEqtrP)
  moreover have "treen_mirror sP' (seqn (x, \<Sigma>))"
    using sP'2 sP
    unfolding treen_mirror_unfold_seqr PS2PTS_def
    by (metis ATree_map.simps(1) PTS2PS_def sP')
  then obtain sP's where sP': "sP' = ATree.seq sP's" 
    unfolding treen_mirror_unfold_seqr by blast
  moreover have "(tru, stable_failures.normal su, \<Sigma>) \<in> parallelFailureToFailure (tru, sP', \<Sigma>)"
    unfolding sP' "parallelFailureToFailure.simps"
    using P[unfolded sP'2[unfolded PS2PTS_def sP' ATree_map.simps, symmetric] parallelTraceStateToTraceState.simps]
    using NCS.simps NsP'[unfolded N\<^sub>p_def sP' leaf1_property.simps] Nnormal
    by fast
  ultimately have g1: "(tru, stable_failures.normal su (* s' *), \<Sigma>) \<in> parallelFailureSetToFailureSet (Collect PreF)"
    using NsP' Pre_imp'
    unfolding treen_mirror_unfold_seqr parallelFailureSetToFailureSet_def by force
  have successor': "\<forall>tr s X. SFsemSigma.sfsem \<Sigma> x {(tr, s, X)} \<noteq> {}"
    using aa3[simplified] .
  have "\<exists> s . (tr, s) \<in> densem x (failureSetToTraceStateSet {(tru, NCstate.normal su, \<Sigma>)})"
    using u
    unfolding failureSetToTraceStateSet_def
    using NCS.simps(1)
    by auto
  hence trInSF: "\<exists>X s'. (tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> x {(tru, NCstate.normal su, \<Sigma>)}"
    using aa5[unfolded wff\<^sub>p_def "leafn1_property.simps", simplified]
      SFsemSigma.tr_of_T_will_be_in_SF[OF _ successor', of "\<union>\<^sub>s\<^sub>c x" tr _ "{(tru, stable_failures.normal su, \<Sigma>)}", simplified]
    unfolding SFsemSigma_def
    by fast
  hence "\<exists>X s'. (tr, s', X) \<in> SFsemSigma.sfsem \<Sigma> x (parallelFailureSetToFailureSet (Collect PreF))"
    using g1[unfolded ] aa5[unfolded wff\<^sub>p_def "leafn1_property.simps", simplified, THEN cj2_2]
      "SFsemSigma.sfsem_monotone_single"[OF _ g1, of "\<union>\<^sub>s\<^sub>c x" "(tr, _, _)" \<Sigma> x]
    unfolding SFsemSigma_def
    by fast
  thus "\<exists>s' X. (tr, s', X)
            \<in> failureSetToParallelFailureSet
                (SFsemSigma.sfsem \<Sigma> x
                  (parallelFailureSetToFailureSet
                    (Collect PreF)))"
    unfolding "sfsem\<^sub>p.simps" failureSetToParallelFailureSet_def parallelFailureSetToFailureSet_def CollectI
    by blast
next
  --"Use the induction hypothesis with custom preconditions, which fix a single starting state."
  fix cl cr :: "('s, 'e) structured_code_parallel" and \<alpha> \<beta> :: "'e definitions.event set" and 
    PreF :: "('s, 'e) parallelFailure \<Rightarrow> bool" and PreT :: "('s, 'e) parallelTraceState \<Rightarrow> bool" and
    tr s1 s2 t1 t2
  assume
    IH1: "\<And>PreF PreT tr s. (tr, s) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T(Collect PreT) \<Longrightarrow> \<forall>tr s. treen_mirror s cl \<longrightarrow> PreT (tr, PS2PTS s) \<and> N\<^sub>p s \<longrightarrow> (\<forall>X. PreF (tr, s, X)) \<Longrightarrow> wff\<^sub>p cl \<Longrightarrow> \<exists>s' X. (tr, s', X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>FCollect PreF" and
    IH2: "\<And>PreF PreT tr s. (tr, s) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T(Collect PreT) \<Longrightarrow> \<forall>tr s. treen_mirror s cr \<longrightarrow> PreT (tr, PS2PTS s) \<and> N\<^sub>p s \<longrightarrow> (\<forall>X. PreF (tr, s, X)) \<Longrightarrow> wff\<^sub>p cr \<Longrightarrow> \<exists>s' X. (tr, s', X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>FCollect PreF" and
    wff: "wff\<^sub>p (parn cl (\<alpha>, \<beta>) cr)" and
    succ_l: "leafn1_property (\<lambda>(c, \<Sigma>). \<forall>tr s X. SFsemSigma.sfsem \<Sigma> c {(tr, s, X)} \<noteq> {}) op \<and> cl" and
    succ_r: "leafn1_property (\<lambda>(c, \<Sigma>). \<forall>tr s X. SFsemSigma.sfsem \<Sigma> c {(tr, s, X)} \<noteq> {}) op \<and> cr" and
    PreT_s12: "PreT ([], par s1 s2)" and
    ex_l: "(tr \<restriction> \<alpha>, t1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{([], s1)}" and
    ex_r: "(tr \<restriction> \<beta>, t2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{([], s2)}" and
    tr_subs: "set tr \<subseteq> \<alpha> \<union> \<beta>" and
    Pre_imp: "\<forall>tr s. treen_mirror s (parn cl (\<alpha>, \<beta>) cr) \<longrightarrow> PreT (tr, PS2PTS s) \<and> N\<^sub>p s \<longrightarrow> (\<forall>X. PreF (tr, s, X))"
  --"First, we prove all easy conditions for the goal."
  have wff_lr: "wff\<^sub>p cl \<and> wff\<^sub>p cr" using wff\<^sub>p_inherit wff by blast
  have tm1: "treen_mirror s1 cl" using ex_l densemp_predecessor wff wff\<^sub>p_inherit by fastforce
  have tm2: "treen_mirror s2 cr" using ex_r densemp_predecessor wff wff\<^sub>p_inherit by fastforce
  have tm1t: "treen_mirror (PTS2PS s1) cl" unfolding PTS2PS_def using tm1 ATree_map_treen_mirror by fast
  have tm2t: "treen_mirror (PTS2PS s2) cr" unfolding PTS2PS_def using tm2 ATree_map_treen_mirror by fast
  have Gtm: "treen_mirror (par (PTS2PS s1) (PTS2PS s2)) (parn cl (\<alpha>, \<beta>) cr)"
    using tm1t tm2t unfolding treen_mirror_inherit ..
  have N12: "N\<^sub>p (par (PTS2PS s1) (PTS2PS s2))"
    unfolding N\<^sub>p_inherit using NpPTS2PS by blast
  have PreT12: "PreT ([], PS2PTS (par (PTS2PS s1) (PTS2PS s2)))"
    by (metis ATree_map.simps(2) PS2PTS2PS_id PS2PTS_def PreT_s12)
  have GPreF: "PreF ([], par (PTS2PS s1) (PTS2PS s2), {})" using Pre_imp PreT12 N12 Gtm by blast
  --"Define the custom conditions for the left and right child"
  define PreTl'  :: "('s, 'e) parallelTraceState \<Rightarrow> bool" where
    PreTl': "PreTl' = (\<lambda> (tr, s) . tr = [] \<and> s = s1)"
  define PreTr' :: "('s, 'e) parallelTraceState \<Rightarrow> bool" where
    PreTr': "PreTr' = (\<lambda> (tr, s) . tr = [] \<and> s = s2)"
  --"With the induction hypothesis, we construct some child in the SF semantics."
  --"First go left."
  have t1_start_in_set: "([], s1) \<in> {x'. PreTl' x'}" unfolding PreTl' by force
  have start_set_conv1: "{(tr, s, X). PreTl' (tr, PS2PTS s) \<and> N\<^sub>p s} = {x . \<exists> X . x = ([], PTS2PS s1, X)}"
    unfolding PreTl'
    by (auto simp add: PS2PTS2PS_id NpPTS2PS Np_PTS2PS2PTS_id)
  have "(tr \<restriction> \<alpha>, t1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>T{x'. PreTl' x'}"
    using densemp_monotone_single[OF ex_l t1_start_in_set] wff_lr
    by blast
  hence "\<exists>s' X. (tr \<restriction> \<alpha>, s', X) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x . \<exists> X . x = ([], PTS2PS s1, X)}"
    using IH1[of "tr \<restriction> \<alpha>" t1 PreTl' "\<lambda>(tr, s, X). PreTl' (tr, PS2PTS s) \<and> N\<^sub>p s", simplified] wff_lr
    unfolding start_set_conv1
    by force
  hence "\<exists> X s1' Xl . (tr \<restriction> \<alpha>, s1', Xl) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], PTS2PS s1, X)}"
    using sfsemp_predecessor wff_lr by blast
  then obtain Xl' s1' where G1: "(tr \<restriction> \<alpha>, s1', Xl') \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], PTS2PS s1, {})}"
    using sfsemp_general_sigma_exchange wff_lr by blast
  --"Next go right."
  have t2_start_in_set: "([], s2) \<in> {x'. PreTr' x'}" unfolding PreTr' by force
  have start_set_conv2: "{(tr, s, X). PreTr' (tr, PS2PTS s) \<and> N\<^sub>p s} = {x . \<exists> X . x = ([], PTS2PS s2, X)}"
    unfolding PreTr'
  by (auto simp add: PS2PTS2PS_id NpPTS2PS Np_PTS2PS2PTS_id)
  have "(tr \<restriction> \<beta>, t2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>T{x'. PreTr' x'}"
    using densemp_monotone_single[OF ex_r t2_start_in_set] wff_lr
    by blast
  hence "\<exists>s' X. (tr \<restriction> \<beta>, s', X) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{x . \<exists> X . x = ([], PTS2PS s2, X)}"
    using IH2[of "tr \<restriction> \<beta>" t2 PreTr' "\<lambda>(tr, s, X). PreTr' (tr, PS2PTS s) \<and> N\<^sub>p s", simplified] wff_lr
    unfolding start_set_conv2
    by force
  hence "\<exists> X s2' Xr . (tr \<restriction> \<beta>, s2', Xr) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], PTS2PS s2, X)}"
    using sfsemp_predecessor wff_lr by blast
  then obtain Xr' s2' where G2: "(tr \<restriction> \<beta>, s2', Xr') \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], PTS2PS s2, {})}"
    using sfsemp_general_sigma_exchange wff_lr by blast
  --"The last step is to construct some valid resulting refusal set."
  obtain Xlr where GX: "Xlr \<inter> (\<alpha> \<union> \<beta>) = Xl' \<inter> \<alpha> \<union> Xr' \<inter> \<beta>" and Xlr: "Xlr = Xl' \<inter> \<alpha> \<union> Xr' \<inter> \<beta>" by auto
  show "\<exists>X X1 X2 X' t1 t2 s1 s2.
        PreF ([], (par s1 s2), X') \<and>
        N\<^sub>p s1 \<and>
        N\<^sub>p s2 \<and> X \<inter> (\<alpha> \<union> \<beta>) = X1 \<inter> \<alpha> \<union> X2 \<inter> \<beta> \<and> (tr \<restriction> \<alpha>, t1, X1) \<in> \<lbrakk>cl\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s1, X')} \<and> (tr \<restriction> \<beta>, t2, X2) \<in> \<lbrakk>cr\<rbrakk>\<^sub>p\<^sub>S\<^sub>F{([], s2, X')}"
    unfolding sfsem\<^sub>p.simps
    using G1 G2 GPreF GX NpPTS2PS[of s1] NpPTS2PS[of s2]
      by blast
qed

end
