theory tree
imports Main
begin

datatype ('lf, 'nd) ATreeNode = seqn (ATreeNode_lf: 'lf) | parn (ATreeNode_l: "('lf, 'nd) ATreeNode") (ATreeNode_n: 'nd) (ATreeNode_r: "('lf, 'nd) ATreeNode")
datatype 'lf ATree = seq (ATree_lf: 'lf) | par (ATree_l: "'lf ATree") (ATree_r: "'lf ATree")
datatype NodeAccessor = Left | Right
datatype 'lf LeafOrNode = Leaf 'lf | Node "'lf ATree" "'lf ATree"
datatype ('lf, 'nd) LeafOrNodeN = LeafN 'lf | NodeN "('lf, 'nd) ATreeNode" 'nd "('lf, 'nd) ATreeNode"

lemma ATree_l: "ATree_l (par l r) = l"
  by auto

lemma ATree_r: "ATree_r (par l r) = r"
  by auto

paragraph {* ATreeNode to ATree *}

fun ATreeNode_to_ATree :: "('lf, 'nd) ATreeNode \<Rightarrow> 'lf ATree" where
  "ATreeNode_to_ATree (seqn lf) = seq lf " |
  "ATreeNode_to_ATree (parn l _ r) = par (ATreeNode_to_ATree l) (ATreeNode_to_ATree r)"

paragraph {* ATree Gathering *}

fun ATree_gather_hlpr :: "NodeAccessor list \<Rightarrow> 'lf ATree \<Rightarrow> (NodeAccessor list \<times> 'lf) set" where
  "ATree_gather_hlpr path (seq lf) = {(path, lf)}" |
  "ATree_gather_hlpr path (par l r) = ATree_gather_hlpr (path @ [Left]) l \<union> ATree_gather_hlpr (path @ [Right]) r"

lemma ATree_gather_hlpr_prefix: "(lrs, lf) \<in> ATree_gather_hlpr pref tree \<Longrightarrow> \<exists> lrs' . lrs = pref @ lrs' \<and> (\<forall> pref' . (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree)"
proof (induct tree arbitrary: pref lrs)
  show "\<And>x pref lrs. (lrs, lf) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' (seq x))" by auto
next
  fix tree1 tree2 pref lrs
  assume
    IH1: "\<And>pref lrs. (lrs, lf) \<in> ATree_gather_hlpr pref tree1 \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree1)" and
    IH2: "\<And>pref lrs. (lrs, lf) \<in> ATree_gather_hlpr pref tree2 \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree2)" and
    inside: "(lrs, lf) \<in> ATree_gather_hlpr pref (par tree1 tree2)"
  have "lrs \<noteq> []" using inside ATree_gather_hlpr.simps(2) IH1 IH2 Nil_is_append_conv Un_iff list.simps(3)
    by auto
  show "\<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' (par tree1 tree2))"
    using inside [unfolded "ATree_gather_hlpr.simps"]
    using IH1[of lrs "pref @ [NodeAccessor.Left]"]
    using IH2[of lrs "pref @ [NodeAccessor.Right]"]
    using ATree_gather_hlpr.simps(2) UnE Un_upper2 append_assoc subsetCE sup_ge1
  proof -
    { fix nns :: "NodeAccessor list \<Rightarrow> NodeAccessor list"
      have ff1: "\<And>p ns a aa. p \<in> ATree_gather_hlpr ns (par (a::'a ATree) aa) \<or> p \<notin> ATree_gather_hlpr (ns @ [NodeAccessor.Left]) a"
        by force
      have "\<And>p ns a aa. p \<in> ATree_gather_hlpr ns (par (a::'a ATree) aa) \<or> p \<notin> ATree_gather_hlpr (ns @ [NodeAccessor.Right]) aa"
        using ATree_gather_hlpr.simps(2) by blast
      then have "\<exists>ns. lrs = pref @ ns \<and> (nns ns @ ns, lf) \<in> ATree_gather_hlpr (nns ns) (par tree1 tree2)"
        using ff1 by (metis UnE \<open>(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) tree1 \<Longrightarrow> \<exists>lrs'. lrs = (pref @ [NodeAccessor.Left]) @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree1)\<close> \<open>(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) tree1 \<union> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) tree2\<close> \<open>(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) tree2 \<Longrightarrow> \<exists>lrs'. lrs = (pref @ [NodeAccessor.Right]) @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree2)\<close> append_assoc) }
    then show ?thesis
      by metis
  qed
qed

lemma ATree_gather_hlpr_prefix_fix: "(lrs, lf) \<in> ATree_gather_hlpr pref tree \<Longrightarrow> \<exists> lrs' . lrs = pref @ lrs' \<and> (pref' @ lrs', lf) \<in> ATree_gather_hlpr pref' tree"
  using ATree_gather_hlpr_prefix by fast

lemma ATree_gather_hlpr_left: "(pref @ Left # lrs, lf) \<in> ATree_gather_hlpr pref (par l r) = ((pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' l)"
proof
  show "(pref @ NodeAccessor.Left # lrs, lf) \<in> ATree_gather_hlpr pref (par l r) \<Longrightarrow> (pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' l"
  unfolding "ATree_gather_hlpr.simps"
  using ATree_gather_hlpr_prefix by fastforce
next
  show "(pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' l \<Longrightarrow> (pref @ NodeAccessor.Left # lrs, lf) \<in> ATree_gather_hlpr pref (par l r)"
    using ATree_gather_hlpr.simps
    by (metis ATree_gather_hlpr_prefix Un_iff append_Cons append_Nil same_append_eq)
qed

lemma ATree_gather_hlpr_right: "(pref @ Right # lrs, lf) \<in> ATree_gather_hlpr pref (par l r) = ((pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' r)"
proof
  show "(pref @ Right # lrs, lf) \<in> ATree_gather_hlpr pref (par l r) \<Longrightarrow> (pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' r"
    unfolding "ATree_gather_hlpr.simps"
    using ATree_gather_hlpr_prefix by fastforce
next
  show "(pref' @ lrs, lf) \<in> ATree_gather_hlpr pref' r \<Longrightarrow> (pref @ NodeAccessor.Right # lrs, lf) \<in> ATree_gather_hlpr pref (par l r)"
    using ATree_gather_hlpr.simps
    by (metis ATree_gather_hlpr_prefix Un_iff append_Cons append_Nil same_append_eq)
qed

definition ATree_gather :: "'lf ATree \<Rightarrow> (NodeAccessor list \<times> 'lf) set" ("\<union>\<^sub>A\<^sub>T _" [70]) where
  "(\<union>\<^sub>A\<^sub>T tree) = (ATree_gather_hlpr ([]::NodeAccessor list) tree)"

lemma ATree_gather_seq[simp]: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T seq lf' = ((lrs, lf) \<in> {([], lf')})"
  unfolding ATree_gather_def "ATree_gather_hlpr.simps" ..

lemma ATree_gather_par[simp]: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T par l r \<Longrightarrow> \<exists> lr lrs' . lrs = lr # lrs' \<and> ((lrs', lf) \<in> \<union>\<^sub>A\<^sub>T l \<and> lr = Left \<or> (lrs', lf) \<in> \<union>\<^sub>A\<^sub>T r \<and> lr = Right)"
  unfolding ATree_gather_def "ATree_gather_hlpr.simps"
  using ATree_gather_hlpr_prefix_fix[of lrs lf "[Left]" "l" "[]"]
  using ATree_gather_hlpr_prefix_fix[of lrs lf "[Right]" "r" "[]"]
  unfolding append_Nil
  by force

lemma ATree_gather_left: "(Left # lrs, lf) \<in> \<union>\<^sub>A\<^sub>T (par l r) = ((lrs, lf) \<in> \<union>\<^sub>A\<^sub>T l)"
  unfolding ATree_gather_def
  using ATree_gather_hlpr_left [of "[]" lrs lf l r "[]", unfolded append_Nil] .

lemma ATree_gather_right: "(Right # lrs, lf) \<in> \<union>\<^sub>A\<^sub>T (par l r) = ((lrs, lf) \<in> \<union>\<^sub>A\<^sub>T r)"
  unfolding ATree_gather_def
  using ATree_gather_hlpr_right [of "[]" lrs lf l r "[]", unfolded append_Nil] .

lemma ATree_gather_same_path:
  assumes
    inside1: "(lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tree" and
    inside2: "(lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tree"
  shows "lf1 = lf2"
  using assms
  unfolding ATree_gather_def
proof -
  {
    fix pref
    have "(lrs, lf1) \<in> ATree_gather_hlpr pref tree \<Longrightarrow> (lrs, lf2) \<in> ATree_gather_hlpr pref tree \<Longrightarrow> lf1 = lf2"
    proof (induct tree arbitrary: pref lrs)
      show "\<And>x pref lrs. (lrs, lf1) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> (lrs, lf2) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> lf1 = lf2" by simp
    next
      fix tree1 tree2 pref lrs
      assume
        IH1: "\<And>pref lrs. (lrs, lf1) \<in> ATree_gather_hlpr pref tree1 \<Longrightarrow> (lrs, lf2) \<in> ATree_gather_hlpr pref tree1 \<Longrightarrow> lf1 = lf2" and
        IH2: "\<And>pref lrs. (lrs, lf1) \<in> ATree_gather_hlpr pref tree2 \<Longrightarrow> (lrs, lf2) \<in> ATree_gather_hlpr pref tree2 \<Longrightarrow> lf1 = lf2" and
        inside1: "(lrs, lf1) \<in> ATree_gather_hlpr pref (par tree1 tree2)" and
        inside2: "(lrs, lf2) \<in> ATree_gather_hlpr pref (par tree1 tree2)"
      from ATree_gather_hlpr_prefix[OF inside1]
      obtain lrs' where lrs: "lrs = pref @ lrs'" and allpref1: "\<forall>pref'. (pref' @ lrs', lf1) \<in> ATree_gather_hlpr pref' (par tree1 tree2)" by blast
      with ATree_gather_hlpr_prefix[OF inside1]
      have allpref2: "\<forall>pref'. (pref' @ lrs', lf2) \<in> ATree_gather_hlpr pref' (par tree1 tree2)" using ATree_gather_hlpr_prefix[OF inside2] by blast
      have inside1: "(lrs', lf1) \<in> ATree_gather_hlpr [] (par tree1 tree2)" using allpref1 append_Nil by metis
      have inside2: "(lrs', lf2) \<in> ATree_gather_hlpr [] (par tree1 tree2)" using allpref2 append_Nil by metis
      have "lrs' \<noteq> []" using ATree_gather_hlpr_prefix inside1 by fastforce
      then obtain lr lrs'' where lrs': "lrs' = lr # lrs''" using list.exhaust by blast
      consider (left) "lr = Left" | (right) "lr = Right" using "NodeAccessor.exhaust" by blast
      thus "lf1 = lf2"
      proof cases
        case left
        have inside1: "(lrs', lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] tree1"
          using inside1[unfolded "ATree_gather_hlpr.simps" append_Nil] ATree_gather_hlpr_prefix
          unfolding lrs' left
          by fastforce
        have inside2: "(lrs', lf2) \<in> ATree_gather_hlpr [NodeAccessor.Left] tree1"
          using inside2[unfolded "ATree_gather_hlpr.simps" append_Nil] ATree_gather_hlpr_prefix
          unfolding lrs' left
          by fastforce
        from IH1[OF inside1 inside2] show ?thesis .
      next
        case right
        have inside1: "(lrs', lf1) \<in> ATree_gather_hlpr [NodeAccessor.Right] tree2"
          using inside1[unfolded "ATree_gather_hlpr.simps" append_Nil] ATree_gather_hlpr_prefix
          unfolding lrs' right
          by fastforce
        have inside2: "(lrs', lf2) \<in> ATree_gather_hlpr [NodeAccessor.Right] tree2"
          using inside2[unfolded "ATree_gather_hlpr.simps" append_Nil] ATree_gather_hlpr_prefix
          unfolding lrs' right
          by fastforce
        from IH2[OF inside1 inside2] show ?thesis .
      qed
    qed
  }
  thus "(lrs, lf1) \<in> ATree_gather_hlpr [] tree \<Longrightarrow> (lrs, lf2) \<in> ATree_gather_hlpr [] tree \<Longrightarrow> lf1 = lf2" by blast
qed

lemma ATree_gather_not_empty[simp]: "\<union>\<^sub>A\<^sub>T s \<noteq> {}"
proof -
  have "\<And> path . ATree_gather_hlpr path s \<noteq> {}"
  proof (induct s, auto) qed
  thus ?thesis unfolding ATree_gather_def by simp
qed

lemma ATree_access_exists:
  shows "\<exists> lrs lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
  using ATree_gather_not_empty by fastforce

paragraph {* ATreeNode Gathering *}

fun ATreeNode_gather_hlpr :: "(NodeAccessor \<times> 'nd) list \<Rightarrow> ('lf, 'nd) ATreeNode \<Rightarrow> ((NodeAccessor \<times> 'nd) list \<times> 'lf) set" where
  "ATreeNode_gather_hlpr path (seqn lf) = {(path, lf)}" |
  "ATreeNode_gather_hlpr path (parn l n r) = ATreeNode_gather_hlpr (path @ [(Left, n)]) l \<union> ATreeNode_gather_hlpr (path @ [(Right, n)]) r"

lemma ATreeNode_gather_hlpr_prefix: "(lrs, lf) \<in> ATreeNode_gather_hlpr pref tree \<Longrightarrow> \<exists> lrs' . lrs = pref @ lrs' \<and> (\<forall> pref' . (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' tree)"
proof (induct tree arbitrary: pref lrs)
  show "\<And>x pref lrs. (lrs, lf) \<in> ATreeNode_gather_hlpr pref (seqn x) \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' (seqn x))" by auto
next
  fix tree1 tree2 pref lrs n
  assume
    IH1: "\<And>pref lrs. (lrs, lf) \<in> ATreeNode_gather_hlpr pref tree1 \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' tree1)" and
    IH2: "\<And>pref lrs. (lrs, lf) \<in> ATreeNode_gather_hlpr pref tree2 \<Longrightarrow> \<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' tree2)" and
    inside: "(lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn tree1 n tree2)"
  have "lrs \<noteq> []" using inside ATreeNode_gather_hlpr.simps(2) IH1 IH2 Nil_is_append_conv Un_iff list.simps(3)
    by auto
  show "\<exists>lrs'. lrs = pref @ lrs' \<and> (\<forall>pref'. (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' (parn tree1 n tree2))"
    using inside [unfolded "ATreeNode_gather_hlpr.simps"]
    using IH1[of lrs "pref @ [(Left, n)]"]
    using IH2[of lrs "pref @ [(Right, n)]"]
    unfolding ATreeNode_gather_hlpr.simps(2) append_assoc
    by (metis Un_iff append_assoc)
qed

lemma ATreeNode_gather_hlpr_prefix_fix: "(lrs, lf) \<in> ATreeNode_gather_hlpr pref tree \<Longrightarrow> \<exists> lrs' . lrs = pref @ lrs' \<and> (pref' @ lrs', lf) \<in> ATreeNode_gather_hlpr pref' tree"
  using ATreeNode_gather_hlpr_prefix by fast

lemma ATreeNode_gather_hlpr_left: "(pref @ (Left, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r) = ((pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' l)"
proof
  show "(pref @ (Left, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r) \<Longrightarrow> (pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' l"
  unfolding "ATreeNode_gather_hlpr.simps"
  using ATreeNode_gather_hlpr_prefix by fastforce
next
  show "(pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' l \<Longrightarrow> (pref @ (Left, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r)"
    using ATreeNode_gather_hlpr.simps
    by (metis ATreeNode_gather_hlpr_prefix Un_iff append_Cons append_Nil same_append_eq)
qed

lemma ATreeNode_gather_hlpr_right: "(pref @ (Right, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r) = ((pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' r)"
proof
  show "(pref @ (Right, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r) \<Longrightarrow> (pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' r"
    unfolding "ATreeNode_gather_hlpr.simps"
    using ATreeNode_gather_hlpr_prefix by fastforce
next
  show "(pref' @ lrs, lf) \<in> ATreeNode_gather_hlpr pref' r \<Longrightarrow> (pref @ (Right, n) # lrs, lf) \<in> ATreeNode_gather_hlpr pref (parn l n r)"
    using ATreeNode_gather_hlpr.simps
    by (metis ATreeNode_gather_hlpr_prefix Un_iff append_Cons append_Nil same_append_eq)
qed

definition ATreeNode_gather :: "('lf, 'nd) ATreeNode \<Rightarrow> ((NodeAccessor \<times> 'nd) list \<times> 'lf) set" ("\<union>\<^sub>A\<^sub>T\<^sub>N _" [70]) where
  "(\<union>\<^sub>A\<^sub>T\<^sub>N tree) = (ATreeNode_gather_hlpr ([]::(NodeAccessor \<times> 'nd) list) tree)"

lemma ATreeNode_gather_seq[simp]: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N seqn lf' = ((lrs, lf) \<in> {([], lf')})"
  unfolding ATreeNode_gather_def "ATreeNode_gather_hlpr.simps" ..

lemma ATreeNode_gather_par[simp]: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn l n r \<Longrightarrow> \<exists> lr lrs' . lrs = (lr, n) # lrs' \<and> ((lrs', lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l \<and> lr = Left \<or> (lrs', lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r \<and> lr = Right)"
  unfolding ATreeNode_gather_def "ATreeNode_gather_hlpr.simps"
  using ATreeNode_gather_hlpr_prefix_fix[of lrs lf "[(Left, n)]" "l" "[]"]
  using ATreeNode_gather_hlpr_prefix_fix[of lrs lf "[(Right, n)]" "r" "[]"]
  unfolding append_Nil
  by force

lemma ATreeNode_gather_left: "((Left, n) # lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r) = ((lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l)"
  unfolding ATreeNode_gather_def
  using ATreeNode_gather_hlpr_left [of "[]" n lrs lf l r "[]", unfolded append_Nil] .

lemma ATreeNode_gather_right: "((Right, n) # lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (parn l n r) = ((lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r)"
  unfolding ATreeNode_gather_def
  using ATreeNode_gather_hlpr_right [of "[]" n lrs lf l r "[]", unfolded append_Nil] .

lemma ATreeNode_gather_same_path:
  fixes tree :: "('lf, 'nd) ATreeNode"
  assumes
    inside1: "(lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tree" and
    inside2: "(lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tree"
  shows "lf1 = lf2"
  using assms
  unfolding ATreeNode_gather_def
proof -
  {
    fix pref
    have "(lrs, lf1) \<in> ATreeNode_gather_hlpr pref tree \<Longrightarrow> (lrs, lf2) \<in> ATreeNode_gather_hlpr pref tree \<Longrightarrow> lf1 = lf2"
    proof (induct tree arbitrary: pref lrs)
      show "\<And>x pref lrs. (lrs, lf1) \<in> ATreeNode_gather_hlpr pref (seqn x) \<Longrightarrow> (lrs, lf2) \<in> ATreeNode_gather_hlpr pref (seqn x) \<Longrightarrow> lf1 = lf2" by simp
    next
      fix tree1 tree2 pref lrs n
      assume
        IH1: "\<And>pref lrs. (lrs, lf1) \<in> ATreeNode_gather_hlpr pref tree1 \<Longrightarrow> (lrs, lf2) \<in> ATreeNode_gather_hlpr pref tree1 \<Longrightarrow> lf1 = lf2" and
        IH2: "\<And>pref lrs. (lrs, lf1) \<in> ATreeNode_gather_hlpr pref tree2 \<Longrightarrow> (lrs, lf2) \<in> ATreeNode_gather_hlpr pref tree2 \<Longrightarrow> lf1 = lf2" and
        inside1: "(lrs, lf1) \<in> ATreeNode_gather_hlpr pref (parn tree1 n tree2)" and
        inside2: "(lrs, lf2) \<in> ATreeNode_gather_hlpr pref (parn tree1 n tree2)"
      from ATreeNode_gather_hlpr_prefix[OF inside1]
      obtain lrs' where lrs: "lrs = pref @ lrs'" and allpref1: "\<forall>pref'. (pref' @ lrs', lf1) \<in> ATreeNode_gather_hlpr pref' (parn tree1 n tree2)" by blast
      with ATreeNode_gather_hlpr_prefix[OF inside1]
      have allpref2: "\<forall>pref'. (pref' @ lrs', lf2) \<in> ATreeNode_gather_hlpr pref' (parn tree1 n tree2)" using ATreeNode_gather_hlpr_prefix[OF inside2] by blast
      have inside1: "(lrs', lf1) \<in> ATreeNode_gather_hlpr [] (parn tree1 n tree2)" using allpref1 append_Nil by metis
      have inside2: "(lrs', lf2) \<in> ATreeNode_gather_hlpr [] (parn tree1 n tree2)" using allpref2 append_Nil by metis
      have "lrs' \<noteq> []" using ATreeNode_gather_hlpr_prefix inside1 by fastforce
      then obtain lr lrs'' where lrs': "lrs' = (lr, n) # lrs''" using list.exhaust
        by (metis ATreeNode_gather_def ATreeNode_gather_par inside1)
      consider (left) "lr = Left" | (right) "lr = Right" using "NodeAccessor.exhaust" by blast
      thus "lf1 = lf2"
      proof cases
        case left
        have inside1: "(lrs', lf1) \<in> ATreeNode_gather_hlpr [(Left, n)] tree1"
          using inside1[unfolded "ATreeNode_gather_hlpr.simps" append_Nil] ATreeNode_gather_hlpr_prefix
          unfolding lrs' left
          by fastforce
        have inside2: "(lrs', lf2) \<in> ATreeNode_gather_hlpr [(Left, n)] tree1"
          using inside2[unfolded "ATreeNode_gather_hlpr.simps" append_Nil] ATreeNode_gather_hlpr_prefix
          unfolding lrs' left
          by fastforce
        from IH1[OF inside1 inside2] show ?thesis .
      next
        case right
        have inside1: "(lrs', lf1) \<in> ATreeNode_gather_hlpr [(Right, n)] tree2"
          using inside1[unfolded "ATreeNode_gather_hlpr.simps" append_Nil] ATreeNode_gather_hlpr_prefix
          unfolding lrs' right
          by fastforce
        have inside2: "(lrs', lf2) \<in> ATreeNode_gather_hlpr [(Right, n)] tree2"
          using inside2[unfolded "ATreeNode_gather_hlpr.simps" append_Nil] ATreeNode_gather_hlpr_prefix
          unfolding lrs' right
          by fastforce
        from IH2[OF inside1 inside2] show ?thesis .
      qed
    qed
  }
  thus "(lrs, lf1) \<in> ATreeNode_gather_hlpr [] tree \<Longrightarrow> (lrs, lf2) \<in> ATreeNode_gather_hlpr [] tree \<Longrightarrow> lf1 = lf2" by blast
qed

lemma ATreeNode_gather_not_empty[simp]: fixes s :: "('lf, 'nd) ATreeNode" shows "\<union>\<^sub>A\<^sub>T\<^sub>N s \<noteq> {}"
proof -
  have "\<And> path . ATreeNode_gather_hlpr path s \<noteq> {}"
  proof (induct s, auto) qed
  thus ?thesis unfolding ATreeNode_gather_def by simp
qed

lemma ATreeNode_access_exists:
  shows "\<exists> nds lrs lf . (zip lrs nds, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr \<and> length lrs = length nds"
    by (metis ATreeNode_gather_not_empty eq_fst_iff equals0I length_map zip_map_fst_snd)

paragraph {* ATree merge *}

fun ATree_merge :: "'lf1 ATree \<Rightarrow> 'lf2 ATree \<Rightarrow> ('lf1 \<times> 'lf2) ATree" ("_ \<times>\<^sub>A\<^sub>T _" [75]) where
  "((seq l) \<times>\<^sub>A\<^sub>T (seq r)) = seq (l, r)" |
  "((par l1 r1) \<times>\<^sub>A\<^sub>T (par l2 r2)) = par (l1 \<times>\<^sub>A\<^sub>T l2) (r1 \<times>\<^sub>A\<^sub>T r2)"

fun ATreen_merge :: "'lf1 ATree \<Rightarrow> ('lf2, 'nd) ATreeNode \<Rightarrow> ('lf1 \<times> 'lf2, 'nd) ATreeNode" ("_ \<times>\<^sub>A\<^sub>T\<^sub>n _" [75]) where
  "((seq l) \<times>\<^sub>A\<^sub>T\<^sub>n (seqn r)) = seqn (l, r)" |
  "((par l1 r1) \<times>\<^sub>A\<^sub>T\<^sub>n (parn l2 n r2)) = parn (l1 \<times>\<^sub>A\<^sub>T\<^sub>n l2) n (r1 \<times>\<^sub>A\<^sub>T\<^sub>n r2)"
  
paragraph {* Leaf properties *}

fun leaf1_property :: "('a \<Rightarrow> bool) \<Rightarrow> (bool \<Rightarrow> bool \<Rightarrow> bool) \<Rightarrow> 'a ATree \<Rightarrow> bool" where
  "leaf1_property P _ (seq a) = P a" |
  "leaf1_property P bl (par s1 s2) =  bl (leaf1_property P bl s1) (leaf1_property P bl s2)"

fun leafn1_property :: "('a \<Rightarrow> bool) \<Rightarrow> (bool \<Rightarrow> bool \<Rightarrow> bool) \<Rightarrow> ('a, _) ATreeNode \<Rightarrow> bool" where
  "leafn1_property P _ (seqn a) = P a" |
  "leafn1_property P bl (parn s1 _ s2) =  bl (leafn1_property P bl s1) (leafn1_property P bl s2)"

lemma leaf1n1_conversion: "leaf1_property f b (ATreeNode_to_ATree tr) = leafn1_property f b tr"
proof (induct tr, auto) qed

fun leaf_property :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> (bool \<Rightarrow> bool \<Rightarrow> bool) \<Rightarrow> 'a ATree \<Rightarrow> 'b ATree \<Rightarrow> bool" where
  "leaf_property P bl (seq a) (seq b) = P a b" |
  "leaf_property P bl (par _ _) (seq _) = False" |
  "leaf_property P bl (seq _) (par _ _) = False" |
  "leaf_property P bl (par s1 s2) (par t1 t2) = bl (leaf_property P bl s1 t1) (leaf_property P bl s2 t2)"

fun leafn_property :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> (bool \<Rightarrow> bool \<Rightarrow> bool) \<Rightarrow> 'a ATree \<Rightarrow> ('b, 'n) ATreeNode \<Rightarrow> bool" where
  "leafn_property P bl (seq a) (seqn b) = P a b" |
  "leafn_property P bl (par _ _) (seqn _) = False" |
  "leafn_property P bl (seq _) (parn _ _ _) = False" |
  "leafn_property P bl (par s1 s2) (parn t1 _ t2) = bl (leafn_property P bl s1 t1) (leafn_property P bl s2 t2)"
  
text {* lf is of type @{text "@{type NodeAccessor list} \<times> 'lf"}, hence we have to extrace @{text 'lf} via snd. *}

lemma "leaf1_property P op \<or> c \<Longrightarrow> \<exists> lf \<in> \<union>\<^sub>A\<^sub>T c . P (snd lf)"
proof -
  have "\<And> path . leaf1_property P op \<or> c \<Longrightarrow> \<exists>lf\<in>ATree_gather_hlpr path c. P (snd lf)"
    proof (induct c, auto) qed
  thus "leaf1_property P op \<or> c \<Longrightarrow> \<exists>lf\<in>\<union>\<^sub>A\<^sub>T c. P (snd lf)"
    unfolding ATree_gather_def
    by blast
qed

lemma "leafn1_property P op \<or> c \<Longrightarrow> \<exists> lf \<in> \<union>\<^sub>A\<^sub>T\<^sub>N c . P (snd lf)"
proof -
  have "\<And> path . leafn1_property P op \<or> c \<Longrightarrow> \<exists>lf\<in>ATreeNode_gather_hlpr path c. P (snd lf)"
    proof (induct c, auto) qed
  thus "leafn1_property P op \<or> c \<Longrightarrow> \<exists>lf\<in>\<union>\<^sub>A\<^sub>T\<^sub>N c. P (snd lf)"
    unfolding ATreeNode_gather_def
    by blast
qed

fun leafnn_property :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> (bool \<Rightarrow> bool \<Rightarrow> bool) \<Rightarrow> ('a, 'm) ATreeNode \<Rightarrow> ('b, 'n) ATreeNode \<Rightarrow> bool" where
  "leafnn_property P bl (seqn a) (seqn b) = P a b" |
  "leafnn_property P bl (parn _ _ _) (seqn _) = False" |
  "leafnn_property P bl (seqn _) (parn _ _ _) = False" |
  "leafnn_property P bl (parn s1 _ s2) (parn t1 _ t2) = bl (leafnn_property P bl s1 t1) (leafnn_property P bl s2 t2)"

lemma leaf1_property_inherit:
  assumes "\<And> l . ud_prop l = leaf1_property P op \<and> l"
  shows "ud_prop (par l1 l2) = (ud_prop l1 \<and> ud_prop l2)"
  using assms by simp

lemma leafn1_property_inherit:
  assumes "\<And> l . ud_prop l = leafn1_property P op \<and> l"
  shows "ud_prop (parn l1 n l2) = (ud_prop l1 \<and> ud_prop l2)"
  using assms by simp

lemma leaf_property_inherit:
  assumes "\<And> l r . ud_prop l r = leaf_property P op \<and> l r"
  shows "ud_prop (par l1 l2) (par r1 r2) = (ud_prop l1 r1 \<and> ud_prop l2 r2)"
  using assms by simp

lemma leafn_property_inherit:
  assumes "\<And> l r . ud_prop l r = leafn_property P op \<and> l r"
  shows "ud_prop (par l1 l2) (parn r1 n r2) = (ud_prop l1 r1 \<and> ud_prop l2 r2)"
  using assms by simp

lemma leafnn_property_inherit:
  assumes "\<And> l r . ud_prop l r = leafnn_property P op \<and> l r"
  shows "ud_prop (parn l1 n1 l2) (parn r1 n2 r2) = (ud_prop l1 r1 \<and> ud_prop l2 r2)"
  using assms by simp

definition tree_mirror: "tree_mirror a b = leaf_property (\<lambda> _ _ . True) (op \<and>) a b"
definition treen_mirror: "treen_mirror a b = leafn_property (\<lambda> _ _ . True) (op \<and>) a b"
definition treenn_mirror: "treenn_mirror a b = leafnn_property (\<lambda> _ _ . True) (op \<and>) a b"

definition "tree_is_seq tree = (\<exists> s . tree = seq s)"
definition "tree_is_par tree = (\<exists> tr1 tr2 . tree = par tr1 tr2)"

text {* There is a problem with using the left hand side in a lemma: As there is no type information
concerning the internal node data type, Isabelle assumes for everyone another type making them unequal. 
This lemma removes this inconvenience. *}

lemma leaf_property_seql: "leaf_property P bl t (seq ss) \<Longrightarrow> \<exists> ts . t = seq ts"
proof (induct t, auto) qed

lemma leaf_property_seqr: "leaf_property P bl (seq ts) s \<Longrightarrow> \<exists> ss . s = seq ss"
proof (induct s, auto) qed

lemma leaf_property_parl: "leaf_property P bl t (par ss1 ss2) \<Longrightarrow> \<exists> ts1 ts2 . t = par ts1 ts2"
proof (induct t, auto) qed

lemma leaf_property_parr: "leaf_property P bl (par ts1 ts2) s \<Longrightarrow> \<exists> ss1 ss2 . s = par ss1 ss2"
proof (induct s, auto) qed

lemma leafn_property_seql: "leafn_property P bl t (seqn ss) \<Longrightarrow> \<exists> ts . t = seq ts"
proof (induct t, auto) qed

lemma leafn_property_seqr: "leafn_property P bl (seq ts) s \<Longrightarrow> \<exists> ss . s = seqn ss"
proof (induct s, auto) qed

lemma leafn_property_parl: "leafn_property P bl t (parn ss1 v ss2) \<Longrightarrow> \<exists> ts1 ts2 . t = par ts1 ts2"
proof (induct t, auto) qed

lemma leafn_property_parr: "leafn_property P bl (par ts1 ts2) s \<Longrightarrow> \<exists> ss1 ss2 v . s = parn ss1 v ss2"
proof (induct s, auto) qed

lemma leafnn_property_seql: "leafnn_property P bl t (seqn ss) \<Longrightarrow> \<exists> ts . t = seqn ts"
proof (induct t, auto) qed

lemma leafnn_property_seqr: "leafnn_property P bl (seqn ts) s \<Longrightarrow> \<exists> ss . s = seqn ss"
proof (induct s, auto) qed

lemma leafnn_property_parl: "leafnn_property P bl t (parn ss1 v ss2) \<Longrightarrow> \<exists> ts1 ts2 v . t = parn ts1 v ts2"
proof (induct t, auto) qed

lemma leafnn_property_parr: "leafnn_property P bl (parn ts1 v ts2) s \<Longrightarrow> \<exists> ss1 ss2 v . s = parn ss1 v ss2"
proof (induct s, auto) qed

lemma leaf_property_reflexive [simp]: "\<forall> x . P x x \<Longrightarrow> bl True True \<Longrightarrow> leaf_property P bl s s"
proof (induct s, auto) qed

lemma leafnn_property_reflexive [simp]: "\<forall> x . P x x \<Longrightarrow> bl True True \<Longrightarrow> leafnn_property P bl s s"
proof (induct s, auto) qed

declare tree_mirror[simp] treen_mirror[simp] treenn_mirror[simp]

lemma tree_mirror_reflexive[simp]: "tree_mirror tr tr"
 by simp

lemma treenn_mirror_reflexive[simp]: "treenn_mirror tr tr"
 by simp

lemma tree_mirror_unfold_seqr[simp]: "tree_mirror a (seq x) = (\<exists> s . a = seq s)"
proof (induct a, auto) qed
lemma tree_mirror_unfold_seql[simp]: "tree_mirror (seq x) a = (\<exists> s . a = seq s)"
proof (induct a, auto) qed
lemma treen_mirror_unfold_seqr[simp]: "treen_mirror a (seqn x) = (\<exists> s . a = seq s)"
proof (induct a, auto) qed
lemma treen_mirror_unfold_seql[simp]: "treen_mirror (seq x) a = (\<exists> s . a = seqn s)"
proof (induct a, auto) qed
lemma treenn_mirror_unfold_seqr[simp]: "treenn_mirror a (seqn x) = (\<exists> s . a = seqn s)"
proof (induct a, auto) qed
lemma treenn_mirror_unfold_seql[simp]: "treenn_mirror (seqn x) a = (\<exists> s . a = seqn s)"
proof (induct a, auto) qed

lemma tree_mirror_unfold_parr[simp]: "tree_mirror a (par c1 c2) \<Longrightarrow> (\<exists> s1 s2 . a = par s1 s2)"
proof (induct a, auto) qed
lemma tree_mirror_unfold_parl[simp]: "tree_mirror (par c1 c2) a \<Longrightarrow> (\<exists> s1 s2 . a = par s1 s2)"
proof (induct a, auto) qed
lemma treen_mirror_unfold_parr[simp]: "treen_mirror a (parn c1 n c2) \<Longrightarrow> (\<exists> s1 s2 . a = par s1 s2)"
proof (induct a, auto) qed
lemma treen_mirror_unfold_parl[simp]: "treen_mirror (par c1 c2) a \<Longrightarrow> (\<exists> s1 s2 n . a = parn s1 n s2)"
proof (induct a, auto) qed
lemma treenn_mirror_unfold_parr[simp]: "treenn_mirror a (parn c1 n c2) \<Longrightarrow> (\<exists> s1 s2 n . a = parn s1 n s2)"
proof (induct a, auto) qed
lemma treenn_mirror_unfold_parl[simp]: "treenn_mirror (parn c1 n c2) a \<Longrightarrow> (\<exists> s1 s2 n . a = parn s1 n s2)"
proof (induct a, auto) qed

lemma tree_mirror_inherit[simp]: "tree_mirror (par a1 a2) (par b1 b2) = (tree_mirror a1 b1 \<and> tree_mirror a2 b2)"
by auto
lemma treen_mirror_inherit[simp]: "treen_mirror (par a1 a2) (parn b1 n b2) = (treen_mirror a1 b1 \<and> treen_mirror a2 b2)"
by auto
lemma treenn_mirror_inherit[simp]: "treenn_mirror (parn a1 n1 a2) (parn b1 n2 b2) = (treenn_mirror a1 b1 \<and> treenn_mirror a2 b2)"
by auto

lemma tree_mirror_symmetric[simp]: "tree_mirror l r \<Longrightarrow> tree_mirror r l"
proof (induct l arbitrary: r, auto simp del: tree_mirror)
  fix l1 l2 :: "'a ATree" and r :: "'b ATree"
  assume
    IH1: "\<And>(r::'b ATree). tree_mirror l1 r \<Longrightarrow> tree_mirror r l1" and
    IH2: "\<And>(r::'b ATree). tree_mirror l2 r \<Longrightarrow> tree_mirror r l2" and
    tm: "tree_mirror (par l1 l2) r"
  then obtain r1 r2 where r: "r = par r1 r2" by force
  have tml: "tree_mirror l1 r1" using tm r by fastforce
  have tmr: "tree_mirror l2 r2" using tm r by fastforce
  show "tree_mirror r (par l1 l2)"
    using IH1[OF tml] IH2[OF tmr] unfolding r by fastforce
qed

lemma treenn_mirror_symmetric[simp]: "treenn_mirror l r \<Longrightarrow> treenn_mirror r l"
proof (induct l arbitrary: r, auto simp del: treenn_mirror)
  fix l1 l2 :: "('a, 'b) ATreeNode" and r :: "('c, 'd) ATreeNode" and nl
  assume
    IH1: "\<And>(r::('c, 'd) ATreeNode). treenn_mirror l1 r \<Longrightarrow> treenn_mirror r l1" and
    IH2: "\<And>(r::('c, 'd) ATreeNode). treenn_mirror l2 r \<Longrightarrow> treenn_mirror r l2" and
    tm: "treenn_mirror (parn l1 nl l2) r"
  then obtain r1 r2 nr where r: "r = parn r1 nr r2" using treenn_mirror_unfold_parl by metis
  have tml: "treenn_mirror l1 r1" using tm r by fastforce
  have tmr: "treenn_mirror l2 r2" using tm r by fastforce
  show "treenn_mirror r (parn l1 nl l2)"
    using IH1[OF tml] IH2[OF tmr] unfolding r by fastforce
qed

lemma tree_mirror_exists_l: "\<exists> l . tree_mirror l r"
proof (induct r, auto)
  show "\<And>x. \<exists>l. leaf_property (\<lambda>_ _. True) op \<and> l (seq x)"
    by (metis leaf_property.simps(1))
next
  fix r1 r2 :: "'b ATree" and l1 l2 :: "'a ATree"
  assume
    lp1: "leaf_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leaf_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leaf_property (\<lambda>_ _. True) op \<and> (par l1 l2) (par r1 r2)"
    unfolding leaf_property.simps ..
  thus "\<exists>(l::'a ATree). leaf_property (\<lambda>_ _. True) op \<and> l (par r1 r2)"
    by blast
qed

lemma tree_mirror_exists_r: "\<exists> r . tree_mirror l r"
proof (induct l, auto)
  show "\<And>x. \<exists>r. leaf_property (\<lambda>_ _. True) op \<and> (seq x) r"
    by (metis leaf_property.simps(1))
next
  fix r1 r2 :: "'a ATree" and l1 l2 :: "'b ATree"
  assume
    lp1: "leaf_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leaf_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leaf_property (\<lambda>_ _. True) op \<and> (par l1 l2) (par r1 r2)"
    unfolding leaf_property.simps ..
  thus "\<exists>(r::'a ATree). leaf_property (\<lambda>_ _. True) op \<and> (par l1 l2) r"
    by blast
qed

lemma treen_mirror_exists_l: "\<exists> l . treen_mirror l r"
proof (induct r, auto)
  show "\<And>x. \<exists>l. leafn_property (\<lambda>_ _. True) op \<and> l (seqn x)"
    by (metis leafn_property.simps(1))
next
  fix r1 r2 :: "('b, 'c) ATreeNode" and l1 l2 :: "'a ATree" and x
  assume
    lp1: "leafn_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leafn_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leafn_property (\<lambda>_ _. True) op \<and> (par l1 l2) (parn r1 x r2)"
    unfolding leaf_property.simps by simp
  thus "\<exists>(l::'a ATree). leafn_property (\<lambda>_ _. True) op \<and> l (parn r1 x r2)"
    by blast
qed

lemma treen_mirror_exists_r: "\<exists> r . treen_mirror l r"
proof (induct l, auto)
  show "\<And>x. \<exists>r. leafn_property (\<lambda>_ _. True) op \<and> (seq x) r"
    by (metis leafn_property.simps(1))
next
  fix r1 r2 :: "('a, 'b) ATreeNode" and l1 l2 :: "'c ATree" and x
  assume
    lp1: "leafn_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leafn_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leafn_property (\<lambda>_ _. True) op \<and> (par l1 l2) (parn r1 x r2)"
    unfolding leaf_property.simps by simp
  thus "\<exists>(r::('a, 'b) ATreeNode). leafn_property (\<lambda>_ _. True) op \<and> (par l1 l2) r"
    by blast
qed

lemma treenn_mirror_exists_l: "\<exists> l . treenn_mirror l r"
proof (induct r, auto)
  show "\<And>x. \<exists>l. leafnn_property (\<lambda>_ _. True) op \<and> l (seqn x)"
    by (metis leafnn_property.simps(1))
next
  fix r1 r2 :: "('c, 'd) ATreeNode" and l1 l2 :: "('a, 'b) ATreeNode" and xl xr
  assume
    lp1: "leafnn_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leafnn_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leafnn_property (\<lambda>_ _. True) op \<and> (parn l1 xl l2) (parn r1 xr r2)"
    unfolding leaf_property.simps by simp
  thus "\<exists>(l::('a, 'b) ATreeNode). leafnn_property (\<lambda>_ _. True) op \<and> l (parn r1 xr r2)"
    by blast
qed

lemma treenn_mirror_exists_r: "\<exists> r . treenn_mirror l r"
proof (induct l, auto)
  show "\<And>x. \<exists>r. leafnn_property (\<lambda>_ _. True) op \<and> (seqn x) r"
    by (metis leafnn_property.simps(1))
next
  fix r1 r2 :: "('a, 'b) ATreeNode" and l1 l2 :: "('c, 'd) ATreeNode" and xl xr
  assume
    lp1: "leafnn_property (\<lambda>_ _. True) op \<and> l1 r1" and
    lp2: "leafnn_property (\<lambda>_ _. True) op \<and> l2 r2"
  hence "leafnn_property (\<lambda>_ _. True) op \<and> (parn l1 xl l2) (parn r1 xr r2)"
    unfolding leaf_property.simps by simp
  thus "\<exists>(r::('a, 'b) ATreeNode). leafnn_property (\<lambda>_ _. True) op \<and> (parn l1 xl l2) r"
    by blast
qed

declare tree_mirror[simp del] treen_mirror[simp del] treenn_mirror[simp del]

lemma leaf_property_symmetric: "\<forall> x y . P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = (op \<or>) \<or> bl = (op \<and>) \<Longrightarrow> leaf_property P1 bl s t \<Longrightarrow> leaf_property P2 bl t s"
proof (induct s arbitrary: t)
  fix x :: 'a and t :: "'b ATree"
  show "\<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P1 bl (seq x) t \<Longrightarrow> leaf_property P2 bl t (seq x)"
  proof (induct t, simp+) qed
next
  fix s1 s2 :: "'a ATree" and t :: "'b ATree"
  assume
    IH1: "\<And>t::'b ATree. \<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P1 bl s1 t \<Longrightarrow> leaf_property P2 bl t s1" and
    IH2: "\<And>t::'b ATree. \<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P1 bl s2 t \<Longrightarrow> leaf_property P2 bl t s2" and
    sym: "\<forall>x y. P1 x y \<longrightarrow> P2 y x" and
    bl: "bl = op \<or> \<or> bl = op \<and>" and
    base: "leaf_property P1 bl (par s1 s2) t"
  obtain t1 t2 where t: "t = par t1 t2" using base leaf_property_parr by blast
  show "leaf_property P2 bl t (par s1 s2)"
    using base IH1[of t1] IH2[of t2] sym bl
    unfolding t
    by auto
qed

lemma leafnn_property_symmetric: "\<forall> x y . P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leafnn_property P1 bl s t \<Longrightarrow> leafnn_property P2 bl t s"
proof (induct s arbitrary: t)
  fix x :: 'a and t :: "('b, 'd) ATreeNode"
  show "\<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leafnn_property P1 bl (seqn x) t \<Longrightarrow> leafnn_property P2 bl t (seqn x)"
  proof (induct t, simp+) qed
next
  fix s1 s2 :: "('a, 'c) ATreeNode" and t :: "('b, 'd) ATreeNode" and x2 :: 'c
  assume
    IH1: "\<And>t::('b, 'd) ATreeNode. \<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leafnn_property P1 bl s1 t \<Longrightarrow> leafnn_property P2 bl t s1" and
    IH2: "\<And>t::('b, 'd) ATreeNode. \<forall>x y. P1 x y \<longrightarrow> P2 y x \<Longrightarrow> bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leafnn_property P1 bl s2 t \<Longrightarrow> leafnn_property P2 bl t s2" and
    sym: "\<forall>x y. P1 x y \<longrightarrow> P2 y x" and
    bl: "bl = op \<or> \<or> bl = op \<and>" and
    base: "leafnn_property P1 bl (parn s1 x2 s2) t"
  obtain t1 t2 \<v> where t: "t = parn t1 \<v> t2" using leafnn_property_parr[OF base] by blast
  show "leafnn_property P2 bl t (parn s1 x2 s2)"
    using base IH1[of t1] IH2[of t2] sym bl
    unfolding t
    by auto
qed

lemma leaf_property_transitive: "\<forall> x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s t \<Longrightarrow> leaf_property P2 (op \<and>) t u \<Longrightarrow> leaf_property P3 (op \<and>) s u"
proof (induct s arbitrary: t u)
  fix x :: 'a and t :: "'b ATree" and u :: "'c ATree"
  show "\<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) (ATree.seq x) t \<Longrightarrow> leaf_property P2 (op \<and>) t u \<Longrightarrow> leaf_property P3 (op \<and>) (ATree.seq x) u"
  proof (induct t, auto)
    fix xa :: 'b
    show "\<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> P1 x xa \<Longrightarrow> leaf_property P2 (op \<and>) (seq xa) u \<Longrightarrow> leaf_property P3 (op \<and>) (seq x) u"
    proof (induct u, simp+) qed    
  qed
next
  fix s1 s2 :: "'a ATree" and t :: "'b ATree" and u :: "'c ATree"
  assume
    IH1: "\<And>(t::'b ATree) (u::'c ATree). \<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s1 t \<Longrightarrow> leaf_property P2 (op \<and>) t u \<Longrightarrow> leaf_property P3 (op \<and>) s1 u" and
    IH2: "\<And>(t::'b ATree) (u::'c ATree). \<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s2 t \<Longrightarrow> leaf_property P2 (op \<and>) t u \<Longrightarrow> leaf_property P3 (op \<and>) s2 u" and
    trans: "\<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z" and
    st: "leaf_property P1 (op \<and>) (par s1 s2) t" and
    tu: "leaf_property P2 (op \<and>) t u"
  obtain t1 t2 where t: "t = par t1 t2" using st leaf_property_parr by blast
  obtain u1 u2 where u: "u = par u1 u2" using tu t leaf_property_parr by blast
  show "leaf_property P3  (op \<and>)(par s1 s2) u"
    using st tu IH1[of t1 u1] IH2[of t2 u2] trans
    unfolding t u
    by simp
qed

lemma leaf_property_transitive2: "\<forall>x y z. P1 x y \<longrightarrow> P2 x y \<Longrightarrow> leaf_property P1 (op \<and>) s t \<Longrightarrow> leaf_property P2 (op \<and>) s t"
proof (induct s arbitrary: t)
  case (seq x)
  then show ?case
    using leaf_property_seqr by fastforce
next
  case (par s1 s2)
  then show ?case
    by (metis (no_types, lifting) leaf_property.simps(4) leaf_property_parr)
qed

lemma leafnn_property_transitive: "\<forall> x y z . P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leafnn_property P1 (op \<and>) s t \<Longrightarrow> leafnn_property P2 (op \<and>) t u \<Longrightarrow> leafnn_property P3 (op \<and>) s u"
proof (induct s arbitrary: t u)
  fix x :: 'a and t :: "('b, 'e) ATreeNode" and u :: "('c, 'f) ATreeNode"
  show "\<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leafnn_property P1 (op \<and>) (seqn x) t \<Longrightarrow> leafnn_property P2 (op \<and>) t u \<Longrightarrow> leafnn_property P3 (op \<and>) (seqn x) u"
  proof (induct t, auto, induct u, auto) qed
next
  fix s1 s2 :: "('a, 'd) ATreeNode" and x2 :: 'd and t :: "('b, 'e) ATreeNode" and u :: "('c, 'f) ATreeNode"
  assume
    IH1: "\<And>(t::('b, 'e) ATreeNode) (u::('c, 'f) ATreeNode). \<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leafnn_property P1 (op \<and>) s1 t \<Longrightarrow> leafnn_property P2 (op \<and>) t u \<Longrightarrow> leafnn_property P3 (op \<and>) s1 u" and
    IH2: "\<And>(t::('b, 'e) ATreeNode) (u::('c, 'f) ATreeNode). \<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leafnn_property P1 (op \<and>) s2 t \<Longrightarrow> leafnn_property P2 (op \<and>) t u \<Longrightarrow> leafnn_property P3 (op \<and>) s2 u" and
    trans: "\<forall>x y z. P1 x y \<and> P2 y z \<longrightarrow> P3 x z" and
    st: "leafnn_property P1 (op \<and>) (parn s1 x2 s2) t" and
    tu: "leafnn_property P2 (op \<and>) t u"
  obtain t1 t2 \<v>t where t: "t = parn t1 \<v>t t2" using leafnn_property_parr[OF st] by blast
  obtain u1 u2 \<v>u where u: "u = parn u1 \<v>u u2" using leafnn_property_parr[OF tu[unfolded t]] by blast
  show "leafnn_property P3 (op \<and>) (parn s1 x2 s2) u"
    using st tu IH1[of t1 u1] IH2[of t2 u2] trans
    unfolding t u
    by simp
qed

lemma leafn_property_transitive1: "\<forall> x y z . P1 x y \<longrightarrow> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s t \<Longrightarrow> leafn_property P2 (op \<and>) t u \<Longrightarrow> leafn_property P3 (op \<and>) s u"
proof (induct s arbitrary: t u)
  fix x :: 'a and t :: "'b ATree" and u :: "('c, 'd) ATreeNode"
  show "\<forall>x y z. P1 x y \<longrightarrow> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) (seq x) t \<Longrightarrow> leafn_property P2 (op \<and>) t u \<Longrightarrow> leafn_property P3 (op \<and>) (seq x) u"
  proof (induct t, auto, induct u, auto) qed
next
  fix s1 s2 :: "'a ATree" and t :: "'b ATree" and u :: "('c, 'd) ATreeNode"
  assume
    IH1: "\<And>(t::'b ATree) (u::('c, 'd) ATreeNode). \<forall>x y z. P1 x y \<longrightarrow> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s1 t \<Longrightarrow> leafn_property P2 (op \<and>) t u \<Longrightarrow> leafn_property P3 (op \<and>) s1 u" and
    IH2: "\<And>(t::'b ATree) (u::('c, 'd) ATreeNode). \<forall>x y z. P1 x y \<longrightarrow> P2 y z \<longrightarrow> P3 x z \<Longrightarrow> leaf_property P1 (op \<and>) s2 t \<Longrightarrow> leafn_property P2 (op \<and>) t u \<Longrightarrow> leafn_property P3 (op \<and>) s2 u" and
    trans: "\<forall>x y z. P1 x y \<longrightarrow> P2 y z \<longrightarrow> P3 x z" and
    st: "leaf_property P1 (op \<and>) (par s1 s2) t" and
    tu: "leafn_property P2 (op \<and>) t u"
  obtain t1 t2 where t: "t = par t1 t2" using leaf_property_parr[OF st] by blast
  obtain u1 u2 \<v>u where u: "u = parn u1 \<v>u u2" using leafn_property_parr[OF tu[unfolded t]] by blast
  show "leafn_property P3 (op \<and>) (par s1 s2) u"
    using st tu IH1[of t1 u1] IH2[of t2 u2] trans
    unfolding t u
    by simp
qed

lemma leafn_property_transitive2: "\<forall>x y z. P1 x y \<longrightarrow> P2 z y \<longrightarrow> P3 z x \<Longrightarrow> leafnn_property P1 (op \<and>) s t \<Longrightarrow> leafn_property P2 (op \<and>) u t \<Longrightarrow> leafn_property P3 (op \<and>) u s"
proof (induct s arbitrary: t u)
  fix x :: 'a and t :: "('b, 'e) ATreeNode" and u :: "'c ATree"
  show "\<forall>x y z. P1 x y \<longrightarrow> P2 z y \<longrightarrow> P3 z x \<Longrightarrow> leafnn_property P1 (op \<and>) (seqn x) t \<Longrightarrow> leafn_property P2 (op \<and>) u t \<Longrightarrow> leafn_property P3 (op \<and>) u (seqn x)"
  proof (induct t, auto, induct u, auto) qed
next
  fix s1 s2 :: "('a, 'd) ATreeNode" and x2 :: 'd and t :: "('b, 'e) ATreeNode" and u :: "'c ATree"
  assume
    IH1: "\<And>(t::('b, 'e) ATreeNode) (u::'c ATree). \<forall>x y z. P1 x y \<longrightarrow> P2 z y \<longrightarrow> P3 z x \<Longrightarrow> leafnn_property P1 (op \<and>) s1 t \<Longrightarrow> leafn_property P2 (op \<and>) u t \<Longrightarrow> leafn_property P3 (op \<and>) u s1" and
    IH2: "\<And>(t::('b, 'e) ATreeNode) (u::'c ATree). \<forall>x y z. P1 x y \<longrightarrow> P2 z y \<longrightarrow> P3 z x \<Longrightarrow> leafnn_property P1 (op \<and>) s2 t \<Longrightarrow> leafn_property P2 (op \<and>) u t \<Longrightarrow> leafn_property P3 (op \<and>) u s2" and
    trans: "\<forall>x y z. P1 x y \<longrightarrow> P2 z y \<longrightarrow> P3 z x" and
    st: "leafnn_property P1 (op \<and>) (parn s1 x2 s2) t" and
    tu: "leafn_property P2 (op \<and>) u t"
  obtain t1 t2 vt where t: "t = parn t1 vt t2" using leafnn_property_parr[OF st] by blast
  obtain u1 u2 where u: "u = par u1 u2" using leafn_property_parl[OF tu[unfolded t]] by blast
  show "leafn_property P3 (op \<and>) u (parn s1 x2 s2)"
    using st tu IH1[of t1 u1] IH2[of t2 u2] trans
    unfolding t u
    by simp
qed

lemma leafn_property_transitive3: "\<forall>x y z. P1 x y \<and> P2 z y \<longrightarrow> P3 x z \<Longrightarrow> leafn_property P1 (op \<and>) t u \<Longrightarrow> leafn_property P2 (op \<and>) s u \<Longrightarrow> leaf_property P3 (op \<and>) t s"
proof (induct s arbitrary: t u)
  fix x :: 'c and t :: "'a ATree" and u :: "('b, 'd) ATreeNode"
  show "\<forall>x y z. P1 x y \<and> P2 z y \<longrightarrow> P3 x z \<Longrightarrow> leafn_property P1 (op \<and>) t u \<Longrightarrow> leafn_property P2 (op \<and>) (seq x) u \<Longrightarrow> leaf_property P3 (op \<and>) t (seq x)"
  proof (induct t, auto, induct u, simp, simp)
    show "\<And> t1 t2 . leafn_property P1 (op \<and>) (par t1 t2) u \<Longrightarrow> leafn_property P2 (op \<and>) (seq x) u \<Longrightarrow> False"
      using leafn_property_seqr by fastforce
  qed
next
  fix s1 s2 :: "'c ATree" and t :: "'a ATree" and u :: "('b, 'd) ATreeNode"
  assume
    IH1: "\<And>(t::'a ATree) (u::('b, 'd) ATreeNode). \<forall>x y z. P1 x y \<and> P2 z y \<longrightarrow> P3 x z \<Longrightarrow> leafn_property P1 (op \<and>) t u \<Longrightarrow> leafn_property P2 (op \<and>) s1 u \<Longrightarrow> leaf_property P3 (op \<and>) t s1" and
    IH2: "\<And>(t::'a ATree) (u::('b, 'd) ATreeNode). \<forall>x y z. P1 x y \<and> P2 z y \<longrightarrow> P3 x z \<Longrightarrow> leafn_property P1 (op \<and>) t u \<Longrightarrow> leafn_property P2 (op \<and>) s2 u \<Longrightarrow> leaf_property P3 (op \<and>) t s2" and
    trans: "\<forall>x y z. P1 x y \<and> P2 z y \<longrightarrow> P3 x z" and
    su: "leafn_property P2 (op \<and>) (par s1 s2) u" and
    tu: "leafn_property P1 (op \<and>) t u"
  obtain u1 u2 vu where u: "u = parn u1 vu u2" using leafn_property_parr[OF su] by blast
  obtain t1 t2 where t: "t = par t1 t2" using leafn_property_parl[OF tu[unfolded u]] by blast
  show "leaf_property P3 (op \<and>) t (par s1 s2)"
    using su tu IH1[of t1 u1] IH2[of t2 u2] trans
    unfolding t u
    by simp
qed

lemma leafn_property_transitive4: "\<forall>x y z. P1 x y \<and> P2 x z \<longrightarrow> P3 y z \<Longrightarrow> leafn_property P1 (op \<and>) u t \<Longrightarrow> leafn_property P2 (op \<and>) u s \<Longrightarrow> leafnn_property P3 (op \<and>) t s"
proof (induct s arbitrary: t u)
  fix x :: 'c and t :: "('b, 'd) ATreeNode" and u :: "'a ATree"
  show "\<forall>x y z. P1 x y \<and> P2 x z \<longrightarrow> P3 y z \<Longrightarrow> leafn_property P1 (op \<and>) u t \<Longrightarrow> leafn_property P2 (op \<and>) u (seqn x) \<Longrightarrow> leafnn_property P3 (op \<and>) t (seqn x)"
  proof (induct t, auto, induct u, auto)
    show "\<And>t1 x2 t2. leafn_property P1 (op \<and>) u (parn t1 x2 t2) \<Longrightarrow> leafn_property P2 (op \<and>) u (seqn x) \<Longrightarrow> False"
      using leafn_property_seql by fastforce
  qed
next
  fix s1 s2 :: "('c, 'e) ATreeNode" and x2 :: 'e and t :: "('b, 'd) ATreeNode" and u :: "'a ATree"
  assume
    IH1: "\<And>(t::('b, 'd) ATreeNode) (u::'a ATree). \<forall>x y z. P1 x y \<and> P2 x z \<longrightarrow> P3 y z \<Longrightarrow> leafn_property P1 (op \<and>) u t \<Longrightarrow> leafn_property P2 (op \<and>) u s1 \<Longrightarrow> leafnn_property P3 (op \<and>) t s1" and
    IH2: "\<And>(t::('b, 'd) ATreeNode) (u::'a ATree). \<forall>x y z. P1 x y \<and> P2 x z \<longrightarrow> P3 y z \<Longrightarrow> leafn_property P1 (op \<and>) u t \<Longrightarrow> leafn_property P2 (op \<and>) u s2 \<Longrightarrow> leafnn_property P3 (op \<and>) t s2" and
    trans: "\<forall>x y z. P1 x y \<and> P2 x z \<longrightarrow> P3 y z" and
    us: "leafn_property P2 (op \<and>) u (parn s1 x2 s2)" and
    ut: "leafn_property P1 (op \<and>) u t"
  obtain u1 u2 where u: "u = par u1 u2" using leafn_property_parl[OF us] by blast
  obtain t1 t2 vt where t: "t = parn t1 vt t2" using leafn_property_parr[OF ut[unfolded u]] by blast
  show "leafnn_property P3 (op \<and>) t (parn s1 x2 s2)"
    using us ut IH1[of u1 t1] IH2[of u2 t2] trans
    unfolding t u
    by simp
qed

lemma leafn_property_transitive5: "\<forall>x y z. P1 x y \<longrightarrow> P2 x y \<Longrightarrow> leafn_property P1 (op \<and>) s t \<Longrightarrow> leafn_property P2 (op \<and>) s t"
proof (induct s arbitrary: t)
  case (seq x)
  then show ?case
    using leafn_property_seqr by fastforce
next
  case (par s1 s2)
  then show ?case
    by (metis (no_types, lifting) leafn_property.simps(4) leafn_property_parr)
qed

lemma leafnn_property_transitive2: "\<forall>x y z. P1 x y \<longrightarrow> P2 x y \<Longrightarrow> leafnn_property P1 (op \<and>) s t \<Longrightarrow> leafnn_property P2 (op \<and>) s t"
proof (induct s arbitrary: t)
  case (seqn x)
  then show ?case
    using leafnn_property_seqr by fastforce
next
  case (parn s1 s2)
  then show ?case
    by (metis (no_types, lifting) leafnn_property.simps(4) leafnn_property_parr)
qed

lemma tree_mirror_transitive:
  "tree_mirror t1 t2 \<Longrightarrow> tree_mirror t2 t3 \<Longrightarrow> tree_mirror t1 t3"
unfolding tree_mirror using leaf_property_transitive[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma treen_mirror_transitive1: "tree_mirror t1 t2 \<Longrightarrow> treen_mirror t2 t3 \<Longrightarrow> treen_mirror t1 t3"
unfolding tree_mirror treen_mirror using leafn_property_transitive1[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma treen_mirror_transitive2: "treenn_mirror t1 t2 \<Longrightarrow> treen_mirror t3 t2 \<Longrightarrow> treen_mirror t3 t1"
unfolding treenn_mirror treen_mirror using leafn_property_transitive2[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma treen_mirror_transitive3: "treen_mirror t1 t2 \<Longrightarrow> treen_mirror t3 t2 \<Longrightarrow> tree_mirror t1 t3"
unfolding tree_mirror treen_mirror using leafn_property_transitive3[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma treen_mirror_transitive4: "treen_mirror t1 t2 \<Longrightarrow> treen_mirror t1 t3 \<Longrightarrow> treenn_mirror t2 t3"
unfolding treenn_mirror treen_mirror using leafn_property_transitive4[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma treenn_mirror_transitive:
  "treenn_mirror t1 t2 \<Longrightarrow> treenn_mirror t2 t3 \<Longrightarrow> treenn_mirror t1 t3"
unfolding treenn_mirror using leafnn_property_transitive[of "\<lambda>_ _. True" "\<lambda>_ _. True" "\<lambda>_ _. True" t1 t2 t3] by blast

lemma ATreeNode_to_ATree_tree_mirror:
  shows "treen_mirror (ATreeNode_to_ATree tr) tr"
proof (induct tr, auto) qed

lemma leafn_property_merge:
  assumes tm: "treen_mirror t tn"
  shows "leafn_property f b t tn = leafn1_property (\<lambda> (t, tn) . f t tn) b (t \<times>\<^sub>A\<^sub>T\<^sub>n tn)"
using assms
proof (induct t arbitrary:tn)
  case (seq x)
  then show ?case
    unfolding treen_mirror_unfold_seql
    by fastforce
next
  case (par t1 t2)
  then obtain tn1 n tn2 where tn: "tn = parn tn1 n tn2"
    using treen_mirror_unfold_parl by metis
  show ?case
    using par(1)[of tn1]
    using par(2)[of tn2]
    using par(3)
    unfolding tn treen_mirror_inherit
  by simp
qed

lemma leaf_property_merge:
  assumes tm: "tree_mirror t t'"
  shows "leaf_property f b t t' = leaf1_property (\<lambda> (t, tn) . f t tn) b (t \<times>\<^sub>A\<^sub>T t')"
using assms
proof (induct t arbitrary:t')
  case (seq x)
  then show ?case
    unfolding treen_mirror_unfold_seql
    by fastforce
next
  case (par t1 t2)
  then obtain tn1 tn2 where tn: "t' = par tn1 tn2"
    using tree_mirror_unfold_parl by metis
  show ?case
    using par(1)[of tn1]
    using par(2)[of tn2]
    using par(3)
    unfolding tn treen_mirror_inherit
  by simp
qed

fun ATreeNode_access_node :: "('lf, 'nd) ATreeNode \<Rightarrow> NodeAccessor list \<Rightarrow> ('lf \<times> 'nd list) option" ("_ $' _") where
"((seqn content) $' []) = Some (content, [])" |
"((parn l n r) $' (Left # lrs)) = (case l $' lrs of Some (a, b) \<Rightarrow> Some (a, n # b) | _ \<Rightarrow> None)" |
"((parn l n r) $' (Right # lrs)) = (case r $' lrs of Some (a, b) \<Rightarrow> Some (a, n # b) | _ \<Rightarrow> None)" |
"((seqn content) $' (v # xs)) = None" |
"((parn l n r) $' []) = None"

fun ATreeNode_update_node :: "('lf, 'nd) ATreeNode \<Rightarrow> NodeAccessor list \<Rightarrow> 'lf \<Rightarrow> ('lf, 'nd) ATreeNode option" ("'(_ $' _ \<leftarrow> _')") where
"((seqn content) $' [] \<leftarrow> content') = (Some (seqn content'))" |
"((parn l n r) $' (Left # lrs) \<leftarrow> content') = (case (l $' lrs \<leftarrow> content') of Some x \<Rightarrow> Some (parn x n r) | _ \<Rightarrow> None)" |
"((parn l n r) $' (Right # lrs) \<leftarrow> content') = (case (r $' lrs \<leftarrow> content') of Some x \<Rightarrow> Some (parn l n x) | _ \<Rightarrow> None)" |
"((seqn _) $' (_ # _) \<leftarrow> _) = None" |
"((parn _ _ _) $' [] \<leftarrow> _) = None"

lemma ATreeNode_access_node_left:
  "Some (lf, n # nds) = ((parn l n r) $' (Left # lrs)) = (Some (lf, nds) = (l $' lrs))"
unfolding ATreeNode_access_node.simps
proof -
  have f1: "the (Some (lf, nds)) = (lf, nds)"
    by auto
  have f2: "\<forall>z f za. if za = None then (case za of None \<Rightarrow> z::('a \<times> 'b list) option | Some (x::'a \<times> 'b list) \<Rightarrow> f x) = z else (case za of None \<Rightarrow> z | Some x \<Rightarrow> f x) = f (the za)"
    by (simp add: option.case_eq_if)
  have "(Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs))) = (Some (lf, nds) = l $' lrs) \<longrightarrow> Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs)) \<and> Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs)) \<or> l $' lrs = None"
    by force
  then have "(Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs))) = (Some (lf, nds) = l $' lrs) \<longrightarrow> Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs)) \<and> Some (lf, n # nds) \<noteq> (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs))"
    by force
  then show "(Some (lf, n # nds) = (case l $' lrs of None \<Rightarrow> None | Some (a, bs) \<Rightarrow> Some (a, n # bs))) = (Some (lf, nds) = l $' lrs)"
    using f2 f1 by (metis (no_types) old.prod.case option.simps(3))
qed

lemma ATreeNode_access_node_right:
  "(Some (lf, n # nds) = ((parn l n r) $' (Right # lrs))) = (Some (lf, nds) = (r $' lrs))"
  using ATreeNode_access_node_left by force

lemma ATreeNode_access_node_eq_len:
  assumes "Some (lf, nds) = tr $' lrs"
  shows "length nds = length lrs"
using assms
proof (induct tr arbitrary: nds lrs)
  case (seqn x)
  then show ?case
    by (metis ATreeNode_access_node.simps(1) ATreeNode_access_node.simps(4) length_0_conv list.exhaust option.sel option.simps(3) snd_conv)
next
  case (parn t1 n t2)
  then obtain lr lrs' where lrs: "lrs = lr # lrs'" using assms
    by (metis ATreeNode_access_node.simps(5) list.collapse option.simps(3))
  moreover
  consider (l) "lr = Left" | (r) "lr = Right" using NodeAccessor.exhaust by blast
  hence "\<exists> nds' . nds = n # nds'"
  proof cases
    case l
    obtain a b where base: "t1 $' lrs' = Some (a, b)"
      using assms
      unfolding lrs l ATreeNode_access_node.simps
    proof -
      assume a1: "\<And>a b. t1 $' lrs' = Some (a, b) \<Longrightarrow> thesis"
      have "Some (lf, nds) \<noteq> None"
        by blast
      then have "\<exists>z. t1 $' lrs' = z \<and> z \<noteq> None"
        by (metis (no_types) ATreeNode_access_node.simps(2) l lrs option.simps(4) parn.prems)
      then show ?thesis
        using a1 by moura
    qed
    show ?thesis
      using parn unfolding lrs l ATreeNode_access_node.simps base
      by simp
  next
    case r
    obtain a b where base: "t2 $' lrs' = Some (a, b)"
      using assms
      unfolding lrs r ATreeNode_access_node.simps
    proof -
      assume a1: "\<And>a b. t2 $' lrs' = Some (a, b) \<Longrightarrow> thesis"
      have "Some (lf, nds) \<noteq> None"
        by blast
      then have "\<exists>z. t2 $' lrs' = z \<and> z \<noteq> None"
        by (metis (no_types) ATreeNode_access_node.simps(3) r lrs option.simps(4) parn.prems)
      then show ?thesis
        using a1 by moura
    qed
    show ?thesis
      using parn unfolding lrs r ATreeNode_access_node.simps base
      by simp
  qed
  then obtain nds' where nds: "nds = n # nds'" by blast
  show ?case
    using parn(1)[of nds' lrs']
    using parn(2)[of nds' lrs']
    using nds lrs
    using NodeAccessor.exhaust[of lr]
    using parn(3)[unfolded nds lrs]
    using ATreeNode_access_node_right
    using ATreeNode_access_node_left
    by (metis length_Cons)
qed

lemma ATreeNode_access_node_parn:
  assumes "Some (lf, nds) = parn t1 n t2 $' lrs"
  shows "\<exists> nds' lr lrs' . nds = n # nds' \<and> lrs = lr # lrs'"
proof -
  obtain lr lrs' where lrs: "lrs = lr # lrs'" using assms
    by (metis ATreeNode_access_node.simps(5) list.collapse option.simps(3))
  moreover
  consider (l) "lr = Left" | (r) "lr = Right" using NodeAccessor.exhaust by blast
  hence "\<exists> nds' . nds = n # nds'"
  proof cases
    case l
    obtain a b where base: "t1 $' lrs' = Some (a, b)"
      using assms
      unfolding lrs l ATreeNode_access_node.simps
    proof -
      assume a1: "\<And>a b. t1 $' lrs' = Some (a, b) \<Longrightarrow> thesis"
      assume "Some (lf, nds) = (case t1 $' lrs' of None \<Rightarrow> None | Some (a, b) \<Rightarrow> Some (a, n # b))"
      then show ?thesis
        using a1 by (metis (lifting) option.case_eq_if option.exhaust surj_pair)
    qed
    show ?thesis
      using assms unfolding lrs l ATreeNode_access_node.simps base
      by simp
  next
    case r
    obtain a b where base: "t2 $' lrs' = Some (a, b)"
      using assms
      unfolding lrs r ATreeNode_access_node.simps
    proof -
      assume a1: "\<And>a b. t2 $' lrs' = Some (a, b) \<Longrightarrow> thesis"
      assume "Some (lf, nds) = (case t2 $' lrs' of None \<Rightarrow> None | Some (a, b) \<Rightarrow> Some (a, n # b))"
      then show ?thesis
        using a1 by (metis (lifting) option.case_eq_if option.exhaust surj_pair)
    qed
    show ?thesis
      using assms unfolding lrs r ATreeNode_access_node.simps base
      by simp
  qed
  ultimately show ?thesis by auto
qed

fun ATree_access_node :: "'lf ATree \<Rightarrow> NodeAccessor list \<Rightarrow> 'lf option" ("_ $ _" [65, 64]) where
"((seq content) $ []) = (Some content)" |
"((par l r) $ (Left # lrs)) = (l $ lrs)" |
"((par l r) $ (Right # lrs)) = (r $ lrs)" |
"((seq content) $ (v # xs)) = None" |
"((par l r) $ []) = None"

fun ATree_update_node :: "'lf ATree \<Rightarrow> NodeAccessor list \<Rightarrow> 'lf \<Rightarrow> 'lf ATree option" ("'(_ $ _ \<leftarrow> _')") where
"((seq content) $ [] \<leftarrow> content') = (Some (seq content'))" |
"((par l r) $ (Left # lrs) \<leftarrow> content') = (case (l $ lrs \<leftarrow> content') of Some x \<Rightarrow> Some (par x r) | _ \<Rightarrow> None)" |
"((par l r) $ (Right # lrs) \<leftarrow> content') = (case (r $ lrs \<leftarrow> content') of Some x \<Rightarrow> Some (par l x) | _ \<Rightarrow> None)" |
"((seq _) $ (_ # _) \<leftarrow> _) = None" |
"((par _ _) $ [] \<leftarrow> _) = None"


lemma ATree_eq_if_all_leafs_eq: "tree_mirror l r \<Longrightarrow> (\<forall> lrs . l $ lrs = r $ lrs) = (l = r)"
proof (induct l arbitrary: r)
  case (seq x)
  then obtain xr where r: "r = seq xr" by auto
  then show ?case
    unfolding r
    by (metis ATree_access_node.simps(1) option.sel) 
next
  case (par l1 l2)
  obtain r1 r2 where r: "r = par r1 r2" using par(3) tree_mirror_unfold_parl by blast
  have "(\<forall>lrs. l1 $ lrs = r1 $ lrs) = (l1 = r1)"
    using par r by simp
  moreover have "(\<forall>lrs. l2 $ lrs = r2 $ lrs) = (l2 = r2)"
    using par r by simp
  ultimately show ?case
    by (metis ATree_access_node.simps(2) ATree_access_node.simps(3) r)
qed

lemma ATree_update_node_tm_and_correct_update:
  assumes
    upd: "Some tr' = (tr $ lrs \<leftarrow> x)"
  shows
    "tree_mirror tr tr' \<and> (\<forall> lrs' t . lrs \<noteq> lrs' \<and> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr \<longrightarrow> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr') \<and> (lrs, x) \<in> \<union>\<^sub>A\<^sub>T tr'"
  using assms
proof (induct tr arbitrary: tr' lrs)
  case (seq tr)
  hence tr': "tr' = seq x"
    by (metis ATree_update_node.simps(1) ATree_update_node.simps(4) list.exhaust option.inject option.simps(3))
  show ?case
    unfolding tr' ATree_update_node.simps ATree_gather_def ATree_gather_hlpr.simps
    by (metis ATree_update_node.simps(4) Pair_inject list.exhaust option.distinct(1) seq.prems singletonD singletonI tree_mirror_unfold_seql)
next
  case (par tr1 tr2)
  obtain lr lrs' where
    lrs: "lrs = lr # lrs'"
    using par
    by (metis ATree_update_node.simps(5) list.exhaust option.simps(3))
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    obtain tr'1 tr'2 where
      tr': "tr' = par tr'1 tr'2"
      using par(3)
      unfolding lrs left
      by (metis (no_types, lifting) ATree_update_node.simps(2) option.case_eq_if option.inject option.simps(3))
    have *: "Some tr'1 = (tr1 $ lrs' \<leftarrow> x)"
      by (metis (no_types, lifting) ATree.inject(2) ATree_update_node.simps(2) left lrs option.case_eq_if option.collapse option.sel option.simps(3) par.prems tr')
    have **: "tree_mirror tr1 tr'1 \<and> (\<forall>lrs'a t. lrs' \<noteq> lrs'a \<and> (lrs'a, t) \<in> \<union>\<^sub>A\<^sub>T tr1 \<longrightarrow> (lrs'a, t) \<in> \<union>\<^sub>A\<^sub>T tr'1) \<and> (lrs', x) \<in> \<union>\<^sub>A\<^sub>T tr'1"
      using par(1)[OF *] .
    have "tree_mirror (par tr1 tr2) tr'"
      by (metis * ATree_update_node.simps(2) ** left lrs option.inject option.simps(5) par.prems tree_mirror_inherit tree_mirror_reflexive)
    moreover have "\<forall>lrs' t. lrs \<noteq> lrs' \<and> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T par tr1 tr2 \<longrightarrow> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr'"
      by (metis (no_types, lifting) "*" "**" ATree_gather_left ATree_gather_par ATree_gather_right ATree_update_node.simps(2) left lrs option.sel option.simps(5) par.prems)
    moreover have "(lrs, x) \<in> \<union>\<^sub>A\<^sub>T tr'"
      by (simp add: "**" ATree_gather_left left lrs tr')
    ultimately show ?thesis by simp
  next
    case right
    obtain tr'1 tr'2 where
      tr': "tr' = par tr'1 tr'2"
      using par(3)
      unfolding lrs right
      by (metis (no_types, lifting) ATree_update_node.simps(3) option.case_eq_if option.inject option.simps(3))
    have *: "Some tr'2 = (tr2 $ lrs' \<leftarrow> x)"
      by (metis (no_types, lifting) ATree.inject(2) ATree_update_node.simps(3) right lrs option.case_eq_if option.collapse option.sel option.simps(3) par.prems tr')
    have **: "tree_mirror tr2 tr'2 \<and> (\<forall>lrs'a t. lrs' \<noteq> lrs'a \<and> (lrs'a, t) \<in> \<union>\<^sub>A\<^sub>T tr2 \<longrightarrow> (lrs'a, t) \<in> \<union>\<^sub>A\<^sub>T tr'2) \<and> (lrs', x) \<in> \<union>\<^sub>A\<^sub>T tr'2"
      using par(2)[OF *] .
    have "tree_mirror (par tr1 tr2) tr'"
      by (metis * ATree_update_node.simps(3) ** right lrs option.inject option.simps(5) par.prems tree_mirror_inherit tree_mirror_reflexive)
    moreover have "\<forall>lrs' t. lrs \<noteq> lrs' \<and> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T par tr1 tr2 \<longrightarrow> (lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr'"
      by (metis (no_types, lifting) "*" "**" ATree_gather_left ATree_gather_par ATree_gather_right ATree_update_node.simps(3) right lrs option.sel option.simps(5) par.prems)
    moreover have "(lrs, x) \<in> \<union>\<^sub>A\<^sub>T tr'"
      by (simp add: "**" ATree_gather_right right lrs tr')
    ultimately show ?thesis by simp
  qed
qed

lemma ATree_update_node_is_Some: "(lrs, t) \<in> \<union>\<^sub>A\<^sub>T tr \<Longrightarrow> \<exists> tr' . Some tr' = (tr $ lrs \<leftarrow> x)"
proof (induct tr arbitrary: lrs)
  case (seq x)
  then show ?case
    unfolding ATree_gather_def ATree_gather_hlpr.simps
    by fastforce
next
  case (par tr1 tr2)
  obtain lr lrs' where
    lrs: "lrs = lr # lrs'"
    using par
    by (metis ATree_gather_par)
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr1: "(lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr1"
      using ATree_gather_left left lrs par.prems by force
    obtain tr' where
      eq_len: "Some tr' = (tr1 $ lrs' \<leftarrow> x)"
      using par(1)[OF in_tr1]
      by blast
    thus ?thesis
      unfolding lrs left
      unfolding ATree_update_node.simps(2)
      by (metis option.simps(5))
  next
    case right
    have in_tr2: "(lrs', t) \<in> \<union>\<^sub>A\<^sub>T tr2"
      using ATree_gather_right right lrs par.prems by force
    obtain tr' where
      eq_len: "Some tr' = (tr2 $ lrs' \<leftarrow> x)"
      using par(2)[OF in_tr2]
      by blast
    thus ?thesis
      unfolding lrs right
      unfolding ATree_update_node.simps(3)
      by (metis option.simps(5))
  qed
qed

fun ATree_map :: "'lf ATree \<Rightarrow> ('lf \<Rightarrow> 'lfn) \<Rightarrow> 'lfn ATree" where
"ATree_map (seq content) f = (seq (f content))" |
"ATree_map (par l r) f = (par (ATree_map l f) (ATree_map r f))"

lemma ATree_map_tree_mirror: "tree_mirror tree (ATree_map tree f)"
  unfolding tree_mirror
proof (induct tree, auto) qed

lemma ATree_map_treen_mirror: "treen_mirror tr1 tr2 = treen_mirror (ATree_map tr1 f) tr2"
  using treen_mirror_transitive1[OF tree_mirror_symmetric[OF ATree_map_tree_mirror]]
  using treen_mirror_transitive1[OF ATree_map_tree_mirror]
  by auto

lemma ATree_map_access_mapped: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr \<Longrightarrow> (lrs, f lf) \<in> \<union>\<^sub>A\<^sub>T (ATree_map tr f)"
proof (induct tr arbitrary:lrs)
  case (seq x)
  then show ?case by auto
next
  case (par tr1 tr2)
  obtain lr lrs' where lrs: "lrs = lr # lrs'" using par(3)
    by (meson ATree_gather_par)
  show ?case
    using par(1)[of lrs'] par(2)[of lrs']
    unfolding ATree_gather_def ATree_gather_hlpr.simps lrs
  proof -
    assume a1: "(lrs', lf) \<in> ATree_gather_hlpr [] tr1 \<Longrightarrow> (lrs', f lf) \<in> ATree_gather_hlpr [] (ATree_map tr1 f)"
    assume a2: "(lrs', lf) \<in> ATree_gather_hlpr [] tr2 \<Longrightarrow> (lrs', f lf) \<in> ATree_gather_hlpr [] (ATree_map tr2 f)"
    have "NodeAccessor.Left # lrs' = lrs \<or> lr = NodeAccessor.Right"
      using NodeAccessor.exhaust lrs by blast
    then show "(lr # lrs', f lf) \<in> ATree_gather_hlpr [] (ATree_map (par tr1 tr2) f)"
      using a2 a1 by (metis (no_types) ATree_gather_def ATree_gather_left ATree_gather_right ATree_map.simps(2) lrs par.prems)
  qed
qed

lemma ATree_unmerge_union: "tree_mirror l r \<Longrightarrow> (lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T (l \<times>\<^sub>A\<^sub>T r) = ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T l \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)"
proof (induct l arbitrary: r lrs)
  fix x :: 'a and r :: "'b ATree" and lrs
  assume
    tm: "tree_mirror (seq x) r"
  obtain rr where r: "r = seq rr" using tm unfolding tree_mirror using leaf_property_seqr by metis
  show "(lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T seq x \<times>\<^sub>A\<^sub>T r = ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T seq x \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)"
    unfolding r "ATree_merge.simps" ATree_gather_def "ATree_gather_hlpr.simps" by fast
next
  fix l1 l2 :: "'a ATree" and r :: "'b ATree" and lrs
  assume
    IH1: "\<And>r lrs. tree_mirror l1 r \<Longrightarrow>
        (lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T l1 \<times>\<^sub>A\<^sub>T r =
        ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T l1 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)" and
    IH2: "\<And>r lrs. tree_mirror l2 r \<Longrightarrow>
        (lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T l2 \<times>\<^sub>A\<^sub>T r =
        ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T l2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)" and
    tm: "tree_mirror (par l1 l2) r"
  obtain r1 r2 where r: "r = par r1 r2" using tm unfolding tree_mirror using leaf_property_parr by metis
  have "(lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T ((par l1 l2) \<times>\<^sub>A\<^sub>T r) \<Longrightarrow> ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)"
    using IH1[of r1] IH2[of r2] tm unfolding r "ATree_merge.simps" ATree_gather_def "ATree_gather_hlpr.simps" tree_mirror leaf_property.simps
      append_Nil
    by (metis (no_types, lifting) ATree_gather_hlpr_prefix Un_iff append_Cons)
  moreover have "((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r) \<Longrightarrow> (lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T ((par l1 l2) \<times>\<^sub>A\<^sub>T r)"
    using IH1[of r1] IH2[of r2] tm unfolding r "ATree_merge.simps" ATree_gather_def "ATree_gather_hlpr.simps" tree_mirror leaf_property.simps
      append_Nil
  proof -
    assume 
      inside: "
      (lrs, lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1 \<union> ATree_gather_hlpr [NodeAccessor.Right] l2 \<and>
      (lrs, lf2) \<in> ATree_gather_hlpr [NodeAccessor.Left] r1 \<union> ATree_gather_hlpr [NodeAccessor.Right] r2" and
      IH1 :"\<And>lrs. leaf_property (\<lambda>_ _. True) op \<and> l1 r1 \<Longrightarrow>
            ((lrs, lf1, lf2) \<in> ATree_gather_hlpr [] l1 \<times>\<^sub>A\<^sub>T r1) =
            ((lrs, lf1) \<in> ATree_gather_hlpr [] l1 \<and>
             (lrs, lf2) \<in> ATree_gather_hlpr [] r1)" and
      IH2: "\<And>lrs. leaf_property (\<lambda>_ _. True) op \<and> l2 r2 \<Longrightarrow>
            ((lrs, lf1, lf2) \<in> ATree_gather_hlpr [] l2 \<times>\<^sub>A\<^sub>T r2) =
            ((lrs, lf1) \<in> ATree_gather_hlpr [] l2 \<and>
             (lrs, lf2) \<in> ATree_gather_hlpr [] r2)" and
      lp: "leaf_property (\<lambda>_ _. True) op \<and> l1 r1 \<and> leaf_property (\<lambda>_ _. True) op \<and> l2 r2"
    obtain lr lrs' where lrs: "lrs = lr # lrs'"
      by (metis ATree_gather_def ATree_gather_hlpr.simps(2) ATree_gather_par append_Nil inside)
    consider (left) "lr = Left" | (right) "lr = Right" using NodeAccessor.exhaust by blast
    thus "(lrs, lf1, lf2) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1 \<times>\<^sub>A\<^sub>T r1 \<union> ATree_gather_hlpr [NodeAccessor.Right] l2 \<times>\<^sub>A\<^sub>T r2"
    proof cases
      case left
      have inside: "(lrs', lf1) \<in> ATree_gather_hlpr [] l1 \<and> (lrs', lf2) \<in> ATree_gather_hlpr [] r1"
        using inside[unfolded lrs left] 
        by (metis ATree_gather_def ATree_gather_hlpr.simps(2) ATree_gather_left append_Nil)
      thus ?thesis
        using IH1[OF lp[THEN conjunct1], of lrs'] "ATree_gather_hlpr.simps" append_Nil
        by (metis ATree_gather_def ATree_gather_left left lrs)
    next
      case right
      have inside: "(lrs', lf1) \<in> ATree_gather_hlpr [] l2 \<and> (lrs', lf2) \<in> ATree_gather_hlpr [] r2"
        using inside[unfolded lrs right] 
        by (metis ATree_gather_def ATree_gather_hlpr.simps(2) ATree_gather_right append_Nil)
      thus ?thesis
        using IH2[OF lp[THEN conjunct2], of lrs'] "ATree_gather_hlpr.simps" append_Nil
        by (metis ATree_gather_def ATree_gather_right lrs right)
    qed
  qed
  ultimately show "(lrs, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T ((par l1 l2) \<times>\<^sub>A\<^sub>T r) = ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T r)"
    by blast
qed


lemma ATreen_unmerge_union: "treen_mirror l r \<Longrightarrow> ((lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N (l \<times>\<^sub>A\<^sub>T\<^sub>n r)) = ((map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T l \<and> (lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r)"
proof (induct l arbitrary: r lrns)
  case (seq x)
  then show ?case 
    unfolding treen_mirror_unfold_seql
    by force
next
  case (par l1 l2)
  obtain r1 n r2 where r: "r = parn r1 n r2" using par(3) treen_mirror_unfold_parl by metis
  have tm1: "treen_mirror l1 r1" using par(3) unfolding r treen_mirror_inherit by simp
  have tm2: "treen_mirror l2 r2" using par(3) unfolding r treen_mirror_inherit by simp
  have "(lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N ((par l1 l2) \<times>\<^sub>A\<^sub>T\<^sub>n r) \<Longrightarrow> ((map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r)"
  proof -
    assume assm: "(lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N par l1 l2 \<times>\<^sub>A\<^sub>T\<^sub>n r"
    hence "lrns \<noteq> []"
      using ATreeNode_gather_par r by force
    then obtain lr n' lrns' where lrns: "lrns = (lr, n') # lrns'"
      by (metis list.collapse surj_pair)
    have n': "n' = n" using assm
      unfolding lrns r ATreen_merge.simps ATreeNode_gather_def ATreeNode_gather_hlpr.simps append_Nil
      using ATreeNode_gather_par assm lrns r by fastforce
    have "((lr, n) # lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Left, n)] l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1 \<union> ATreeNode_gather_hlpr [(NodeAccessor.Right, n)] l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2"
      using assm unfolding r lrns n'
        ATreeNode_gather_def ATreeNode_gather_hlpr.simps ATreen_merge.simps append_Nil
      by simp
    have "((lr, n) # lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Left, n)] (l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1) \<Longrightarrow> ?thesis"
    proof -
      assume left: "((lr, n) # lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Left, n)] (l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1)"
      hence lr: "lr = Left"
        using ATreeNode_gather_hlpr_prefix by fastforce
      have "((lrns', lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1)"
        using ATreeNode_gather_def ATreeNode_gather_left assm lr lrns n' r by fastforce
      hence "((NodeAccessor.Left, n) # lrns', lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Left, n)] r1"
        using par(1)[of r1 lrns', OF tm1]
        unfolding lrns n' lr r ATreeNode_gather_def ATree_gather_def 
      proof -
        assume a1: "((lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [] l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1) = ((map fst lrns', lf1) \<in> ATree_gather_hlpr [] l1 \<and> (lrns', lf2) \<in> ATreeNode_gather_hlpr [] r1)"
        assume "(lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [] l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1"
        then show ?thesis
          using a1 by (metis (no_types) ATreeNode_gather_hlpr_prefix_fix Cons_eq_appendI)
      qed
      moreover have "(map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2"
        using par(1)[of r1 lrns'] par(2)[of r2 lrns'] par(3) assm unfolding r lrns n'
        unfolding treen_mirror_inherit ATree_gather_def "ATree_gather_hlpr.simps" append_Nil
        using ATree_gather_hlpr_prefix 
        by (metis ATree_gather_def Cons_eq_appendI UnI1 \<open>(lrns', lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1\<close> \<open>treen_mirror l1 r1 \<Longrightarrow> ((lrns', lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1) = ((map fst lrns', lf1) \<in> \<union>\<^sub>A\<^sub>T l1 \<and> (lrns', lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r1)\<close> fst_conv list.simps(9) lr)
      ultimately show ?thesis
        unfolding r lrns n' ATreeNode_gather_def append_Nil lr
        by (simp)
    qed
    moreover have "((lr, n) # lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Right, n)] (l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2) \<Longrightarrow> ?thesis"
    proof -
      assume left: "((lr, n) # lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Right, n)] (l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2)"
      hence lr: "lr = Right"
        using ATreeNode_gather_hlpr_prefix by fastforce
      have "((lrns', lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2)"
        using ATreeNode_gather_def ATreeNode_gather_right assm lr lrns n' r by fastforce
      hence "((NodeAccessor.Right, n) # lrns', lf2) \<in> ATreeNode_gather_hlpr [(NodeAccessor.Right, n)] r2"
        using par(2)[of r2 lrns', OF tm2]
        unfolding lrns n' lr r ATreeNode_gather_def ATree_gather_def 
      proof -
        assume a1: "((lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [] l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2) = ((map fst lrns', lf1) \<in> ATree_gather_hlpr [] l2 \<and> (lrns', lf2) \<in> ATreeNode_gather_hlpr [] r2)"
        assume "(lrns', lf1, lf2) \<in> ATreeNode_gather_hlpr [] l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2"
        then show ?thesis
          using a1 by (metis (no_types) ATreeNode_gather_hlpr_prefix_fix Cons_eq_appendI)
      qed
      moreover have "(map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2"
        using par(1)[of r1 lrns'] par(2)[of r2 lrns'] par(3) assm unfolding r lrns n'
        unfolding treen_mirror_inherit ATree_gather_def "ATree_gather_hlpr.simps" append_Nil
        using ATree_gather_hlpr_prefix 
        by (metis Cons_eq_appendI UnI2 \<open>(lrns', lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2\<close> fst_conv list.simps(9) lr)
      ultimately show ?thesis
        unfolding r lrns n' ATreeNode_gather_def append_Nil lr
        by (simp)
    qed
    ultimately show "(map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r"
      using assm unfolding r lrns n' ATreeNode_gather_def ATreeNode_gather_hlpr.simps ATreen_merge.simps append_Nil
      by fast
  qed
  moreover have "((map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r) \<Longrightarrow> (lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N ((par l1 l2) \<times>\<^sub>A\<^sub>T\<^sub>n r)"
    using par(1)[of r1 "tl lrns"] par(2)[of r2 "tl lrns"] tm1 tm2 unfolding r "ATree_merge.simps" ATree_gather_def "ATree_gather_hlpr.simps" tree_mirror leaf_property.simps
      append_Nil
  proof clarsimp
    assume 
      l_or_r: "((map fst lrns, lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1 \<or> (map fst lrns, lf1) \<in> ATree_gather_hlpr [NodeAccessor.Right] l2)" and
      in_r: "(lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn r1 n r2" and
      IH1: "((tl lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1) = ((map fst (tl lrns), lf1) \<in> ATree_gather_hlpr [] l1 \<and> (tl lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r1)" and
      IH2: "((tl lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2) = ((map fst (tl lrns), lf1) \<in> ATree_gather_hlpr [] l2 \<and> (tl lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r2)"
    obtain lr n' lrns' where lrs: "lrns = (lr, n') # lrns'"
      by (meson ATreeNode_gather_par in_r)
    consider (left) "lr = Left" | (right) "lr = Right" using NodeAccessor.exhaust by blast
    thus "(lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn l1 \<times>\<^sub>A\<^sub>T\<^sub>n r1 n l2 \<times>\<^sub>A\<^sub>T\<^sub>n r2"
    proof cases
      case left
      have inside: "(map fst lrns', lf1) \<in> ATree_gather_hlpr [] l1 \<and> (lrns', lf2) \<in> ATreeNode_gather_hlpr [] r1"
        using l_or_r in_r unfolding lrs left
      proof -
        assume a1: "(map fst ((NodeAccessor.Left, n') # lrns'), lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1 \<or> (map fst ((NodeAccessor.Left, n') # lrns'), lf1) \<in> ATree_gather_hlpr [NodeAccessor.Right] l2"
        assume a2: "((NodeAccessor.Left, n') # lrns', lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn r1 n r2"
        obtain nns :: "NodeAccessor list \<Rightarrow> 'a ATree \<Rightarrow> NodeAccessor list \<Rightarrow> 'a \<Rightarrow> NodeAccessor list \<Rightarrow> NodeAccessor list" where
          "\<forall>x0 x1 x2 x3 x4. (\<exists>v5. x4 = x2 @ v5 \<and> (x0 @ v5, x3) \<in> ATree_gather_hlpr x0 x1) = (x4 = x2 @ nns x0 x1 x2 x3 x4 \<and> (x0 @ nns x0 x1 x2 x3 x4, x3) \<in> ATree_gather_hlpr x0 x1)"
          by moura
        then have f3: "\<forall>ns a nsa aa nsb. (ns, a) \<notin> ATree_gather_hlpr nsa aa \<or> ns = nsa @ nns nsb aa nsa a ns \<and> (nsb @ nns nsb aa nsa a ns, a) \<in> ATree_gather_hlpr nsb aa"
          by (meson ATree_gather_hlpr_prefix_fix)
        have "\<forall>ps b a c aa. (ps, b::'b) \<notin> \<union>\<^sub>A\<^sub>T\<^sub>N parn a (c::'c) aa \<or> (\<exists>n psa. ps = (n, c) # psa \<and> ((psa, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N a \<and> n = NodeAccessor.Left \<or> (psa, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N aa \<and> n = NodeAccessor.Right))"
          by simp
        then have f4: "(lrns', lf2) \<in> ATreeNode_gather_hlpr [] r1"
          using a2 by (metis (no_types) ATreeNode_gather_def NodeAccessor.distinct(1) fst_conv list.inject)
        have f5: "(NodeAccessor.Left # [] @ nns [] l1 [NodeAccessor.Left] lf1 (map fst ((NodeAccessor.Left, n') # lrns')) = NodeAccessor.Left # map fst lrns') = ([] @ nns [] l1 [NodeAccessor.Left] lf1 (map fst ((NodeAccessor.Left, n') # lrns')) = map fst lrns')"
          by simp
        have "(map fst ((NodeAccessor.Left, n') # lrns'), lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1"
          using f3 a1 by fastforce
        then show ?thesis
          using f5 f4 f3 by (metis (no_types) append_Cons fst_conv list.simps(9))
      qed
      thus ?thesis
        using ATreeNode_gather_left ATreeNode_gather_par IH1 in_r left lrs by fastforce
    next
      case right
      have inside: "(map fst lrns', lf1) \<in> ATree_gather_hlpr [] l2 \<and> (lrns', lf2) \<in> ATreeNode_gather_hlpr [] r2"
        using l_or_r in_r unfolding lrs right
      proof -
        assume a1: "((NodeAccessor.Right, n') # lrns', lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn r1 n r2"
        assume a2: "(map fst ((NodeAccessor.Right, n') # lrns'), lf1) \<in> ATree_gather_hlpr [NodeAccessor.Left] l1 \<or> (map fst ((NodeAccessor.Right, n') # lrns'), lf1) \<in> ATree_gather_hlpr [NodeAccessor.Right] l2"
        have f3: "\<forall>ps b a c aa. (ps, b::'b) \<notin> \<union>\<^sub>A\<^sub>T\<^sub>N parn a (c::'c) aa \<or> (\<exists>n psa. ps = (n, c) # psa \<and> ((psa, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N a \<and> n = NodeAccessor.Left \<or> (psa, b) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N aa \<and> n = NodeAccessor.Right))"
          by simp
        have f4: "(map fst ((NodeAccessor.Right, n') # lrns'), lf1) \<notin> ATree_gather_hlpr [NodeAccessor.Left] l1"
          by (metis (no_types) ATree_gather_hlpr_prefix_fix NodeAccessor.distinct(1) fst_conv hd_append list.sel(1) list.simps(9) not_Cons_self2)
        have "NodeAccessor.Right # map fst lrns' = map fst ((NodeAccessor.Right, n') # lrns')"
          by auto
        then show ?thesis
          using f4 f3 a2 a1 by (metis ATreeNode_gather_def ATree_gather_hlpr_prefix_fix NodeAccessor.distinct(1) fst_conv list.sel(1) list.sel(3) not_Cons_self2 tl_append2)
      qed
      thus ?thesis
        using ATreeNode_gather_par ATreeNode_gather_right IH2 in_r lrs right by fastforce
    qed
  qed
  ultimately show "(lrns, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N ((par l1 l2) \<times>\<^sub>A\<^sub>T\<^sub>n r) = ((map fst lrns, lf1) \<in> \<union>\<^sub>A\<^sub>T par l1 l2 \<and> (lrns, lf2) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r)"
    by blast
qed

fun ATree_mapfoldl :: "'lf ATree \<Rightarrow> ('lf \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a" where
"ATree_mapfoldl (seq content) lf f = (lf content)" |
"ATree_mapfoldl (par l r) lf f = (f (ATree_mapfoldl l lf f) (ATree_mapfoldl r lf f))"

fun ATreeNode_mapfoldl :: "('lf, 'nd) ATreeNode \<Rightarrow> ('lf \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'nd \<Rightarrow> 'a) \<Rightarrow> 'a" where
"ATreeNode_mapfoldl (seqn content) lf f = (lf content)" |
"ATreeNode_mapfoldl (parn l n r) lf f = (f (ATreeNode_mapfoldl l lf f) (ATreeNode_mapfoldl r lf f) n)"

lemma leaf_property_merge_to1: "tree_mirror c s \<Longrightarrow> (leaf_property P bl c s) = (leaf1_property (\<lambda> (c, s) . P c s) bl (c \<times>\<^sub>A\<^sub>T s))"
proof (induct c arbitrary: s)
  fix x :: 'a and s :: "'b ATree"
  assume tm: "tree_mirror (seq x) s"
  then obtain ss where s: "s = seq ss" unfolding tree_mirror using leaf_property_seqr by metis
  show "leaf_property P bl (seq x) s = leaf1_property (\<lambda>(x, y). P x y) bl seq x \<times>\<^sub>A\<^sub>T s"
    using tm
    unfolding s leaf_property.simps "ATree_merge.simps" "leaf1_property.simps" by simp
next
  fix c1 c2 :: "'a ATree" and s :: "'b ATree"
  assume
    IH1: "\<And>s. tree_mirror c1 s \<Longrightarrow>
             leaf_property P bl c1 s =
             leaf1_property (\<lambda>(x, y). P x y) bl c1 \<times>\<^sub>A\<^sub>T s" and
    IH2: "\<And>s. tree_mirror c2 s \<Longrightarrow>
             leaf_property P bl c2 s =
             leaf1_property (\<lambda>(x, y). P x y) bl c2 \<times>\<^sub>A\<^sub>T s" and
    tm: "tree_mirror (par c1 c2) s"
  then obtain s1 s2 where s: "s = par s1 s2" unfolding tree_mirror using leaf_property_parr by metis
  show "leaf_property P bl (par c1 c2) s =
       leaf1_property (\<lambda>(x, y). P x y) bl par c1 c2 \<times>\<^sub>A\<^sub>T s "
    using IH1[of s1] IH2[of s2] tm 
    unfolding leaf_property.simps "leaf1_property.simps" "ATree_merge.simps" s tree_mirror by simp
qed

lemma leaf1_property_or_ex: "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl tr \<Longrightarrow> \<exists> path lf . (path, lf) \<in> \<union>\<^sub>A\<^sub>T (tr) \<and> P lf"
  unfolding ATree_gather_def
proof -
  {
    fix path
    have "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl tr \<Longrightarrow> \<exists>path' lf . (path', lf) \<in> ATree_gather_hlpr path tr \<and> P lf"
    proof (induct tr arbitrary: path)
      show "\<And>x path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl (seq x) \<Longrightarrow> \<exists>path' lf. (path', lf) \<in> ATree_gather_hlpr path (seq x) \<and> P lf"
        unfolding "leaf1_property.simps" ATree_gather_hlpr.simps by simp
    next
      fix c1 c2 path
      assume
        IH1: "\<And> path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl c1 \<Longrightarrow> \<exists>path' lf. (path', lf) \<in> ATree_gather_hlpr path c1 \<and> P lf" and
        IH2: "\<And> path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl c2 \<Longrightarrow> \<exists>path' lf. (path', lf) \<in> ATree_gather_hlpr path c2 \<and> P lf" and
        bl: "bl = op \<or> \<or> bl = op \<and>" and
        lp: "leaf1_property P bl (par c1 c2)"
      show "\<exists>path' lf. (path', lf) \<in> ATree_gather_hlpr path (par c1 c2) \<and> P lf"
        unfolding "ATree_merge.simps" "ATree_gather_hlpr.simps"
        using bl 
      by (metis IH1 IH2 UnI1 UnI2 leaf1_property.simps(2) lp)
    qed
  }
  thus "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf1_property P bl tr \<Longrightarrow> \<exists>path lf. (path, lf) \<in> ATree_gather_hlpr [] tr \<and> P lf"
    by presburger
qed

lemma leaf_property_or_ex: "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl c s \<Longrightarrow> \<exists> path lf1 lf2 . (path, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T (c \<times>\<^sub>A\<^sub>T s) \<and> P lf1 lf2"
  unfolding ATree_gather_def
proof -
  {
    fix path
    have "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl c s \<Longrightarrow> \<exists>path' lf1 lf2. (path', lf1, lf2) \<in> ATree_gather_hlpr path c \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2"
    proof (induct c arbitrary: s path)
      show "\<And>x s path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl (seq x) s \<Longrightarrow> \<exists>path' lf1 lf2. (path', lf1, lf2) \<in> ATree_gather_hlpr path seq x \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2"
        using leaf_property_seqr by fastforce
    next
      fix c1 c2 s path
      assume
        IH1: "\<And>s path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl c1 s \<Longrightarrow> \<exists>path' lf1 lf2. (path', lf1, lf2) \<in> ATree_gather_hlpr path c1 \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2" and
        IH2: "\<And>s path. bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl c2 s \<Longrightarrow> \<exists>path' lf1 lf2. (path', lf1, lf2) \<in> ATree_gather_hlpr path c2 \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2" and
        bl: "bl = op \<or> \<or> bl = op \<and>" and
        lp: "leaf_property P bl (par c1 c2) s"
      obtain s1 s2 where s: "s = par s1 s2" using leaf_property_parr lp by blast
      show "\<exists>path' lf1 lf2. (path', lf1, lf2) \<in> ATree_gather_hlpr path par c1 c2 \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2"
        unfolding s "ATree_merge.simps" "ATree_gather_hlpr.simps"
        using bl 
      by (metis IH1 IH2 UnI1 UnI2 leaf_property.simps(4) lp s)
    qed
  }
  thus "bl = op \<or> \<or> bl = op \<and> \<Longrightarrow> leaf_property P bl c s \<Longrightarrow> \<exists>path lf1 lf2. (path, lf1, lf2) \<in> ATree_gather_hlpr [] c \<times>\<^sub>A\<^sub>T s \<and> P lf1 lf2"
    by presburger
qed

lemma leaf1_property_or_ex2: "(path, lf) \<in> \<union>\<^sub>A\<^sub>T tr \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P (op \<or>) tr"
  unfolding ATree_gather_def
proof -
  {
    fix pref
    have "(path, lf) \<in> ATree_gather_hlpr pref tr \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P op \<or> tr"
    proof (induct tr arbitrary: pref)
      show "\<And>x pref. (path, lf) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P op \<or> (seq x)" 
        unfolding tree_mirror ATree_gather_def using "leaf1_property.simps" by force
    next
      fix c1 c2 pref
      assume
        IH1: "\<And>pref. (path, lf) \<in> ATree_gather_hlpr pref c1 \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P op \<or> c1" and
        IH2: "\<And>pref. (path, lf) \<in> ATree_gather_hlpr pref c2 \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P op \<or> c2" and
        inside: "(path, lf) \<in> ATree_gather_hlpr pref (par c1 c2)" and
        P: "P lf"
      have "(path, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) c1 \<union> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) c2"
        using inside[unfolded ATree_merge.simps ATree_gather_def "ATree_gather_hlpr.simps"] by simp
      thus "leaf1_property P op \<or> (par c1 c2)"
        using IH1 IH2 P
        unfolding "leaf1_property.simps" tree_mirror
        by blast
    qed
  }
  thus "(path, lf) \<in> ATree_gather_hlpr [] tr \<Longrightarrow> P lf \<Longrightarrow> leaf1_property P op \<or> tr"
    by blast
qed

lemma leaf_property_or_ex2: "tree_mirror c s \<Longrightarrow> (path, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T (c \<times>\<^sub>A\<^sub>T s) \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P (op \<or>) c s"
  unfolding ATree_gather_def
proof -
  {
    fix pref
    have "tree_mirror c s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P op \<or> c s"
    proof (induct c arbitrary: s pref)
      show "\<And>x s pref. tree_mirror (seq x) s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref seq x \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P op \<or> (seq x) s" 
        unfolding tree_mirror ATree_gather_def using leaf_property_seqr "leaf_property.simps" by force
    next
      fix c1 c2 s pref
      assume
        IH1: "\<And>s pref. tree_mirror c1 s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c1 \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P op \<or> c1 s" and
        IH2: "\<And>s pref. tree_mirror c2 s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c2 \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P op \<or> c2 s" and
        tm: "tree_mirror (par c1 c2) s" and
        inside: "(path, lf1, lf2) \<in> ATree_gather_hlpr pref par c1 c2 \<times>\<^sub>A\<^sub>T s" and
        P: "P lf1 lf2"
      obtain s1 s2 where s: "s = par s1 s2" using tm leaf_property_parr unfolding tree_mirror by blast
      have "(path, lf1, lf2) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) c1 \<times>\<^sub>A\<^sub>T s1 \<union> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) c2 \<times>\<^sub>A\<^sub>T s2"
        using inside[unfolded s ATree_merge.simps ATree_gather_def "ATree_gather_hlpr.simps"] by simp
      thus "leaf_property P op \<or> (par c1 c2) s"
        using IH1[of s1] IH2[of s2] tm P
        unfolding s "leaf_property.simps" tree_mirror
        by blast
    qed
  }
  thus "tree_mirror c s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr [] c \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P op \<or> c s"
    by blast
qed

lemma leaf1_property_and_all: "leaf1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> \<union>\<^sub>A\<^sub>T tr \<Longrightarrow> P lf"
unfolding ATree_gather_def
proof -
  {
    fix pref
    have "\<And> pref . leaf1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> ATree_gather_hlpr pref tr \<Longrightarrow> P lf"
    proof (induct tr arbitrary: pref)
      show "\<And>x pref . leaf1_property P op \<and> (seq x) \<Longrightarrow> (path, lf) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> P lf"
        by fastforce
    next
      fix c1 c2 pref
      assume
        IH1: "\<And>pref. leaf1_property P op \<and> c1 \<Longrightarrow> (path, lf) \<in> ATree_gather_hlpr pref c1 \<Longrightarrow> P lf" and
        IH2: "\<And>pref. leaf1_property P op \<and> c2 \<Longrightarrow> (path, lf) \<in> ATree_gather_hlpr pref c2 \<Longrightarrow> P lf" and
        lp: "leaf1_property P op \<and> (par c1 c2)" and
        ing: "(path, lf) \<in> ATree_gather_hlpr pref (par c1 c2)"
      thus "P lf" by auto
    qed
  }
  thus "leaf1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> ATree_gather_hlpr [] tr \<Longrightarrow> P lf"
    by blast
qed

lemma leafn1_property_and_all: "leafn1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr \<Longrightarrow> P lf"
unfolding ATreeNode_gather_def
proof -
  {
    fix pref
    have "\<And> pref . leafn1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr pref tr \<Longrightarrow> P lf"
    proof (induct tr arbitrary: pref)
      show "\<And>x pref . leafn1_property P op \<and> (seqn x) \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr pref (seqn x) \<Longrightarrow> P lf"
        by fastforce
    next
      fix c1 c2 pref n
      assume
        IH1: "\<And>pref. leafn1_property P op \<and> c1 \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr pref c1 \<Longrightarrow> P lf" and
        IH2: "\<And>pref. leafn1_property P op \<and> c2 \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr pref c2 \<Longrightarrow> P lf" and
        lp: "leafn1_property P op \<and> (parn c1 n c2)" and
        ing: "(path, lf) \<in> ATreeNode_gather_hlpr pref (parn c1 n c2)"
      thus "P lf" by auto
    qed
  }
  thus "leafn1_property P op \<and> tr \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr [] tr \<Longrightarrow> P lf"
    by blast
qed

lemma leaf_property_and_all: "leaf_property P op \<and> c s \<Longrightarrow> (path, lf1, lf2) \<in> \<union>\<^sub>A\<^sub>T (c \<times>\<^sub>A\<^sub>T s) \<Longrightarrow> P lf1 lf2"
unfolding ATree_gather_def
proof -
  {
    fix pref
    have "\<And> pref . leaf_property P op \<and> c s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2"
    proof (induct c arbitrary: s pref)
      show "\<And>x pref s. leaf_property P op \<and> (seq x) s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref seq x \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2"
        using leaf_property_seqr by fastforce
    next
      fix c1 c2 s pref
      assume
        IH1: "\<And>pref s. leaf_property P op \<and> c1 s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c1 \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2" and
        IH2: "\<And>pref s. leaf_property P op \<and> c2 s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr pref c2 \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2" and
        lp: "leaf_property P op \<and> (par c1 c2) s" and
        ing: "(path, lf1, lf2) \<in> ATree_gather_hlpr pref par c1 c2 \<times>\<^sub>A\<^sub>T s"
      moreover obtain s1 s2 where s: "s = par s1 s2" using leaf_property_parr lp by blast
      ultimately show "P lf1 lf2" by auto
    qed
  }
  thus "leaf_property P op \<and> c s \<Longrightarrow> (path, lf1, lf2) \<in> ATree_gather_hlpr [] c \<times>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2"
    by blast
qed

lemma leaf_property_and_all': "leaf_property P op \<and> c s \<Longrightarrow> (path, lf1) \<in> \<union>\<^sub>A\<^sub>T c \<Longrightarrow> (path, lf2) \<in> \<union>\<^sub>A\<^sub>T s \<Longrightarrow> P lf1 lf2"
proof -
  assume a1: "leaf_property P op \<and> c s"
  assume a2: "(path, lf1) \<in> \<union>\<^sub>A\<^sub>T c"
  assume a3: "(path, lf2) \<in> \<union>\<^sub>A\<^sub>T s"
  have "tree_mirror s c"
    using a1 by (simp add: leaf_property_symmetric tree_mirror)
  then have "tree_mirror c s"
    by (simp add: leaf_property_symmetric)
  then show ?thesis
    using a3 a2 a1 by (meson ATree_unmerge_union leaf_property_and_all)
qed

lemma leaf_property_and_all2: "tree_mirror c s \<Longrightarrow> \<forall> lrs lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T c \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T s \<longrightarrow> P lf1 lf2 \<Longrightarrow> leaf_property P (op \<and>) c s"
proof (induct c arbitrary: s)
  case (seq x)
  then obtain ss where s: "s = seq ss" unfolding tree_mirror_unfold_seql by blast
  show ?case
    using seq(2)
    unfolding s
    unfolding leaf_property.simps ATree_gather_def ATree_gather_hlpr.simps
    by blast
next
  case (par c1 c2)
  then obtain s1 s2 where s: "s = par s1 s2" using tree_mirror_unfold_parl by metis
  have "\<forall>lrs lf1 lf2. (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T c1 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T s1 \<longrightarrow> P lf1 lf2"
    by (metis ATree_gather_left par.prems(2) s)
  moreover have "\<forall>lrs lf1 lf2. (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T c2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T s2 \<longrightarrow> P lf1 lf2"
    by (metis ATree_gather_right par.prems(2) s)
  ultimately show ?case
    using par(1,2,3)
    unfolding s tree_mirror_inherit
    using leaf_property_inherit[of "leaf_property P op \<and>" P c1 c2 s1 s2]
    by presburger
qed

lemma leaf1_property_and_all2: "\<forall> lrs lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T c \<longrightarrow> P lf \<Longrightarrow> leaf1_property P (op \<and>) c"
proof (induct c)
  case (seq x)
  then show ?case
    unfolding leaf1_property.simps ATree_gather_def ATree_gather_hlpr.simps
    by blast
next
  case (par c1 c2)
  then show ?case
    using ATree_gather_right ATree_gather_left
    by fastforce
qed

lemma ATree_access_imp_in_set:
  fixes tr:: "'lf ATree"
  assumes "Some lf = (tr $ lrs)"
  shows "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
using assms
unfolding ATree_gather_def
proof -
  {
    fix pref
    have "Some lf = (tr $ lrs) \<Longrightarrow> (pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr"
    proof (induct tr arbitrary: lrs pref)
      fix x lrs pref
      assume lf: "Some lf = seq x $ lrs"
      hence lrs: "lrs = []"
        by (metis ATree_access_node.simps(4) list.exhaust option.simps(3))
      (* with ex have pref: "pref = []" by simp *)
      show "(pref @ lrs, lf) \<in> ATree_gather_hlpr pref (seq x)"
        using lf[unfolded lrs "ATree_access_node.simps"] lrs
        unfolding "ATree_gather_hlpr.simps"
        by simp
    next
      fix tr1 tr2 lrs pref
      assume
        IH1: "\<And>lrs pref. Some lf = tr1 $ lrs \<Longrightarrow> (pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr1" and
        IH2: "\<And>lrs pref. Some lf = tr2 $ lrs \<Longrightarrow> (pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr2" and
        lf: "Some lf = par tr1 tr2 $ lrs"
      hence "\<exists> lr lrs' . lrs = lr # lrs'"
        by (metis ATree_access_node.simps(5) list.exhaust option.simps(3))
      then obtain lr lrs' where lrs: "lrs = lr # lrs'" by blast
      consider (l) "lr = Left" | (r) "lr = Right" using NodeAccessor.exhaust by blast
      thus "(pref @ lrs, lf) \<in> ATree_gather_hlpr pref (par tr1 tr2) "
      proof cases
        case l
        show ?thesis
          using IH1[of lrs' "pref @ [NodeAccessor.Left]"] lf unfolding l "ATree_gather_hlpr.simps" lrs "ATree_access_node.simps" by simp
      next
        case r
        show ?thesis
          using IH2[of lrs' "pref @ [NodeAccessor.Right]"] lf unfolding r "ATree_gather_hlpr.simps" lrs "ATree_access_node.simps" by simp
      qed
    qed
  }
  hence "Some lf = (tr $ lrs) \<Longrightarrow> ([]@lrs, lf) \<in> ATree_gather_hlpr [] tr"  by blast
  thus "Some lf = (tr $ lrs) \<Longrightarrow> (lrs, lf) \<in> ATree_gather_hlpr [] tr" by simp
qed

lemma ATree_in_set_imp_access:
  fixes tr:: "'lf ATree"
  assumes "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
  shows "Some lf = (tr $ lrs)"
using assms unfolding ATree_gather_def
proof -
  {
    fix pref
    have "(pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr \<Longrightarrow> Some lf = tr $ lrs"
    proof (induct tr arbitrary: pref lrs)
      show "\<And>x pref lrs . (pref @ lrs, lf) \<in> ATree_gather_hlpr pref (seq x) \<Longrightarrow> Some lf = seq x $ lrs" by auto
    next
      fix tr1 tr2 pref lrs
      assume
        IH1: "\<And>pref lrs. (pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr1 \<Longrightarrow> Some lf = tr1 $ lrs" and
        IH2: "\<And>pref lrs. (pref @ lrs, lf) \<in> ATree_gather_hlpr pref tr2 \<Longrightarrow> Some lf = tr2 $ lrs" and
        lf: "(pref @ lrs, lf) \<in> ATree_gather_hlpr pref (par tr1 tr2)"
      have "(lrs, lf) \<in> ATree_gather_hlpr [] (par tr1 tr2)" using ATree_gather_hlpr_prefix[OF lf]
        by (metis append_self_conv2 same_append_eq)
      have "lrs \<noteq> []" using ATree_gather_hlpr_prefix[OF lf] ATree_gather_hlpr_prefix by fastforce
      then obtain lr lrs' where lrs: "lrs = lr # lrs'" using list.exhaust by blast
      consider (left) "lr = Left" | (right) "lr = Right" using NodeAccessor.exhaust by blast
      thus "Some lf = par tr1 tr2 $ lrs"
      proof cases
        case left
        have ins: "(pref @ NodeAccessor.Left # lrs', lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) tr1"
          using lf
          unfolding lrs left
          unfolding "ATree_gather_hlpr.simps"
          using ATree_gather_hlpr_prefix
          by fastforce
        hence ins: "([]@lrs', lf) \<in> ATree_gather_hlpr ([]) tr1"
          using ATree_gather_hlpr_prefix[OF ins]
          by (metis append_Cons append_Nil append_assoc same_append_eq)
        show ?thesis
          using IH1[OF ins]
          unfolding  "ATree_access_node.simps" left lrs .
      next
        case right
        have ins: "(pref @ NodeAccessor.Right # lrs', lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) tr2"
          using lf
          unfolding lrs right
          unfolding "ATree_gather_hlpr.simps"
          using ATree_gather_hlpr_prefix
          by fastforce
        hence ins: "([]@lrs', lf) \<in> ATree_gather_hlpr ([]) tr2"
          using ATree_gather_hlpr_prefix[OF ins]
          by (metis append_Cons append_Nil append_assoc same_append_eq)
        show ?thesis
          using IH2[OF ins]
          unfolding  "ATree_access_node.simps" right lrs .
      qed
    qed
  }
  hence "([]@lrs, lf) \<in> ATree_gather_hlpr [] tr \<Longrightarrow> Some lf = tr $ lrs" by blast
  thus "(lrs, lf) \<in> ATree_gather_hlpr [] tr \<Longrightarrow> Some lf = tr $ lrs" by simp
qed

lemma ATree_in_set_eq_access:
  shows "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr = (Some lf = (tr $ lrs))"
using ATree_in_set_imp_access ATree_access_imp_in_set by metis


lemma ATree_eq_eq_gather_access:
  assumes
    tm: "tree_mirror tr1 tr2" and
    lf1: "(lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr1" and
    lf2: "(lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr2"
  shows "(lf1 = lf2) = ((tr1 $ lrs = tr2 $ lrs))"
by (metis ATree_in_set_imp_access lf1 lf2 option.inject)

lemma ATree_eq_eq_gather_access2:
  assumes "(tr1 $ lrs = tr2 $ lrs)"
  shows "\<exists> lf1 lf2 . ((lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr1 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr2 \<and> lf1 = lf2) \<or> (\<not> (\<exists> lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr1) \<and> \<not> (\<exists> lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr2))"
  by (metis ATree_in_set_eq_access assms) 

lemma ATreeNode_all_leafs_eq_imp_eq:
  assumes "(\<forall>lrns_s lf. ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1) = ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2))"
  shows "tr1 = tr2"
    using assms
proof (induct tr1 arbitrary: tr2)
  case (seqn x)
  then show ?case
    by (metis ATreeNode.exhaust ATreeNode_gather_par ATreeNode_gather_seq insertI1 list.distinct(1) prod.sel(2) singletonD)
next
  case (parn l1 x1 r1)
  obtain l2 x2 r2 where tr2: "tr2 = parn l2 x2 r2"
    using parn(3)
    by (metis ATreeNode.exhaust ATreeNode_gather_def ATreeNode_gather_hlpr.simps(1) ATreeNode_gather_par insertI1 list.distinct(1))
  have "\<forall>lrns_s lf. ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1) = ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2)"
  proof auto
    show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2"
      using parn(3)[unfolded tr2]
      by (metis (no_types, lifting) ATreeNode_gather_left ATreeNode_gather_par fst_conv list.sel(1))
    show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l2 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N l1"
      using parn(3)[unfolded tr2]
      by (metis (no_types, lifting) ATreeNode_gather_left ATreeNode_gather_par fst_conv list.sel(1))
  qed
  moreover have "\<forall>lrns_s lf. ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r1) = ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r2)"
  proof auto
    show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r1 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r2"
      using parn(3)[unfolded tr2]
      by (metis (no_types, lifting) ATreeNode_gather_right ATreeNode_gather_par fst_conv list.sel(1))
    show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r2 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N r1"
      using parn(3)[unfolded tr2]
      by (metis (no_types, lifting) ATreeNode_gather_right ATreeNode_gather_par fst_conv list.sel(1))
  qed
  moreover have "x1 = x2"
  proof -
    obtain lrns lf where "(lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn l1 x1 r1" using ATreeNode_access_exists by fast
    thus ?thesis
    using parn(3)[unfolded tr2] 
      by (metis (no_types, lifting) ATreeNode_gather_par Pair_inject list.inject)
  qed
  ultimately show ?case
    using parn unfolding tr2 by simp
qed

lemma ATreeNode_access_imp_in_set:
  fixes tr:: "('lf, 'nd) ATreeNode"
  assumes "Some (lf, nds) = (tr $' (lrs))"
  shows "(zip lrs nds, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr"
using assms
unfolding ATreeNode_gather_def
proof -
  {
    fix pref
    have "Some (lf, nds) = (tr $' (lrs)) \<Longrightarrow> (pref @ zip lrs nds, lf) \<in> ATreeNode_gather_hlpr pref tr"
    proof (induct tr arbitrary: lrs nds pref)
      case (seqn x)
      (* fix x lrs pref
      assume lf: "Some lf = seq x $ lrs" *)
      hence lrs: "lrs = [] \<and> nds = []"
        by (metis ATreeNode_access_node.simps(1) ATreeNode_access_node.simps(4) list.exhaust option.inject option.simps(3) prod.inject)
      show ?case
        using seqn[unfolded lrs[THEN conjunct1] lrs[THEN conjunct2] "ATreeNode_access_node.simps"] lrs
        unfolding "ATreeNode_gather_hlpr.simps"
        by simp
    next
      case (parn tr1 n tr2)
      obtain lr lrs' nds' where
        lrs: "lrs = lr # lrs'" and nds: "nds = n # nds'"
        using ATreeNode_access_node_parn parn(3) by metis
      then obtain nds' where nds: "nds = n # nds'" by blast
      consider (l) "lr = Left" | (r) "lr = Right" using NodeAccessor.exhaust by blast
      thus ?case
      proof cases
        case l
        have hlpr: "Some (lf, nds') = tr1 $' lrs'"
          using parn(3)
          unfolding l lrs nds
          by (meson ATreeNode_access_node_left)
        show ?thesis
          using parn(1)[OF hlpr, of "pref @ [(Left, n)]"] unfolding l "ATreeNode_gather_hlpr.simps" lrs nds "ATreeNode_access_node.simps" by simp
      next
        case r
        have hlpr: "Some (lf, nds') = tr2 $' lrs'"
          using parn(3)
          unfolding r lrs nds
          by (meson ATreeNode_access_node_right)
        show ?thesis
          using parn(2)[OF hlpr, of "pref @ [(Right, n)]"] unfolding r "ATreeNode_gather_hlpr.simps" lrs nds "ATreeNode_access_node.simps" by simp
      qed
    qed
  }
  hence "Some (lf, nds) = (tr $' lrs) \<Longrightarrow> ([]@zip lrs nds, lf) \<in> ATreeNode_gather_hlpr [] tr"  by blast
  thus "Some (lf, nds) = tr $' lrs \<Longrightarrow> (zip lrs nds, lf) \<in> ATreeNode_gather_hlpr [] tr" by simp
qed

lemma ATreeNode_access_imp_in_set':
  fixes tr:: "('lf, 'nd) ATreeNode"
  assumes "Some (lf, map snd lrns) = (tr $' (map fst lrns))"
  shows "(lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr"
  by (metis ATreeNode_access_imp_in_set assms zip_map_fst_snd)

lemma ATreeNode_in_set_imp_access:
  fixes tr:: "('lf, 'nd) ATreeNode"
  assumes "(zip lrs nds, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr" and "length lrs = length nds"
  shows "Some (lf, nds) = (tr $' lrs)"
using assms unfolding ATreeNode_gather_def
proof -
  {
    fix pref
    have "(pref @ zip lrs nds, lf) \<in> ATreeNode_gather_hlpr pref tr \<Longrightarrow> length lrs = length nds \<Longrightarrow> Some (lf, nds) = tr $' lrs"
    proof (induct tr arbitrary: pref lrs nds)
      case (seqn x)
      have lrs: "lrs = []" using seqn unfolding ATreeNode_gather_hlpr.simps
        using map_fst_zip by fastforce
      hence nds: "nds = []" using seqn by simp
      have lf: "lf = x" using seqn unfolding ATreeNode_gather_hlpr.simps by simp
      show ?case unfolding lf lrs nds ATreeNode_access_node.simps ..
    next 
      case (parn tr1 x tr2)
      hence "lrs \<noteq> []"
        using ATreeNode_gather_hlpr_prefix by fastforce
      then obtain lr lrs' nds' n where
        lrs: "lrs = lr # lrs'" and
        nds: "nds = n # nds'" using parn(3-4)
        by (metis length_0_conv list.collapse)
      have n: "n = x"
        using parn(3) unfolding ATreeNode_gather_hlpr.simps lrs nds
          zip.simps
        using ATreeNode_gather_hlpr_prefix by force
      consider (l) "lr = Left" | (r) "lr = Right" using NodeAccessor.exhaust by blast
      thus ?case
      proof cases
        case l
        have tmp: "(pref @ zip lrs' nds', lf) \<in> ATreeNode_gather_hlpr pref (tr1)"
          using parn(3) unfolding lrs nds n l
          using ATreeNode_gather_hlpr_left by force
        show ?thesis
          using parn(1)[OF tmp] parn(4)
          unfolding lrs l nds n ATreeNode_access_node_left by simp
      next
        case r
        have tmp: "(pref @ zip lrs' nds', lf) \<in> ATreeNode_gather_hlpr pref (tr2)"
          using parn(3) unfolding lrs nds n r
          using ATreeNode_gather_hlpr_right by force
        show ?thesis
          using parn(2)[OF tmp] parn(4)
          unfolding lrs r nds n ATreeNode_access_node_right by simp
      qed
    qed
  }
  hence "([] @ zip lrs nds, lf) \<in> ATreeNode_gather_hlpr [] tr \<Longrightarrow> length lrs = length nds \<Longrightarrow> Some (lf, nds) = tr $' lrs" by blast
  thus "(zip lrs nds, lf) \<in> ATreeNode_gather_hlpr [] tr \<Longrightarrow> length lrs = length nds \<Longrightarrow> Some (lf, nds) = tr $' lrs" by simp
qed

lemma ATreeNode_in_set_imp_access':
  fixes tr:: "('lf, 'nd) ATreeNode"
  assumes "(lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr"
  shows "Some (lf, map snd lrns) = (tr $' (map fst lrns))"
  by (simp add: ATreeNode_in_set_imp_access assms zip_map_fst_snd)

lemma ATreeNode_in_set_eq_access:
  shows "(lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr = ((Some (lf, map snd lrns)) = (tr $' (map fst lrns)))"
using ATreeNode_in_set_imp_access' ATreeNode_access_imp_in_set' by metis

lemma ATreeNode_to_ATree_leafs_remain:
  assumes
    in_atn: "(zip lrs nds, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr" and
    len_eq: "length lrs = length nds"
  shows "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T (ATreeNode_to_ATree tr)"
    using assms
proof (induct tr arbitrary: lrs nds)
  case (seqn x)
  then show ?case
    unfolding ATreeNode_to_ATree.simps
    unfolding ATree_gather_def ATreeNode_gather_def
    unfolding ATree_gather_hlpr.simps
    unfolding ATreeNode_gather_hlpr.simps
    using map_fst_zip by fastforce
next
  case (parn tr1 x2 tr2)
  have "zip lrs nds  \<noteq> []"
    using parn(3)
    using ATreeNode_gather_par by fastforce
  then obtain lr lrs' nd nds' where
    lrs: "lrs = lr # lrs'" and
    nds: "nds = nd # nds'" and
    eq_len: "length lrs' = length nds'"
    using parn(3)
    by (metis length_Suc_conv neq_Nil_conv parn(4) zip_Nil)
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr'1: "(zip lrs' nds', lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1"
      by (metis ATreeNode_access_imp_in_set ATreeNode_access_node_left ATreeNode_access_node_parn ATreeNode_in_set_imp_access' left list.sel(1) lrs map_fst_zip map_snd_zip nds parn.prems(1) parn.prems(2))
    then show ?thesis
      using ATree_gather_left left lrs
      using eq_len parn.hyps(1) by fastforce
  next
    case right
    have in_tr'2: "(zip lrs' nds', lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2"
      by (metis ATreeNode_access_imp_in_set ATreeNode_access_node_right ATreeNode_access_node_parn ATreeNode_in_set_imp_access' right list.sel(1) lrs map_fst_zip map_snd_zip nds parn.prems(1) parn.prems(2))
    then show ?thesis
      using ATree_gather_right right lrs
      using eq_len parn.hyps(2) by fastforce
  qed
qed

lemma ATreeNode_to_ATree_leafs_remain':
  assumes "(lrns, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr"
  shows "(map fst lrns, lf) \<in> \<union>\<^sub>A\<^sub>T (ATreeNode_to_ATree tr)"
using ATreeNode_to_ATree_leafs_remain
  by (metis assms length_map zip_map_fst_snd)

lemma ATree_gather_same_path_ex:
  assumes
    inside1: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tree1" and
    inside2: "tree_mirror tree1 tree2"
  shows "\<exists> lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tree2"
proof -
  {
    fix pref
    have "(lrs, lf) \<in> ATree_gather_hlpr pref tree1 \<Longrightarrow> tree_mirror tree1 tree2 \<Longrightarrow> \<exists> lf . (lrs, lf) \<in> ATree_gather_hlpr pref tree2"
    proof (induct tree1 arbitrary: lrs tree2 lf pref)
      fix x :: 'a and lrs and tree2 :: "'b ATree" and lf pref
      assume
        inside: "(lrs, lf) \<in> ATree_gather_hlpr pref (seq x)" and
        tm: "tree_mirror (seq x) tree2"
      then obtain t2 where tree2: "tree2 = seq t2" using tm[unfolded tree_mirror] leaf_property_seqr by fast
      show "\<exists>lf. (lrs, lf) \<in> ATree_gather_hlpr pref tree2"
        using inside
        unfolding tree2 "ATree_gather_hlpr.simps" by simp
    next
      fix tree11 tree12 :: "'a ATree" and lrs pref and tree2 :: "'b ATree" and lf :: "'a"
      assume
        IH1: "\<And>lrs (tree2::'b ATree) lf pref.
           (lrs, lf) \<in> ATree_gather_hlpr pref tree11 \<Longrightarrow>
           tree_mirror tree11 tree2 \<Longrightarrow>
           \<exists>lf. (lrs, lf) \<in> ATree_gather_hlpr pref tree2" and
        IH2: "\<And>lrs (tree2::'b ATree) lf pref.
           (lrs, lf) \<in> ATree_gather_hlpr pref tree12 \<Longrightarrow>
           tree_mirror tree12 tree2 \<Longrightarrow>
           \<exists>lf. (lrs, lf) \<in> ATree_gather_hlpr pref tree2" and
        inside: "(lrs, lf) \<in> ATree_gather_hlpr pref (par tree11 tree12)" and
        tm: "tree_mirror (par tree11 tree12) tree2"
      obtain t21 t22 where tree2: "tree2 = par t21 t22" using tm[unfolded tree_mirror] leaf_property_parr by fast
      have tm': "tree_mirror tree11 t21 \<and> tree_mirror tree12 t22" using tm[unfolded tree2] unfolding tree_mirror "leaf_property.simps" by simp
      have "lrs \<noteq> []"
        using ATree_gather_hlpr_prefix inside by fastforce
      then obtain lr lrs' where lrs: "lrs = lr # lrs'" using list.exhaust by blast
      from inside [unfolded ATree_gather_hlpr.simps]
      consider (left) "(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) tree11" | (right) "(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) tree12" by blast
      thus "\<exists>lf. (lrs, lf) \<in> ATree_gather_hlpr pref tree2"
      proof cases
        case left
        obtain lf where p: "(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Left]) t21"
          using IH1[OF left tm'[THEN conjunct1]] by blast
        thus ?thesis unfolding tree2 using ATree_gather_hlpr.simps by auto
      next
        case right
        obtain lf where p: "(lrs, lf) \<in> ATree_gather_hlpr (pref @ [NodeAccessor.Right]) t22"
          using IH2[OF right tm'[THEN conjunct2]] by blast
        thus ?thesis unfolding tree2 using ATree_gather_hlpr.simps by auto
      qed
    qed
  }
  hence "(lrs, lf) \<in> ATree_gather_hlpr [] tree1 \<Longrightarrow> tree_mirror tree1 tree2 \<Longrightarrow> \<exists> lf . (lrs, lf) \<in> ATree_gather_hlpr [] tree2"
  by blast
  thus ?thesis using assms unfolding ATree_gather_def by presburger
qed

lemma ATree_eq_eq_gather_access_all:
  assumes tm: "tree_mirror tr1 tr2" and
    "\<forall> lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr1 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr2 \<longrightarrow> lf1 = lf2"
  shows "((tr1 $ lrs = tr2 $ lrs))"
proof -
  consider (is_some) "\<exists> lf1 . Some lf1 = tr1 $ lrs" | (is_none) "None = tr1 $ lrs"
    by (metis not_None_eq)
  thus ?thesis
  proof (cases)
    case is_some
    then obtain lf1 where ac1: "Some lf1 = tr1 $ lrs" by blast
    hence in_tr1: "(lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr1" using ATree_in_set_eq_access by fast
    then obtain lf2 where in_tr2: "(lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr2"
      using tm ATree_gather_same_path_ex by metis
    hence ac2: "Some lf2 = tr2 $ lrs"
      using ATree_in_set_eq_access by fast
    have "lf1 = lf2" using in_tr1 in_tr2 assms(2) by blast
    thus ?thesis using ac1 ac2 by argo
  next
    case is_none
    then show ?thesis
      by (metis ATree_access_imp_in_set ATree_gather_same_path_ex ATree_in_set_imp_access option.collapse option.simps(3) tm tree_mirror_symmetric)
  qed
qed

lemma ATree_eq_eq_gather_access3:
  assumes tm: "tree_mirror tt tt'"
  shows "(\<forall>lrs' lf lf'. lrs \<noteq> lrs' \<and> (lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tt \<and> (lrs', lf') \<in> \<union>\<^sub>A\<^sub>T tt' \<longrightarrow> lf = lf') = (\<forall>lrs'. lrs \<noteq> lrs' \<longrightarrow> tt $ lrs' = tt' $ lrs')"
  by (meson ATree_eq_eq_gather_access ATree_eq_eq_gather_access_all tm)

lemma ATree_merge_left_tree_mirror:
  assumes
    tml: "tree_mirror l c" and
    tmr: "tree_mirror r c"
  shows "tree_mirror (l \<times>\<^sub>A\<^sub>T r) c"
using assms
proof (induct l arbitrary: r c)
  fix x::"'a" and r::"'c ATree" and c::"'b ATree"
  assume
    tml: "tree_mirror (seq x) c" and
    tmr: "tree_mirror r c"
  obtain cs where c: "c = seq cs" using tml
    unfolding tree_mirror
    using leaf_property_seqr
    by fast
  obtain rs where r: "r = seq rs" using tmr unfolding c
    unfolding tree_mirror
    using leaf_property_seql
    by fast
  show "tree_mirror ((seq x) \<times>\<^sub>A\<^sub>T r) c"
    using tmr
    unfolding c r tree_mirror leaf_property.simps "ATree_merge.simps"
    .
next
  fix l1 l2::"'a ATree" and r::"'c ATree" and c::"'b ATree"
  assume
    IH1: "\<And>(r::'c ATree) (c::'b ATree). tree_mirror l1 c \<Longrightarrow> tree_mirror r c \<Longrightarrow> tree_mirror (l1 \<times>\<^sub>A\<^sub>T r) c" and
    IH2: "\<And>(r::'c ATree) (c::'b ATree). tree_mirror l2 c \<Longrightarrow> tree_mirror r c \<Longrightarrow> tree_mirror (l2 \<times>\<^sub>A\<^sub>T r) c" and
    tml: "tree_mirror (par l1 l2) c" and
    tmr: "tree_mirror r c"
  obtain c1 c2 where c: "c = par c1 c2" using tml[unfolded tree_mirror] leaf_property_parr by fast
  obtain r1 r2 where r: "r = par r1 r2" using tmr[unfolded tree_mirror c] leaf_property_parl by fast
  show "tree_mirror ((par l1 l2) \<times>\<^sub>A\<^sub>T r) c"
    using IH1[of c1 r1] IH2[of c2 r2] tml tmr
    unfolding c r tree_mirror leaf_property.simps ATree_merge.simps by blast
qed

lemma ATree_merge_right_tree_mirror:
  assumes
    tml: "tree_mirror c l" and
    tmr: "tree_mirror c r"
  shows "tree_mirror c (l \<times>\<^sub>A\<^sub>T r)"
using assms ATree_merge_left_tree_mirror[of l c r] leaf_property_symmetric[of "(\<lambda>_ _. True)" "(\<lambda>_ _. True)" "op \<and>", simplified] unfolding tree_mirror by metis

lemma ATree_merge_tree_mirror_distributive_left:
  assumes
    tml: "tree_mirror c l" and
    tmr: "tree_mirror c r"
  shows "tree_mirror (c \<times>\<^sub>A\<^sub>T l) (c \<times>\<^sub>A\<^sub>T r)"
  unfolding tree_mirror
using leaf_property_transitive[OF _
 ATree_merge_right_tree_mirror[OF _ tml, of c, unfolded tree_mirror, OF leaf_property_reflexive, simplified, THEN leaf_property_symmetric[of "\<lambda> _ _ . True" "\<lambda> _ _ . True" "op \<and>", simplified]]
 ATree_merge_right_tree_mirror[OF _ tmr, of c, unfolded tree_mirror, OF leaf_property_reflexive, simplified]
]
by simp

primrec node1_property :: "('nd \<Rightarrow> bool) \<Rightarrow> ('lf, 'nd) ATreeNode \<Rightarrow> bool"
  where
"node1_property f (seqn lf) = True" |
"node1_property f (parn l n r) = ((f n) \<and> node1_property f l \<and> node1_property f r)"

primrec subtreeN_property :: "(('lf, 'nd) ATreeNode \<Rightarrow> 'nd \<Rightarrow> ('lf, 'nd) ATreeNode \<Rightarrow> bool) \<Rightarrow> ('lf, 'nd) ATreeNode \<Rightarrow> bool"
  where
"subtreeN_property f (seqn lf) = True" |
"subtreeN_property f (parn l n r) = ((f l n r) \<and> subtreeN_property f l \<and> subtreeN_property f r)"

fun subtreeN2_property :: "('lf1 ATree \<Rightarrow> 'lf1 ATree \<Rightarrow> ('lf2, 'nd) ATreeNode \<Rightarrow> 'nd \<Rightarrow> ('lf2, 'nd) ATreeNode \<Rightarrow> bool) \<Rightarrow> 'lf1 ATree \<Rightarrow> ('lf2, 'nd) ATreeNode \<Rightarrow> bool"
  where
"(subtreeN2_property _ (seq _) (seqn _)) = True" |
"subtreeN2_property f (par l1 r1) (parn l2 n r2) = ((f l1 r1 l2 n r2) \<and> subtreeN2_property f l1 l2 \<and> subtreeN2_property f r1 r2)" |
"(subtreeN2_property _ (par _ _) (seqn _)) = False" |
"(subtreeN2_property _ (seq _) (parn _ _ _)) = False"


lemma node1_property_and_all: "node1_property P tr \<Longrightarrow> (path, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr \<Longrightarrow> \<forall> x \<in> set path . P (snd x)"
unfolding ATreeNode_gather_def
proof -
  {
    have "\<And> pref . node1_property P tr \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr pref tr \<Longrightarrow> \<forall>x\<in>set pref. P (snd x) \<Longrightarrow> \<forall>x\<in>set path. P (snd x)"
    proof (induct tr)
      case (seqn x)
      then show ?case by fastforce
    next
      case (parn tr1 n tr2)
      have "(\<forall>x\<in>set (pref @ [(Left, n)]). P (snd x)) \<and> (\<forall>x\<in>set (pref @ [(Right, n)]). P (snd x))" using parn by simp
      thus "\<forall>x\<in>set path. P (snd x)" using parn
        by (metis ATreeNode_gather_hlpr.simps(2) UnE node1_property.simps(2))
    qed
  }
  thus "node1_property P tr \<Longrightarrow> (path, lf) \<in> ATreeNode_gather_hlpr [] tr \<Longrightarrow> \<forall>x\<in>set path. P (snd x)"
    by fastforce
qed

lemma ATree_map_access:
  "(\<exists> p . ATree_map tr f $ lrs = Some p) = (\<exists> p . tr $ lrs = Some p)"
proof(induct tr arbitrary: lrs)
  case (seq x)
  then show ?case
    unfolding ATree_map.simps
    by (metis ATree_access_node.simps(1) ATree_access_node.simps(4) hd_Cons_tl option.simps(3))
next
  case (par tr1 tr2)
  consider
    (empty) "lrs = []" |
    (left) "\<exists> lrs' . lrs = Left # lrs'" |
    (right) "\<exists> lrs' . lrs = Right # lrs'"
    by (metis NodeAccessor.exhaust hd_Cons_tl)
  thus ?case
  proof (cases)
    case empty
    show ?thesis
      unfolding empty ATree_map.simps ATree_access_node.simps
      by simp
  next
    case left
    then obtain lrs' where lrs: "lrs = Left # lrs'" by blast
    show ?thesis
      unfolding lrs ATree_map.simps ATree_access_node.simps
      using par(1) by presburger
  next
    case right
    then obtain lrs' where lrs: "lrs = Right # lrs'" by blast
    show ?thesis
      unfolding lrs ATree_map.simps ATree_access_node.simps
      using par(2) by presburger
  qed
qed

lemma ATree_merge_access_exchange_r:
  assumes
    "(l \<times>\<^sub>A\<^sub>T r) $ lrs = Some (ll, rr)" and
    "r $ lrs = r' $ lrs" and
    "tree_mirror l r" and
    "tree_mirror l r'"
  shows "(l \<times>\<^sub>A\<^sub>T r') $ lrs = Some (ll, rr)"
by (metis (no_types, lifting) ATree_access_imp_in_set ATree_in_set_imp_access ATree_unmerge_union assms)

lemma ATree_merge_access_exchange_l:
  assumes
    "(l \<times>\<^sub>A\<^sub>T r) $ lrs = Some (ll, rr)" and
    "l $ lrs = l' $ lrs" and
    "tree_mirror l r" and
    "tree_mirror l' r"
  shows "(l' \<times>\<^sub>A\<^sub>T r) $ lrs = Some (ll, rr)"
by (metis (no_types, lifting) ATree_access_imp_in_set ATree_in_set_imp_access ATree_unmerge_union assms)
  
lemma ATree_map_access_eq_distr:
  assumes "tr $ lrs = tr' $ lrs"
  shows "ATree_map tr f $ lrs = ATree_map tr' f $ lrs"
    using assms
proof(induct tr arbitrary: lrs tr')
  case (seq x)
  then show ?case
    unfolding ATree_map.simps
    by (metis (no_types, lifting) ATree.disc(1) ATree.disc(1) ATree.exhaust ATree.simps(1) ATree.simps(1) ATree_access_node.simps(1) ATree_access_node.simps(1) ATree_access_node.simps(4) ATree_access_node.simps(4) ATree_access_node.simps(5) ATree_map.simps(1) ATree_map_access is_seq_def is_seq_def list.exhaust not_None_eq option.distinct(1) option.distinct(1) option.inject option.inject)
next
  case (par tr1 tr2)
  consider
    (empty) "lrs = []" |
    (left) "\<exists> lrs' . lrs = Left # lrs'" |
    (right) "\<exists> lrs' . lrs = Right # lrs'"
    by (metis NodeAccessor.exhaust hd_Cons_tl)
  thus ?case
  proof (cases)
    case empty
    show ?thesis
      unfolding empty ATree_map.simps ATree_access_node.simps
      by (metis ATree_access_node.simps(5) ATree_map_access empty not_None_eq par.prems)
  next
    case left
    then obtain lrs' where lrs: "lrs = Left # lrs'" by blast
    consider (some) "\<exists> p . (par tr1 tr2) $ lrs = Some p" | (none) "(par tr1 tr2) $ lrs = None" by blast
    thus ?thesis
    proof cases
      case some
      then obtain p where p: "(par tr1 tr2) $ lrs = Some p" by blast
      hence tr'_acc: "tr' $ lrs = Some p" using par(3) by simp
      have "\<exists> tr'1 tr'2 . tr' = par tr'1 tr'2"
      proof (rule ccontr, auto)
        assume "\<forall>tr'1 tr'2. tr' \<noteq> par tr'1 tr'2"
        then obtain t where tr': "tr' = seq t"
          by (metis (full_types) ATree.rel_cases ATree.rel_eq)
        thus False
          using lrs tr'_acc by auto
      qed
      then obtain tr'1 tr'2 where tr': "tr' = par tr'1 tr'2"
        by blast
      show ?thesis
        using p tr'_acc
        unfolding tr' lrs ATree_map.simps ATree_access_node.simps
        using par(1)
        by presburger
    next
      case none
      then show ?thesis
        by (metis ATree_map_access not_None_eq par.prems)
    qed
  next
    case right
    then obtain lrs' where lrs: "lrs = Right # lrs'" by blast
    consider (some) "\<exists> p . (par tr1 tr2) $ lrs = Some p" | (none) "(par tr1 tr2) $ lrs = None" by blast
    thus ?thesis
    proof cases
      case some
      then obtain p where p: "(par tr1 tr2) $ lrs = Some p" by blast
      hence tr'_acc: "tr' $ lrs = Some p" using par(3) by simp
      have "\<exists> tr'1 tr'2 . tr' = par tr'1 tr'2"
      proof (rule ccontr, auto)
        assume "\<forall>tr'1 tr'2. tr' \<noteq> par tr'1 tr'2"
        then obtain t where tr': "tr' = seq t"
          by (metis (full_types) ATree.rel_cases ATree.rel_eq)
        thus False
          using lrs tr'_acc by auto
      qed
      then obtain tr'1 tr'2 where tr': "tr' = par tr'1 tr'2"
        by blast
      show ?thesis
        using p tr'_acc
        unfolding tr' lrs ATree_map.simps ATree_access_node.simps
        using par(2)
        by presburger
    next
      case none
      then show ?thesis
        by (metis ATree_map_access not_None_eq par.prems)
    qed
  qed
qed

lemma ATree_eq:
  assumes
    tm: "tree_mirror tr tr'" and
    all_paths_equal: "\<And> lrs lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr' \<Longrightarrow> lf1 = lf2"
  shows "tr = tr'"
using assms
proof (induct tr arbitrary: tr')
  case (seq x)
  then show ?case
    unfolding tree_mirror_unfold_seql
    unfolding ATree_map.simps
    unfolding ATree_gather_def ATree_gather_hlpr.simps
    by force
next
  case (par tr1 tr2)
  obtain tr'1 tr'2 where tr': "tr' = par tr'1 tr'2"
    using par(3) tree_mirror_unfold_parl by metis
  have tm1: "tree_mirror tr1 tr'1"
    using tr' par(3) tree_mirror_inherit by fast
  have tm2: "tree_mirror tr2 tr'2"
    using tr' par(3) tree_mirror_inherit by fast
  have left_eq: "\<And> lrs lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr1 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr'1 \<Longrightarrow> lf1 = lf2"
    using par(4) ATree_gather_left tr' by fastforce
  have right_eq: "\<And> lrs lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr2 \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr'2 \<Longrightarrow> lf1 = lf2"
    using par(4) ATree_gather_right tr' by fastforce
  have "tr1 = tr'1"
    using par(1)[OF tm1] left_eq .
  moreover have "tr2 = tr'2"
    using par(2)[OF tm2] right_eq .
  ultimately show ?case using tr' by simp
qed

lemma ATree_map_access_is_map:
  assumes
    "(lrs, flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr f"
  shows "\<exists> lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr \<and> flf = f lf"
using assms
proof (induct tr arbitrary: lrs)
  case (seq x)
  then show ?case
    unfolding ATree_map.simps
    unfolding ATree_gather_def ATree_gather_hlpr.simps
    by blast
next
  case (par tr1 tr2)
  then obtain lr lrs' where
    lrs: "lrs = lr # lrs'"
    by (metis ATree_gather_par ATree_map.simps(2))
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr1_m: "(lrs', flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr1 f"
      using ATree_gather_left left lrs par.prems by fastforce
    obtain lf where
      in_tr1: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr1" and
      map_eq: "flf = f lf"
      using par(1)[OF in_tr1_m] by blast
    thus ?thesis
      using ATree_gather_left left lrs by fastforce
  next
    case right
    have in_tr2_m: "(lrs', flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr2 f"
      using ATree_gather_right right lrs par.prems by fastforce
    obtain lf where
      in_tr2: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr2" and
      map_eq: "flf = f lf"
      using par(2)[OF in_tr2_m] by blast
    thus ?thesis
      using ATree_gather_right right lrs by fastforce
  qed
qed

lemma ATree_map_access_is_map':
  assumes
    "(lrs, flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr f" and
    "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
  shows "flf = f lf"
using assms
proof (induct tr arbitrary: lrs)
  case (seq x)
  then show ?case
    unfolding ATree_map.simps
    unfolding ATree_gather_def ATree_gather_hlpr.simps
    by blast
next
  case (par tr1 tr2)
  obtain lr lrs' where
    lrs: "lrs = lr # lrs'"
    using par
    by (metis ATree_gather_par)
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr1_m: "(lrs', flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr1 f"
      using ATree_gather_left left lrs par.prems by fastforce
    have in_tr1: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr1"
      using ATree_gather_left left lrs par.prems by fastforce
    show ?thesis
      using par(1)[OF in_tr1_m in_tr1] .
  next
    case right
    have in_tr2_m: "(lrs', flf) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr2 f"
      using ATree_gather_right right lrs par.prems by fastforce
    have in_tr2: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr2"
      using ATree_gather_right right lrs par.prems by fastforce
    show ?thesis
      using par(2)[OF in_tr2_m in_tr2] .
  qed
qed

lemma ATree_map_eq:
  assumes
    tm: "tree_mirror tr tr'" and
    lf_map_eq: "\<And> lrs lf1 lf2 . (lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr \<and> (lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr' \<Longrightarrow> f lf1 = f' lf2"
  shows "ATree_map tr f = ATree_map tr' f'"
proof (rule ATree_eq, auto)
  show "tree_mirror (ATree_map tr f) (ATree_map tr' f')"
    by (meson ATree_map_tree_mirror tm tree_mirror_symmetric tree_mirror_transitive)
next
  fix lrs lf1_m lf2_m
  assume 
    lf1_m: "(lrs, lf1_m) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr f" and
    lf2_m: "(lrs, lf2_m) \<in> \<union>\<^sub>A\<^sub>T ATree_map tr' f'"
  obtain lf1 where lf1: "(lrs, lf1) \<in> \<union>\<^sub>A\<^sub>T tr" by (meson ATree_map_access_is_map lf1_m)
  obtain lf2 where lf2: "(lrs, lf2) \<in> \<union>\<^sub>A\<^sub>T tr'" by (meson ATree_map_access_is_map lf2_m)
  show "lf1_m = lf2_m"
    using lf1 lf1_m lf2 lf2_m
    by (metis ATree_map_access_is_map' lf_map_eq)
qed

lemma ATree_access_node_eq_distr_r:
  assumes "r $ lrs = r' $ lrs" and tm: "tree_mirror l r" and tm': "tree_mirror l r'"
  shows "(l \<times>\<^sub>A\<^sub>T r) $ lrs = (l \<times>\<^sub>A\<^sub>T r') $ lrs"
    using assms
proof (induct r arbitrary: l r' lrs)
  case (seq rr)
  obtain ll where l: "l = seq ll" using seq(2) unfolding tree_mirror_unfold_seqr by blast
  obtain rr' where r': "r' = seq rr'" using seq(3) unfolding l tree_mirror_unfold_seql by blast
  consider (none) "seq rr $ lrs = None" | (some) "\<exists> p . seq rr $ lrs = Some p" by blast
  thus ?case(*  using seq(1) unfolding l r' ATree_merge.simps ATree_access_node.simps *)
  proof cases
    case none
    hence "lrs \<noteq> []" by auto
    hence "\<And> p . seq p $ lrs = None"
      by (metis ATree_access_node.simps(4) list.exhaust)
    then show ?thesis using seq(1) unfolding l r' ATree_merge.simps ATree_access_node.simps by metis
  next
    case some
    hence lrs: "lrs = []"
      using ATree_access_node.elims by blast
    then show ?thesis using seq(1) unfolding l r' ATree_merge.simps ATree_access_node.simps lrs by simp
  qed
next
  case (par r1 r2)
  obtain l1 l2 where l: "l = par l1 l2" using par(4) tree_mirror_unfold_parr by blast
  obtain r1' r2' where r': "r' = par r1' r2'" using par(5) tree_mirror_unfold_parl l by blast
  consider
    (empty) "lrs = []" |
    (left) "\<exists> lrs' . lrs = Left # lrs'" |
    (right) "\<exists> lrs' . lrs = Right # lrs'"
    by (metis NodeAccessor.exhaust hd_Cons_tl)
  thus ?case
  proof (cases)
    case empty
    show ?thesis unfolding l r' empty ATree_merge.simps ATree_access_node.simps ..
  next
    case left
    then obtain lrs' where lrs: "lrs = Left # lrs'" by blast
    then show ?thesis
      using par(1) par(3) par(5,4)
      unfolding l r' lrs ATree_merge.simps ATree_access_node.simps tree_mirror_inherit
      by blast
  next
    case right
    then obtain lrs' where lrs: "lrs = Right # lrs'" by blast
    then show ?thesis
      using par(2-5)
      unfolding l r' lrs ATree_merge.simps ATree_access_node.simps tree_mirror_inherit
      by blast
  qed
qed

lemma ATree_gather_imp_ATreeNode_gather:
  assumes
    tm: "treen_mirror tr tr'" and
    gather: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
  shows "\<exists> nds lf' . length lrs = length nds \<and> (zip lrs nds, lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'"
using assms
proof (induct tr arbitrary: lrs tr')
  case (seq x)
  obtain x' where tr': "tr' = seqn x'"
    using seq(1) treen_mirror_unfold_seql by fastforce
  then show ?case
    using seq(2)
    unfolding tr' ATreeNode_to_ATree.simps ATree_gather_def ATree_gather_hlpr.simps
    unfolding ATreeNode_gather_def ATreeNode_gather_hlpr.simps
    by force
next
  case (par tr1 tr2)
  obtain tr'1 tr'2 n where tr': "tr' = parn tr'1 n tr'2"
    using par(3) treen_mirror_unfold_parl by metis
  have tm1: "treen_mirror tr1 tr'1"
    using tr' par(3) treen_mirror_inherit by fast
  have tm2: "treen_mirror tr2 tr'2"
    using tr' par(3) treen_mirror_inherit by fast
  obtain lr lrs' where
    lrs: "lrs = lr # lrs'"
    using par
    by (metis ATree_gather_par)
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr1: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr1"
      using ATree_gather_left left lrs par.prems by force
    obtain nds lf' where
      eq_len: "length lrs' = length nds" and in_tr'1: "(zip lrs' nds, lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'1"
      using par(1)[OF tm1 in_tr1]
      by blast
    have "\<exists> nd . (zip lrs (nd # nds), lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn tr'1 n tr'2"
      using ATreeNode_gather_left in_tr'1 left lrs by fastforce
    thus ?thesis
      using eq_len lrs tr'
      by (metis length_Cons)
  next
    case right
    have in_tr2: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr2"
      using ATree_gather_right right lrs par.prems by force
    obtain nds lf' where
      eq_len: "length lrs' = length nds" and in_tr'2: "(zip lrs' nds, lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'2"
      using par(2)[OF tm2 in_tr2]
      by blast
    have "\<exists> nd . (zip lrs (nd # nds), lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N parn tr'1 n tr'2"
      using ATreeNode_gather_right in_tr'2 right lrs by fastforce
    thus ?thesis
      using eq_len lrs tr'
      by (metis length_Cons)
  qed
qed

lemma ATree_gather_imp_ATreeNode_gather_conversion:
  assumes "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T (ATreeNode_to_ATree cuc)"
  shows "\<exists> nds . length lrs = length nds \<and> (zip lrs nds, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N cuc"
using assms using ATree_gather_imp_ATreeNode_gather
  by (metis ATreeNode_to_ATree_leafs_remain ATreeNode_to_ATree_tree_mirror ATree_gather_same_path)

lemma ATreeNode_gather_imp_ATree_gather:
  assumes
    tm: "treen_mirror tr tr'" and
    len_eq: "length lrs = length nds" and
    gather: "(zip lrs nds, lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'"
  shows "\<exists> lf . (lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tr"
using assms
proof (induct tr' arbitrary: tr lrs nds)
  case (seqn x)
  then obtain s where tr: "tr = seq s"
    unfolding treen_mirror_unfold_seqr
    by fast
  show ?case
    using seqn(2-3)
    unfolding ATreeNode_gather_def ATreeNode_gather_hlpr.simps
    unfolding tr ATree_gather_def ATree_gather_hlpr.simps
    using map_fst_zip by fastforce
next
  case (parn tr'1 x2 tr'2)
  obtain tr1 tr2 where tr: "tr = par tr1 tr2"
    using parn(3) treen_mirror_unfold_parr by metis
  have tm1: "treen_mirror tr1 tr'1"
    using tr parn(3) treen_mirror_inherit by fast
  have tm2: "treen_mirror tr2 tr'2"
    using tr parn(3) treen_mirror_inherit by fast
  have "zip lrs nds  \<noteq> []"
    using parn(5)
    using ATreeNode_gather_par by fastforce
  then obtain lr lrs' nd nds' where
    lrs: "lrs = lr # lrs'" and
    nds: "nds = nd # nds'" and
    eq_len: "length lrs' = length nds'"
    using parn(4)
    by (metis ATreeNode_access_node_parn ATreeNode_in_set_imp_access length_Suc_conv parn.prems(3))
  consider (left) "lr = Left" | (right) "lr = Right"
    using NodeAccessor.exhaust by meson
  thus ?case
  proof cases
    case left
    have in_tr'1: "(zip lrs' nds', lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'1"
      by (metis ATreeNode_access_imp_in_set ATreeNode_access_node_left ATreeNode_access_node_parn ATreeNode_in_set_imp_access left list.sel(1) lrs nds parn.prems(2) parn.prems(3))
    obtain lf where lf: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr1"
      using parn(1)[OF tm1 eq_len in_tr'1]
      by presburger
    then show ?thesis
      using ATree_gather_left left lrs tr by fastforce
  next
    case right
    have in_tr'2: "(zip lrs' nds', lf') \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr'2"
      by (metis ATreeNode_access_imp_in_set ATreeNode_access_node_right ATreeNode_access_node_parn ATreeNode_in_set_imp_access right list.sel(1) lrs nds parn.prems(2) parn.prems(3))
    obtain lf where lf: "(lrs', lf) \<in> \<union>\<^sub>A\<^sub>T tr2"
      using parn(2)[OF tm2 eq_len in_tr'2]
      by presburger
    then show ?thesis
      using ATree_gather_right right lrs tr by fastforce
  qed
qed

lemma ATreen_gather_same_path_ex:
  fixes tree2 :: "('lf, 'nd) ATreeNode"
  assumes
    inside1: "(lrs, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tree2" and
    inside2: "treen_mirror tree1 tree2"
  shows "\<exists> lf . (map fst lrs, lf) \<in> \<union>\<^sub>A\<^sub>T tree1"
  by (metis ATreeNode_gather_imp_ATree_gather assms(1) assms(2) length_map zip_map_fst_snd)

definition is_subtreeN_of
  where
"is_subtreeN_of tr1 n tr2 lrns TR = (
  (\<forall> lrns_s lf . ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1) = (((lrns) @ (Left, n) # (lrns_s), lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR)) \<and>
  (\<forall> lrns_s lf . ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2) = (((lrns) @ (Right, n) # (lrns_s), lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR)))"

definition is_subtree_of
  where
"is_subtree_of tr1 tr2 lrs TR = (
  (\<forall> lrs_s lf . ((lrs_s, lf) \<in> \<union>\<^sub>A\<^sub>T tr1) = (((lrs) @ Left # (lrs_s), lf) \<in> \<union>\<^sub>A\<^sub>T TR)) \<and>
  (\<forall> lrs_s lf . ((lrs_s, lf) \<in> \<union>\<^sub>A\<^sub>T tr2) = (((lrs) @ Right # (lrs_s), lf) \<in> \<union>\<^sub>A\<^sub>T TR)))"

lemma subtreeN2_hlpr:
  assumes "\<And> tr1_1 tr1_2 tr2_1 tr2_2 lrs_p ns_p n.
    length lrs_p = length ns_p \<Longrightarrow>
    treen_mirror tr1_1 tr2_1 \<Longrightarrow>
    treen_mirror tr1_2 tr2_2 \<Longrightarrow>
    is_subtree_of tr1_1 tr1_2 lrs_p TR1 \<Longrightarrow>
    is_subtreeN_of tr2_1 n tr2_2 (zip lrs_p ns_p) TR2 \<Longrightarrow>
    f tr1_1 tr1_2 tr2_1 n tr2_2" and
    tm: "treen_mirror TR1 TR2"
  shows "subtreeN2_property f TR1 TR2"
using assms
proof(induct TR1 arbitrary: TR2)
  case (seq x1)
  obtain x2 where TR2: "TR2 = seqn x2" using seq(2) unfolding treen_mirror_unfold_seql by blast
  show ?case unfolding TR2 subtreeN2_property.simps ..
next
  case (par l1 r1)
  then obtain l2 n r2 where TR2: "TR2 = parn l2 n r2"
    using par(2) treen_mirror_unfold_parl by metis
  have "\<And>lrs_p ns_p tr1_1 tr2_1 tr1_2 tr2_2 n.
        length lrs_p = length ns_p \<Longrightarrow>
        treen_mirror tr1_1 tr2_1 \<Longrightarrow>
        treen_mirror tr1_2 tr2_2 \<Longrightarrow>
        is_subtree_of tr1_1 tr1_2 lrs_p l1 \<Longrightarrow>
        is_subtreeN_of tr2_1 n tr2_2 (zip lrs_p ns_p) l2 \<Longrightarrow>
        f tr1_1 tr1_2 tr2_1 n tr2_2"
  proof -
    fix lrs_p ns_p tr1_1 tr2_1 tr1_2 tr2_2 n_p
    assume
      len_eq: "length lrs_p = length ns_p" and
      tm1: "treen_mirror tr1_1 tr2_1" and
      tm2: "treen_mirror tr1_2 tr2_2" and
      is_subtree1: "is_subtree_of tr1_1 tr1_2 lrs_p l1" and
      is_subtree2: "is_subtreeN_of tr2_1 n_p tr2_2 (zip lrs_p ns_p) l2"
    have "length (NodeAccessor.Left # lrs_p) = length (n # ns_p)"
      by (simp add: len_eq)
    moreover have "is_subtree_of tr1_1 tr1_2 (Left # lrs_p) (par l1 r1)"
      using is_subtree1
      unfolding is_subtree_of_def
      by (simp add: ATree_gather_left)
    moreover have "is_subtreeN_of tr2_1 n_p tr2_2 (zip (Left # lrs_p) (n # ns_p)) TR2"
      unfolding TR2
      using is_subtree2
      unfolding is_subtreeN_of_def
      by (simp add: ATreeNode_gather_left)
    ultimately show "f tr1_1 tr1_2 tr2_1 n_p tr2_2"
      using par(3) tm1 tm2 by blast
  qed
  moreover have "\<And>lrs_p ns_p tr1_1 tr2_1 tr1_2 tr2_2 n.
        length lrs_p = length ns_p \<Longrightarrow>
        treen_mirror tr1_1 tr2_1 \<Longrightarrow>
        treen_mirror tr1_2 tr2_2 \<Longrightarrow>
        is_subtree_of tr1_1 tr1_2 lrs_p r1 \<Longrightarrow>
        is_subtreeN_of tr2_1 n tr2_2 (zip lrs_p ns_p) r2 \<Longrightarrow>
        f tr1_1 tr1_2 tr2_1 n tr2_2"
  proof -
    fix lrs_p ns_p tr1_1 tr2_1 tr1_2 tr2_2 n_p
    assume
      len_eq: "length lrs_p = length ns_p" and
      tm1: "treen_mirror tr1_1 tr2_1" and
      tm2: "treen_mirror tr1_2 tr2_2" and
      is_subtree1: "is_subtree_of tr1_1 tr1_2 lrs_p r1" and
      is_subtree2: "is_subtreeN_of tr2_1 n_p tr2_2 (zip lrs_p ns_p) r2"
    have "length (NodeAccessor.Right # lrs_p) = length (n # ns_p)"
      by (simp add: len_eq)
    moreover have "is_subtree_of tr1_1 tr1_2 (Right # lrs_p) (par l1 r1)"
      using is_subtree1
      unfolding is_subtree_of_def
      by (simp add: ATree_gather_right)
    moreover have "is_subtreeN_of tr2_1 n_p tr2_2 (zip (Right # lrs_p) (n # ns_p)) TR2"
      unfolding TR2
      using is_subtree2
      unfolding is_subtreeN_of_def
      by (simp add: ATreeNode_gather_right)
    ultimately show "f tr1_1 tr1_2 tr2_1 n_p tr2_2"
      using par(3) tm1 tm2 by blast
  qed
  moreover have "treen_mirror l1 l2"
    using TR2 par.prems(2) by auto
  moreover have "treen_mirror r1 r2"
    using TR2 par.prems(2) by auto
  ultimately have "subtreeN2_property f l1 l2 \<and> subtreeN2_property f r1 r2"
    using par(1,2) by blast
  moreover have "f l1 r1 l2 n r2"
      using par(3)[of "[]" "[]" l1 l2 r1 r2 n, unfolded is_subtreeN_of_def is_subtree_of_def TR2, simplified]
      by (simp add: ATreeNode_gather_left ATreeNode_gather_right ATree_gather_left ATree_gather_right \<open>treen_mirror l1 l2\<close> \<open>treen_mirror r1 r2\<close>)
  ultimately show ?case
    unfolding TR2
    unfolding subtreeN2_property.simps
  by auto
qed

lemma subtreeN_hlpr:
  assumes
    st_prop: "subtreeN_property f TR" and
    is_subtree: "is_subtreeN_of tr1 n tr2 lrns_p TR"
  shows "f tr1 n tr2"
using assms
proof (induct TR arbitrary: lrns_p)
  case (seqn x)
  then show ?case
    unfolding subtreeN_property.simps
    unfolding is_subtreeN_of_def ATreeNode_gather_def ATreeNode_gather_hlpr.simps
    by (metis ATreeNode_gather_def ATreeNode_gather_not_empty Nil_is_append_conv equals0I list.distinct(1) old.prod.inject singleton_iff surj_pair)
next
  case (parn TR1 x2 TR2)
  consider (lrns_p) "lrns_p = []" | (not_empty) "\<exists> lrn lrns_p' . lrns_p = lrn # lrns_p'"
    by (meson list.exhaust)
  thus ?thesis
  proof cases
    case lrns_p
    have "(\<forall>lrns_s lf. ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1) = ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR1))"
      using parn(4)[unfolded lrns_p is_subtreeN_of_def  append_Nil, simplified, THEN conjunct1]
      by (metis (no_types, hide_lams) ATreeNode_gather_left ATreeNode_gather_not_empty ATreeNode_gather_par equals0I list.sel(1) old.prod.inject surj_pair)
    hence tr1_eq_TR1: "tr1 = TR1"
      by (simp add: ATreeNode_all_leafs_eq_imp_eq)
    moreover have "(\<forall>lrns_s lf. ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2) = ((lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR2))"
      using parn(4)[unfolded lrns_p is_subtreeN_of_def  append_Nil, simplified, THEN conjunct2]
      by (metis (no_types, hide_lams) ATreeNode_gather_right ATreeNode_gather_not_empty ATreeNode_gather_par equals0I list.sel(1) old.prod.inject surj_pair)
    hence "tr2 = TR2"
      by (simp add: ATreeNode_all_leafs_eq_imp_eq)
    moreover have "x2 = n"
      by (metis ATreeNode_gather_not_empty ATreeNode_gather_par Pair_inject append_Nil equals0I is_subtreeN_of_def list.sel(1) lrns_p parn.prems(2) prod.collapse)
    ultimately have "(parn TR1 x2 TR2) = (parn tr1 n tr2)"
      by argo
    then show ?thesis
      using parn(3)[unfolded subtreeN_property.simps]
      by blast
  next
    case not_empty
    then obtain lrn lrns_p' where lrns_p: "lrns_p = lrn # lrns_p'"
      using parn(4)[unfolded is_subtreeN_of_def] by blast
    obtain lr n2 where lrn: "lrn = (lr, n2)" by fastforce
    consider (left) "lr = Left" | (right) "lr = Right"
      using NodeAccessor.exhaust by meson
    thus ?thesis
    proof cases
      case left
      have *: "is_subtreeN_of tr1 n tr2 lrns_p' TR1"
        unfolding is_subtreeN_of_def lrns_p lrn left
      proof auto
        show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1 \<Longrightarrow> (lrns_p' @ (NodeAccessor.Left, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR1"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn left
          using ATreeNode_gather_par by fastforce
        show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2 \<Longrightarrow> (lrns_p' @ (NodeAccessor.Right, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR1"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn left
          using ATreeNode_gather_par by fastforce
        show "\<And>lrns_s lf. (lrns_p' @ (NodeAccessor.Left, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR1 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn left
          using ATreeNode_gather_par
          by (metis (no_types, lifting) ATreeNode_access_exists ATreeNode_gather_left append_Cons list.inject prod.sel(2))
        show "\<And>lrns_s lf. (lrns_p' @ (NodeAccessor.Right, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR1 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn left
          using ATreeNode_gather_par
          by (metis (no_types, lifting) ATreeNode_access_exists ATreeNode_gather_left append_Cons list.inject prod.sel(2))
      qed
      then show ?thesis
        using parn(1)[OF _ *] parn(3)[unfolded subtreeN_property.simps] by argo
    next
      case right
      have *: "is_subtreeN_of tr1 n tr2 lrns_p' TR2"
        unfolding is_subtreeN_of_def lrns_p lrn right
      proof auto
        show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1 \<Longrightarrow> (lrns_p' @ (NodeAccessor.Left, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR2"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn right
          using ATreeNode_gather_par by fastforce
        show "\<And>lrns_s lf. (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2 \<Longrightarrow> (lrns_p' @ (NodeAccessor.Right, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR2"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn right
          using ATreeNode_gather_par by fastforce
        show "\<And>lrns_s lf. (lrns_p' @ (NodeAccessor.Left, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR2 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn right
          using ATreeNode_gather_par
          by (metis (no_types, lifting) ATreeNode_access_exists ATreeNode_gather_right append_Cons list.inject prod.sel(2))
        show "\<And>lrns_s lf. (lrns_p' @ (NodeAccessor.Right, n) # lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N TR2 \<Longrightarrow> (lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2"
          using parn(4)
          unfolding is_subtreeN_of_def lrns_p lrn right
          using ATreeNode_gather_par
          by (metis (no_types, lifting) ATreeNode_access_exists ATreeNode_gather_right append_Cons list.inject prod.sel(2))
      qed
      then show ?thesis
        using parn(2)[OF _ *] parn(3)[unfolded subtreeN_property.simps] by argo
    qed
  qed
qed

lemma all_subtreesN:
  assumes
    path: "(lrns_p @ lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr" and
    not_empty: "lrns_s \<noteq> []"
  shows "\<exists> tr1 n tr2 lrns_s' lr. is_subtreeN_of tr1 n tr2 lrns_p tr \<and> lrns_s = (lr, n) # lrns_s'"
using path
proof (induct tr arbitrary: lrns_p)
  case (seqn x)
  then show ?case
    using not_empty
    unfolding ATreeNode_gather_def ATreeNode_gather_hlpr.simps
    by blast
next
  case (parn tr1 x2 tr2)
  consider (empty) "lrns_p = []" | (multiple) "\<exists> a xs. lrns_p = a # xs"
    using not_empty
    by (metis Nil_is_append_conv neq_Nil_conv)
  then show ?case
  proof cases
    case empty  
    hence "is_subtreeN_of tr1 x2 tr2 lrns_p (parn tr1 x2 tr2)"
      unfolding empty is_subtreeN_of_def
      by (simp add: ATreeNode_gather_left ATreeNode_gather_right)
    thus ?thesis
      using not_empty
      using ATreeNode_gather_par empty parn.prems by fastforce
  next
    case multiple
    then obtain a xs where lrns_p: "lrns_p = a # xs"
      by blast
    then obtain lr n where a: "a = (lr, n)" by fastforce
    have n: "n = x2"
      using parn(3)[unfolded lrns_p a]
      using ATreeNode_gather_par by fastforce
    consider (left) "lr = Left" | (right) "lr = Right"
      using NodeAccessor.exhaust by meson
    thus ?thesis
    proof cases
      case left
      hence *: "(xs @ lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr1"
        using ATreeNode_gather_par \<open>lrns_p = a # xs\<close> a parn.prems by force
      obtain tr1a n' tr2a lrns_s' lr' where
        subtree: "is_subtreeN_of tr1a n' tr2a xs tr1" and
        lrns: "lrns_s = (lr', n') # lrns_s'"
        using parn(1)[OF *] by blast
      have "is_subtreeN_of tr1a n' tr2a lrns_p (parn tr1 x2 tr2)"
        unfolding is_subtreeN_of_def
        using subtree
        unfolding lrns_p left a is_subtreeN_of_def n
        by (simp add: ATreeNode_gather_left)
      thus ?thesis
        using lrns by blast
    next
      case right
      hence *: "(xs @ lrns_s, lf) \<in> \<union>\<^sub>A\<^sub>T\<^sub>N tr2"
        using ATreeNode_gather_par \<open>lrns_p = a # xs\<close> a parn.prems by force
      obtain tr1a n' tr2a lrns_s' lr' where
        subtree: "is_subtreeN_of tr1a n' tr2a xs tr2" and
        lrns: "lrns_s = (lr', n') # lrns_s'"
        using parn(2)[OF *] by blast
      have "is_subtreeN_of tr1a n' tr2a lrns_p (parn tr1 x2 tr2)"
        unfolding is_subtreeN_of_def
        using subtree
        unfolding lrns_p right a is_subtreeN_of_def n
        by (simp add: ATreeNode_gather_right)
      thus ?thesis
        using lrns by blast
    qed
  qed
qed

end